package utilities;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;


/** ------------------------------------------------------------------------------------------------------------------
 * start of ARFFfile class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class ARFFfile {
	
	
	
	
	/*  global parameters  */
	private String path_filename;
	private ArrayList<String> header;
	private String relation;
	private ArrayList<String> attributeSpecifications;
	private ArrayList<String> dataVectors;
	/*  end of global parameters  */
	
	
	
	
	/*  start of getPathFilename method  */
	/**
	 * @return a String of the filename (including the path prefix) that was loaded to create the current ARFFfile object
	 */
	public String getPathFilename(){
		return this.path_filename;
	}/*  end of getPathFilename method  */
	
	
	
	
	/*  start of getFilename method  */
	/**
	 * @return a String of the filename (excluding the path prefix if there is one) that was loaded to create the current ARFFfile object
	 */
	public String getFilename(){
		if(this.path_filename.contains("/")){
			return this.path_filename.substring(this.path_filename.lastIndexOf("/")+1);
		}
		return this.path_filename;
	}/*  end of getFilename method  */

	
	
	
	/*  start of getNumberOfAttributeVectors method  */
	/**
	 * @return the number of data vectors stored in the current ARFFfile object
	 */
	public int getNumberOfAttributeVectors(){
		return this.dataVectors.size();
	}/*  end of getNumberOfAttributeVectors method  */
	
	
	
	
	/*  start of loadFromFile method  */
	/**
	 * @param path_filename the filename (including the prefix path) of the input ARFF file to load
	 */
	private void loadFromFile(String path_filename){
		if(path_filename == null){
			System.err.println("Error: the specified filename is null and the input ARFF file cannot be loaded.\nSystem.exit(0)");
			System.exit(0);
		}
		
		this.path_filename = path_filename;
		
		this.header = new ArrayList<String>();
		
		this.relation = new String("@RELATION 'default_name'");
		
		this.attributeSpecifications = new ArrayList<String>();
		
		this.dataVectors = new ArrayList<String>();
		
		

		String line = null;
		try{
			BufferedReader br = new BufferedReader(new FileReader(path_filename));
			
			boolean loadingHeaderPhase = true;
			boolean relationFound = false;
			boolean loadingDataPhase = false;
			
			while((line = br.readLine()) != null){
				
				
				//  a comment line
				if(line.startsWith("%")){
					
					//  comment lines part of the header
					if(loadingHeaderPhase){
						this.header.add(line);
					}
					else{
						//  skip the line
					}
				}
				
				//  the line declaring the relation
				else if(line.startsWith("@RELATION")){
					loadingHeaderPhase = false;
					
					if(relationFound){
						System.err.println("Error: can only have one relation specified per file (already found " + this.relation + ").\nSystem.exit(0)");
						System.exit(0);
					}
					
					this.relation = line;
					relationFound = true;
				}
				
				//  the lines defining the list of attributes
				else if(line.startsWith("@ATTRIBUTE")){
					loadingHeaderPhase = false;
					
					this.attributeSpecifications.add(line);
					
				}
				
				//  starting the section of lines that define the data vectors
				else if(line.startsWith("@DATA")){
					loadingHeaderPhase = false;
					
					if(!relationFound){
						System.err.println("Error: the @RELATION has not been specified before the @DATA.\nSystem.exit(0)");
						System.exit(0);
					}
					
					if(this.attributeSpecifications.size() < 1){
						System.err.println("Error: the @ATTRIBUTE have not been specified before the @DATA.\nSystem.exit(0)");
						System.exit(0);
					}
					
					loadingDataPhase = true;
				}
				
				else{
					if(loadingDataPhase){
						if(line.split(",").length == this.attributeSpecifications.size()){
							this.dataVectors.add(line);
						}
					}
					else{
						//  skip the line
					}
				}
				
				
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		
		this.header.trimToSize();
		this.attributeSpecifications.trimToSize();
		this.dataVectors.trimToSize();
		
		System.out.println("Loaded input file: " + path_filename);
	}/*  end of loadFromFile method  */
	
	
	
	
	/*  start of displayData method  */
	/**
	 * @return void
	 */
	public void displayData(){
		
		//  first add the header lines
		for(String currentHeaderLine: this.header){
			System.out.println(currentHeaderLine);
		}
		
		//  then the relation line
		System.out.println(this.relation);
		
		//  then skip a line
		System.out.println();
		
		//  then the attributes
		for(String currentAttributeLine: this.attributeSpecifications){
			System.out.println(currentAttributeLine);
		}
		
		//  the skip a line
		System.out.println();
		
		//  then the data
		System.out.println("@DATA");
		for(String currentVector: this.dataVectors){
			System.out.println(currentVector);
		}
		
	}/*  end of displayData method  */
	
	
	
	
	/*  start of outputToFileSkippingSpecifiedVectors  */
	/**
	 * @param path_filename the filename (including the path prefix) to use to output the current ARFFfile object to file
	 * @param skipIndex_start the start index (inclusive) of the data vectors to skip when outputting to file
	 * @param skipIndex_end the end index (inclusive) of the data vectors to skip when outputting to file
	 */
	public void outputToFileSkippingSpecifiedVectors(String path_filename, int skipIndex_start, int skipIndex_end){
		if(path_filename == null){
			System.err.println("Error: the output filename is null and the data cannot be output to file.\nSystem.exit(0)");
			System.exit(0);
		}
		

		try{
			BufferedWriter bw = new BufferedWriter(new FileWriter(path_filename));
			

			//  first the header lines
			for(String currentHeaderLine: this.header){
				bw.write(currentHeaderLine);
				bw.newLine();
				bw.flush();
			}
			
			//  then the relation line
			bw.write(this.relation);
			bw.newLine();
			bw.flush();
			
			//  then skip a line
			bw.newLine();
			bw.flush();
			
			//  then the attributes
			for(String currentAttributeLine: this.attributeSpecifications){
				bw.write(currentAttributeLine);
				bw.newLine();
				bw.flush();
			}
			
			//  the skip a line
			bw.newLine();
			bw.flush();
			
			//  then the data
			bw.write("@DATA");
			bw.newLine();
			bw.flush();
			for(int i=1; i<=this.dataVectors.size(); i++){
				if((i >= skipIndex_start) && (i <= skipIndex_end)){
					//  skip the vector
				}
				else{
					bw.write(this.dataVectors.get(i-1));		//  need to correct the index to start counting at zero
					bw.newLine();
					bw.flush();
				}
			}
			
			bw.flush();
			bw.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error: cannot write file with name \"" + path_filename + "\"\nSystem.exit(0)");
			System.exit(0);
		}
		System.out.println("\tOutput file: " + path_filename);
	}/*  end of outputToFileSkippingSpecifiedVectors method  */
	
	

	
	/*  start of outputToFileSpecifiedVectors  */
	/**
	 * @param path_filename the filename (including the path prefix) to use to output the current ARFFfile object to file
	 * @param index_start the start index (inclusive) of the data vectors to output to file
	 * @param index_end the end index (inclusive) of the data vectors to output to file
	 */
	public void outputToFileSpecifiedVectors(String path_filename, int index_start, int index_end){
		if(path_filename == null){
			System.err.println("Error: the output filename is null and the data cannot be output to file.\nSystem.exit(0)");
			System.exit(0);
		}
		

		try{
			BufferedWriter bw = new BufferedWriter(new FileWriter(path_filename));
			

			//  first the header lines
			for(String currentHeaderLine: this.header){
				bw.write(currentHeaderLine);
				bw.newLine();
				bw.flush();
			}
			
			//  then the relation line
			bw.write(this.relation);
			bw.newLine();
			bw.flush();
			
			//  then skip a line
			bw.newLine();
			bw.flush();
			
			//  then the attributes
			for(String currentAttributeLine: this.attributeSpecifications){
				bw.write(currentAttributeLine);
				bw.newLine();
				bw.flush();
			}
			
			//  the skip a line
			bw.newLine();
			bw.flush();
			
			//  then the data
			bw.write("@DATA");
			bw.newLine();
			bw.flush();
			for(int i=1; i<=this.dataVectors.size(); i++){
				if((i >= index_start) && (i <= index_end)){
					bw.write(this.dataVectors.get(i-1));		//  need to correct the index to start counting at zero
					bw.newLine();
					bw.flush();
				}
				else{
					//  skip the vector
				}
			}
			
			bw.flush();
			bw.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error: cannot write file with name \"" + path_filename + "\"\nSystem.exit(0)");
			System.exit(0);
		}
		System.out.println("\tOutput file: " + path_filename);
	}/*  end of outputToFileSpecifiedVectors method  */
	
	
	
	
	/*  start of outputToFile method  */
	/**
	 * @param path_filename the filename (including the path prefix) to use to output the current ARFFfile object to file
	 */
	public void outputToFile(String path_filename){
		this.outputToFileSpecifiedVectors(path_filename, 1, this.getNumberOfAttributeVectors());
	}/*  end of outputToFile method  */
	
	
	
	
	/*  start of toString method  */
	/**
	 * @return a String representation of the data contained in the ARFFfile object
	 */
	public String toString(){
		
		String ARFFasString = "";
		
		//  first add the header lines
		for(String currentHeaderLine: this.header){
			ARFFasString += currentHeaderLine;
			ARFFasString += "/n";
		}
		
		//  then the relation line
		ARFFasString += this.relation;
		
		//  then skip a line
		ARFFasString += "\n\n";
		
		//  then the attributes
		for(String currentAttributeLine: this.attributeSpecifications){
			ARFFasString += currentAttributeLine;
			ARFFasString += "\n";
		}
		
		//  the skip a line
		ARFFasString += "\n";
		
		//  then the data
		ARFFasString += "@DATA\n";
		for(String currentVector: this.dataVectors){
			ARFFasString += currentVector;
			ARFFasString += "\n";
		}
		
		return ARFFasString;
	}/*  end of toString method  */
	
	
	
	
	/*  start of constructor method  */
	/**
	 * @param path_filename the filename (including the prefix path) of the input ARFF file to load 
	 */
	public ARFFfile(String path_filename){
		if(path_filename == null){
			System.err.println("Error: the specified filename is null and the input ARFFfile object cannot be created.\nSystem.exit(0)");
			System.exit(0);
		}
		
		this.loadFromFile(path_filename);
	}/*  end of constructor method  */
	
	
	
	
	
	
	
}/*  end of ARFFfile class  */
