package utilities;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;


/** ------------------------------------------------------------------------------------------------------------------
 * start of ARFFutility class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class ARFFutility {

	

	
	/*  global parameters  */
	private String relation;
	private HashMap<Integer, String> classNames;
	private ArrayList<String> possibleClassLabels;
	private ArrayList<String> ARFFheader;
	private ArrayList<String> attributeNames;
	private HashMap<Integer, ArrayList<String>> attributeVectors;
	/*  end of global parameters  */
	
	
	
	
	/*  start of constructor method  */
	/**
	 * @return void
	 */
	public ARFFutility(){
		this("default_relation_name");
	}/*  end of constructor method  */
	
	
	
	
	/*  start of constructor method  */
	/**
	 * @param relation a String specifying the name of the dataset
	 */
	public ARFFutility(String relation){
		if(relation == null){
			System.err.println("Error: the relation String is null and cannot initialize the ARFFutility object.\nSystem.exit(0)");
			System.exit(0);
		}
		
		this.relation = relation;
	}/*  end of constructor method  */
	
	
	

	/*  start of generateARFFfile method  */
	/**
	 * @param path_filename_data the filename (including the path prefix) of the file containing the input data matrix
	 * @param path_filename_class_labels the filename (including the path prefix) of the file containing the class labels
	 */
	public void generateARFFfile(String path_filename_data, String path_filename_class_labels){
		this.generateARFFfile(path_filename_data, path_filename_class_labels, null);
	}/*  end of generateARFFfile method  */
	
	
	
	
	/*  start of generateARFFfile method  */
	/**
	 * @param path_filename_data the filename (including the path prefix) of the file containing the input data matrix
	 * @param path_filename_class_labels the filename (including the path prefix) of the file containing the class labels
	 * @param path_filename_ARFF_header the filename (including the path prefix) of the file containing the header to use for the output ARFF file
	 */
	public void generateARFFfile(String path_filename_data, String path_filename_class_labels, String path_filename_ARFF_header){
		if((path_filename_data == null) || (path_filename_class_labels == null)){
			System.err.println("Error: all of the parameters must be non-null to generate the ARFF file.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		this.loadClassLabels(path_filename_class_labels);
//		this.displayClassLabels();
		

		this.loadARFFheader(path_filename_ARFF_header);
//		this.displayARFFheader();
		
		
		this.loadDataMatrix(path_filename_data);
//		this.displayAttributes();
		
		
		this.outputARFFfile(this.changeFilenameExtension(path_filename_data, "arff"));
	}/*  end of generateARFFfile method  */
	
	
	
	
	/*  start of generateARFFfile method  */
	/**
	 * @param attributeVectors the HashMap containing all of the data as vectors of attributes
	 * @param path_filename_class_labels the filename (including the path prefix) of the file containing the class labels
	 * @param outputPath the filesystem path to use as a prefix for the output file
	 */
//	public void generateARFFfile(HashMap<Integer, ArrayList<String>> attributeVectors, String path_filename_class_labels, String outputPath){
//		this.generateARFFfile(attributeVectors, path_filename_class_labels, null, outputPath);
//	}/*  end of generateARFFfile method  */
	
	
	
	
	/*  start of generateARFFfile method  */
	/**
	 * @param attributeVectors the HashMap containing all of the data as vectors of attributes
	 * @param path_filename_class_labels the filename (including the path prefix) of the file containing the class labels
	 * @param path_filename_ARFF_header the filename (including the path prefix) of the file containing the header to use for the output ARFF file
	 * @param path_filename_attribute_names the filename (including the path prefix) of the file containing the name for each attribute
	 * @param path_filename_output the the filename (including the path prefix) for the output file
	 */
	public void generateARFFfile(HashMap<Integer, ArrayList<String>> attributeVectors, String path_filename_class_labels, String path_filename_ARFF_header, String path_filename_attribute_names, String path_filename_output){
		if((attributeVectors == null) || (path_filename_class_labels == null)){
			System.err.println("Error: all of the parameters must be non-null to generate the ARFF file.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		this.loadClassLabels(path_filename_class_labels);
//		this.displayClassLabels();
		

		this.loadARFFheader(path_filename_ARFF_header);
//		this.displayARFFheader();
		
		
		this.loadAttributeNames(path_filename_attribute_names);
//		this.displayAttributeNames();
		
		
		this.attributeVectors = attributeVectors;
//		this.displayAttributes();
		
		
		this.outputARFFfile(this.changeFilenameExtension(path_filename_output, "arff"));
	}/*  end of generateARFFfile method  */
	
	
	
	
	/*  start of loadClassLabels method  */
	/**
	 * @param filename_class_labels the name of the file containing the class labels for each of the structures
	 */
	private void loadClassLabels(String filename_class_labels){
		if(filename_class_labels == null){
			System.err.println("Error: the class labels filename is null and cannot be opened.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		classNames = new HashMap<Integer, String>();
		possibleClassLabels = new ArrayList<String>();
		

		String line = null;
		int lineCounter = -1;			//  there is one header line
		try{
			BufferedReader br = new BufferedReader(new FileReader(filename_class_labels));
			while((line = br.readLine()) != null){
				if(lineCounter >= 0){
					
					String[] values = line.split("\t");
					if(values.length != 3){
						System.err.println("Error: there are not exactly three entries in each row of the class labels file.\nSystem.exit(0)");
						System.exit(0);
					}
					
					int startIndex = Integer.parseInt(values[1]);
					int endIndex = Integer.parseInt(values[2]);
					
					String currentClassName = new String(values[0]);
					
					possibleClassLabels.add(currentClassName);
					
					for(int i=startIndex; i<=endIndex; i++){
						int currentStructureIndex = i;
						
						if(classNames.containsKey(currentStructureIndex)){
							System.err.println("Error: the class names HashMap already contains a class label for structure index " + currentStructureIndex + ".\nSystem.exit(0)");
							System.exit(0);
						}
						
						classNames.put(currentStructureIndex, currentClassName);
					}
				}
				else{
					//  skip and print the header line
//					System.out.println(line);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		possibleClassLabels.trimToSize();
		System.out.println("Loaded input file: " + filename_class_labels);
	}/*  end of loadClassLabels method  */
	
	
	
	
	/*  start of displayClassLabels method  */
	/**
	 * @return void
	 */
	private void displayClassLabels(){
		if(classNames == null){
			System.err.println("Error: the class names HashMap is null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<Integer> structureIndices = new ArrayList<Integer>(classNames.keySet());
		Collections.sort(structureIndices);
		
		System.out.println("===== Class Labels =====");
		System.out.print("Possible class labels: {");
		for(int i=0; i<possibleClassLabels.size(); i++){
			System.out.print(possibleClassLabels.get(i));
			if(i != possibleClassLabels.size()-1){
				System.out.print(",");
			}
		}
		System.out.println("}");
		for(int currentStructureIndex: structureIndices){
			System.out.println(currentStructureIndex + ":\t" + classNames.get(currentStructureIndex));
		}
		System.out.println("Total number of structures = " + classNames.size());
	}/*  end of displayClassLabels method  */
	
	
	
	
	/*  start of loadAttributeNames method  */
	/**
	 * @param filename_attribute_names
	 */
	private void loadAttributeNames(String filename_attribute_names){
		
		
		attributeNames = new ArrayList<String>();
		
		
		if(filename_attribute_names == null){
			return;
		}
		

		String line = null;
		try{
			BufferedReader br = new BufferedReader(new FileReader(filename_attribute_names));
			while((line = br.readLine()) != null){
				attributeNames.add(line);
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		attributeNames.trimToSize();
		System.out.println("Loaded input file: " + filename_attribute_names);
	}/*  end of loadAttributeNames method  */
	


	
	/*  start of displayAttributeNames method  */
	/**
	 * @return void
	 */
	private void displayAttributeNames(){
		if(attributeNames == null){
			System.err.println("Error: the attribute names list is null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		System.out.println("===== Attribute Names =====");
		for(String currentLine: attributeNames){
			System.out.println(currentLine);
		}
	}/*  end of displayAttributeNames method  */
	
	
	
	
	/*  start of loadARFFheader method  */
	/**
	 * @param filename_ARFF_header
	 */
	private void loadARFFheader(String filename_ARFF_header){
		
		
		ARFFheader = new ArrayList<String>();
		
		
		if(filename_ARFF_header == null){
			return;
		}
		

		String line = null;
		try{
			BufferedReader br = new BufferedReader(new FileReader(filename_ARFF_header));
			while((line = br.readLine()) != null){
				ARFFheader.add(line);
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		ARFFheader.trimToSize();
		System.out.println("Loaded input file: " + filename_ARFF_header);
	}/*  end of loadARFFheader method  */
	
	
	
	
	/*  start of displayARFFheader method  */
	/**
	 * @return void
	 */
	private void displayARFFheader(){
		if(ARFFheader == null){
			System.err.println("Error: the ARFF header is null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		System.out.println("===== ARFF header =====");
		for(String currentLine: ARFFheader){
			System.out.println(currentLine);
		}
	}/*  end of displayARFFheader method  */
	
	
	
	
	/*  start of outputARFFfile method  */
	/**
	 * @param filename_ARFF the filename to use for the output ARFF file
	 */
	private void outputARFFfile(String filename_ARFF){
		if(filename_ARFF == null){
			System.err.println("Error: the ARFF output filename is null.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		if(classNames.size() != attributeVectors.size()){
			System.err.println("Error: there are not the same number of entries specified by the input class labels file and the input data matrix file.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		outputARFFheader(filename_ARFF);
		
		outputARFFrelation(filename_ARFF);
		
		outputARFFattributes(filename_ARFF);
		
		outputARFFdata(filename_ARFF);
		
		System.out.println("Output file: " + filename_ARFF);
	}/*  end of outputARFFfile method  */
	
	
	
	
	/*  start of outputARFFheader method  */
	/**
	 * @param filename_ARFF the filename to use for the output ARFF file
	 */
	private void outputARFFheader(String filename_ARFF){
		if(filename_ARFF == null){
			System.err.println("Error: the ARFF output filename is null.\nSystem.exit(0)");
			System.exit(0);
		}
		
		try{
			BufferedWriter bw = new BufferedWriter(new FileWriter(filename_ARFF));
			
			//  write the header lines
			for(String currentHeaderLine: ARFFheader){
				bw.write(currentHeaderLine);
				bw.newLine();
				bw.flush();
			}
			
			bw.flush();
			bw.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error: cannot write file with name \"" + filename_ARFF + "\"\nSystem.exit(0)");
			System.exit(0);
		}
//		System.out.println("\tOutput file: " + outputPath + filename_ARFF);
	}/*  end of outputARFFheader method  */
	
	
	

	/*  start of outputARFFrelation method  */
	/**
	 * @param filename_ARFF the filename to use for the output ARFF file
	 */
	private void outputARFFrelation(String filename_ARFF){
		if(filename_ARFF == null){
			System.err.println("Error: the ARFF output filename is null.\nSystem.exit(0)");
			System.exit(0);
		}
		
		try{
			BufferedWriter bw = new BufferedWriter(new FileWriter(filename_ARFF, true));
			
			//  write the relation line
			bw.write("@RELATION '" + relation + "'");
			bw.newLine();
			bw.flush();
			
			bw.flush();
			bw.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error: cannot write file with name \"" + filename_ARFF + "\"\nSystem.exit(0)");
			System.exit(0);
		}
//		System.out.println("\tOutput file: " + outputPath + filename_ARFF);
	}/*  end of outputARFFrelation method  */
	
	

	
	/*  start of outputARFFattributes method  */
	/**
	 * @param filename_ARFF the filename to use for the output ARFF file
	 */
	private void outputARFFattributes(String filename_ARFF){
		if(filename_ARFF == null){
			System.err.println("Error: the ARFF output filename is null.\nSystem.exit(0)");
			System.exit(0);
		}
		
		try{
			BufferedWriter bw = new BufferedWriter(new FileWriter(filename_ARFF, true));
			
			bw.newLine();
			bw.flush();
			
			
			if(attributeNames.size() > 0){
				System.out.println("Using the attribute names loaded from file.");
				

				if(attributeNames.size() != attributeVectors.get(attributeVectors.keySet().iterator().next()).size()){
					System.err.println("Error: the number of attribute names loaded from file (" + attributeNames.size() + ") does not match the number of calculated attribute vectors (" + attributeVectors.get(attributeVectors.keySet().iterator().next()).size() + ").\nSystem.exit(0)");
					System.exit(0);
				}
				
				
				for(String currentAttributeName: attributeNames){
					bw.write("@ATTRIBUTE " + currentAttributeName + " NUMERIC");
					bw.newLine();
					bw.flush();
				}
			}
			else{
				System.out.println("Using the default attribute names (none loaded from file).");
				
				for(int i=1; i<=attributeVectors.size(); i++){
					bw.write("@ATTRIBUTE " + i + " NUMERIC");
					bw.newLine();
					bw.flush();
				}
			}
			
			
			//  write all of the possible class labels
			bw.write("@ATTRIBUTE class {");
			for(int i=0; i<possibleClassLabels.size(); i++){
				bw.write(possibleClassLabels.get(i));
				if(i != possibleClassLabels.size()-1){
					bw.write(",");
				}
			}
			bw.write("}");
			bw.newLine();
			bw.flush();
			
			
			bw.flush();
			bw.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error: cannot write file with name \"" + filename_ARFF + "\"\nSystem.exit(0)");
			System.exit(0);
		}
//		System.out.println("\tOutput file: " + outputPath + filename_ARFF);
	}/*  end of outputARFFattributes method  */
	
	
	
	
	/*  start of outputARFFdata method  */
	/**
	 * @param filename_ARFF the filename to use for the output ARFF file
	 */
	private void outputARFFdata(String filename_ARFF){
		if(filename_ARFF == null){
			System.err.println("Error: the ARFF output filename is null.\nSystem.exit(0)");
			System.exit(0);
		}
		
		try{
			BufferedWriter bw = new BufferedWriter(new FileWriter(filename_ARFF, true));
			
			bw.newLine();
			bw.flush();
			
			
			bw.write("@DATA");
			bw.newLine();
			bw.flush();

			ArrayList<Integer> structureIndices = new ArrayList<Integer>(attributeVectors.keySet());
			Collections.sort(structureIndices);
			
			for(int currentStructureIndex: structureIndices){
				ArrayList<String> currentStructureAttributesVector = attributeVectors.get(currentStructureIndex);
				for(String currentAttributeValue: currentStructureAttributesVector){
					bw.write(currentAttributeValue + ",");
				}
				bw.write(classNames.get(currentStructureIndex));
				bw.newLine();
				bw.flush();
			}
			
			bw.flush();
			bw.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error: cannot write file with name \"" + filename_ARFF + "\"\nSystem.exit(0)");
			System.exit(0);
		}
//		System.out.println("\tOutput file: " + outputPath + filename_ARFF);
	}/*  end of outputARFFdata method  */
	
	
	
	
	/*  start of changeFilenameExtension method  */
	/**
	 * @param original_filename the original filename for which to replace the extension
	 * @param new_extension the new extension to use for the return filename
	 * @return the original filename with the extension replaced by the new extension passed as a parameter
	 */
	private String changeFilenameExtension(String original_filename, String new_extension){
		if((original_filename == null) || (new_extension == null)){
			System.err.println("Error: the original filename and the new extension must be non-null to change the extension.\nSystem.exit(0)");
			System.exit(0);
		}
		if(!original_filename.contains(".")){
			return original_filename + "." + new_extension;
		}
		return original_filename.substring(0, original_filename.lastIndexOf(".")) + "." + new_extension;
	}/*  end of changeFilenameExtension method  */
	
	
	
	
	/*  start of loadDataMatrix method  */
	/**
	 * @param filename_data the name of the file containing all of the data stored in matrix format
	 */
	private void loadDataMatrix(String filename_data){
		if(filename_data == null){
			System.err.println("Error: the filename is null and the data cannot be loaded from file.\nSystem.exit(0)");
			System.exit(0);
		}
		

		attributeVectors = new HashMap<Integer, ArrayList<String>>();
		
		String line = null;
		int lineCounter = -1;			//  there is one header line with the column labels
		try{
			BufferedReader br = new BufferedReader(new FileReader(filename_data));
			while((line = br.readLine()) != null){
				if(lineCounter >= 0){
					
					String[] values = line.split("\t");
					
					int currentStructureIndex = Integer.parseInt(values[0]);
					
					ArrayList<String> currentStructureAttributesVector = new ArrayList<String>();
					for(int i=1; i<values.length; i++){
						currentStructureAttributesVector.add(values[i]);
					}
					currentStructureAttributesVector.trimToSize();
					
					if(attributeVectors.containsKey(currentStructureIndex)){
						System.err.println("Error: there already exists an attribute vector for structure " + currentStructureIndex + ".\nSystem.exit(0)");
						System.exit(0);
					}
					attributeVectors.put(currentStructureIndex, currentStructureAttributesVector);
				}
				else{
					//  skip and print the header line
//					System.out.println(line);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		System.out.println("Loaded input file: " + filename_data);
	}/*  end of loadDataMatrix method  */
	
	
	
	
	/*  start of displayAttributes method  */
	/**
	 * @return void
	 */
	private void displayAttributes(){
		if(attributeVectors == null){
			System.err.println("Error: the attributes vector is null and the data cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<Integer> structureIndices = new ArrayList<Integer>(attributeVectors.keySet());
		Collections.sort(structureIndices);
		
		for(int currentStructureIndex: structureIndices){
			System.out.print(currentStructureIndex + ":\t");
			ArrayList<String> currentStructureAttributesVector = attributeVectors.get(currentStructureIndex);
			for(String currentAttributeValue: currentStructureAttributesVector){
				System.out.print(currentAttributeValue + ",");
			}
			System.out.println(classNames.get(currentStructureIndex));
		}
	}/*  end of displayAttributes method  */
	
	
	
	
	/*  start of convertDoubleVectorsToStringVectors  */
	/**
	 * @param attributeVectorsAsDoubles the HashMap of attributes stored as arrays of doubles to be converted to ArrayLists of Strings
	 */
	public HashMap<Integer, ArrayList<String>> convertDoubleVectorToStringVector(HashMap<Integer, double[]> attributeVectorsAsDoubles){
		if(attributeVectorsAsDoubles == null){
			System.err.println("Error: the attribute vectors HashMap is null and cannot be converted from double arrays to ArrayLists of Strings.\nSystem.exit(0)");
			System.exit(0);
		}
		
		HashMap<Integer, ArrayList<String>> attributeVectorsAsStrings = new HashMap<Integer, ArrayList<String>>();
		
		ArrayList<Integer> attribute_indices = new ArrayList<Integer>(attributeVectorsAsDoubles.keySet());
		Collections.sort(attribute_indices);
		
		for(Integer current_attribute_index: attribute_indices){
			double[] current_values_array = attributeVectorsAsDoubles.get(current_attribute_index);
			
			ArrayList<String> new_string_array = new ArrayList<String>();
			for(int i=0; i<current_values_array.length; i++){
				new_string_array.add(String.valueOf(current_values_array[i]));
			}
			new_string_array.trimToSize();
			attributeVectorsAsStrings.put(current_attribute_index, new_string_array);
		}
		
		if(attributeVectorsAsDoubles.size() != attributeVectorsAsStrings.size()){
			System.err.println("Error: the doubles HashMap and the Strings HashMap do not have the same number of entries (the conversion failed).\nSystem.exit(0)");
			System.exit(0);
		}
		
		return attributeVectorsAsStrings;
	}/*  end of convertDoubleVectorsToStringVectors method  */
	
	
	
	
	/*  start of loadARFFfile method  */
	/**
	 * @param path_filename the filename (including the prefix path) of the input ARFF file to load
	 * @return an ARFFfile object representation of the file loaded from disk
	 */
	public ARFFfile loadARFFfile(String path_filename){
		if(path_filename == null){
			System.err.println("Error: the specified filename is null and the input ARFF file cannot be loaded.\nSystem.exit(0)");
			System.exit(0);
		}
		
		return new ARFFfile(path_filename);
	}/*  end of loadARFFfile method  */
	
	
	
	
	
}/*  end of ARFFutility class  */
