package utilities;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;



/** ------------------------------------------------------------------------------------------------------------------
 * start of AnnotateIFs class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class AnnotateIFs {

	
	
	
	/*  global parameters  */
	private static boolean verbose = true;
	private static String inputPath = "";
	private static String outputPath = "";
	/*  end of global parameters  */
	
	
	
	
	/*  start of loadIFsFromFile method  */
	/**
	 * @param inputFilename the input IF filename
	 * @return a list containing the IF values in a HashMap<String, Double> (first entry) and the fragments list in an ArrayList<Integer> (second entry)
	 */
	private static ArrayList<Object> loadIFsFromFile(String inputFilename){		
		if(inputFilename == null){
			System.err.println("Error: the input filename is null and the IF values cannot be loaded from file.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		HashMap<String, Double> IFs =  new HashMap<String, Double>();
		
		ArrayList<Integer> fragmentsExistInIF = new ArrayList<Integer>();
		
		
		String line = null;
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + inputFilename));			
			while((line = br.readLine()) != null){
				String values[] = line.split("\t");
				//  make sure that there are exactly five (5) entries in each row
				if(values.length == 5){
					
					int forwardIndex = Integer.valueOf(values[0]);
					int reverseIndex = Integer.valueOf(values[1]);
					double IFvalue = Double.valueOf(values[2]);
					
					String keyCurrentPairIF = forwardIndex + "-" + reverseIndex;
					String keyCurrentPairIFMirror = reverseIndex + "-" + forwardIndex;
					
					IFs.put(keyCurrentPairIF, IFvalue);
					IFs.put(keyCurrentPairIFMirror, IFvalue);
					
					if(!fragmentsExistInIF.contains(forwardIndex)){
						fragmentsExistInIF.add(forwardIndex);
					}
					if(!fragmentsExistInIF.contains(reverseIndex)){
						fragmentsExistInIF.add(reverseIndex);
					}
				}
				else{
					System.err.println("Error: there are not five (5) entries in each row of the file \"" + inputPath + inputFilename + "\"\nSystem.exit(0)");
					System.exit(0);
				}
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error: cannot find file with name \"" + inputPath + inputFilename + "\"\nSystem.exit(0)");
			System.exit(0);
		}
		fragmentsExistInIF.trimToSize();
		Collections.sort(fragmentsExistInIF);
		
		if(verbose){
			System.out.println("\tOpened and read file: " + inputPath + inputFilename);
		}
		
		ArrayList<Object> returnList = new ArrayList<Object>();
		returnList.add(IFs);
		returnList.add(fragmentsExistInIF);
		returnList.trimToSize();
		
		return returnList;
	}/*  end of loadIFsFromFile method  */
	
	
	

	/*  start of displayIFs method  */
	/**
	 * @param IFs the HashMap containing all of the IF values to display
	 * @return void
	 */
	private static void displayIFs(HashMap<String, Double> IFs){
		if(IFs == null){
			System.err.println("Error: the IFs HashMap is null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<String> allKeys = new ArrayList<String>(IFs.keySet());
		Collections.sort(allKeys);
		for(String currentKey: allKeys){
			System.out.println(currentKey + ":\t" + IFs.get(currentKey));
		}		
	}/*  end of displayIFs method  */
	
	
	
	
	/*  start of loadFragmentsFromFile method  */
	/**
	 * @param inputFilename the input fragments filename
	 * @return a HashMap containing the fragment details
	 */
	private static HashMap<Integer, String> loadFragmentsFromFile(String inputFilename){
		if(inputFilename == null){
			System.err.println("Error: the fragments filename is null and the fragments cannot be loaded from file.\nSystem.exit(0)");
			System.exit(0);
		}
		
		HashMap<Integer, String> fragmentLabels = new HashMap<Integer, String>();
		
		String line = null;
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + inputFilename));			
			while((line = br.readLine()) != null){
				String values[] = line.split("\t");
				//  make sure that there are exactly five (5) entries in each row
				if(values.length == 5){
					
					String currentEntry = values[0] + "|" + values[1] + "|" + values[2] + ":" + values[3] + "-" + values[4];
					
					fragmentLabels.put(Integer.parseInt(values[0]), currentEntry);						
				}
				else{
					System.err.println("Error: there are not five (5) entries in each row of the file \"" + inputPath + inputFilename + "\"\nSystem.exit(0)");
					System.exit(0);
				}
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error: cannot find file with name \"" + inputPath + inputFilename + "\"\nSystem.exit(0)");
			System.exit(0);
		}
		
		if(verbose){
			System.out.println("\tOpened and read file: " + inputPath + inputFilename);
		}
		
		return fragmentLabels;
	}/*  end of loadFragmentsFromFile method  */
	
	
	
	
	/*  start of displayFragmentLabels method  */
	/**
	 * @param fragmentLabels a HashMap containing the fragment labels to display
	 * @return void
	 */
	private static void displayFragmentLabels(HashMap<Integer, String> fragmentLabels){
		if(fragmentLabels == null){
			System.err.println("Error: the HashMap of fragment labels is null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<Integer> allKeys = new ArrayList<Integer>(fragmentLabels.keySet());
		Collections.sort(allKeys);
		for(Integer currentKey: allKeys){
			System.out.println(fragmentLabels.get(currentKey));
		}
	}/*  end of displayFragmentLabels method  */
	
	
	
	
	/*  start of displayFragmentsExistInIF method  */
	/**
	 * @param fragmentsExistInIF the list of fragments to display
	 * @return void
	 */
	private static void displayFragmentsExistInIF(ArrayList<Integer> fragmentsExistInIF){
		if(fragmentsExistInIF == null){
			System.err.println("Error: the fragments list is null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		for(Integer currentFragment: fragmentsExistInIF){
			System.out.println(currentFragment);
		}
	}/*  end of displayFragmentsExistInIF method  */
	
	
	
	
	/*  start of outputIFs method  */
	/**
	 * @param filename_IFs the original IF input filename that will be modified and used as the output filename for the matrix
	 * @param IFs the HashMap containing the IF values
	 * @param fragmentsExistInIF the list of fragments that exist in the input data
	 * @param fragmentLabels the HashMap containing the fragment labels
	 * @return void
	 */
	private static void outputIFs(String filename_IFs, HashMap<String, Double> IFs, ArrayList<Integer> fragmentsExistInIF, HashMap<Integer, String> fragmentLabels){
		if((filename_IFs == null) || (IFs == null) || (fragmentsExistInIF == null) || (fragmentLabels == null)){
			System.err.println("Error: all the parameters need to be non-null to output to file.\nSystem.exit(0)");
			System.exit(0);
		}
		
		String outputFilename = outputPath + filename_IFs.substring(0, filename_IFs.lastIndexOf(".")) + "_my5C.txt";
		
		try{
			BufferedWriter bw = new BufferedWriter(new FileWriter(outputFilename));

			//  write the first line
			bw.write(filename_IFs);
			for(int currentFragmentIndex: fragmentsExistInIF){
				bw.write("\t" + fragmentLabels.get(currentFragmentIndex));
			}
			bw.newLine();
			bw.flush();
			
			//  write all of the values to file
			for(int currentFragmentIndexRow: fragmentsExistInIF){
				bw.write(fragmentLabels.get(currentFragmentIndexRow));
				
				for(int currentFragmentIndexColumn: fragmentsExistInIF){
					String searchKeyIF = currentFragmentIndexRow + "-" + currentFragmentIndexColumn;
					if(IFs.containsKey(searchKeyIF)){
						bw.write("\t" + IFs.get(searchKeyIF));
					}
					else{
						bw.write("\t0");						
					}					
				}
				bw.newLine();
				bw.flush();
			}
			bw.close();
		}catch (IOException e) {
			e.printStackTrace();
		}//end of try-catch block
	
		if(verbose){
			System.out.println("\nOutput file: " + outputFilename);
		}
	}/*  end of outputIFs method  */
	
	
	
	
	/*  start of getFilenames method  */
	/**
	 * @param filename_experiment_list the file containing the list of experiments
	 * @return a list of String objects representing the loaded filenames
	 */
	private static ArrayList<String> getFilenames(String filename_experiment_list){
		if(filename_experiment_list == null){
			System.err.println("Error: the filename list is null.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<String> experimentList = new ArrayList<String>();

		String line = null;
		int lineCounter = -1;			//  there is one header line
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename_experiment_list));
			while((line = br.readLine()) != null){
				if(lineCounter >= 0){
					experimentList.add(line);
				}
				else{
					//  skip and the header line
//					System.out.println(line);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		experimentList.trimToSize();
		
		if(verbose){
			System.out.println("Loaded input file: " + inputPath + filename_experiment_list);
		}
		
		return experimentList;
	}/*  end of getFilenames method  */
	
	
	
	
	/*  start of displayFilenames method  */
	/**
	 * @param experimentList the list of filenames to display
	 */
	private static void displayFilenames(ArrayList<String> experimentList){
		if(experimentList == null){
			System.err.println("Error: the experiment list is null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);	
		}
		
		System.out.println("Filename");
		for(String currentFilename: experimentList){
			System.out.println(currentFilename);
		}
	}/*  end of displayFilenames method  */
	
	
	
	
	/*  start of main method  */
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		
		
		
		boolean usingCommandLine = false;
		
		String filename_Fragments = "default.txt";
		String filename_file_list = "default.txt";
		ArrayList<String> filenames = new ArrayList<String>();
		
		
		//  load the parameters from the command line
		if(args.length >= 2){
			//  first parameter is the filename of the fragment genomic coordinates
			filename_Fragments = args[0];
			
			//  all subsequent parameters are the filenames of the experiment files
			for(int i=1; i<args.length; i++){
				filenames.add(args[i]);
			}
			filenames.trimToSize();
			
			usingCommandLine = true;
			verbose = false;
		}
		else{
			System.err.println("Error: there need to be at least two (2) command line parameters.\n" +
					"\t\tfragment_genomic_coordinate_list.txt\n" +
					"\t\texperiment_1_filename.txt\n" +
					"\t\texperiment_2_filename.txt\n" +
					"\t\texperiment_..._filename.txt\n" +
					"System.exit(0)");
//			System.exit(0);
		}
		
		
		
		
		/******************************************************
		 * 													  *
		 * 				SET THE PARAMETERS HERE				  *
		 * 													  *
		 ******************************************************/
		
		
		if(!usingCommandLine){
		
			
			inputPath = "AnnotateIFs/Input/";
			outputPath = "AnnotateIFs/Output/";
			
//			filename_file_list = "file_list.txt";
//			filename_file_list = "file_list_[HoxA].txt";
//			filename_file_list = "file_list_[GD].txt";
//			filename_file_list = "file_list_[test].txt";
			filename_file_list = "file_list.txt";
			
			filename_Fragments = "fragments_HoxA_hg19.txt";
//			filename_Fragments = "fragments_GeneDesert_hg19.txt";
		
//			verbose = false;
			
			
		}
		

		/******************************************************
		 * 													  *
		 * 				END OF SETTING THE PARAMETERS		  *
		 * 													  * 
		 ******************************************************/
		
		
		
		
		if(!usingCommandLine){
			filenames = getFilenames(filename_file_list);
		}
//		displayFilenames(filenames);
		
		
		for(String filename_IFs: filenames){
		
		
			//  display the parameters
			if(verbose){
				System.out.println("IF input file = " + filename_IFs);
				System.out.println("fragments input file = " + filename_Fragments);
			}
		
		
			//  load the IF values from file
			ArrayList<Object> IFsListAndExistingFragmentList = loadIFsFromFile(filename_IFs);
			HashMap<String, Double> IFs = (HashMap<String, Double>)IFsListAndExistingFragmentList.get(0);
			ArrayList<Integer> fragmentsExistInIF = (ArrayList<Integer>)IFsListAndExistingFragmentList.get(1);
			
			
			//  display the IF values
//			displayIFs(IFs);
		
		
			//  display the fragments that exist in the IF file
//			displayFragmentsExistInIF(fragmentsExistInIF);
		

			//  load the fragments from file
			HashMap<Integer, String> fragmentLabels = loadFragmentsFromFile(filename_Fragments);
		
		
			//  display the fragment labels
//			displayFragmentLabels(fragmentLabels);
		
		
			//  output the IFs as a matrix with the fragment labels
			outputIFs(filename_IFs, IFs, fragmentsExistInIF, fragmentLabels);
		
			
		}
		
		if(verbose){
			System.out.println("Done!");
		}
	}/*  end of main method  */

}/*  end of AnnotateIFs class  */
