package analysis;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;

import utilities.PDButility;


/** ------------------------------------------------------------------------------------------------------------------
 * start of BurnInAnalyzer class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class BurnInAnalyzer {

	
	
	
	/*  global variables  */
	private boolean verbose = true;
	private String inputPath = "";
	private String outputPath = "";
	private int windowSize = 10;
	private int numberOfStructures = 0;
	private PDButility pdb = new PDButility();
	private int numberOfPoints = 0;
	private String experimentNameOne = "default";
	private String experimentNameTwo = "default";
	private ArrayList<double[][]> structuresRunOne;
	private ArrayList<double[][]> structuresRunTwo;
	private ArrayList<double[][]> intraStructureDistancesRunOne;
	private ArrayList<double[][]> intraStructureDistancesRunTwo;
	private double[][] interStructureDistancesRunOne;
	private double[][] interStructureDistancesRunTwo;
	private double[][] interStructureDistancesCrossRuns;
	private double[] averageInterStructureDistancesRunOne;
	private double[] averageInterStructureDistancesRunTwo;
	private double[] averageInterStructureDistancesCrossRuns;
	private String output_filename_suffix = "";
	/*  end of global variables  */
	
	
	
	
	/*  start of constructor  */
	/**
	 * @param
	 * @return void
	 */
	public BurnInAnalyzer(){
		this("", "", false);
	}/*  end of constructor  */
	
	
	
	
	/*  start of constructor  */
	/**
	 * @param inputPath the String to use as the input path for the current BurnInAnalyzer object
	 * @param outputPath the String to use as the output path for the current BurnInAnalyzer object
	 * @param verbose the verbose boolean flag to specify if output should be sent to the console (true) or not (false)
	 * @return void
	 */
	public BurnInAnalyzer(String inputPath, String outputPath, boolean verbose){
		
		//  set the inputPath
		this.setInputPath(inputPath);
		
		//  set the outputPath
		this.setOutputPath(outputPath);
		
		//  set the verbose flag
		this.setVerbose(verbose);
		
	}/*  end of constructor  */
	
	
	
	
	/*  start of setInputPath method  */
	/**
	 * @param inputPath the String to use as the input path for the current BurnInAnalyzer object
	 * @return void
	 */
	public void setInputPath(String inputPath){
		if(inputPath == null){
			System.err.println("Error: cannot set the input path to null.\nSystem.exit(0)");
			System.exit(0);
		}
		this.inputPath = inputPath;
	}/*  end of setInputPath method  */
	
	
	
	
	/*  start of getInputPath method  */
	/**
	 * @return the String used as the input path for the current BurnInAnalyzer object
	 */
	public String getInputPath(){
		return this.inputPath;
	}/*  end of getInputPath method  */
	
	
	
	
	/*  start of setOutputPath method  */
	/**
	 * @param outputPath the String to use as the output path for the current BurnInAnalyzer object
	 * @return void
	 */
	public void setOutputPath(String outputPath){
		if(outputPath == null){
			System.err.println("Error: cannot set the output path to null.\nSystem.exit(0)");
			System.exit(0);
		}
		this.outputPath = outputPath;
	}/*  end of setOutputPath method  */
	
	
	
	
	/*  start of getOutputPath method  */
	/**
	 * @return the String used as the output path for the current BurnInAnalyzer object
	 */
	public String getOutputPath(){
		return this.outputPath;
	}/*  end of getOutputPath method  */
	
	
	
	
	/*  start of setVerbose method  */
	/**
	 * @param verbose the verbose boolean flag to specify if output should be sent to the console (true) or not (false)
	 * @return void
	 */
	public void setVerbose(boolean verbose){
		this.verbose = verbose;
	}/*  end of setVerbose method  */
	
	
	
	
	/*  start of getVerbose method  */
	/**
	 * @return the verbose boolean flag that specifies if output is sent to the console (true) or not (false)
	 */
	public boolean getVerbose(){
		return this.verbose;
	}/*  end of getVerbose method  */
	
	
	
	
	/*  start of getStructuresFromFile method  */
	private void getStructuresFromFile(String experimentName, ArrayList<double[][]> structures){
		//  initialize the list that will hold all of the structures
		for(int i=0; i<numberOfStructures; i++){
			double structureInitializer[][] = new double[numberOfPoints][numberOfPoints];
			structures.add(structureInitializer);
		}

		
		//  define the filename of each of the structures to be read from file
		String structureFilename;
		//  use the convention that the structure file names start at index 1
		for(int i=1; i<=numberOfStructures; i++){
			if(i <= 9){
				structureFilename = "structure_000" + i + "_" + experimentName + ".pdb";
			}
			else if(i <= 99){
				structureFilename = "structure_00" + i + "_" + experimentName + ".pdb";
			}
			else if(i <= 999){
				structureFilename = "structure_0" + i + "_" + experimentName + ".pdb";
			}
			else{
				structureFilename = "structure_" + i + "_" + experimentName + ".pdb";
			}
//			System.out.println("Tyring to read structure file: " + inputPath + structureFilename);
			
			//  get the structure from file and store it in the list of structures
			//  remember that the indices in the array list start at zero (0) and not one (1) like the filenames
			structures.set(i-1, pdb.getStructure(inputPath + structureFilename));
		}
	}/*  end of getStructuresFromFile method  */
	
	
	
	
	/*  start of calculateIntraStructureDistances method  */
	private void calculateIntraStructureDistances(ArrayList<double[][]> structures, ArrayList<double[][]> intraStructureDistances){
		for(int structureIndex=0; structureIndex<numberOfStructures; structureIndex++){
			double[][] currentStructureIntraDistances = new double[numberOfPoints][numberOfPoints];
//			System.out.println("\tcalculating the intra distances for structure " + (structureIndex+1));
			for(int i=0; i<numberOfPoints-1; i++){
				for(int j=i+1; j<numberOfPoints; j++){
					currentStructureIntraDistances[i][j] = calculateDistanceBetweenPoints(structures.get(structureIndex), i, j);
				}
			}		
			intraStructureDistances.add(currentStructureIntraDistances);			
		}
	}/*  end of calculateIntraStructureDistances method  */

	
	
	
	/*  start of calculateInterStructureDistances method  */
	private void calculateInterStructureDistances(ArrayList<double[][]> intraStructureDistancesFirstList, ArrayList<double [][]> intraStructureDistancesSecondList, double[][] interStructureDistances){
		for(int firstStructureIndex=0; firstStructureIndex<numberOfStructures-1; firstStructureIndex++){
			for(int secondStructureIndex=firstStructureIndex+1; secondStructureIndex<numberOfStructures; secondStructureIndex++){
				interStructureDistances[firstStructureIndex][secondStructureIndex] = calculateDistanceBetweenStructures(intraStructureDistancesFirstList.get(firstStructureIndex), intraStructureDistancesSecondList.get(secondStructureIndex));
			}
		}
	}/*  end of calculateInterStructureDistances method  */
	
	
	
	
	/*  start of calculateAverageInterStructureDistance method  */
	private void calculateAverageInterStructureDistance(double[][] interStructureDistances, double[] averageInterStructureDistances){
		//  calculate the number of pairs = windowSize! / (2! * (windowSize-2)! )
		double numberOfPairs = windowSize * (windowSize-1) / 2.0;
//		System.out.println("The number of pairs = " + numberOfPairs);
		
		
		//  calculate the average inter structure difference across all pairs of structures within the window
		for(int i=0; i<numberOfStructures-windowSize+1; i++){
			//  go through all the possible pairs of structures in the windows starting at index i
			for(int index1=i; index1<i+windowSize-1; index1++){
				for(int index2=index1+1; index2<i+windowSize; index2++){
					averageInterStructureDistances[i] += interStructureDistances[index1][index2];
				}
			}
			if(numberOfPairs > 1){
				averageInterStructureDistances[i] /= numberOfPairs;
			}
		}		
	}/*  end of calculateAverageInterStructureDistance method  */
	
	
	
	
	/*  start of calculateDistanceBetweenPoints method  */
	private double calculateDistanceBetweenPoints(double[][] structure, int coord1, int coord2){
		double term1 = structure[coord2][0] - structure[coord1][0];
		term1 *= term1;
		double term2 = structure[coord2][1] - structure[coord1][1];
		term2 *= term2;
		double term3 = structure[coord2][2] - structure[coord1][2];
		term3 *= term3;
		return Math.sqrt(term1 + term2 + term3);
	}/*  end of calculateDistanceBetweenPoints method  */
	
	
	
	
	/*  start of calculateDistanceBetweenStructures method  */
	private double calculateDistanceBetweenStructures(double[][] intraFirstStructureDistances, double[][] intraSecondStructureDistances){
		double distance = 0;
		for(int i=0; i<numberOfPoints-1; i++){
			for(int j=i+1; j<numberOfPoints; j++){
				double term = intraSecondStructureDistances[i][j] - intraFirstStructureDistances[i][j];
				term *= term;
				distance += term;
			}
		}
		return distance;
	}/*  end of calculateDistanceBetweenStructures method  */
	
	
	
	
	/*  start of outputInterStructureDistancesToFile method  */
	private void outputInterStructureDistancesToFile(double[][] array, String outputFilename){
		try{
			BufferedWriter bw = new BufferedWriter(new FileWriter(outputPath + outputFilename));
			
			if(array.length == 0){
				bw.write("Note: the array to output had size zero (0). There is no data to output for this array.");
				bw.flush();
			}
			else{
				bw.write("\t");
				bw.flush();
				for(int i=2; i<=numberOfStructures; i++){
					bw.write(i + "\t");
				}
				bw.newLine();
				bw.flush();
				for(int i=0; i<numberOfStructures-1; i++){
					bw.write(String.valueOf(i+1));
					bw.flush();
					for(int j=0; j<numberOfStructures; j++){
						if(j<=i){
							bw.write("\t");
							bw.flush();
						}
						else if(j<numberOfStructures-1){
							bw.write(array[i][j] + "\t");
							bw.flush();
						}
						else{
							bw.write(String.valueOf(array[i][j]));
							bw.flush();
						}						
					}
					bw.newLine();
					bw.flush();
				}
			}
			bw.flush();
			bw.close();
			
			if(verbose){
				System.out.println("\tOutput file: " + outputPath + outputFilename);			
			}
		}catch (IOException e) {
			e.printStackTrace();
		}//end of try-catch block
	}/*  end of outputInterStructureDistancesToFile method  */
	
	
	
	
	/*  start of outputAverageInterStructureDistancesToFile method  */
	private void outputAverageInterStructureDistancesToFile(double[] arrayRunOne, double[] arrayRunTwo, double[] arrayCrossRuns, String outputFilename){
		try{
			BufferedWriter bw = new BufferedWriter(new FileWriter(outputPath + outputFilename));
			
			if(arrayCrossRuns.length == 0){
				bw.write("Note: the array to output had size zero (0). There is no data to output for this array.");
				bw.flush();
			}
			else{
				bw.write("\tRun One\tRun Two\tCross Runs");
				bw.newLine();
				bw.flush();
				for(int i=0; i<numberOfStructures-windowSize+1; i++){
					bw.write("[" + (i+1) + "-" + (i+windowSize) + "]\t");
					bw.flush();
					bw.write(arrayRunOne[i] + "\t" + arrayRunTwo[i] + "\t" + arrayCrossRuns[i]);
					bw.newLine();
					bw.flush();
				}
			}
			bw.flush();
			bw.close();
			
			if(verbose){
				System.out.println("\tOutput file: " + outputPath + outputFilename);			
			}
		}catch (IOException e) {
			e.printStackTrace();
		}//end of try-catch block
	}/*  end of outputAverageInterStructureDistancesToFile method  */
	
	
	
	
	/*  start of run method  */
	/**
	 * @param args the first parameter is the name of the first experiment, the second parameter is the name of the second experiment, the third parameter is the number of structures per run, the fourth parameter is the window size
	 */
	public void run(String[] args) {
		switch (args.length) {
		case 4:
			experimentNameOne = args[0];
			experimentNameTwo = args[1];
			numberOfStructures = Integer.parseInt(args[2]);
			windowSize = Integer.parseInt(args[3]);
			output_filename_suffix = "_" + experimentNameOne + "_" + experimentNameTwo;
			break;
		default:
			System.err.println("Error: there need to be exactly four (4) command line parameters.\n" +
					"\t\texperiment_1_name\t\t\t(just the name substring in 'structure_####_experimentName[i].pdb')\n" +
					"\t\texperiment_2_name\n" +
					"\t\tnumber_of_structures_per_run\n" +
					"\t\twindow_size\n" +
					"System.exit(0)");
			System.exit(0);
			break;
		}
		
		if(numberOfStructures < 2){
			System.err.println("Error: cannot compare less than two (2) structures per run.\nSystem.exit(0)");
			System.exit(0);
		}
		
		//  display the parameters of the run
		if(verbose){
			System.out.println("Parameters:");
			System.out.println("\tUsing a pool of " + numberOfStructures + " structures");
			System.out.println("\twindow size = " + windowSize);
		}
		
		
		//  output the parameters to file
		if(verbose){
			try{
				String parameterFilename = "parameters" + output_filename_suffix + ".txt";
				BufferedWriter bw = new BufferedWriter(new FileWriter(outputPath + parameterFilename));

				bw.write("experimentNameOne=\t" + experimentNameOne);
				bw.newLine();
				bw.flush();
				bw.write("experimentNameTwo=\t" + experimentNameTwo);
				bw.newLine();
				bw.flush();
				bw.write("numberOfStructures=\t" + numberOfStructures);
				bw.newLine();
				bw.flush();
				bw.write("windowSize=\t" + windowSize);
				bw.newLine();
				bw.flush();
				bw.write("command line parameters:");
				for(String arg: args){
					bw.write("\t" + arg);
				}
				bw.newLine();
				bw.flush();

				bw.close();

				System.out.println("\tOutput file: " + outputPath + parameterFilename);				
			}catch (IOException e) {
				e.printStackTrace();
			}//end of try-catch block
		}

		//  calculate how long running BurnInAnalyzer takes (time point before run)
		long timeBefore = System.currentTimeMillis();

		
		if(verbose){
			System.out.println("\nRunning...\n");
		}
		
		
		//  make sure that the window size is smaller than the number of structures
		if(windowSize >= numberOfStructures){
			System.err.println("Error: the window size (" + windowSize + ") is NOT smaller than the number of structures (" + numberOfStructures + ").");
			System.err.println("System.exit(0)");
			System.exit(0);
		}
		
		
		//  get the number of points in each structure
		numberOfPoints = pdb.getNumberOfPoints(inputPath + "structure_0001_" + experimentNameOne + ".pdb");
//		System.out.println("\tNumber of points = " + numberOfPoints);
		
		
		//  load all of the structures from file
		structuresRunOne = new ArrayList<double[][]>(numberOfStructures);
		getStructuresFromFile(experimentNameOne, structuresRunOne);
//		System.out.println("\tstructuresRunOne.size() = " + structuresRunOne.size());
		structuresRunTwo = new ArrayList<double[][]>(numberOfStructures);
		getStructuresFromFile(experimentNameTwo, structuresRunTwo);
//		System.out.println("\tstructuresRunTwo.size() = " + structuresRunTwo.size());
		
		
		//  calculate the intra structure distances for all of the structures
		intraStructureDistancesRunOne = new ArrayList<double[][]>(numberOfStructures);
		intraStructureDistancesRunTwo = new ArrayList<double[][]>(numberOfStructures);
		calculateIntraStructureDistances(structuresRunOne, intraStructureDistancesRunOne);
		calculateIntraStructureDistances(structuresRunTwo, intraStructureDistancesRunTwo);
//		System.out.println("\tintraStructureDistancesRunOne.size() = " + intraStructureDistancesRunOne.size());
//		System.out.println("\tintraStructureDistancesRunTwo.size() = " + intraStructureDistancesRunTwo.size());
		
		
		//  no longer need the structures so free the memory
		structuresRunOne = null;
		structuresRunTwo = null;
		
		
		//  calculate the inter structure distances for the first run, the second run, and across the two runs
		interStructureDistancesRunOne = new double[numberOfStructures][numberOfStructures];
		interStructureDistancesRunTwo = new double[numberOfStructures][numberOfStructures];
		interStructureDistancesCrossRuns = new double[numberOfStructures][numberOfStructures];
		calculateInterStructureDistances(intraStructureDistancesRunOne, intraStructureDistancesRunOne, interStructureDistancesRunOne);
		calculateInterStructureDistances(intraStructureDistancesRunTwo, intraStructureDistancesRunTwo, interStructureDistancesRunTwo);
		calculateInterStructureDistances(intraStructureDistancesRunOne, intraStructureDistancesRunTwo, interStructureDistancesCrossRuns);
		
		
		//  no longer need the intra structure distances so free the memory
		intraStructureDistancesRunOne = null;
		intraStructureDistancesRunTwo = null;
		
		
		//  calculate the average inter structure distance within each window
		averageInterStructureDistancesRunOne = new double[numberOfStructures];
		averageInterStructureDistancesRunTwo = new double[numberOfStructures];
		averageInterStructureDistancesCrossRuns = new double[numberOfStructures];
		calculateAverageInterStructureDistance(interStructureDistancesRunOne, averageInterStructureDistancesRunOne);
		calculateAverageInterStructureDistance(interStructureDistancesRunTwo, averageInterStructureDistancesRunTwo);
		calculateAverageInterStructureDistance(interStructureDistancesCrossRuns, averageInterStructureDistancesCrossRuns);
		
		
		//  output the inter structure distances to file
		if(verbose){
			outputInterStructureDistancesToFile(interStructureDistancesRunOne, "interStructureDistancesRunOne" + output_filename_suffix + ".txt");
			outputInterStructureDistancesToFile(interStructureDistancesRunTwo, "interStructureDistancesRunTwo" + output_filename_suffix + ".txt");
			outputInterStructureDistancesToFile(interStructureDistancesCrossRuns, "interStructureDistancesCrossRuns" + output_filename_suffix + ".txt");
		}
		
		
		//  output the average inter structure distances to file
		outputAverageInterStructureDistancesToFile(averageInterStructureDistancesRunOne, averageInterStructureDistancesRunTwo, averageInterStructureDistancesCrossRuns, "averageInterStructureDistances" + output_filename_suffix + ".txt");
		
		
		//  output the time taken by BurnInAnalyzer
		long timeTaken = System.currentTimeMillis() - timeBefore;
		timeTaken = timeTaken / (long)(1000.0);			//	convert from milliseconds to seconds		
		if(timeTaken > 3600){		//	(1hr = 3600 sec)
			int numberOfHours = (int)timeTaken / (int)3600;
			long leftOver = timeTaken % 3600;
			int numberOfMinutes = (int)leftOver / (int)60;
			leftOver = leftOver % 60;
			int numberOfSeconds = (int)leftOver;
			if(verbose){
				System.out.println("Amount of time taken = " + numberOfHours + "hr" + numberOfMinutes + "min" + numberOfSeconds + "sec");
			}
		}
		else if(timeTaken > 60){		//	(1min = 60 sec)
			int numberOfMinutes = (int)timeTaken / (int)60;
			long leftOver = timeTaken % 60;
			int numberOfSeconds = (int)leftOver;
			if(verbose){
				System.out.println("Amount of time taken = " + numberOfMinutes + "min" + numberOfSeconds + "sec");
			}
		}
		else{
			int numberOfSeconds = (int)timeTaken;
			if(verbose){
				System.out.println("Amount of time taken = " + numberOfSeconds + "sec");
			}
		}
	}/*  end of run method  */
}/*  end of BurnInAnalyzer class  */