package utilities;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;


/** ------------------------------------------------------------------------------------------------------------------
 * start of CalculateAverageBAC class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class CalculateAverageBAC {

	
	
	
	/*  global parameters  */
	private static boolean verbose = true;
	private static String inputPath = "";
	private static String outputPath = "";
	
	private static int index_Min = Integer.MAX_VALUE;
	private static int index_Max = Integer.MIN_VALUE;
	private static ArrayList<String> allFragmentPairKeys;
	/*  end of global parameters  */
	
	
	
	
	/*  start of loadInputFiles method  */
	/**
	 * @param filenames_BAC the list of BAC filenames to load from file
	 * @return a list of HashMaps, each containing the data for one of the specified BAC experiment files
	 */
	private static ArrayList<HashMap<String, double[]>> loadInputFiles(ArrayList<String> filenames_BAC){
		//  make sure that there is more than one input file specified 
		if(filenames_BAC.size() < 2){
			System.err.println("Error: there are less than two input files specified. No need to calculate average.\nSystem.exit(0)");
			System.exit(0);
		}
		
		allFragmentPairKeys = new ArrayList<String>();
		ArrayList<HashMap<String, double[]>> inputFiles = new ArrayList<HashMap<String, double[]>>(filenames_BAC.size());
		
		
		for(int filename_index=0; filename_index<filenames_BAC.size(); filename_index++){
			
			HashMap<String, double[]> currentInputFileHashMap = new HashMap<String, double[]>();
			
			String line = null;
			int lineCounter = 0;
			try{
				BufferedReader br = new BufferedReader(new FileReader(inputPath + filenames_BAC.get(filename_index)));
				while((line = br.readLine()) != null){
					String[] values = line.split("\t");
					if(values.length == 10){
						//  parse the values once that will be used to save time
						int firstIndex = Integer.parseInt(values[0]);
						int secondIndex = Integer.parseInt(values[1]);

						//  check for a new minimum index
						if(firstIndex < index_Min){
							index_Min = firstIndex;
						}
						if(secondIndex < index_Min){
							index_Min = secondIndex;
						}

						//  check for a new maximum index
						if(firstIndex > index_Max){
							index_Max = firstIndex;
						}
						if(secondIndex > index_Max){
							index_Max = secondIndex;
						}
						
						double[] currentRowValues = new double[10];
						currentRowValues[0] = Double.parseDouble(values[0]);
						currentRowValues[1] = Double.parseDouble(values[1]);
						currentRowValues[2] = Double.parseDouble(values[2]);
						currentRowValues[3] = Double.parseDouble(values[3]);
						currentRowValues[4] = Double.parseDouble(values[4]);
						currentRowValues[5] = Double.parseDouble(values[5]);
						currentRowValues[6] = Double.parseDouble(values[6]);
						currentRowValues[7] = Double.parseDouble(values[7]);
						currentRowValues[8] = Double.parseDouble(values[8]);
						currentRowValues[9] = Double.parseDouble(values[9]);
						
						//  always have the indices in the order min-max
						if(firstIndex == secondIndex){
							System.err.println("Error: the fragment indices are the same.\nSystem.exit(0)");
							System.exit(0);
						}
						
						if(firstIndex > secondIndex){
							int tempIndex = firstIndex;
							firstIndex = secondIndex;
							secondIndex = tempIndex;
						}
						
						String currentRowKey = firstIndex + "-" + secondIndex;
						
						if(currentInputFileHashMap.containsKey(currentRowKey)){
							System.err.println("Error: there is already an entry for fragment pair " + currentRowKey + ".\nSystem.exit(0)");
							System.exit(0);
						}
						
						currentInputFileHashMap.put(currentRowKey, currentRowValues);
						
						if(!allFragmentPairKeys.contains(currentRowKey)){
							allFragmentPairKeys.add(currentRowKey);
						}
					}
					else{
						System.err.println("Error: there are not ten (10) entries in each row of the file \"" + filenames_BAC.get(filename_index) + "\"\nSystem.exit(0)");
						System.exit(0);
					}
					lineCounter++;
				}
				br.close();
			}catch (IOException e) {
				e.printStackTrace();
				System.err.println("Error: cannot find file with name \"" + filenames_BAC.get(filename_index) + "\"\nSystem.exit(0)");
				System.exit(0);
			}
			inputFiles.add(currentInputFileHashMap);
			
			if(verbose){
				System.out.println("\tOpened and read file: " + filenames_BAC.get(filename_index));			
			}
		}
		allFragmentPairKeys.trimToSize();
		
		if(verbose){
			System.out.println("min index = " + index_Min);
			System.out.println("max index = " + index_Max);		
			for(int i=0; i<filenames_BAC.size(); i++){
				System.out.println("number of entries in " + filenames_BAC.get(i) + " = " + inputFiles.get(i).size());
			}
		}
		
		return inputFiles;
	}/*  end of loadInputFiles method  */
	
	
	
	
	/*  start of calculateAverageValues method  */
	/**
	 * @param inputFiles a list of HashMaps, each containing the data for one of the BAC experiment files
	 * @param filenames_BAC the list of BAC filenames matching the entries in the list of HashMaps
	 * @return a HashMap containing the average value for each probe pair across all of the given input BAC experiment files
	 */
	private static HashMap<String, double[]> calculateAverageValues(ArrayList<HashMap<String, double[]>> inputFiles, ArrayList<String> filenames_BAC){
		if(inputFiles == null){
			System.err.println("Error: trying to calculate average BAC values with null pointer.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		//  create the HashMap that will hold all of the average values
		HashMap<String, double[]> averageValues = new HashMap<String, double[]>();
		
		//  go through all of the fragment pairs seen in at least one of the input files
		for(String currentKey: allFragmentPairKeys){
			
			double[] averageValuesForCurrentKey = new double[10];
			
			double[][] valuesForCurrentKey = new double[filenames_BAC.size()][10];

			for(int inputFileIndex=0; inputFileIndex<filenames_BAC.size(); inputFileIndex++){
				if(inputFiles.get(inputFileIndex).containsKey(currentKey)){
					valuesForCurrentKey[inputFileIndex] = inputFiles.get(inputFileIndex).get(currentKey);
				}
			}
			int[] counts = new int[10];


			for(int i=0; i<filenames_BAC.size(); i++){
				for(int j=0; j<10; j++){
					if(valuesForCurrentKey[i][j] > 0){
						averageValuesForCurrentKey[j] += valuesForCurrentKey[i][j];
						counts[j]++;
					}
				}
			}

			for(int i=0; i<10; i++){
				if(counts[i] > 1){
					averageValuesForCurrentKey[i] /= (double)counts[i];
				}
			}

			
			averageValues.put(currentKey, averageValuesForCurrentKey);
		}
		
		return averageValues;
	}/*  end of calculateAverageValues method  */
	
	
	
	
	/*  start of outputAverageValuesToFile method  */
	/**
	 * @param averageValues the HashMap containing the average value for each probe pair across all of the given input BAC experiment files
	 * @param filenames_BAC the list of BAC filenames
	 * @return void
	 */
	private static void outputAverageValuesToFile(HashMap<String, double[]> averageValues, ArrayList<String> filenames_BAC){
		String outputFilename = "average";
		for(int i=0; i<filenames_BAC.size(); i++){
			if(i<filenames_BAC.size()-1){
				String prefix = filenames_BAC.get(i).substring(0, filenames_BAC.get(i).lastIndexOf("."));
				outputFilename += "_" + prefix;
			}
			else{
				outputFilename += "_" + filenames_BAC.get(i);
			}
		}
		try{
			BufferedWriter bw = new BufferedWriter(new FileWriter(outputPath + outputFilename));
			
			int counter = 0;
			for(int i=index_Min; i<=index_Max; i++){
				for(int j=index_Min; j<=index_Max; j++){
					String currentKey = i + "-" + j;
					if(averageValues.containsKey(currentKey)){

						double[] currentEntryValues = averageValues.get(currentKey);
						
						for(int k=0; k<10; k++){
							if(k<2){
								bw.write((int)currentEntryValues[k] + "\t");
							}
							else if(k<9){
								bw.write(currentEntryValues[k] + "\t");
							}
							else{
								bw.write(String.valueOf(currentEntryValues[k]));
							}
						}
						bw.newLine();
						bw.flush();
//						System.out.println(currentKey);
						counter++;
					}					
				}
			}
//			System.out.println(counter);
			bw.flush();
			bw.close();
			
			if(verbose){
				System.out.println("\tOutput file: " + outputPath + outputFilename);			
			}
		}catch (IOException e) {
			e.printStackTrace();
		}//end of try-catch block
	}/*  end of outputAverageValuesToFile method  */
	
	
	
	
	/*  start of getFilenames method  */
	/**
	 * @param filename_file_list the file containing the list of files
	 * @return a list of the BAC filenames
	 */
	private static ArrayList<String> getFilenames(String filename_file_list){
		if(filename_file_list == null){
			System.err.println("Error: the filename list is null.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<String> filenames_BAC = new ArrayList<String>();

		String line = null;
		int lineCounter = -1;			//  there is one header line
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename_file_list));
			while((line = br.readLine()) != null){
				if(lineCounter >= 0){
					filenames_BAC.add(line);
				}
				else{
					//  skip and the header line
//					System.out.println(line);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		filenames_BAC.trimToSize();
		
		if(verbose){
			System.out.println("Loaded input file: " + inputPath + filename_file_list);
		}
		
		return filenames_BAC;
	}/*  end of getFilenames method  */
	
	

	
	/*  start of displayFilenames method  */
	/**
	 * @param filenames_BAC the list of BAC filenames to display
	 * @return void
	 */
	private static void displayFilenames(ArrayList<String> filenames_BAC){
		if(filenames_BAC == null){
			System.err.println("Error: the file list is null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);	
		}
		
		System.out.println("Filename");
		for(String currentFilename: filenames_BAC){
			System.out.println(currentFilename);
		}
	}/*  end of displayFilenames method  */
	
	
	
	
	/*  start of main method  */
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		
		
		
		boolean usingCommandLine = false;
		
		String filename_file_list = "default.txt";
		ArrayList<String> filenames_BAC = new ArrayList<String>();
		
		
		//  load the parameters from the command line
		if(args.length >= 2){
			//  all subsequent parameters are the filenames of the experiment files
			for(int i=0; i<args.length; i++){
				filenames_BAC.add(args[i]);
			}
			filenames_BAC.trimToSize();
			
			usingCommandLine = true;
			verbose = false;
		}
		else{
			System.err.println("Error: there need to be at least two (2) command line parameters.\n" +
					"\t\tBAC_1_filename.txt\n" +
					"\t\tBAC_2_filename.txt\n" +
					"\t\tBAC_..._filename.txt\n" +
					"System.exit(0)");
//			System.exit(0);
		}
		
		
		
		/******************************************************
		 * 													  *
		 * 				SET THE PARAMETERS HERE				  *
		 * 													  *
		 ******************************************************/
		
		
		if(!usingCommandLine){
			
			
			inputPath = "CalculateAverageBAC/Input/";
			outputPath = "CalculateAverageBAC/Output/";
		
			
			filename_file_list = "file_list.txt";
		
			
//			verbose = false;
			
			
		}
		
		
		/******************************************************
		 * 													  *
		 * 				END OF SETTING THE PARAMETERS		  *
		 * 													  * 
		 ******************************************************/
		
		
		
		
		if(!usingCommandLine){
			filenames_BAC = getFilenames(filename_file_list);
		}
//		displayFilenames(filenames_BAC);

		
		ArrayList<HashMap<String, double[]>> inputFiles = loadInputFiles(filenames_BAC);
		
		
		HashMap<String, double[]> averageValues = calculateAverageValues(inputFiles, filenames_BAC);
		
		
		outputAverageValuesToFile(averageValues, filenames_BAC);
		
		
		if(verbose){
			System.out.println("Done!");
		}
	}/*  end of main method  */

}/*  end of CalculateAverageBAC class  */
