package modeling;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Random;
import javax.vecmath.Point3d;
import org.openscience.cdk.Atom;
import org.openscience.cdk.AtomContainer;

import com.ibm.icu.text.DecimalFormat;


/** ------------------------------------------------------------------------------------------------------------------
 * start of Chromosome class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class Chromosome implements Comparable<Chromosome>{
	
	
	
	
	/*  static global parameters  */
	private static int id_counter = 0;
	private static Random random = new Random();
	/*  end of static global parameters  */
	
	
	
	
	/*  global parameters  */
	private int id;
	private Atom[] atoms;
	private double fitness = 0;
	private int[] genomicCoordinates;
	/*  end of global parameters  */
	
	
	

	/*  start of constructor method  */
	/** 
	 * @param numberOfPoints the number of points to initialize the chromosome with
	 */
	public Chromosome(int size){
		this.id = id_counter++;
		
		initializeChromosome(size);
	}/*  end of constructor method  */

	
	

	/*  start of constructor method  */
	/** 
	 * @param atoms the atoms of the chromosome
	 */
	public Chromosome(Atom[] atoms){
		this.id = id_counter++;
		
		setAllAtoms(atoms);
	}/*  end of constructor method  */
	
	
	
	
	/*  start of constructor method  */
	/**
	 * @param the coordinates of the Atoms in the Chromosome
	 */
	public Chromosome(double[][] coordinates){
		if(coordinates == null){
			System.err.println("Error: the passed reference is null.\nSystem.exit(0)");
			System.exit(0);
		}
		if(coordinates.length < 1){
			System.err.println("Error: there is less than one point in the chromosome.\nSystem.exit(0)");
			System.exit(0);
		}
		
		this.id = id_counter++;
		
		this.atoms = new Atom[coordinates.length];
		for(int i=0; i<coordinates.length; i++){
			this.atoms[i] = new Atom("C", new Point3d(coordinates[i][0], coordinates[i][1], coordinates[i][2]));
		}
	}/*  end of constructor method  */
	
	
	
	
	/*  start of constructor method  */
	/**
	 * @param chromosome the chromosome to deep copy to create a new chromosome from
	 * the Atom objects and the Point3d objects are both deep copied
	 * the returned chromosome has a new unique ID
	 */
	public Chromosome(Chromosome chromosome){
		if(chromosome == null){
			System.err.println("Error: trying to create a new chromosome from a null pointer.\nSystem.exit(0)");
			System.exit(0);
		}
		if(chromosome.getSize() < 1){
			System.err.println("Error: trying to create a new chromosome from an empty chromosome.\nSystem.exit(0)");
			System.exit(0);
		}
		
		this.id = id_counter++;
		
		this.atoms = new Atom[chromosome.getSize()];
		for(int i=0; i<this.atoms.length; i++){
			this.atoms[i] = chromosome.getAtomDeepCopy(i);
		}
		
		this.fitness = chromosome.getFitness();
	}/*  end of constructor method  */
	
	
	
	
	/*  start of setFragmentGenomicCoordinates method  */
	/**
	 * @param filename_genomicCoordinates the filename of the file containing the genomic coordinates for each atom in the chromosome
	 */
	public void setFragmentGenomicCoordinates(String filename_genomicCoordinates){
		if(filename_genomicCoordinates == null){
			System.err.println("Error: the filename is null and the genomic coordinates cannot be loaded.\nSystem.exit(0)");
			System.exit(0);
		}
		
		this.genomicCoordinates = new int[this.getSize()];
		
		String line = null;
		int lineCounter = 0;
		try{
			BufferedReader br = new BufferedReader(new FileReader(filename_genomicCoordinates));
			while((line = br.readLine()) != null){
				int genomicLocation = Integer.parseInt(line);
				if(lineCounter >= genomicCoordinates.length){
					System.err.println("Error: there are more entries in the genomic coordinates file than in the structure.\nSystem.exit(0)");
					System.exit(0);
				}
				genomicCoordinates[lineCounter] = genomicLocation;
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		if(lineCounter != genomicCoordinates.length){
			System.err.println("Error: there are not the same number of entries in the structure and the genomic coordinates file.\nSystem.exit(0)");
			System.exit(0);
		}
//		System.out.println("Loaded input file: " + filename_genomicCoordinates);
	}/*  end of setFragmentGenomicCoordinates method  */
	
	
	
	
	/*  start of getPoint3dCoordinatesOfGenomicLocation method  */
	/**
	 * @param genomicLocation the genomic index of the location to determine the x,y,z coordinates for
	 * @return a Point3d object containing the x,y,z coordinates of the specified genomic location
	 */
	public Point3d getPoint3dCoordinatesOfGenomicLocation(int genomicLocation){
		return this.getCoordinatesAndFragmentOfGenomicLocation(genomicLocation)[0];
	}/*  end of getPoint3dCoordinatesOfGenomicLocation method  */
	
	
	
	
	/*  start of getPoint3dCoordinatesOfGenomicLocation method  */
	/**
	 * @param genomicLocation the genomic index of the location to determine the x,y,z coordinates for
	 * @return an array of Point3d objects containing the x,y,z coordinates of the specified genomic location, the coordinates of the start point of the containing fragment, and the end point of the containing fragment
	 */
	public Point3d[] getCoordinatesAndFragmentOfGenomicLocation(int genomicLocation){
		//  make sure that the genomic coordinate mappings match the current chromosome size
		if((genomicCoordinates == null) || (this.getSize() != genomicCoordinates.length)){
			System.err.println("Error: the genomic coordinates were not properly loaded for the current chromosome and the specified genomic location cannot be determined.\nSystem.exit(0)");
			System.exit(0);
		}
		
		Point3d coordinates = new Point3d();
		
		//  make sure that the genomic position is within the range of the structure
		if((genomicLocation < this.genomicCoordinates[0]) || (genomicLocation > this.genomicCoordinates[this.getSize()-1])){
			System.err.println("Error: the genomic position " + genomicLocation + " is not contained within the chromosome.");
			System.err.println("System.exit(0)");
			System.exit(0);
		}
		else{
//			System.out.println("The genomic position " + genomicLocation + " is contained within the chromosome.");
		}
		

		//  find the endpoints of the fragment that contains the genomic position
		if(this.getSize() < 2){
			System.err.println("Error: there is less than one fragment in the structure.\nSystem.exit(0)");
			System.exit(0);
		}
		
		int fragmentStartIndex = 0;
		for(int i=0; i<this.getSize()-1; i++){
			if((genomicLocation >= this.genomicCoordinates[i]) && (genomicLocation <= this.genomicCoordinates[i+1])){
				fragmentStartIndex = i;
				break;
			}			
		}
		int fragmentEndIndex = fragmentStartIndex + 1;
		//  at this point the genomic position MUST be contained within the region defined by [startPosition,endPosition]
		int fragmentStartGenomicLocation = this.genomicCoordinates[fragmentStartIndex];
		int fragmentEndGenomicLocation = this.genomicCoordinates[fragmentEndIndex];
//		System.out.println("Genomic location " + genomicLocation + " should lie in the fragment that starts at location " + fragmentStartGenomicLocation + " and ends at location " + fragmentEndGenomicLocation);
		
		int sizeOfFragment = fragmentEndGenomicLocation - fragmentStartGenomicLocation;
//		System.out.println("size of fragment = " + sizeOfFragment);
		int distanceIntoFragment = genomicLocation - fragmentStartGenomicLocation;
//		System.out.println("distance into fragment = " + distanceIntoFragment);
		double fractionIntoFragment = distanceIntoFragment / (double)sizeOfFragment;
//		System.out.println("fraction into fragment = " + fractionIntoFragment);
		
		Point3d startPoint3d = this.getAtomCoordinates(fragmentStartIndex);
		Point3d endPoint3d = this.getAtomCoordinates(fragmentEndIndex);
		
		double xVector = endPoint3d.x - startPoint3d.x;
		double yVector = endPoint3d.y - startPoint3d.y;
		double zVector = endPoint3d.z - startPoint3d.z;
		
		coordinates.x = startPoint3d.x + fractionIntoFragment * xVector;
		coordinates.y = startPoint3d.y + fractionIntoFragment * yVector;
		coordinates.z = startPoint3d.z + fractionIntoFragment * zVector;
		
		Point3d[] returnPoints = new Point3d[3];
		returnPoints[0] = coordinates;
		returnPoints[1] = startPoint3d;
		returnPoints[2] = endPoint3d;
		return returnPoints;
	}/*  end of getPoint3dCoordinatesOfGenomicLocation method  */
	
	
	
	
	/*  start of initializeChromosome method  */
	/**
	 * @param size the number of points to initialize the chromosome with
	 */
	private void initializeChromosome(int size){
		//  initialize the local array that will store the chromosome coordinates
		atoms = new Atom[size];
		
		//  initialize the chromosome in a cube with side length [0,1[
		for(int i=0; i<size; i++){
			atoms[i] = new Atom("C", new Point3d(random.nextDouble(), random.nextDouble(), random.nextDouble()));
//			atoms[i] = new Atom("C", new Point3d(1+random.nextDouble(), 1+random.nextDouble(), 1+random.nextDouble()));		//  use the 1+random.nextDouble() to avoid NaN problems with very small double values
		}
	}/*  end of initializeChromosome method  */
	
	
	
	
	/*  start of setAllAtoms method  */
	/** 
	 * @param atoms the atoms of the chromosome
	 */
	public void setAllAtoms(Atom[] atoms){
		//  make sure that the passed chromosome properly exists
		if(atoms == null){
			System.err.println("Error: trying to set chromosome with a null pointer.\nSystem.exit(0)");
			System.exit(0);
		}
		if(atoms.length < 1){
			System.err.println("Error: trying to set chromosome with an empty array.\nSystem.exit(0)");
			System.exit(0);
		}
		if(atoms[0] == null){
			System.err.println("Error: trying to set chromosome with uninitialized Atoms.\nSystem.exit(0)");
			System.exit(0);
		}
		
		//  set the local array that will store the chromosome atoms
		this.atoms = atoms;
	}/*  end of setAllAtoms method  */
	
	
	
	
	/*  start of getAtom method  */
	/**
	 * @return atom the atom of the chromosome at the specified index
	 * the range is from 0 (inclusively) to chromosome_Size (exclusively)
	 */
	public Atom getAtom(int index){
		if(atoms == null){
			System.err.println("Error: the chromosome is null.\nSystem.exit(0)");
			System.exit(0);
		}
		if((index < 0) || (index >= this.getSize())){
			System.err.println("Error: the index is out of range.\nSystem.exit(0)");
			System.exit(0);
		}
		return atoms[index];
	}/*  end of getAtom method  */
	
	
	
	
	/*  start of getAtomDeepCopy method  */
	/**
	 * @return atom a new Atom object containing a deep copy of the atom at the specified index
	 * the Point3d object of the Atom object is also a deep copy
	 * the range is from 0 (inclusively) to chromosome_Size (exclusively)
	 */
	public Atom getAtomDeepCopy(int index){
		if(this.atoms == null){
			System.err.println("Error: the chromosome is null.\nSystem.exit(0)");
			System.exit(0);
		}
		if((index < 0) || (index >= this.getSize())){
			System.err.println("Error: the index is out of range.\nSystem.exit(0)");
			System.exit(0);
		}
		return new Atom("C", this.getAtomCoordinatesDeepCopy(index));
	}/*  end of getAtomDeepCopy method  */
	
	
	
	
	/*  start of getAtoms method  */
	/**
	 * @return atoms the atoms of the chromosome
	 */
	public Atom[] getAtoms(){
		if(atoms == null){
			System.err.println("Error: the chromosome is null.\nSystem.exit(0)");
			System.exit(0);
		}
		return atoms;
	}/*  end of getAtoms method  */
	
	
	
	
	/*  start of getAtoms method  */
	/**
	 * @return atoms the subset of atoms of the chromosome as specified by the range [minIndex, maxIndex[
	 * the atoms are numbered starting from zero (0), inclusively, to chromosomeSize, inclusively (the maxIndex is not returned as part of the subset)
	 */
	public Atom[] getAtoms(int minIndex, int maxIndex){
		//  make sure that the chromosome is defined
		if(atoms == null){
			System.err.println("Error: the chromosome is null.\nSystem.exit(0)");
			System.exit(0);
		}
		
		//  make sure that the indices are within bound
		if((minIndex < 0) || (minIndex >= maxIndex)){
			System.err.println("Error: the minIndex is out of range.\nSystem.exit(0)");
			System.exit(0);
		}
		if((maxIndex > this.getSize())){
			System.err.println("Error: the maxIndex is out of range.\nSystem.exit(0)");
			System.exit(0);
		}
		
		//  create the array that will store the subset of atoms
		Atom[] atomSubset = new Atom[maxIndex - minIndex];
		for(int i=minIndex; i<maxIndex; i++){
			atomSubset[i-minIndex] = atoms[i];
		}
		return atomSubset;
	}/*  end of getAtoms method  */
	
	
	
	
	/*  start of getAsAtomContainer method  */
	/**
	 * @return Chromosome the Chromosome as represented by an AtomContainer
	 */
	public AtomContainer getAsAtomContainer(){
		AtomContainer container = new AtomContainer();
		for(int i=0; i<atoms.length; i++){
			container.addAtom(atoms[i]);
		}
		return container;
	}/*  end of getAsAtomContainer method  */
	
	
	
	
	/*  start of getAsDoubleArray method  */
	/**
	 * @return doubleArray the Chromosome as represented by an array containing the coordinates of the Atoms
	 */
	public double[][] getAsDoubleArray(){
		double[][] structure = new double[atoms.length][3];
		Point3d currentPoint3d;
		for(int i=0; i<atoms.length; i++){
			currentPoint3d = atoms[i].getPoint3d();
			structure[i][0] = currentPoint3d.x;
			structure[i][1] = currentPoint3d.y;
			structure[i][2] = currentPoint3d.z;
		}
		return structure;
	}/*  end of getAsDoubleArray method  */
	
	
	
	
	/*  start of getSize method  */
	/**
	 * @return size the number of atoms in the chromosome
	 */
	public int getSize(){
		if(atoms == null){
			System.err.println("Error: the chromosome is null.\nSystem.exit(0)");
			System.exit(0);
		}
		return this.atoms.length;
	}/*  end of getSize method  */




	/*  start of getID method  */
	/**
	 * @return the id of the chromosome
	 */
	public int getID() {
		return this.id;
	}/*  end of getID method  */




	/*  start of getFitness method  */
	/**
	 * @return fitness the fitness of the chromosome formatted to two decimal places
	 */
	public double getFitness() {
		return this.fitness;
	}/*  end of getFitness method  */
	
	
	
	
	/*  start of setFitness method  */
	/**
	 * @param fitness the fitness to set
	 */
	public void setFitness(double fitness) {
		//  makes sure that the fitness is a number
		if(fitness != fitness){
			System.err.println("Error: the fitness (" + fitness + ") is Not-a-Number (NaN).\nSystem.exit(0)");
			System.exit(0);
		}
		this.fitness = fitness;
	}/*  end of setFitness method  */
	
	
	
	
	/*  start of setAtom method  */
	/**
	 * @param index the index of the atom to set (range is [0,size-1])
	 * @param atom the atom used to replace the current atom at the specified index
	 */
	public void setAtom(int index, Atom atom){
		if((index < 0) || (index >= atoms.length)){
			System.err.println("Error: the index (" + index + ") is out of range.\nSystem.exit(0)");
			System.exit(0);
		}
		this.atoms[index] = atom;
	}/*  end of setAtom method  */
	
	
	
	
	/*  start of setAtomCoordinates method  */
	/**
	 * @param index the index of the atom to set the coordinates of (range is [0,size-1])
	 * @param Point3d the 3d point used to replace the Point3d of the atom at the specified index
	 */
	public void setAtomCoordinates(int index, Point3d coordinates){
		if((index < 0) || (index >= atoms.length)){
			System.err.println("Error: the index (" + index + ") is out of range.\nSystem.exit(0)");
			System.exit(0);
		}
		this.atoms[index].setPoint3d(coordinates);		
	}/*  end of setAtomCoordinates method  */
	
	
	
	
	/*  start of getCoordinates method  */
	/**
	 * @return an array of references to the Point3d objects containing the coordinates of the Atoms of the chromosome
	 */
	public Point3d[] getCoordinates(){
		Point3d[] coordinates = new Point3d[atoms.length];
		for(int i=0; i<atoms.length; i++){
			coordinates[i] = atoms[i].getPoint3d();
		}
		return coordinates;
	}/*  end of getCoordinates method  */
	
	
	
	
	/*  start of getAtomCoordinates method  */
	/**
	 * @param index the index of the atom to get the coordinates of (range is 0 (inclusively) to chromosomeSize (exclusively))
	 * @return Point3d a reference to the Point3d object containing the coordinates of the Atom at the specified index
	 */
	public Point3d getAtomCoordinates(int index){
		if((index < 0) || (index >= this.getSize())){
			System.err.println("Error: the index is out of range.\nSystem.exit(0)");
			System.exit(0);
		}
		return atoms[index].getPoint3d();
	}/*  end of getAtomCoordinates method  */
	
	
	
	
	/*  start of getAtomCoordinatesDeepCopy method  */
	/**
	 * @param index the index of the atom to get the coordinates of (range is 0 (inclusively) to chromosomeSize (exclusively))
	 * @return Point3d a new Point3d object containing a deep copy of the coordinates of the Atom at the specified index
	 */
	public Point3d getAtomCoordinatesDeepCopy(int index){
		if((index < 0) || (index >= this.getSize())){
			System.err.println("Error: the index is out of range.\nSystem.exit(0)");
			System.exit(0);
		}
		return new Point3d(this.atoms[index].getPoint3d().x, this.atoms[index].getPoint3d().y, this.atoms[index].getPoint3d().z);
	}/*  end of getAtomCoordinatesDeepCopy method  */
	
	
	
	
	/*  start of testForNaNCoordinates method  */
	/**
	 * @void
	 * prints an error and exits if there exists a coordinate that has a value that is Not-a-Number (NaN)
	 */
	public void testForNaNCoordinates(){
		Point3d coordinatesOfCurrentAtom;
		for(int i=0; i<atoms.length; i++){
			coordinatesOfCurrentAtom = atoms[i].getPoint3d();
			if((coordinatesOfCurrentAtom.x != coordinatesOfCurrentAtom.x) || (coordinatesOfCurrentAtom.y != coordinatesOfCurrentAtom.y) || (coordinatesOfCurrentAtom.z != coordinatesOfCurrentAtom.z)){
				System.err.println("Error: there is a coordinate that is Not-a-Number (NaN) in chromosome " + this.getID() + ".\nSystem.exit(0)");
				System.exit(0);
			}
		}
		System.out.println("\t" + this.getID() + ":\tpassed NaN test");
	}/*  end of testForNaNCoordinates method  */
	
	
	
	
	/*  start of getGenomicCoordinates method  */
	/**
	 * @return an array containing the genomic coordinates for each fragment of the chromosome
	 */
	public int[] getGenomicCoordinates(){
		if(genomicCoordinates == null){
			System.err.println("Error: the genomic coordinates are null and cannot be returned.\nSystem.exit(0)");
		}
		return genomicCoordinates;
	}/*  end of getGenomicCoordinates method  */
	
	
	
	
	/*  start of compareTo method  */
	public int compareTo(Chromosome chromosome) {
		
		//  the current chromosome has a higher fitness
		if(this.fitness > chromosome.getFitness()){
			return 1;
		}
		//  the current chromosome has a lower fitness
		else if(this.fitness < chromosome.getFitness()){
			return -1;
		}
		//  the two chromosomes being compared have the same fitness
		else{
			return 0;		
		}
	}/*  end of compareTo method  */
	
	
	
	
	/*  start of toString method  */
	@Override
	public String toString(){
		String chromosomeAsString = "\t";
		chromosomeAsString += "ID=" + this.getID() + "\tFITNESS=" + this.getFitness() + "\n";
		Point3d coordinatesOfCurrentAtom;
		for(int i=0; i<atoms.length; i++){
			coordinatesOfCurrentAtom = atoms[i].getPoint3d();
			chromosomeAsString += (i+1) + "\t" + coordinatesOfCurrentAtom.x + "\t" + coordinatesOfCurrentAtom.y + "\t" + coordinatesOfCurrentAtom.z + "\n";
//			chromosomeAsString += (i+1) + "\t" + coordinatesOfCurrentAtom.x + "\t" + coordinatesOfCurrentAtom.y + "\t" + coordinatesOfCurrentAtom.z + "\t" + this.genomicCoordinates[i] + "\n";
		}
		return chromosomeAsString;
	}/*  end of toString method  */
	
	
	
	
}/*  end of Chromosome class  */
