package analysis;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;

import utilities.LOOCVresults;
import weka.classifiers.Evaluation;
import weka.classifiers.functions.SMO;
import weka.classifiers.trees.J48;
import weka.core.Instances;
import weka.core.converters.ConverterUtils.DataSource;


/** ------------------------------------------------------------------------------------------------------------------
 * start of ClassifierLOOCV class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class ClassifierLOOCV {
	
	
	
	
	/*  global parameters  */
	private boolean verbose = false;
	
	private String inputPath = "";
	private String outputPath = "";
	/*  end of global parameters  */
	
	
	
	
	/*  start of isVerbose method  */
	/**
	 * @return the verbose
	 */
	public boolean isVerbose() {
		return verbose;
	}
	/*  end of isVerbose method  */
	
	
	
	
	/*  start of setVerbose method  */
	/**
	 * @param verbose the verbose to set
	 */
	public void setVerbose(boolean verbose) {
		this.verbose = verbose;
	}/*  end of setVerbose method  */
	
	
	
	
	/*  start of getInputPath method  */
	/**
	 * @return the inputPath
	 */
	public String getInputPath() {
		return inputPath;
	}/*  end of getInputPath method  */
	
	
	
	
	/*  start of setInputPath method  */
	/**
	 * @param inputPath the inputPath to set
	 */
	public void setInputPath(String inputPath) {
		this.inputPath = inputPath;
	}/*  end of setInputPath method  */
	
	
	
	
	/*  start of getOutputPath method  */
	/**
	 * @return the outputPath
	 */
	public String getOutputPath() {
		return outputPath;
	}/*  end of getOutputPath method  */

	
	
	
	/*  start of setOutputPath method  */
	/**
	 * @param outputPath the outputPath to set
	 */
	public void setOutputPath(String outputPath) {
		this.outputPath = outputPath;
	}/*  end of setOutputPath method  */
	
	
	
	
	/*  start of loadExperimentNamesAndLOOCVindices method  */
	/**
	 * @param filename_file_list the name of the file containing the list of matched cellular--gene desert experiments
	 * @return a list of length two (2) containing the list of experiment names (the first list (ArrayList of String)), and the LOOCV indices (the second list (ArrayList of int[]))
	 */
	private ArrayList loadExperimentNamesAndLOOCVindices(String filename_experiment_names_and_LOOCV_indices){
		if(filename_experiment_names_and_LOOCV_indices == null){
			System.err.println("Error: the filename is null and the experiment names and LOOCV indices cannot be loaded.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		ArrayList<String> experimentNames = new ArrayList<String>();
		ArrayList<int[]> LOOCVindices = new ArrayList<int[]>();
		
		
		String line = null;
		int lineCounter = -1;			//  there is one header line
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename_experiment_names_and_LOOCV_indices));
			while((line = br.readLine()) != null){
				if(lineCounter >= 0){
					String[] values = line.split("\t");
					if(values.length != 3){
						System.err.println("Error: there are not exactly three entries in each row of the " + filename_experiment_names_and_LOOCV_indices + " file.\nSystem.exit(0)");
						System.exit(0);
					}
					experimentNames.add(values[0]);
					
					int[] currentLOOCVindices = new int[2];
					currentLOOCVindices[0] = Integer.parseInt(values[1]);
					currentLOOCVindices[1] = Integer.parseInt(values[2]);
					LOOCVindices.add(currentLOOCVindices);
				}
				else{
					//  skip the header line
//					System.out.println(line);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		
		
		experimentNames.trimToSize();
		LOOCVindices.trimToSize();
		
		
		if(experimentNames.size() != LOOCVindices.size()){
			System.err.println("Error: there are not a matching number of experiment names and LOOCV indices.\nSystem.exit(0)");
			System.exit(0);
		}
		
		//  build the return array
		ArrayList return_tuple_list = new ArrayList();
		return_tuple_list.add(experimentNames);
		return_tuple_list.add(LOOCVindices);
		return_tuple_list.trimToSize();
		
		return return_tuple_list;
	}/*  end of loadExperimentNamesAndLOOCVindices method  */
	
	
	
	
	/*  start of displayExperimentNamesAndIndices method  */
	/**
	 * @param cellularFilenames the list of cellular filenames
	 * @param geneDesertFilenames the list of gene desert filenames
	 * @param classLabels the list of class labels that corresponding to each of the matched cellular-gene desert filename pairs
	 * @param relation the relation label
	 */
	private void displayExperimentNamesAndIndices(ArrayList<String> experimentNames, ArrayList<int[]> LOOCVindices){
		if((experimentNames == null) || (LOOCVindices == null)){
			System.err.println("Error: both the experiment names and LOOCV indices must be non-null to display.\nSystem.exit(0)");
			System.exit(0);
		}
		
		System.out.println("\n==========  Experiment names\t--\tLOOCV indices  ==========");
		for(int i=0; i<experimentNames.size(); i++){
			System.out.println("\t\t" + experimentNames.get(i) + "\t\t--\t\t" + LOOCVindices.get(i)[0] + "-" + LOOCVindices.get(i)[1]);
		}
		for(int i=0; i<(24+41); i++){
			System.out.print("=");
		}
		System.out.println();
	}/*  end of displayExperimentNamesAndIndices method  */
	
	
	
	
	/*  Start of LOOCV method  */
	/**
	 * @param complete_dataset
	 * @param experimentNames
	 * @param LOOCVindices
	 * @return and LOOCVresults object containing the results of the LOOCV run
	 */
	private LOOCVresults LOOCV(Instances complete_dataset, ArrayList<String> experimentNames, ArrayList<int[]> LOOCVindices) throws Exception {
		if((complete_dataset == null) || (experimentNames == null) || (LOOCVindices == null)){
			System.err.println("Error: all the parameters need to be non-null to perform the LOOCV.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		//  create the object that will store the LOOCV results
		LOOCVresults results_of_LOOCV = new LOOCVresults();
		//  specify the relation name
		results_of_LOOCV.setRelation(complete_dataset.relationName());
		//  specify the properties of the complete dataset
		results_of_LOOCV.setCompleteDatasetProperties(complete_dataset.numInstances(), complete_dataset.numAttributes(), complete_dataset.numClasses());
		
		
		
		double[][] confusionMatrix = new double[complete_dataset.numClasses()][complete_dataset.numClasses()];
		
		
		//  leave-out each experiment, one-at-a-time
		for(int i=0; i<experimentNames.size(); i++){
			
			String currentExperimentLeftOutName = experimentNames.get(i);
			
			int[] currentExperimentLeftOutInstanceIndexRange = LOOCVindices.get(i);
			int currentSizeOfTestSet = currentExperimentLeftOutInstanceIndexRange[1] - currentExperimentLeftOutInstanceIndexRange[0] + 1;
			if(currentSizeOfTestSet < 1){
				System.err.println("Error: the current test set has less than one instance.\nSystem.exit(0)");
				System.exit(0);
			}
			Instances currentTestDataset = new Instances(complete_dataset, currentSizeOfTestSet);
			
			Instances currentTrainingDataset = new Instances(complete_dataset, complete_dataset.numInstances()-currentSizeOfTestSet);
			
			//  assemble the current training and test datasets as subsets of the complete dataset
			for(int j=0; j<complete_dataset.numInstances(); j++){
				//  check if the current instance index is contained within the range specified for the current test set
				if((j >= (currentExperimentLeftOutInstanceIndexRange[0]-1)) && (j <= (currentExperimentLeftOutInstanceIndexRange[1]-1))){
					//  add the instance to the test set
					currentTestDataset.add(complete_dataset.get(j));
				}
				//  otherwise, add the instance to the training set
				else{
					currentTrainingDataset.add(complete_dataset.get(j));
				}
			}
			
			

			//  create the SVM object
			SMO SVMclassifier = new SMO();
			//  set the parameters for the SVM (note that "-N 2" turns off the default normalization of the input data)
			SVMclassifier.setOptions(weka.core.Utils.splitOptions("-C 1.0 -L 0.0010 -P 1.0E-12 -N 2 -V -1 -W 1 -K \"weka.classifiers.functions.supportVector.PolyKernel -C 250007 -E 1.0\""));
			//  train the SVM on the input data
			SVMclassifier.buildClassifier(currentTrainingDataset);
			
			Evaluation evaluationOfClassifier = new Evaluation(currentTrainingDataset);
			evaluationOfClassifier.evaluateModel(SVMclassifier, currentTestDataset);
			
			
			
			
			//  use a decision tree instead
//			J48 j48 = new J48();
//			j48.buildClassifier(currentTrainingDataset);
//			
//			Evaluation evaluationOfClassifier = new Evaluation(currentTrainingDataset);
//			evaluationOfClassifier.evaluateModel(j48, currentTestDataset);
			
			
			
			
			results_of_LOOCV.addResult(currentExperimentLeftOutName, evaluationOfClassifier.toMatrixString("Confusion matrix for " + currentExperimentLeftOutName));
			
			//  if the test dataset only has one instance, add it to the appropriate result list (correctly classified or incorrectly classified)
//			if(currentTestDataset.numInstances() == 1){
//				double[][] currentConfusionMatrix = evaluationOfClassifier.confusionMatrix();
//				int numCorrectlyPredicted = 0;
//				int numIncorrectlyPredicted = 0;
//				for(int i=0; i<currentConfusionMatrix.length; i++){
//					for(int j=0; j<currentConfusionMatrix.length; j++){
//						if(i == j){
//							numCorrectlyPredicted += currentConfusionMatrix[i][j];
//						}
//						else{
//							numIncorrectlyPredicted += currentConfusionMatrix[i][j];
//						}
//					}
//				}
				if((evaluationOfClassifier.correct() + evaluationOfClassifier.incorrect()) == 1){
					if(evaluationOfClassifier.correct() == 1){
						results_of_LOOCV.addTestDatasetToCorrectlyPredictedList(currentExperimentLeftOutName);
					}
					else if(evaluationOfClassifier.incorrect() == 1){
						results_of_LOOCV.addTestDatasetToIncorrectlyPredictedList(currentExperimentLeftOutName);
					}
				}
//			}
			
			
			if(this.isVerbose()){
				System.out.println(evaluationOfClassifier.toSummaryString("\n\n========================================================================\nResults for left-out: \"" + currentExperimentLeftOutName + "\"\n", false));
				System.out.println(evaluationOfClassifier.toMatrixString("Confusion matrix"));
				System.out.println("Relation name = " + currentTrainingDataset.relationName() + "  " + currentTestDataset.relationName());
				System.out.println("Number of Attributes = " + currentTrainingDataset.numAttributes() + "  " + currentTestDataset.numAttributes());
				System.out.println("Number of Instances = " + currentTrainingDataset.numInstances() + "  " + currentTestDataset.numInstances());
				System.out.println("========================================================================");
			}
			
			double[][] currentConfusionMatrix = evaluationOfClassifier.confusionMatrix();
			
			for(int m=0; m<confusionMatrix.length; m++){
				for(int n=0; n<confusionMatrix.length; n++){
					confusionMatrix[m][n] += currentConfusionMatrix[m][n];
				}
			}
		}
		
		
		//  display the final confusion matrix
		if(this.isVerbose()){
			System.out.println("\n\n========================================================================\nSummary confusion matrix");
			for(int i=0; i<confusionMatrix.length; i++){
				for(int j=0; j<confusionMatrix.length; j++){
					System.out.print((int)confusionMatrix[i][j] + "\t");
				}
				System.out.println();
			}
			System.out.println("========================================================================");
		}
		
		
		results_of_LOOCV.setConfusionMatrix(confusionMatrix);
		
		return results_of_LOOCV;
	}/*  end of LOOCV method  */
	
	
	
	
	/*  start of run method  */
	/**
	 * @param args an array containing exactly two Strings: [0]=data_arff_file; [1]=matched_experiment_names_and_indices
	 */
	public void run(String[] args) throws Exception{
		if(args == null){
			System.err.println("Error: the args are null and cannot be used.\nSystem.exit(0)");
			System.exit(0);
		}
		if(args.length != 2){
			System.err.println("Error: there are not exactly two (2) arguments.\nSystem.exit(0)");
			System.exit(0);
		}
		
				
		//  load the experiment names and LOOCV indices
		ArrayList name_indices_tuple = this.loadExperimentNamesAndLOOCVindices(args[1]);
		ArrayList<String> experimentNames = (ArrayList<String>)name_indices_tuple.get(0);
		ArrayList<int[]> LOOCVindices = (ArrayList<int[]>)name_indices_tuple.get(1);
		name_indices_tuple = null;
		if(this.isVerbose()){
			this.displayExperimentNamesAndIndices(experimentNames, LOOCVindices);
		}
					
		
		//  load the cellular experiment data specified by the arff filename
		Instances complete_dataset = DataSource.read(inputPath + args[0]);
		if(complete_dataset.classIndex() == -1){
			complete_dataset.setClassIndex(complete_dataset.numAttributes()-1);
		}
		if(this.isVerbose()){
			System.out.println("\n\n==========================================================\nComplete dataset details:");
			System.out.println("Relation name = " + complete_dataset.relationName());
			System.out.println("Number of Attributes = " + complete_dataset.numAttributes());
			System.out.println("Number of Instances = " + complete_dataset.numInstances());
			System.out.println("==========================================================\n");
		}
		
		
		
		//  perform the LOOCV
		LOOCVresults results_of_LOOCV = this.LOOCV(complete_dataset, experimentNames, LOOCVindices);
		
		
		//  output the LOOCV results to file
		results_of_LOOCV.outputToFile(this.getOutputPath() + "LOOCV_results_", this.isVerbose());
		
		
		if(this.isVerbose()){
			System.out.println("\nDone!");
		}
	}/*  end of run method  */
	
	
	
	
	/*  start of main method  */
	/**
	 * @param args
	 */
	public static void main(String[] args) throws Exception {
		
		
		
		
		boolean usingCommandLine = false;
		
		
		//  load the parameters from the command line
		if(args.length == 2){
			usingCommandLine = true;
		}
		else{
			System.err.println("Error: there need to be exactly two (2) command line parameters.\n" +
					"\t\texperiment_data.arff\n" +
					"\t\tmatched_experiment_names_and_LOOCV_indices.txt\n" +
					"System.exit(0)");
//			System.exit(0);
		}
		
		
		
		
		/******************************************************
		 * 													  *
		 * 				SET THE PARAMETERS HERE				  *
		 * 													  *
		 ******************************************************/
		
		
		if(!usingCommandLine){
			
			
			args = new String[2];
			
//			args[0] = "IFs_[expression_profile][1].arff";
//			args[1] = "experiment_list_[expression_profile][1].txt";
			
//			args[0] = "IFs_[expression_profile][2].arff";
//			args[1] = "experiment_list_[expression_profile][2].txt";
			
//			args[0] = "IFs_[MLL_state].arff";
//			args[1] = "experiment_list_[MLL_state].txt";
			
//			args[0] = "IFs_[MLL_subtype].arff";
//			args[1] = "experiment_list_[MLL_subtype].txt";
			
			
//			args[0] = "IFs_[expression_profile].arff";
//			args[0] = "IFs_[MLL_fusion_type].arff";
//			args[0] = "IFs_[MLL_state].arff";
			
//			args[0] = "IFs_[HoxA9].arff";
//			args[0] = "IFs_[MLL].arff";
			
//			args[0] = "IFs_[HoxA9]_KARPAS_ML2.arff";
//			args[0] = "IFs_[MLL]_KARPAS_ML2.arff";
			
//			args[1] = "experiment_list.txt";
			
			
//			args[0] = "IFs_[MLL_state_training].arff";
//			args[1] = "experiment_list_[MLL_state_training].txt";
			
			
//			args[0] = "IFs_[MLL_state][all].arff";
//			args[1] = "experiment_list[all].txt";
			
			
//			args[0] = "Gene_expression_for_ARFF_[MLL_state].arff";
//			args[1] = "experiment_list.txt";
			
			
//			args[0] = "Gene_expression_for_ARFF_[MLL_subtype].arff";
//			args[1] = "experiment_list.txt";
			
			
//			args[0] = "GD34-39.arff";
//			args[0] = "GD35-40.arff";
//			args[0] = "GD36-49.arff";
//			args[1] = "experiment_list.txt";
			
			
			
//			args[0] = "5C_2012-10-19.arff";
//			args[0] = "Gene_expression_2012-10-19.arff";
//			args[0] = "Gene_expression_and_5C_2012-06-28.arff";
			args[0] = "HoxA9_2012-10-19.arff";
			args[1] = "experiment_list.txt";
			
			
		}
		
		
		/******************************************************
		 * 													  *
		 * 				END OF SETTING THE PARAMETERS		  *
		 * 													  * 
		 ******************************************************/
		
		
		
		
		//  create the object to perform the LOOCV
		ClassifierLOOCV brain = new ClassifierLOOCV();
		
		if(!usingCommandLine){
			brain.setInputPath("ClassifierLOOCV/Input/");
			brain.setOutputPath("ClassifierLOOCV/Output/");
			brain.setVerbose(true);
		}
		
		
		brain.run(args);
		
		
		
		
	}/*  end of main method  */

}/*  end of ClassifierLOOCV class  */
