package utilities;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;

import modeling.Chromosome;


/** ------------------------------------------------------------------------------------------------------------------
 * start of ConvertIFtoAttributes class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class ConvertIFtoAttributes {

	
	
	
	/*  global parameters  */
	private static final String inputPath = "ConvertIFtoAttributes/Input/";
	private static final String outputPath = "ConvertIFtoAttributes/Output/";
	/*  end of global parameters  */
	

	
	
	/*  start of load_filenames method  */
	/**
	 * @param filename_file_list the filename of the file containing the list of experiment filenames
	 */
	private static ArrayList<String> load_filenames(String filename_file_list){
		if(filename_file_list == null){
			System.err.println("Error: the filename for the file list is null and the files cannot be loaded.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<String> filenames = new ArrayList<String>();
		

		String line = null;
		int lineCounter = -1;			//  there is one header line
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename_file_list));
			while((line = br.readLine()) != null){
				if(lineCounter >= 0){
					filenames.add(line);
				}
				else{
					//  skip and print the header line
//					System.out.println(line);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		filenames.trimToSize();
		
		System.out.println("Loaded input file: " + inputPath + filename_file_list);
		
		return filenames;
	}/*  end of load_filenames method  */
	
	
	
	
	/*  start of displayFilenames method  */
	/**
	 * @param filenames an ArrayList of Strings representing the filenames
	 */
	private static void displayFilenames(ArrayList<String> filenames){
		if(filenames == null){
			System.err.println("Error: the filenames list is null and the names cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		System.out.println("=== Filenames ===");
		for(String current_filename: filenames){
			System.out.println(current_filename);
		}
	}/*  end of displayFilenames method  */
	
	
	
	
	/*  start of loadFiles method  */
	/**
	 * @param filenames an ArrayList of Strings specifying all of the experiment filenames
	 * @return an ArrayList of IF objects corresponding to each of the input experiment files
	 */
	private static ArrayList<IFs> loadFiles(ArrayList<String> filenames){
		if(filenames == null){
			System.err.println("Error: the filenames list is null and the experiments cannot be loaded from file.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<IFs> inputData = new ArrayList<IFs>();
		
		for(String current_filename: filenames){
			inputData.add(new IFs(inputPath + current_filename));
		}
		inputData.trimToSize();
		
		return inputData;
	}/*  end of loadFiles method  */
	
	
	
	
	/*  start of displayInputData method  */
	/**
	 * @param dataList an ArrayList of IFs objects containing the data to display
	 */
	private static void displayInputData(ArrayList<IFs> dataList){
		if(dataList == null){
			System.err.println("Error: the list of IFs objects is null and the data cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		for(int i=0; i<dataList.size(); i++){
			System.out.println();
			System.out.println(dataList.get(i));
		}
	}/*  end of displayInputData method  */
	
	

	
	/*  start of generateAttributeNames method  */
	/**
	 * @param IFs_list an ArrayList of IFs objects containing all of the input data
	 * @return an ArrayList of Strings representing each of the attribute names (each pair of points in a structure)
	 */
	private static ArrayList<String> generateAttributeNames(ArrayList<IFs> IFs_list){
		if(IFs_list == null){
			System.err.println("Error: the IFs list is null and the attribute names cannot be generated.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		//  find the minimum and maximum indices in the input data
		ArrayList<Integer> indexList = new ArrayList<Integer>();
		for(IFs currentIFs: IFs_list){
			indexList.add(currentIFs.getMinIndex());
			indexList.add(currentIFs.getMaxIndex());
		}
		
		int minIndex = Collections.min(indexList);
		int maxIndex = Collections.max(indexList);
		
		ArrayList<String> attributeNames = new ArrayList<String>();
		
		for(int i=minIndex; i<maxIndex; i++){
			for(int j=i+1; j<=maxIndex; j++){
				attributeNames.add("IF_" + i + "_" + j);
			}
		}
		attributeNames.trimToSize();
		
		return attributeNames;
	}/*  end of generateAttributeNames method  */
	
	
	

	/*  start of displayAttributeNames method  */
	/**
	 * @param attributeNames an ArrayList of Strings containing the attribute names
	 */
	private static void displayAttributeNames(ArrayList<String> attributeNames){
		if(attributeNames == null){
			System.err.println("Error: the attribute names list is null and the attribute names cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		for(String currentAttributeName: attributeNames){
			System.out.println(currentAttributeName);
		}
	}/*  end of displayAttributeNames method  */
	
	

	
	/*  start of outputAttributeNames method  */
	/**
	 * @param filename_output the filename to use to write the attribute names data to file
	 * @param attributeNames an ArrayList of Strings containing the attribute names
	 */
	private static void outputAttributeNames(String filename_output, ArrayList<String> attributeNames){
		if(attributeNames == null){
			System.err.println("Error: the attribute names list is null and the attribute names cannot be output to file.\nSystem.exit(0)");
			System.exit(0);
		}
		if(filename_output == null){
			System.err.println("Error: the output filename is null and the attribute names cannot be written to file.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		try{
			BufferedWriter bw = new BufferedWriter(new FileWriter(inputPath + filename_output));
			

			for(String currentAttributeName: attributeNames){
				bw.write(currentAttributeName);
				bw.newLine();
				bw.flush();
			}
			
			bw.flush();
			bw.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error: cannot write file with name \"" + filename_output + "\"\nSystem.exit(0)");
			System.exit(0);
		}
		System.out.println("\tOutput file: " + inputPath + filename_output);
	}/*  end of outputAttributeNames method  */
	
	
	
	
	/*  start of formatIFsAsAttributeVectors method  */
	/**
	 * @param list_of_experiment_data an ArrayList of IFs objects containing the input data for each of the experiments
	 * @return a HashMap containing the IFs formatted as an attribute vector for each experiment
	 */
	private static HashMap<Integer, ArrayList<String>> formatIFsAsAttributeVectors(ArrayList<IFs> list_of_experiment_data, ArrayList<String> attributeNames){
		if(list_of_experiment_data == null){
			System.err.println("Error: the list of experiment data is null and the data cannot be formatted as attribute vectors.\nSystem.exit(0)");
			System.exit(0);
		}
		
		HashMap<Integer, ArrayList<String>> attributeVectors = new HashMap<Integer, ArrayList<String>>();
		
		for(int i=0; i<list_of_experiment_data.size(); i++){
			int indexOfExperiment = i + 1;
			
			IFs currentExperimentIFs = list_of_experiment_data.get(i);
			
			ArrayList<String> currentAttributeVector = new ArrayList<String>();
			
			for(int j=0; j<attributeNames.size(); j++){
				
				String currentAttributeName = attributeNames.get(j);
				
				int firstIndex = Integer.parseInt(currentAttributeName.substring(currentAttributeName.indexOf("_")+1, currentAttributeName.lastIndexOf("_")));
				int secondIndex = Integer.parseInt(currentAttributeName.substring(currentAttributeName.lastIndexOf("_")+1, currentAttributeName.length()));
				
//				System.out.println(firstIndex + "-" + secondIndex);
				
				currentAttributeVector.add(currentExperimentIFs.getIFforPair(firstIndex, secondIndex));
				
			}
			
			currentAttributeVector.trimToSize();
			
			if(attributeVectors.containsKey(indexOfExperiment)){
				System.err.println("Error: the HashMap of attribute vectors already contains an entry for experiment " + indexOfExperiment + ".\nSystem.exit(0)");
				System.exit(0);
			}
			attributeVectors.put(indexOfExperiment, currentAttributeVector);
		}
		return attributeVectors;
	}/*  end of formatIFsAsAttributeVectors method  */
	
	
	
	
	/*  start of displayAttributeVectors method  */
	/**
	 * @param attributeVectors a HashMap containing the attribute vectors to display
	 * @param filenames an ArrayList of Strings specifying the experiment filenames
	 */
	private static void displayAttributeVectors(HashMap<Integer, ArrayList<String>> attributeVectors, ArrayList<String> filenames){
		if(attributeVectors == null){
			System.err.println("Error: the attributes vector HashMap is null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		for(int i=1; i<=attributeVectors.size(); i++){
			ArrayList<String> currentAttributeVector = attributeVectors.get(i);
			System.out.println("\n===  " + filenames.get(i-1) + "  ===");
			for(String currentValue: currentAttributeVector){
				System.out.print("\t" + currentValue);
			}
			System.out.println();
		}		
	}/*  end of displayAttributeVectors method  */
	
	
	
	
	/*  start of outputAttributeVectorsToFile method  */
	/**
	 * @param attributeVectors a HashMap containing the attribute vectors to output
	 * @param relation a String describing the name of the data set
	 * @param filename_class_labels the filename for the file containing the class label for each data vector
	 * @param filename_ARFF_header the filename for the file containing the header for the output ARFF file
	 * @param filename_attribute_names the filename for the file containing the name of each attribute vector
	 */
	private static void outputAttributeVectorsToFile(HashMap<Integer, ArrayList<String>> attributeVectors, String relation, String filename_class_labels, String filename_ARFF_header, String filename_attribute_names){
		if(filename_class_labels == null){
			System.err.println("Error: the class labels filename must be non-null to output the attribute vectors to file.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		ARFFutility utility_ARFF;
		if(relation != null){
			utility_ARFF = new ARFFutility(relation);
		}
		else{
			utility_ARFF = new ARFFutility();
		}
		
		utility_ARFF.generateARFFfile(attributeVectors, inputPath + filename_class_labels, inputPath + filename_ARFF_header, inputPath + filename_attribute_names, outputPath + relation);
	}/*  end of outputAttributeVectorsToFile method  */
	
	
	
	
	/*  start of main method  */
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		


		
		/******************************************************
		 * 													  *
		 * 				SET THE PARAMETERS HERE				  *
		 * 													  *
		 ******************************************************/
		
		
		String filename_attribute_names = "AttributeNames.txt";
		
		
//		String filename_file_list = "file_list.txt";
		
		
		
		
		//  TRAINING SET
//		String filename_file_list = "file_list_[training].txt";
//		String filename_class_labels = "ClassLabels_[MLL_state][training].txt";
//		String relation = "IFs_[MLL_state_training]";
////		
//		String filename_ARFF_header = "ARFF_header.txt";
		
		
		
		
		//  TRAINING SET FOR MLL SUBTYPE
//		String filename_file_list = "file_list_[training].txt";
//		String filename_class_labels = "ClassLabels_[MLL_subtype][training_subset].txt";
//		String relation = "IFs_[MLL_subtype]";
//		
//		String filename_ARFF_header = "ARFF_header.txt";
		
		
		
		//  TEST SET
//		String filename_file_list = "file_list_[HL60-1_test].txt";
//		String relation = "IFs_[MLL_state_HL60-1_test]";

//		String filename_file_list = "file_list_[Jurkat_test].txt";
//		String relation = "IFs_[MLL_state_Jurkat_test]";
		
//		String filename_file_list = "file_list_[NT2-D1-24h-1_test].txt";
//		String relation = "IFs_[MLL_state_NT2-D1-24h-1_test]";
		
//		String filename_file_list = "file_list_[MOLT4-1_test].txt";
//		String relation = "IFs_[MLL_state_MOLT4-1_test]";
		
//		String filename_file_list = "file_list_[U937-1_test].txt";
//		String relation = "IFs_[MLL_state_U937-1_test]";
//		
//		String filename_class_labels = "ClassLabels_[MLL_state_wt][test].txt";
//		String filename_ARFF_header = "ARFF_header.txt";
		
//		String filename_file_list = "file_list_[KARPAS-45-1_test].txt";
//		String relation = "IFs_[MLL_state_KARPAS-45-1_test]";
		
//		String filename_file_list = "file_list_[ML-2-1_test].txt";
//		String relation = "IFs_[MLL_state_ML-2-1_test]";
		
//		String filename_file_list = "file_list_[RS4-11-1_test].txt";
//		String relation = "IFs_[MLL_state_RS4-11-1_test]";
		
//		String filename_file_list = "file_list_[KNOP-8-1_test].txt";
//		String relation = "IFs_[MLL_state_KNOP-8-1_test]";
		
//		String filename_file_list = "file_list_[KNOP-8-2_test].txt";
//		String relation = "IFs_[MLL_state_KNOP-8-2_test]";
//		
//		String filename_file_list = "file_list_[THP-1-0hr_biological_test].txt";
//		String relation = "IFs_[MLL_state_THP-1-0hr_biological_test]";
		
//		String filename_file_list = "file_list_[THP-1-4hr_biological_test].txt";
//		String relation = "IFs_[MLL_state_THP-1-4hr_biological_test]";
		
//		String filename_file_list = "file_list_[THP-1-12hr-3_biological_test].txt";
//		String relation = "IFs_[MLL_state_THP-1-12hr-3_biological_test]";
		
//		String filename_file_list = "file_list_[THP-1-48hr_biological_test].txt";
//		String relation = "IFs_[MLL_state_THP-1-48hr_biological_test]";
		
//		String filename_file_list = "file_list_[THP-1-96hr_biological_test].txt";
//		String relation = "IFs_[MLL_state_THP-1-96hr_biological_test]";
		
//		String filename_file_list = "file_list_[THP-1-0h_biological2_test].txt";
//		String relation = "IFs_[MLL_state_THP-1-0h_biological2_test]";
		
//		String filename_file_list = "file_list_[THP-1-4h_biological2_test].txt";
//		String relation = "IFs_[MLL_state_THP-1-4h_biological2_test]";
		
//		String filename_file_list = "file_list_[THP-1-12h_biological2_test].txt";
//		String relation = "IFs_[MLL_state_THP-1-12h_biological2_test]";
		
//		String filename_file_list = "file_list_[THP-1-48h_biological2_test].txt";
//		String relation = "IFs_[MLL_state_THP-1-48h_biological2_test]";
		
		String filename_file_list = "file_list_[THP-1-96h_biological2_test].txt";
		String relation = "IFs_[MLL_state_THP-1-96h_biological2_test]";
		
//		String filename_file_list = "file_list_[NOMO-1-1_test].txt";
//		String relation = "IFs_[MLL_state_NOMO-1-1_test]";
//		
//		
		String filename_class_labels = "ClassLabels_[MLL_state_fusion][test].txt";
		String filename_ARFF_header = "ARFF_header.txt";
		
		
		
				
		
		
		//  ALL IN ONE
//		String filename_file_list = "file_list_[all].txt";
//		String filename_ARFF_header = "ARFF_header.txt";
//
//		String relation = "IFs_[MLL_state][all]";
//		String filename_class_labels = "ClassLabels_[MLL_state][all].txt";
//		
//		String relation = "IFs_[MLL_fusion_type]";
//		String filename_class_labels = "ClassLabels_[MLL_fusion_type].txt";
		
		
		
		
		
		
		/******************************************************
		 * 													  *
		 * 				END OF SETTING THE PARAMETERS		  *
		 * 													  * 
		 ******************************************************/
		
		
		
		
		ArrayList<String> filenames = load_filenames(filename_file_list);
//		displayFilenames(filenames);
		
		
		ArrayList<IFs> input_data = loadFiles(filenames);
//		displayInputData(input_data);
		
		
		ArrayList<String> attributeNames = generateAttributeNames(input_data);
//		displayAttributeNames(attributeNames);
		outputAttributeNames(filename_attribute_names, attributeNames);
		
		
		HashMap<Integer, ArrayList<String>> attributeVectors = formatIFsAsAttributeVectors(input_data, attributeNames);
//		displayAttributeVectors(attributeVectors, filenames);
		
		
		outputAttributeVectorsToFile(attributeVectors, relation, filename_class_labels, filename_ARFF_header, filename_attribute_names);
		
		
		System.out.println("Done!");
	}/*  end of main method  */

}/*  end of ConvertIFtoAttributes class  */
