package utilities;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;


/** ------------------------------------------------------------------------------------------------------------------
 * start of ExcludeProbes class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class ExcludeProbes {

	
	

	/*  global parameters  */
	private static boolean verbose = true;
	private static String inputPath = "";
	private static String outputPath = "";
	/*  end of global parameters  */
	
	

	
	/*  start of getFilenames method  */
	/**
	 * @param filename_experiment_list the file containing the list of experiments
	 * @return a list of String objects representing the loaded filenames
	 */
	private static ArrayList<String> getFilenames(String filename_experiment_list){
		if(filename_experiment_list == null){
			System.err.println("Error: the filename list is null.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<String> experimentList = new ArrayList<String>();

		String line = null;
		int lineCounter = -1;			//  there is one header line
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename_experiment_list));
			while((line = br.readLine()) != null){
				if(lineCounter >= 0){
					experimentList.add(line);
				}
				else{
					//  skip and the header line
//					System.out.println(line);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		experimentList.trimToSize();
		if(verbose){
			System.out.println("Loaded input file: " + inputPath + filename_experiment_list);
		}
		
		return experimentList;
	}/*  end of getFilenames method  */
	
	

	
	/*  start of displayFilenames method  */
	/**
	 * @param experimentList the list of filenames to display
	 */
	private static void displayFilenames(ArrayList<String> experimentList){
		if(experimentList == null){
			System.err.println("Error: the experiment list is null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);	
		}
		
		System.out.println("Filename");
		for(String currentFilename: experimentList){
			System.out.println(currentFilename);
		}
	}/*  end of displayFilenames method  */
	
		
	
	
	/*  start of loadInputData method  */
	/**
	 * @param experimentList the list of filenames to load
	 * @return a list containing the loaded data loaded as FiveC objects 
	 */
	private static ArrayList<FiveC> loadInputData(ArrayList<String> experimentList){
		if(experimentList == null){
			System.err.println("Error: the experiment list is null and the data cannot be loaded.\nSystem.exit(0)");
			System.exit(0);	
		}
		
		ArrayList<FiveC> FiveC_inputData = new ArrayList<FiveC>();
		
		
		for(String filename: experimentList){
			FiveC_inputData.add(new FiveC(filename, inputPath, outputPath));
		}
		
		FiveC_inputData.trimToSize();
		
		return FiveC_inputData;
	}/*  end of loadInputData method  */
	
	

	
	/*  start of displayData method  */
	/**
	 * @param FiveC_inputData the list of FiveC objects to display
	 */
	private static void displayData(ArrayList<FiveC> FiveC_inputData){
		if(FiveC_inputData == null){
			System.err.println("The input data arrays are null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		for(FiveC currentFiveC: FiveC_inputData){
			System.out.println("\n\n" + currentFiveC);
		}
	}/*  end of displayData method  */
	
	

	
	/*  start of loadExcludedProbeList method  */
	/**
	 * @param filename_probes_to_exclude the filename of the file containing the list of probes to exclude
	 * @return a list containing the probe pairs to exclude
	 */
	private static ArrayList<int[]> loadExcludedProbeList(String filename_probes_to_exclude){
		if(filename_probes_to_exclude == null){
			System.err.println("Error: the filename is null and the excluded probes list cannot be loaded.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<int[]> excludedProbes = new ArrayList<int[]>();
		

		String line = null;
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename_probes_to_exclude));
			while((line = br.readLine()) != null){
				String[] values = line.split("\t");
				if(values.length != 2){
					System.err.println("Error: there are not exactly two (2) entries in each row of the excluded probe list.\nSystem.exit(0)");
					System.exit(0);
				}
				
				int firstIndex = Integer.parseInt(values[0]);
				int secondIndex = Integer.parseInt(values[1]);
				
				//  doesn't matter if the two indices are equal, if the entry is not present, then it will be skipped
//				if(firstIndex == secondIndex){
//					System.err.println("Error: the two indices are equal (" + firstIndex + " == " + secondIndex +").\nSystem.exit(0)");
//					System.exit(0);
//				}
				
				int[] currentProbes = new int[2];
				currentProbes[0] = firstIndex;
				currentProbes[1] = secondIndex;
				
				excludedProbes.add(currentProbes);
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		if(verbose){
			System.out.println("Loaded input file: " + inputPath + filename_probes_to_exclude);
		}
		
		excludedProbes.trimToSize();
		
		return excludedProbes;
	}/*  end of loadExcludedProbeList method  */
	
	
	
	
	/*  start of displayProbeList method  */
	/**
	 * @param excludedProbes the list of probe pairs to exclude to display
	 */
	private static void displayProbeList(ArrayList<int[]> excludedProbes){
		if(excludedProbes == null){
			System.err.println("Error: the probe list is null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		for(int[] currentPair: excludedProbes){
			System.out.println(currentPair[0] + " - " + currentPair[1]);
		}
	}/*  end of displayProbeList method  */
	
	
	
	
	/*  start of removeExcludedProbes method  */
	/**
	 * @param FiveC_inputData the list of FiveC objects to clean
	 * @param excludedProbes the list of probe pairs to exclude
	 * @return the list of FiveC objects with the specified probe pairs removed
	 */
	private static ArrayList<FiveC> removeExcludedProbes(ArrayList<FiveC> FiveC_inputData, ArrayList<int[]> excludedProbes){
		if((excludedProbes == null) || (FiveC_inputData == null)){
			System.err.println("Error: both the probe list and input data need to be non-null to remove the specified probes.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		if(excludedProbes.size() < 1){
			if(verbose){
				System.out.println("The list of probes to exclude is empty and no probes need to be removed.");
			}
		}
		else{
			for(FiveC currentExperiment: FiveC_inputData){
				currentExperiment.removeProbes(excludedProbes);
			}
		}
		
		return FiveC_inputData;
	}/*  end of removeExcludedProbes method  */
	
	
	
	
	/*  start of outputDataToFile method  */
	/**
	 * @param FiveCdata the list of FiveC objects to output to file
	 */
	private static void outputDataToFile(ArrayList<FiveC> FiveCdata){
		if(FiveCdata == null){
			System.err.println("Error: the FiveC data is null and cannot be output to file.\nSystem.exit(0)");
			System.exit(0);
		}
		
		for(FiveC currentFiveC: FiveCdata){
			currentFiveC.outputToFile();
		}
	}/*  end of outputDataToFile method  */
		
	
	
	
	/*  start of main method  */
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		
		
		
		boolean usingCommandLine = false;
		
		String filename_probes_to_exclude = "default.txt";
		String filename_experiment_list = "default.txt";
		ArrayList<String> experimentList = new ArrayList<String>();
		
		
		//  load the parameters from the command line
		if(args.length >= 2){
			//  first parameter is the filename of the probes to exclude list
			filename_probes_to_exclude = args[0];
			
			//  all subsequent parameters are the filenames of the experiment files
			for(int i=1; i<args.length; i++){
				experimentList.add(args[i]);
			}
			experimentList.trimToSize();
			
			usingCommandLine = true;
			verbose = false;
		}
		else{
			System.err.println("Error: there need to be at least two (2) command line parameters.\n" +
					"\t\tprobes_to_exclude_list_filename.txt\n" +
					"\t\texperiment_1_filename.txt\n" +
					"\t\texperiment_2_filename.txt\n" +
					"\t\texperiment_..._filename.txt\n" +
					"System.exit(0)");
//			System.exit(0);
		}
		
		
		
		
		/******************************************************
		 * 													  *
		 * 				SET THE PARAMETERS HERE				  *
		 * 													  *
		 ******************************************************/
		
		
		if(!usingCommandLine){
		
			
			inputPath = "ExcludeProbes/Input/";
			outputPath = "ExcludeProbes/Output/";

		
			filename_experiment_list = "experiment_list_GD.txt";
			filename_probes_to_exclude = "excluded_probes_GD.txt";
		
		
//			filename_experiment_list = "experiment_list_HoxA.txt";
//			filename_probes_to_exclude = "excluded_probes_HoxA.txt";
		
		
		}
		
		
		/******************************************************
		 * 													  *
		 * 				END OF SETTING THE PARAMETERS		  *
		 * 													  * 
		 ******************************************************/
		
		
		
		
		if(!usingCommandLine){
			experimentList = getFilenames(filename_experiment_list);
		}
//		displayFilenames(experimentList);
		

		ArrayList<FiveC> FiveC_inputData = loadInputData(experimentList);
//		displayData(FiveC_inputData);
		
		
		ArrayList<int[]> excludedProbes = loadExcludedProbeList(filename_probes_to_exclude);
//		displayProbeList(excludedProbes);
		
		
		ArrayList<FiveC> FiveC_cleaned = removeExcludedProbes(FiveC_inputData, excludedProbes);
		
		
		outputDataToFile(FiveC_cleaned);
		if(verbose){
			System.out.println("Output cleaned files.");
		}
		
		
		if(verbose){
			System.out.println("Done!");
		}
	}/*  end of main method  */

}/*  end of ExcludeProbes class  */
