package utilities;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;


/** ------------------------------------------------------------------------------------------------------------------
 * start of ExtractIFvalues class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class ExtractIFvalues {

	
	
	
	/*  global parameters  */
	private static boolean verbose = true;
	
	private static String inputPath = "";
	private static String outputPath = "";
	/*  end of global parameters  */
	
	

	
	/*  start of loadFilenames method  */
	/**
	 * @param filename_file_list the filename of the file containing the list of experiment filenames
	 * @return a list of the filenames contained in the specified file
	 */
	private static ArrayList<String> loadFilenames(String filename_file_list){
		if(filename_file_list == null){
			System.err.println("Error: the filename for the file list is null and the files cannot be loaded.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<String> filenames = new ArrayList<String>();
		

		String line = null;
		int lineCounter = -1;			//  there is one header line
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename_file_list));
			while((line = br.readLine()) != null){
				if(lineCounter >= 0){
					filenames.add(line);
				}
				else{
					//  skip and print the header line
//					System.out.println(line);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		filenames.trimToSize();
		
		if(verbose){
			System.out.println("Loaded input file: " + inputPath + filename_file_list);
		}
		
		return filenames;
	}/*  end of loadFilenames method  */
	
	
	

	/*  start of displayFilenames method  */
	/**
	 * @param filenames an ArrayList of Strings representing the filenames
	 */
	private static void displayFilenames(ArrayList<String> filenames){
		if(filenames == null){
			System.err.println("Error: the filenames list is null and the names cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		System.out.println("=== Filenames ===");
		for(String current_filename: filenames){
			System.out.println(current_filename);
		}
	}/*  end of displayFilenames method  */
	
	

	
	/*  start of loadFragments method  */
	/**
	 * @param filename_fragment_list the filename of the file containing the list of fragment pairs
	 * @return a list of the fragment pairs contained in the specified file
	 */
	private static ArrayList<int[]> loadFragments(String filename_fragment_list){
		if(filename_fragment_list == null){
			System.err.println("Error: the filename for the fragment list is null and the fragments cannot be loaded.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<int[]> fragments = new ArrayList<int[]>();
		

		String line = null;
		int lineCounter = -1;			//  there is one header line
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename_fragment_list));
			while((line = br.readLine()) != null){
				if(lineCounter >= 0){
					String[] values = line.split("\t");
					if(values.length != 2){
						System.err.println("Error: there are not exactly two entries in each row of the fragment list.\nSystem.exit(0)");
						System.exit(0);
					}
					int[] currentFragmentPair = new int[2];
					currentFragmentPair[0] = Integer.parseInt(values[0]);
					currentFragmentPair[1] = Integer.parseInt(values[1]);
					fragments.add(currentFragmentPair);
				}
				else{
					//  skip and print the header line
//					System.out.println(line);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		fragments.trimToSize();
		
		if(verbose){
			System.out.println("Loaded input file: " + inputPath + filename_fragment_list);
		}
		
		return fragments;
	}/*  end of loadFragments method  */
	
	


	/*  start of displayFragments method  */
	/**
	 * @param fragments an ArrayList of int arrays representing the fragment pairs
	 */
	private static void displayFragments(ArrayList<int[]> fragments){
		if(fragments == null){
			System.err.println("Error: the fragments list is null and the fragments cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		System.out.println("=== Fragments ===");
		for(int[] current_fragment_pair: fragments){
			System.out.println(current_fragment_pair[0] + "\t" + current_fragment_pair[1]);
		}
	}/*  end of displayFragments method  */
	
	
	
	
	/*  start of extractValues method  */
	/**
	 * @param filenames the list of filenames
	 * @param fragments the list of fragment pairs to extract
	 * @return a list of extracted values for each file
	 */
	private static ArrayList<ArrayList<Double>> extractValues(ArrayList<String> filenames, ArrayList<int[]> fragments){
		if((filenames == null) || (fragments == null)){
			System.err.println("Error: both the file list and the fragment list must be non-null to extract the IF values.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<ArrayList<Double>> extractedValues = new ArrayList<ArrayList<Double>>();
		
		for(String currentFilename: filenames){
			
			ArrayList<Double> currentExtractedIFs = new ArrayList<Double>();
			for(int i=0; i<fragments.size(); i++){
				currentExtractedIFs.add(0.0);
			}
			
			String line = null;
			try{
				BufferedReader br = new BufferedReader(new FileReader(inputPath + currentFilename));
				while((line = br.readLine()) != null){
					
					String[] values = line.split("\t");
					if(values.length != 5){
						System.err.println("Error: each line of the IF file does not contain exactly five (5) entries.\nSystem.exit(0)");
						System.exit(0);
					}
					
					int fragment1 = Integer.parseInt(values[0]);
					int fragment2 = Integer.parseInt(values[1]);
					double IF = Double.parseDouble(values[2]);
					
					for(int i=0; i<fragments.size(); i++){
						int[] currentFragmentPair = fragments.get(i);
						if( ((fragment1 == currentFragmentPair[0]) && (fragment2 == currentFragmentPair[1])) || ((fragment2 == currentFragmentPair[0]) && (fragment1 == currentFragmentPair[1])) ){
							currentExtractedIFs.set(i, IF);
						}
					}
				}
				br.close();
			}catch (IOException e) {
				e.printStackTrace();
			}
			
			currentExtractedIFs.trimToSize();
			extractedValues.add(currentExtractedIFs);
		}
		
		return extractedValues;
	}/*  end of extractValues method  */
	
	
	
	
	/*  start of displayExtractedValues method  */
	/**
	 * @param filenames the list of filenames
	 * @param fragments the list of fragment pairs
	 * @param extractedValues the list of extracted values to display
	 */
	private static void displayExtractedValues(ArrayList<String> filenames, ArrayList<int[]> fragments, ArrayList<ArrayList<Double>> extractedValues){
		if((filenames == null) || (fragments == null) || (extractedValues == null)){
			System.err.println("Error: all of the parameters need to be non-null to display the extracted values.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		//  display the header
		System.out.println("=== Extracted values ===");
		
		
		//  display the fragment pairs
		for(int[] currentFragmentPair: fragments){
			System.out.print("\t" + currentFragmentPair[0] + "-" + currentFragmentPair[1]);
		}
		System.out.println();
		
		
		//  display the extracted values
		for(int i=0; i<filenames.size(); i++){
			System.out.print(filenames.get(i));
			
			for(Double currentExtractedValue: extractedValues.get(i)){
				System.out.print("\t" + currentExtractedValue);
			}
			System.out.println();
		}		
	}/*  end of displayExtractedValues method  */
	

	
	
	/*  start of outputExtractedValues method  */
	/**
	 * @param filenames the list of filenames
	 * @param fragments the list of fragment pairs
	 * @param extractedValues the list of extracted values to display
	 */
	private static void outputExtractedValues(ArrayList<String> filenames, ArrayList<int[]> fragments, ArrayList<ArrayList<Double>> extractedValues){
		if((filenames == null) || (fragments == null) || (extractedValues == null)){
			System.err.println("Error: all of the parameters need to be non-null to output the extracted values.\nSystem.exit(0)");
			System.exit(0);
		}
		
		String filename_output = "extracted_IFs.txt";
		
		try{
			BufferedWriter bw = new BufferedWriter(new FileWriter(outputPath + filename_output));
	
			
			//  output the fragment pairs
			for(int[] currentFragmentPair: fragments){
				bw.write("\tfeature_" + currentFragmentPair[0] + "_" + currentFragmentPair[1]);
			}
			bw.newLine();
			bw.flush();
			
			
			//  output the extracted values
			for(int i=0; i<filenames.size(); i++){
				bw.write(filenames.get(i));
				
				for(Double currentExtractedValue: extractedValues.get(i)){
					bw.write("\t" + currentExtractedValue);
				}
				bw.newLine();
				bw.flush();
			}		
			
			bw.flush();		
			bw.close();
		}catch (IOException e) {
			e.printStackTrace();
		}//end of try-catch block
	
		if(verbose){
			System.out.println("\nOutput file: " + outputPath + filename_output);
		}
	}/*  end of outputExtractedValues method  */
	
	
	
	
	/*  start of main method  */
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		
		

		/******************************************************
		 * 													  *
		 * 				SET THE PARAMETERS HERE				  *
		 * 													  *
		 ******************************************************/
		
		
		inputPath = "ExtractIFvalues/Input/";
		outputPath = "ExtractIFvalues/Output/";
		
		
//		String file_list = "file_list.txt";
//		String file_list = "file_list_[training].txt";
//		String file_list = "file_list_[test].txt";
//		String file_list = "file_list_[AF9].txt";
//		String file_list = "file_list_[ENL].txt";
//		String file_list = "file_list_[fusion].txt";
		String file_list = "file_list_[wt].txt";
		
//		String fragment_list = "fragment_list_[MLL_state].txt";
//		String fragment_list = "fragment_list_[MLL_fusion].txt";
//		String fragment_list = "fragment_list_[3D-SP][20_features].txt";
//		String fragment_list = "fragment_list_[information_gain].txt";
//		String fragment_list = "fragments_[infogain]_[AF9_vs_ENL].txt";
		String fragment_list = "fragments_[infogain]_[wt_vs_fusion].txt";
		

		/******************************************************
		 * 													  *
		 * 				END OF SETTING THE PARAMETERS		  *
		 * 													  * 
		 ******************************************************/
		
		
		
		
		ArrayList<String> filenames = loadFilenames(file_list);
		displayFilenames(filenames);
		
		
		ArrayList<int[]> fragments = loadFragments(fragment_list);
		displayFragments(fragments);
		

		ArrayList<ArrayList<Double>> extractedValues = extractValues(filenames, fragments);
		displayExtractedValues(filenames, fragments, extractedValues);
		
		
		outputExtractedValues(filenames, fragments, extractedValues);
		
		
		if(verbose){
			System.out.println("Done!");
		}
	}/*  end of main method  */

}/*  end of ExtractIFvalues class  */
