package utilities;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;

import org.apache.commons.math3.stat.correlation.PearsonsCorrelation;
import org.apache.commons.math3.stat.correlation.SpearmansCorrelation;
import org.apache.commons.math3.stat.descriptive.rank.Median;


/** ------------------------------------------------------------------------------------------------------------------
 * start of GeneDesertCorrelation class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class GeneDesertCorrelation {

	
	
	
	/*  global parameters  */
	private static boolean verbose = true;
	
	private static String inputPath = "";
	private static String outputPath = "";
	
	private static final int[] FEATURE_LENGTH_TO_INDEX_MAPPING = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,0,-1,-1,1,2,3,4,5,6,7};		//  feature lengths are 15, 18, 19, 20, 21, 22, 23, and 24 nt
	/*  end of global parameters  */
	

	
	
	/*  start of loadFilenames method  */
	/**
	 * @param filename_file_list the filename of the file containing the list of experiment filenames
	 * @return a list of the filenames contained in the specified file
	 */
	private static ArrayList<String> loadFilenames(String filename_file_list){
		if(filename_file_list == null){
			System.err.println("Error: the filename for the file list is null and the files cannot be loaded.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<String> filenames = new ArrayList<String>();
		

		String line = null;
		int lineCounter = -1;			//  there is one header line
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename_file_list));
			while((line = br.readLine()) != null){
				if(lineCounter >= 0){
					filenames.add(line);
				}
				else{
					//  skip and print the header line
//					System.out.println(line);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		filenames.trimToSize();
		
		if(verbose){
			System.out.println("Loaded input file: " + inputPath + filename_file_list);
		}
		
		return filenames;
	}/*  end of loadFilenames method  */
	
	
	

	/*  start of displayFilenames method  */
	/**
	 * @param filenames an ArrayList of Strings representing the filenames
	 */
	private static void displayFilenames(ArrayList<String> filenames){
		if(filenames == null){
			System.err.println("Error: the filenames list is null and the names cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		System.out.println("=== Filenames ===");
		for(String current_filename: filenames){
			System.out.println(current_filename);
		}
	}/*  end of displayFilenames method  */
	
	
	
	
	/*  start of loadExperiments method  */
	/**
	 * @param filenames an ArrayList of Strings specifying all of the experiment filenames
	 * @return an ArrayList of FiveC objects corresponding to each of the input experiment files
	 */
	private static ArrayList<FiveC> loadExperiments(ArrayList<String> filenames){
		if(filenames == null){
			System.err.println("Error: the filenames list is null and the experiments cannot be loaded from file.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<FiveC> inputData = new ArrayList<FiveC>();
		
		for(String current_filename: filenames){
			inputData.add(new FiveC(inputPath + current_filename));
		}
		inputData.trimToSize();
		
		return inputData;
	}/*  end of loadExperiments method  */
	
	
	
	
	/*  start of displayExperiments method  */
	/**
	 * @param experiments the list of FiveC objects containing the data to display
	 */
	private static void displayExperiments(ArrayList<FiveC> experiments){
		if(experiments == null){
			System.err.println("Error: the experiments list needs to be non-null to be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		for(FiveC currentExperiment: experiments){
			System.out.println(currentExperiment);
		}
	}/*  end of displayExperiments method  */
	
	
	
	
	/*  start of getQuantitySignals method  */
	/**
	 * @param experiments the list of FiveC objects containing the data to obtain the quantity signal from
	 * @return an array containing the quantity signal for each of the experiments (quantity signal is the median signal across all neighbours and next-neighbours (max gap = 3))
	 */
	private static double[] getQuantitySignals(ArrayList<FiveC> experiments, int featureLength){
		if(experiments == null){
			System.err.println("Error: the experiments list needs to be non-null to determine the quantity signal.\nSystem.exit(0)");
			System.exit(0);
		}
		
		double quantitySignals[] = new double[experiments.size()];
		
		
		for(int i=0; i<experiments.size(); i++){
			FiveC currentExperiment = experiments.get(i);

//			System.out.println("\n" + currentExperiment.getExperimentName());
			
			//  get all of the neighbours and next-neighbours for the current experiment
			ArrayList<Double> signalValues = currentExperiment.getAllFragmentPairValues(featureLength, 1, 3);
			//  convert the list to an array of doubles
			double signalValuesDoubleArray[] = new double[signalValues.size()];
			for(int j=0; j<signalValues.size(); j++){
				signalValuesDoubleArray[j] = (double)signalValues.get(j);
			}
			
			Median medianCalculator = new Median();
			quantitySignals[i] = medianCalculator.evaluate(signalValuesDoubleArray);
//			System.out.println("median = " + quantitySignals[i]);
		}
		
		
		return quantitySignals;
	}/*  end of getQuantitySignals method  */
	
	
	
	
	/*  start of displayQuantitySignals method  */
	/**
	 * @param quantitySignals the array of quantity signals to display
	 * @param filenames the list of filenames matching the entries in the specified quantity signals array
	 */
	private static void displayQuantitySignals(double[] quantitySignals, ArrayList<String> filenames){
		if((quantitySignals == null) || (filenames == null)){
			System.err.println("Error: the parameters must be non-null to display the quantity signals.\nSystem.exit(0)");
			System.exit(0);
		}
		
		System.out.println("\nFilename\t\tQuantity signal");
		for(int i=0; i<filenames.size(); i++){
			System.out.println(filenames.get(i) + "\t\t" + quantitySignals[i]);
		}
	}/*  end of displayQuantitySignals method  */
	
	
	
	
	/*  start of getPearsonCorrelationCoefficients method  */
	/**
	 * @param experiments the list of FiveC objects containing the experiment data
	 * @param featureLength the feature length to use the signal from in calculating the Pearsons correlation coefficient (should be the same as the one used to determine the quantity signals)
	 * @param minGapBetweenFragmentPairIndices the minimum gap between fragment pair indices to consider
	 * @param maxGapBetweenFragmentPairIndices the maximum gap between fragment pair indices to consider
	 * @param quantitySignals the array of quantity signals for the given experiments
	 * @return a list containing the Pearsons correlation coefficient for each pair of fragments tested
	 */
	private static ArrayList<double[]> getPearsonCorrelationCoefficients(ArrayList<FiveC> experiments, int featureLength, int minGapBetweenFragmentPairIndices, int maxGapBetweenFragmentPairIndices, double[] quantitySignals){
		if((experiments == null) || (quantitySignals == null)){
			System.err.println("Error: the parameters need to be non-null to determine the Pearsons correlation coefficients.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		ArrayList<double[]> pearsonsCorrelationValues = new ArrayList<double[]>();
		
		//  get all of the fragment pair indices that have a gap matching the specified value
		ArrayList<int[]> allFragmentPairIndices = experiments.get(0).getAllFragmentPairIndices(minGapBetweenFragmentPairIndices, maxGapBetweenFragmentPairIndices);
		
		
		//  for each fragment pair, extract the signal value for the specified feature length across all of the experiments
		for(int[] currentFragmentPair: allFragmentPairIndices){
//			System.out.println(currentFragmentPair[0] + "\t" + currentFragmentPair[1]);
			
			//  the vector of signal values for the current fragment pair across all of the experiments
			double[] signalValuesForCurrentFragmentPairAcrossAllExperiments = new double[experiments.size()];
			
			for(int i=0; i<experiments.size(); i++){
				signalValuesForCurrentFragmentPairAcrossAllExperiments[i] = experiments.get(i).getFragmentPairValues(currentFragmentPair[0], currentFragmentPair[1])[FEATURE_LENGTH_TO_INDEX_MAPPING[featureLength]];
			}
			
			//  index with the current fragment pair and the vector array of values across all of the experiments
			
			//  calculate the Pearsons correlation coefficient between the vector of quantity signals and the vector of signals for the current fragment pair across all of the experiments
			PearsonsCorrelation calculator_PearsonsCorrelation = new PearsonsCorrelation();
			double currentPearsonsCorrelationCoefficient = calculator_PearsonsCorrelation.correlation(quantitySignals, signalValuesForCurrentFragmentPairAcrossAllExperiments);
			
//			System.out.println("Pearsons\t" + currentFragmentPair[0] + "-" + currentFragmentPair[1] + ":\t" + currentPearsonsCorrelationCoefficient);
			
			
			double[] currentPearsonsCorrelationEntry = new double[3];
			currentPearsonsCorrelationEntry[0] = currentFragmentPair[0];
			currentPearsonsCorrelationEntry[1] = currentFragmentPair[1];
			currentPearsonsCorrelationEntry[2] = currentPearsonsCorrelationCoefficient;
			
			
			pearsonsCorrelationValues.add(currentPearsonsCorrelationEntry);
			
			
			//  calculate the Spearmans correlation coefficient between the vector of quantity signals and the vector of signals for the current fragment pair across all of the experiments
//			SpearmansCorrelation calculator_SpearmansCorrelation = new SpearmansCorrelation();
//			double currentSpearmansCorrelationCoefficient = calculator_SpearmansCorrelation.correlation(quantitySignals, signalValuesForCurrentFragmentPairAcrossAllExperiments);
//			
//			System.out.println("Spearmans\t" + currentFragmentPair[0] + "-" + currentFragmentPair[1] + ":\t" + currentSpearmansCorrelationCoefficient);
			
		}
		pearsonsCorrelationValues.trimToSize();
		
		return pearsonsCorrelationValues;
	}/*  end of getPearsonCorrelationCoefficients method  */
	
	
	
	
	/*  start of displayPearsonsCorrelationCoefficients method  */
	/**
	 * @param pearsonsCorrelationCoefficients the list of Pearsons correlation coefficients
	 */
	private static void displayPearsonsCorrelationCoefficients(ArrayList<double[]> pearsonsCorrelationCoefficients){
		if(pearsonsCorrelationCoefficients == null){
			System.err.println("Error: the list of Pearsons correlation coefficients needs to be non-null to display.\nSystem.exit(0)");
			System.exit(0);
		}
		
		for(double[] currentPearsonsCorrelationEntry: pearsonsCorrelationCoefficients){
			System.out.println((int)currentPearsonsCorrelationEntry[0] + "-" + (int)currentPearsonsCorrelationEntry[1] + ":\t" + currentPearsonsCorrelationEntry[2]);
		}
	}/*  end of displayPearsonsCorrelationCoefficients method  */
	
	
	
	
	/*  start of main method  */
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		

		
		
		/******************************************************
		 * 													  *
		 * 				SET THE PARAMETERS HERE				  *
		 * 													  *
		 ******************************************************/
		
		
		inputPath = "GeneDesertCorrelation/Input/";
		outputPath = "GeneDesertCorrelation/Output/";
		
		
		String filename_file_list = "experiment_list.txt";
		
		int featureLength = 24;
		
		int minGapBetweenFragmentPairIndices = -1;
		int maxGapBetweenFragmentPairIndices = -1;
		
		
		/******************************************************
		 * 													  *
		 * 				END OF SETTING THE PARAMETERS		  *
		 * 													  * 
		 ******************************************************/
		
		
		
		
		ArrayList<String> filenames = loadFilenames(filename_file_list);
		if(verbose){
//			displayFilenames(filenames);
		}
		
		
		ArrayList<FiveC> experiments = loadExperiments(filenames);
//		displayExperiments(experiments);
		
		
		double[] quantitySignals = getQuantitySignals(experiments, featureLength);
//		displayQuantitySignals(quantitySignals, filenames);
		
		
		ArrayList<double[]> pearsonsCorrelationCoefficients = getPearsonCorrelationCoefficients(experiments, featureLength, minGapBetweenFragmentPairIndices, maxGapBetweenFragmentPairIndices, quantitySignals);
		displayPearsonsCorrelationCoefficients(pearsonsCorrelationCoefficients);
		
		
		
		if(verbose){
			System.out.println("\nDone!");
		}
	}/*  end of main method  */

}/*  end of GeneDesertCorrelation class  */
