package utilities;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;


/** ------------------------------------------------------------------------------------------------------------------
 * start of IFcalculator class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class IFcalculator {

	
	
	
	/*  global parameters  */
	private static boolean verbose = true;
	private static String inputPath = "";
	private static String outputPath = "";
	/*  end of global parameters  */
	
	
	
	
	/*  start of loadSignalsFromFile method  */
	/**
	 * @param filename the input filename
	 * @return a HashMap containing the data loaded from file with primer pair indices used as key values (i.e. "index1-index2")
	 */
	private static HashMap<String, double[]> loadSignalsFromFile(String filename){		
		if(filename == null){
			System.err.println("Error: the filename is null and the data cannot be loaded from file.\nSystem.exit(0)");
			System.exit(0);
		}
		
		HashMap<String, double[]> signalsHashMap = new HashMap<String, double[]>();
		
		String line = null;
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename));			
			while((line = br.readLine()) != null){
				String[] values = line.split("\t");
				if(values.length < 10){
					System.err.println("Error: there is an entry in \"" + inputPath + filename + "\" that does not have ten (10) values" +
							" (i.e. index1, index2, 15bp, 18bp, 19bp, 20bp, 21bp, 22bp, 23bp, 24bp)\nSystem.exit(0)");
					System.exit(0);
				}
				
				double[] rowEntry = new double[10];
				for(int i=0; i<10; i++){
					rowEntry[i] = Double.parseDouble(values[i]);
				}
				
				String keyForCurrentRow = values[0] + "-" + values[1];
				signalsHashMap.put(keyForCurrentRow, rowEntry);
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		
		if(verbose){
			System.out.println("Loaded signals file: " + inputPath + filename);
		}
		
		return signalsHashMap;
	}/*  end of loadSignalsFromFile method  */
	
	

	
	/*  start of displaySignals method  */
	/**
	 * @param signalsToDisplay the ArrayList of signals to display
	 */
	private static void displaySignals(HashMap<String, double[]> signalsToDisplay){
		ArrayList<String> allKeys = new ArrayList<String>(signalsToDisplay.keySet());
		Collections.sort(allKeys);
		for(String currentKey: allKeys){
			double[] currentEntry = signalsToDisplay.get(currentKey);
			System.out.print(currentKey);
			for(double currentValue: currentEntry){
				System.out.print("\t" + currentValue);
			}
			System.out.println();
		}
	}/*  end of displaySignals method  */
	
	
	
	
	/*  start of calculateIFs method  */
	/**
	 * @param BACSignals the HashMap containing the BAC signals to use as the denominator in calculating the IF values
	 * @param cellularSignals the HashMap containing the cellular signals to use as the numerator in calculating the IF values
	 * @return an ArrayList containing the calculated IF values
	 */
	private static ArrayList<double[]> calculateIFs(HashMap<String, double[]> BACSignals, HashMap<String, double[]> cellularSignals){
		if((cellularSignals == null) || (BACSignals == null)){
			System.err.println("Error: the cellular signals and BAC signals must be non-null to calculate the IF values.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<double[]> IFs = new ArrayList<double[]>();
		
		ArrayList<String> keys_cellular = new ArrayList<String>(cellularSignals.keySet());
		Collections.sort(keys_cellular);
		for(String currentKey: keys_cellular){
			if(BACSignals.containsKey(currentKey)){
				double[] cellularCurrentRow = cellularSignals.get(currentKey);
				double[] BACCurrentRow = BACSignals.get(currentKey);
				
				//  make sure that the fragment indices are the same in both files
				if((cellularCurrentRow[0] != BACCurrentRow[0]) || (cellularCurrentRow[1] != BACCurrentRow[1])){
					System.err.println("Error: the fragment indices do not match in the cellular and BAC files.\nSystem.exit(0)");
					System.exit(0);
				}
				
				//  calculate the IF and stdev for the current row
				ArrayList<Double> rowIFs = new ArrayList<Double>(0);
				for(int j=3; j<10; j++){
					double cellularSignal = cellularCurrentRow[j];
					double BACSignal = BACCurrentRow[j];
					
					if((cellularSignal != 0) && (BACSignal != 0)){
						rowIFs.add(cellularSignal / BACSignal);
					}
				}
				rowIFs.trimToSize();
				double rowIF = 0;
				double rowStdev = 0;
				for(Double ratio: rowIFs){
					rowIF += ratio;
				}
				if(rowIFs.size() > 1){
					rowIF /= (double)rowIFs.size();
					
					for(Double ratio: rowIFs){
						rowStdev += Math.pow(ratio - rowIF, 2);					
					}
					rowStdev /= (double)(rowIFs.size() - 1);
					rowStdev = Math.sqrt(rowStdev);
				}			
//				System.out.println((int)cellularCurrentRow[0] + "\t" + (int)cellularCurrentRow[1] + "\t" + rowIF + "\t" + rowStdev + "\t" + rowIFs.size());
				
				
				//  save the IF for the current row
				double[] IFentry = new double[5];
				IFentry[0] = cellularCurrentRow[0];
				IFentry[1] = cellularCurrentRow[1];
				IFentry[2] = rowIF;
				IFentry[3] = rowStdev;
				IFentry[4] = rowIFs.size();
				IFs.add(IFentry);
			}
			else{
				//  the entry has been excluded from the BAC, skip it
			}
		}
		IFs.trimToSize();
		
		
//		for(int i=0; i<cellularSignals.size(); i++){
//			double[] cellularCurrentRow = cellularSignals.get(i);
//			double[] BACCurrentRow = BACSignals.get(i);
//			
//			//  make sure that the fragment indices are the same in both files
//			if((cellularCurrentRow[0] != BACCurrentRow[0]) || (cellularCurrentRow[1] != BACCurrentRow[1])){
//				System.err.println("Error: the fragment indices do not match in the cellular and BAC files.\nSystem.exit(0)");
//				System.exit(0);
//			}
//			
//			//  calculate the IF and stdev for the current row
//			ArrayList<Double> rowIFs = new ArrayList<Double>(0);
//			for(int j=3; j<10; j++){
//				double cellularSignal = cellularCurrentRow[j];
//				double BACSignal = BACCurrentRow[j];
//				
//				if((cellularSignal != 0) && (BACSignal != 0)){
//					rowIFs.add(cellularSignal / BACSignal);
//				}
//			}
//			rowIFs.trimToSize();
//			double rowIF = 0;
//			double rowStdev = 0;
//			for(Double ratio: rowIFs){
//				rowIF += ratio;
//			}
//			if(rowIFs.size() > 1){
//				rowIF /= (double)rowIFs.size();
//				
//				for(Double ratio: rowIFs){
//					rowStdev += Math.pow(ratio - rowIF, 2);					
//				}
//				rowStdev /= (double)(rowIFs.size() - 1);
//				rowStdev = Math.sqrt(rowStdev);
//			}			
////			System.out.println((int)cellularCurrentRow[0] + "\t" + (int)cellularCurrentRow[1] + "\t" + rowIF + "\t" + rowStdev + "\t" + rowIFs.size());
//			
//			
//			//  save the IF for the current row
//			double[] IFentry = new double[5];
//			IFentry[0] = cellularCurrentRow[0];
//			IFentry[1] = cellularCurrentRow[1];
//			IFentry[2] = rowIF;
//			IFentry[3] = rowStdev;
//			IFentry[4] = rowIFs.size();
//			IFs.add(IFentry);
//		}
//		IFs.trimToSize();
		
		return IFs;
	}/*  end of calculateIFs method  */
	
	
	

	/*  start of displayIFs method  */
	/**
	 * @param IFsToDisplay the interaction frequencies to display
	 */
	private static void displayIFs(ArrayList<double[]> IFsToDisplay){
		for(double[] currentRow: IFsToDisplay){
			for(int i=0; i<5; i++){
				if(i < 2){
					System.out.print((int)currentRow[i] + "\t");
				}
				else if(i<4){
					System.out.print(currentRow[i] + "\t");
				}
				else{
					System.out.print(currentRow[i]);
				}
			}
			System.out.println();
		}
	}/*  end of displayIFs method  */
	
	
	
	
	/*  start of outputIFs method  */
	/**
	 * @param outputFilename the filename for the output file
	 * @param IFsToOutput the ArrayList of IFs to output to file
	 */
	private static void outputIFs(String outputFilename, ArrayList<double[]> IFsToOutput){
		try{
			BufferedWriter bw = new BufferedWriter(new FileWriter(outputPath + outputFilename));

			for(double[] currentRow: IFsToOutput){
				for(int i=0; i<5; i++){
					if(i < 2){
						bw.write((int)currentRow[i] + "\t");
					}
					else if(i<4){
						bw.write(currentRow[i] + "\t");
					}
					else{
						bw.write(String.valueOf(currentRow[i]));
					}
				}
				bw.newLine();
				bw.flush();
			}			
			bw.close();
		}catch (IOException e) {
			e.printStackTrace();
		}//end of try-catch block
	
		if(verbose){
			System.out.println("\nOutput file: " + outputPath + outputFilename);
		}
	}/*  end of outputIFs method  */
	
	
	
	
	/*  start of cleanIFs method  */
	/**
	 * @param stdev_cutoff the cutoff value to use to clean the IF values (only IF values that are greater than the stdev_cutoff percentage * the IF value are kept)
	 * @param IFs the list of interaction frequency values to clean based on the specified standard deviation cutoff
	 * @return a list containing both the list of removed IF entries (first entry) and the list of cleaned IF entries (second entry)
	 */
	private static ArrayList<ArrayList<double[]>> cleanIFs(int stdev_cutoff, ArrayList<double[]> IFs){
		if(IFs == null){
			System.err.println("Error: the IFs are null and cannot be cleaned.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<double[]> IFsCleaned = new ArrayList<double[]>();
		ArrayList<double[]> IFsRemoved = new ArrayList<double[]>();
		
		for(double[] currentRow: IFs){
			//  if the IF value is zero, then remove the row
			if(currentRow[2] == 0){
				IFsRemoved.add(currentRow);
			}
			//  if the IF value is less than or equal to the specified cutoff of the stdev, then remove the row
			else if((currentRow[2] * stdev_cutoff / 100.0) < currentRow[3]){
				IFsRemoved.add(currentRow);
			}
			//  otherwise keep the IF value
			else{
				IFsCleaned.add(currentRow);
			}
		}
		IFsCleaned.trimToSize();
		IFsRemoved.trimToSize();
		
		ArrayList<ArrayList<double[]>> returnList = new ArrayList<ArrayList<double[]>>();
		returnList.add(IFsRemoved);
		returnList.add(IFsCleaned);
		returnList.trimToSize();
		
		return returnList;
	}/*  end of cleanIFs method  */
	
	
	
	
	/*  start of renumberFragments method  */
	/**
	 * @param IFsCleaned the list of IF entries that have been cleaned and need to be renumbered
	 * @param cellularInputFilename the filename of the cellular input data
	 * @return a list containing the IF entries renumbered to start at index one (1) and contain no gaps
	 */
	private static ArrayList<double[]> renumberFragments(ArrayList<double[]> IFsCleaned, String cellularInputFilename){
		
		ArrayList<double[]> IFsRenumbered = new ArrayList<double[]>(IFsCleaned.size());
		
		//  keep a list of all the existing fragment indices
		ArrayList<Integer> existingFragmentIndices = new ArrayList<Integer>();
		
		for(double[] currentRow: IFsCleaned){
			int index1 = (int)currentRow[0];
			int index2 = (int)currentRow[1];
			
			//  check if the first index is already in the existing list and if not, then add it
			if(!existingFragmentIndices.contains(index1)){
				existingFragmentIndices.add(index1);
			}
			
			//  check if the second index is already in the existing list and if not, then add it			
			if(!existingFragmentIndices.contains(index2)){
				existingFragmentIndices.add(index2);
			}
		}
		existingFragmentIndices.trimToSize();
		
		//  sort the existing fragment indices in increasing order
		Collections.sort(existingFragmentIndices);
		
//		for(Integer index: existingFragmentIndices){
//			System.out.println(index);
//		}
		
		ArrayList<Integer> missingIndices = new ArrayList<Integer>(0);
		int minIndex = existingFragmentIndices.get(0);
		int maxIndex = existingFragmentIndices.get(existingFragmentIndices.size()-1);
//		System.out.println("minIndex = " + minIndex + "\tmaxIndex = " + maxIndex);
		for(int indexToTestFor=minIndex; indexToTestFor<=maxIndex; indexToTestFor++){
			//  if the index to test for does not exist, then add it to the list of missing indices
			if(!existingFragmentIndices.contains(indexToTestFor)){
				missingIndices.add(indexToTestFor);
			}
		}
		missingIndices.trimToSize();
		
//		for(Integer missingIndex: missingIndices){
//			System.out.println(missingIndex);
//		}
		
		
		int fragmentMapping[] = new int[maxIndex-minIndex+1];	//  Note: this assumes that the indices do not start at zero in the input file
		int currentNewIndex = existingFragmentIndices.size();
		for(int i=maxIndex-minIndex; i>=minIndex-minIndex; i--){
			fragmentMapping[i] = currentNewIndex;
			if(existingFragmentIndices.contains(i+minIndex)){
				currentNewIndex--;
			}			
		}
		
		
//		for(int i=0; i<fragmentMapping.length; i++){
//			System.out.println((i+minIndex) + "\t" + fragmentMapping[i]);
//		}
		
		
		for(double[] currentRow: IFsCleaned){
			double[] newRow = new double[5];
			
			newRow[0] = fragmentMapping[(int)currentRow[0]-minIndex];
			newRow[1] = fragmentMapping[(int)currentRow[1]-minIndex];
			newRow[2] = currentRow[2];
			newRow[3] = currentRow[3];
			newRow[4] = currentRow[4];
			
			IFsRenumbered.add(newRow);
		}
		IFsRenumbered.trimToSize();
		
		
		//  output the fragment mapping to file
		String filenameFragmentMapping = outputPath + "fragmentMapping_" + cellularInputFilename;
		try{
			BufferedWriter bw = new BufferedWriter(new FileWriter(filenameFragmentMapping));

			for(int i=0; i<fragmentMapping.length; i++){
				if(i == 0){
					bw.write(fragmentMapping[i] + "\t" + (i+minIndex));
				}
				else if(fragmentMapping[i] == fragmentMapping[i-1]){
					bw.write("\t" + (i+minIndex));
				}
				else{
					bw.newLine();
					bw.flush();
					bw.write(fragmentMapping[i] + "\t" + (i+minIndex));
				}
			}
			bw.flush();			
			bw.close();
		}catch (IOException e) {
			e.printStackTrace();
		}//end of try-catch block
		
		if(verbose){
			System.out.println("\nOutput file: " + filenameFragmentMapping);		
		}
		
		return IFsRenumbered;
	}/*  end of renumberFragments method  */
	
	
	
	
	/*  start of getFilenames method  */
	/**
	 * @param filename_file_list the file containing the list of files
	 * @return a list of Objects containing the BAC filename (first entry, a String), and the list of cellular filenames (second entry, an ArrayList<String>)
	 */
	private static ArrayList<Object> getFilenames(String filename_file_list){
		if(filename_file_list == null){
			System.err.println("Error: the filename list is null.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<String> filenames_cellular = new ArrayList<String>();
		String filename_BAC = new String();

		String line = null;
		int lineCounter = -1;			//  there is one header line
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename_file_list));
			while((line = br.readLine()) != null){
				if(lineCounter == -1){
				//  skip and the header line
//					System.out.println(line);
				}
				else if(lineCounter == 0){				//  the first line after the header has the BAC filename
					String[] values = line.split("\t");
					filenames_cellular.add(values[0]);
					filename_BAC = values[1];
				}
				else{
					filenames_cellular.add(line.split("\t")[0]);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		filenames_cellular.trimToSize();
		
		if(verbose){
			System.out.println("Loaded input file: " + inputPath + filename_file_list);
		}
		
		ArrayList<Object> returnList = new ArrayList<Object>();
		returnList.add(filename_BAC);
		returnList.add(filenames_cellular);
		returnList.trimToSize();
		
		return returnList;
	}/*  end of getFilenames method  */
	
	

	
	/*  start of displayFilenames method  */
	/**
	 * @param filename_BAC the filename for the file containing the BAC experimental data
	 * @param filenames_cellular the list of filenames for the files containing the cellular experimental data
	 */
	private static void displayFilenames(String filename_BAC, ArrayList<String> filenames_cellular){
		if((filenames_cellular == null) || (filename_BAC == null)){
			System.err.println("Error: the file list and BAC must be non-null to be displayed.\nSystem.exit(0)");
			System.exit(0);	
		}
		
		System.out.println("\nBAC filename =\t" + filename_BAC);
		
		System.out.println("\nCellular filenames");
		for(String currentFilename: filenames_cellular){
			System.out.println(currentFilename);
		}
	}/*  end of displayFilenames method  */
	
	
	
	
	/*  start of main method  */
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		
		
		
		boolean usingCommandLine = false;
		
		int stdevCutoff = 75;
		String filename_BAC = "default.txt";
		ArrayList<String> filenames_cellular = new ArrayList<String>();
		String filename_file_list = "default.txt";
		
		
		//  load the parameters from the command line
		if(args.length >= 3){
			//  first parameter is the standard deviation cutoff value
			stdevCutoff = Integer.parseInt(args[0]);
			
			//  the second parameter is the BAC experiment filename
			filename_BAC = args[1];
			
			//  all subsequent parameters are the filenames of the experiment files
			for(int i=2; i<args.length; i++){
				filenames_cellular.add(args[i]);
			}
			filenames_cellular.trimToSize();
			
			usingCommandLine = true;
			verbose = false;
		}
		else{
			System.err.println("Error: there need to be at least three (3) command line parameters.\n" +
					"\t\tstandard_deviation_cutoff_percentage_value\n" +
					"\t\tBAC_experiment_filename.txt\n" +
					"\t\tGenomic_experiment_1_filename.txt\n" +
					"\t\tGenomic_experiment_2_filename.txt\n" +
					"\t\tGenomic_experiment_..._filename.txt\n" +
					"System.exit(0)");
//			System.exit(0);
		}
		
		
		
		
		
		/******************************************************
		 * 													  *
		 * 				SET THE PARAMETERS HERE				  *
		 * 													  *
		 ******************************************************/
		
		
		if(!usingCommandLine){
			
			
			inputPath = "IFcalculator/Input/";
			outputPath = "IFcalculator/Output/";
			
			
//			filename_file_list = "experiment_list.txt";
			
//			filename_file_list = "file_list_[HoxA].txt";
			filename_file_list = "file_list_[GeneDesert].txt";
		
		
			stdevCutoff = 75;
			
			
//			verbose = false;
			
			
		}
		
		
		/******************************************************
		 * 													  *
		 * 				END OF SETTING THE PARAMETERS		  *
		 * 													  * 
		 ******************************************************/
		
		
		
		
		if(!usingCommandLine){
			ArrayList<Object> listBACthenGenomicFilenames = getFilenames(filename_file_list);
			filename_BAC = (String)listBACthenGenomicFilenames.get(0);
			filenames_cellular = (ArrayList<String>)listBACthenGenomicFilenames.get(1);
		}
		if(verbose){
			displayFilenames(filename_BAC, filenames_cellular);
		}
		
		
		for(String filename_CellularSignals: filenames_cellular){
		
		
			//  display the parameters
			if(verbose){
				System.out.println("cellular input file = " + filename_CellularSignals);
				System.out.println("BAC input file = " + filename_BAC);
				System.out.println("stdev cutoff = " + stdevCutoff + "%");
			}
		
		
			//  load the cellular signals from file
			HashMap<String, double[]> cellularSignals = loadSignalsFromFile(filename_CellularSignals);
		

			//  display the cellular signals
//			displaySignals(cellularSignals);
		

			//  load the BAC signals from file
			HashMap<String, double[]> BACSignals = loadSignalsFromFile(filename_BAC);
		
			
			//  display the BAC signals
//			displaySignals(BACSignals);
		
			
			//  calculate the IF values
			ArrayList<double[]> IFs = calculateIFs(BACSignals, cellularSignals);
		
		
			//  display the IF values
//			displayIFs(IFs);
			
		
			//  output the IF values to file before removing the values that fail the standard deviation cutoff
			if(verbose){
				outputIFs("IFs_without_stdev_clean_" + filename_CellularSignals, IFs);
			}
			
		
			//  remove the IF values that fall below the stdev cutoff
			ArrayList<ArrayList<double[]>> removedAndCleanedIFlists = cleanIFs(stdevCutoff, IFs);
			ArrayList<double[]> IFsRemoved = removedAndCleanedIFlists.get(0);
			ArrayList<double[]> IFsCleaned = removedAndCleanedIFlists.get(1);
		
		
			//  display the removed IF values
//			displayIFs(IFsRemoved);
		

			//  display the cleaned IF values
//			displayIFs(IFsCleaned);
		
			
			//  output the removed IF values to file
			outputIFs("IFs_removed_" + filename_CellularSignals, IFsRemoved);
		

			//  output the cleaned IF values to file
			outputIFs("IFs_original_numbering_" + filename_CellularSignals, IFsCleaned);
			
		
			//  renumber the fragments to start at index one and have no gaps
			//  this method also outputs the fragment mappings to file
			ArrayList<double[]> IFsRenumbered = renumberFragments(IFsCleaned, filename_CellularSignals);
		
		
			//  display the renumbered fragments
//			displayIFs(IFsRenumbered);
		
		
			//  output the renumbered IF values to file
			outputIFs("IFs_renumbered_" + filename_CellularSignals, IFsRenumbered);
		}
		
		if(verbose){
			System.out.println("Done!");
		}
	}/*  end of main method  */

}/*  end of IFcalculator class  */
