package utilities;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;


/** ------------------------------------------------------------------------------------------------------------------
 * start of IFs class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class IFs {
	
	
	
	
	/*  global parameters  */
	private String path_filename;
	private double[][] IF_values;
	private int min_index = -1;
	private int max_index = -1;
	private int numberOfPoints = -1;
	/*  end of global parameters  */
	
	
	
	
	/*  start of constructor method  */
	/**
	 * @param path_filename the filename (including the path prefix) of the file containing the IF values to instantiate the current IF object with
	 */
	public IFs(String path_filename){
		if(path_filename == null){
			System.err.println("Error: cannot create an IFs object with a null filename.\nSystem.exit(0)");
			System.exit(0);
		}
		
		this.loadIFvalues(path_filename);
	}/*  end of constructor method  */
	
	
	

	/*  start of getFilename method  */
	/**
	 * @return a String of the filename that specified the file containing the input data for the current IF object
	 */
	public String getFilename(){
		return this.path_filename;
	}/*  end of getFilename method  */
	
	
	
	
	/*  start of getIFs method  */
	/**
	 * @return a matrix of doubles containing all of the IF values (note that missing values are specified by a value of '-1')
	 */
	public double[][] getIFs(){
		return this.IF_values;
	}/*  end of getIFs method  */

	
	
	
	/*  start of getMinIndex method  */
	/**
	 * @return the integer value of the minimum index found in the input file
	 */
	public int getMinIndex(){
		return this.min_index;
	}/*  end of getMinIndex method  */
	
	
	
	
	/*  start of getMaxIndex method  */
	/**
	 * @return the integer value of the maximum index found in the input file
	 */
	public int getMaxIndex(){
		return this.max_index;
	}/*  end of getMaxIndex method  */
	
	
	

	/*  start of getNumberOfPoints method  */
	/**
	 * @return the integer value of the number of points in the input file
	 */
	public int getNumberOfPoints(){
		return this.numberOfPoints;
	}/*  end of getNumberOfPoints method  */
	
	
	
	
	/*  start of loadIFvalues method  */
	/**
	 * @param path_filename the filename (including the path prefix) of the file containing the IF values to load
	 */
	public void loadIFvalues(String path_filename){
		if(path_filename == null){
			System.err.println("Error: the filename is null and the IF values cannot be loaded from file.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		//  store the filename used to load the input data
		this.path_filename = path_filename;
		
		
		//  determine the number of points in the file and the minimum and maximum indices
		String line = null;
		int lineCounter = 0;
		int index_Min = Integer.MAX_VALUE;
		int index_Max = Integer.MIN_VALUE;
		try{
			BufferedReader br = new BufferedReader(new FileReader(path_filename));
			while((line = br.readLine()) != null){
				String[] values = line.split("\t");
				if(values.length == 5){
					//  parse the values once that will be used to save time
					int firstIndex = Integer.parseInt(values[0]);
					int secondIndex = Integer.parseInt(values[1]);

					//  check for a new minimum index
					if(firstIndex < index_Min){
						index_Min = firstIndex;
					}
					if(secondIndex < index_Min){
						index_Min = secondIndex;
					}

					//  check for a new maximum index
					if(firstIndex > index_Max){
						index_Max = firstIndex;
					}
					if(secondIndex > index_Max){
						index_Max = secondIndex;
					}
				}
				else{
					//  each row must have in order: index1    index2    IF    stdev    count
					System.err.println("Error: there are not five (5) entries in each row of the file \"" + path_filename + "\"\nSystem.exit(0)");
					System.exit(0);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error: cannot find file with name \"" + path_filename + "\"\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		//  store the min and max indices
		this.min_index = index_Min;
		this.max_index = index_Max;
		
		
		//  determine the number of points in the file
		this.numberOfPoints = index_Max - index_Min + 1;
		
		
		//  initialize the array for the IF values
		this.IF_values = new double[this.numberOfPoints][this.numberOfPoints];
		for(int i=0; i<this.numberOfPoints; i++){
			for(int j=0; j<this.numberOfPoints; j++){
				this.IF_values[i][j] = -1;
			}
		}

		
		//  initialize the array for the variance values
//		array_Variances = new double[numberOfPoints][numberOfPoints];
//		for(int i=0; i<numberOfPoints; i++){
//			for(int j=0; j<numberOfPoints; j++){
//				array_Variances[i][j] = -1;
//			}
//		}

		
		//  re-read the file and store the values
		line = null;
		try{
			BufferedReader br = new BufferedReader(new FileReader(path_filename));			
			while((line = br.readLine()) != null){
				String values[] = line.split("\t");
				//  make sure that there are exactly five (5) entries in each row
				if(values.length == 5){
					int firstIndex = Integer.parseInt(values[0]) - index_Min;
					int secondIndex = Integer.parseInt(values[1]) - index_Min;
					
					//  exclude IF entries that do not have an associated standard deviation measure
					double stdev = Double.parseDouble(values[3]);
					if(stdev > 0){
//						double variance = stdev * stdev;
//						//  keep the IFvariances matrix symmetric
//						array_Variances[firstIndex][secondIndex] = variance;
//						array_Variances[secondIndex][firstIndex] = array_Variances[firstIndex][secondIndex];

						//  keep the IFvalues matrix symmetric
						this.IF_values[firstIndex][secondIndex] = Double.parseDouble(values[2]);
						this.IF_values[secondIndex][firstIndex] = this.IF_values[firstIndex][secondIndex];
					}									
				}
				else{
					System.err.println("Error: there are not five (5) entries in each row of the file \"" + path_filename + "\"\nSystem.exit(0)");
					System.exit(0);
				}
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error: cannot find file with name \"" + path_filename + "\"\nSystem.exit(0)");
			System.exit(0);
		}
		System.out.println("\tOpened and read file: " + path_filename);
	}/*  end of loadIFvalues method  */
	
	
	
	
	/*  start of getIFforPair method  */
	/**
	 * @param firstIndex the index of the first point in the pair for which to retrieve the IF value
	 * @param secondIndex the index of the second point in the pair for which to retrieve the IF value
	 * @return a String representation of the IF value for the pair specified by the method parameters (or '?' if the value is missing)
	 */
	public String getIFforPair(int firstIndex, int secondIndex){
		//  if either of the indices are out of bounds, then return they symbol for missing values in an ARFF file ('?')
		if((firstIndex > this.max_index) || (firstIndex < this.min_index) || (secondIndex > this.max_index) || (secondIndex < this.min_index)){
//			System.err.println("Error: the indices must both be within bounds [" + this.min_index + "-" + this.max_index + "].\nSystem.exit(0)");
//			System.exit(0);
			return "?";
		}
		
		
		//  the array is symmetric, but still maintain that the first index should be the smaller of the two
		if(firstIndex > secondIndex){
			int tempIndex = secondIndex;
			secondIndex = firstIndex;
			firstIndex = tempIndex;
		}
		
		
		//  fetch the IF value for the specified pair from the array
		double IFvalue = this.IF_values[firstIndex - this.min_index][secondIndex - this.min_index];
		
		
		//  if the value is missing then return the symbol for missing values in an ARFF file ('?')
		if(IFvalue <= 0){
			return "?";
		}
		
		
		return String.valueOf(IFvalue);
	}/*  end of getIFforPair method  */
	
	
	
	
	/*  start of toString method  */
	/**
	 * @return a String representation of the IF object
	 */
	public String toString(){
		if(IF_values == null){
			System.err.println("Error: the IF data is null and cannot be converted to String format.\nSystem.exit(0)");
			System.exit(0);
		}
		
		String IF_string = "=== " + this.path_filename + " ===\n";
		
		//  write the header line
		for(int i=this.min_index; i<this.min_index+this.numberOfPoints; i++){
			IF_string += "\t" + i;
		}
		IF_string += "\n";
		for(int i=0; i<this.numberOfPoints; i++){
			IF_string += String.valueOf(i+this.min_index);
			for(int j=0; j<this.numberOfPoints; j++){
				if(j>i){
					if(IF_values[i][j] <= 0){
						IF_string += "\t?";
					}
					else{
						IF_string += "\t" + IF_values[i][j];
					}
				}
				else{
					IF_string += "\t";
				}
			}
			IF_string += "\n";
		}
		return IF_string;
	}/*  end of toString method  */
	
	
	
	
	
	
	
	
	
	
}/*  end of IFs class  */
