package utilities;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import modeling.Chromosome;


/** ------------------------------------------------------------------------------------------------------------------
 * start of IFutility class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class IFutility {
	
	
	
	
	/*  global parameters  */
	private double[][] array_IFs;
	private double[][] array_Variances;	
	private double exponent;
	private boolean exponentSet = false;
	/*  end of global parameters  */
	
	
	
	
	/*  start of loadFile method  */
	/**
	 * @param filename the name of the file to load
	 */
	public void loadFile(String filename){
		//  determine the number of points in the file and the minimum and maximum indices
		String line = null;
		int lineCounter = 0;
		int index_Min = Integer.MAX_VALUE;
		int index_Max = Integer.MIN_VALUE;
		try{
			BufferedReader br = new BufferedReader(new FileReader(filename));
			while((line = br.readLine()) != null){
				String[] values = line.split("\t");
				if(values.length == 5){
					//  parse the values once that will be used to save time
					int firstIndex = Integer.parseInt(values[0]);
					int secondIndex = Integer.parseInt(values[1]);

					//  check for a new minimum index
					if(firstIndex < index_Min){
						index_Min = firstIndex;
					}
					if(secondIndex < index_Min){
						index_Min = secondIndex;
					}

					//  check for a new maximum index
					if(firstIndex > index_Max){
						index_Max = firstIndex;
					}
					if(secondIndex > index_Max){
						index_Max = secondIndex;
					}
				}
				else{
					//  each row must have in order: index1    index2    IF    stdev    count
					System.err.println("Error: there are not five (5) entries in each row of the file \"" + filename + "\"\nSystem.exit(0)");
					System.exit(0);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error: cannot find file with name \"" + filename + "\"\nSystem.exit(0)");
			System.exit(0);
		}
		
		//  determine the number of points in the file
		int numberOfPoints = index_Max - index_Min + 1;
		
		//  initialize the array for the IF values
		array_IFs = new double[numberOfPoints][numberOfPoints];
		for(int i=0; i<numberOfPoints; i++){
			for(int j=0; j<numberOfPoints; j++){
				array_IFs[i][j] = -1;
			}
		}
		
		//  initialize the array for the variance values
		array_Variances = new double[numberOfPoints][numberOfPoints];
		for(int i=0; i<numberOfPoints; i++){
			for(int j=0; j<numberOfPoints; j++){
				array_Variances[i][j] = -1;
			}
		}

		
		//  re-read the file and store the values
		line = null;
		try{
			BufferedReader br = new BufferedReader(new FileReader(filename));			
			while((line = br.readLine()) != null){
				String values[] = line.split("\t");
				//  make sure that there are exactly five (5) entries in each row
				if(values.length == 5){
					int firstIndex = Integer.parseInt(values[0]) - index_Min;
					int secondIndex = Integer.parseInt(values[1]) - index_Min;
					
					//  exclude IF entries that do not have an associated standard deviation measure
					double stdev = Double.parseDouble(values[3]);
					if(stdev > 0){
						double variance = stdev * stdev;
						//  keep the IFvariances matrix symmetric
						array_Variances[firstIndex][secondIndex] = variance;
						array_Variances[secondIndex][firstIndex] = array_Variances[firstIndex][secondIndex];

						//  keep the IFvalues matrix symmetric
						array_IFs[firstIndex][secondIndex] = Double.parseDouble(values[2]);
						array_IFs[secondIndex][firstIndex] = array_IFs[firstIndex][secondIndex];
					}									
				}
				else{
					System.err.println("Error: there are not five (5) entries in each row of the file \"" + filename + "\"\nSystem.exit(0)");
					System.exit(0);
				}
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error: cannot find file with name \"" + filename + "\"\nSystem.exit(0)");
			System.exit(0);
		}
//		System.out.println("\tOpened and read file: " + filename);
	}/*  end of loadFile method  */

	
	

	/*  start of getIFs method  */
	/**
	 * @return the IFs
	 */
	public double[][] getIFs() {
		if(array_IFs == null){
			System.out.println("Returning a null pointer...");
		}
		return array_IFs;
	}/*  end of getIFs method  */
	
	
	

	/*  start of getVariances method  */
	/**
	 * @return the Variances
	 */
	public double[][] getVariances() {
		if(array_Variances == null){
			System.out.println("Returning a null pointer...");
		}
		return array_Variances;
	}/*  end of getVariances method  */


	
	
	/*  start of getChromosomeSize method  */
	/**
	 * @return chrosomomeSize the number of points in the chromosome as defined by the input data
	 */
	public int getChromosomeSize(){
		if(array_IFs == null){
			System.err.println("Error: the input data has not yet been loaded.\nSystem.exit(0)");
			System.exit(0);
		}
		return array_IFs.length;
	}/*  end of getChromosomeSize method  */
	
	
	
	
	/*  start of getExponent method  */
	/**
	 * @return the exponent
	 */
	public double getExponent() {
		return exponent;
	}/*  end of getExponent method  */




	/*  start of setExponent method  */
	/**
	 * @param exponent the exponent to set
	 */
	public void setExponent(double exponent) {
		if(exponentSet){
			System.err.println("Error: the exponent has previously been set to a value of " + exponent + ".\nSystem.exit(0)");
			System.exit(0);
		}
		this.exponent = exponent;
		exponentSet = true;
	}/*  end of setExponent method  */


	
	
	/*  start of calculateProbability method  */
	/**
	 * @param structure the structure for which to calculate the fitness
	 */
	public double calculateProbability(double[][] structure){
		return this.calculateFitness(new Chromosome(structure));
	}/*  end of calculateProbability method  */


	
	
	/*  start of calculateFitness method  */
	/**
	 * @param chromosome the chromosome for which to calculate the fitness
	 */
	public double calculateFitness(Chromosome chromosome){
		
		//  make sure that the exponent for the IF modeling function has been set
		if(!exponentSet){
			System.err.println("Error: the exponent for the IF modeling function has not yet been set.\nSystem.exit(0)");
			System.exit(0);
		}
		
		//  make sure that the IF data has been loaded from file
		if(array_IFs == null){
			System.err.println("Error: the IF data has not been loaded from file.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		int chromosomeSize = chromosome.getSize();
		
		
		//  make sure that there are matching values in the structure and input IF data
		if(chromosomeSize != array_IFs.length){
			System.err.println("Error: the size of the structure and the input IF data do not match.System.exit(0)");
			System.exit(0);
		}
		
		
		//  do not impose any prior probabilities on the chromosomes
		double prob_C = 1.0;
		double prob_IF = 1.0;

		//  ln( Pr[C|IF] ) = ln( Pr[IF|C] ) + ln( Pr[C] ) - ln( Pr[IF] )
		//  ln( Pr[C|IF] ) = ln( Pr[IF|C] ) +      0      +      0
		//  ln( Pr[C|IF] ) = ln( Pr[IF|C] )
		//	ln( Pr[IF|C] ) = (-0.5) * SUM(i,j){ [ ( ( IF(i,j) - (1 / (d(i,j)^exponent) ) )^2 / variance) + ln( 2 * PI * variance ) ] }
		
		double fitness = 0;
		
		for(int i=0; i<chromosomeSize-1; i++){
			for(int j=i+1; j<chromosomeSize; j++){
				
				if(array_IFs[i][j] != -1.0){

					//  calculate the distance in the chromosome between point i and point j
					double distanceTerm = chromosome.getAtomCoordinates(j).x - chromosome.getAtomCoordinates(i).x;		//	 x2 - x1
					double distance = distanceTerm * distanceTerm;														//  (x2 - x1)^2
					distanceTerm = chromosome.getAtomCoordinates(j).y - chromosome.getAtomCoordinates(i).y;				//	 y2 - y1
					distance += distanceTerm * distanceTerm;															//	(y2 - y1)^2
					distanceTerm = chromosome.getAtomCoordinates(j).z - chromosome.getAtomCoordinates(i).z;				//	 z2 - z1
					distance += distanceTerm * distanceTerm;															//	(z2 - z1)^2
					
					distance = Math.sqrt(distance);																		//	sqrt[ (x2 - x1)^2 + (y2 - y1)^2 + (z2 - z1)^2 ]
					
					
					double IFmodel = 1.0 / Math.pow(distance, exponent);
					
					
					//  calculate the two terms and add the product to the sum
					double term = array_IFs[i][j] - IFmodel;	//  the numerator
					term *= term;
					term /= array_Variances[i][j];		//  divide the numerator by the denominator
					term += Math.log(2.0 * Math.PI * array_Variances[i][j]);

					fitness += term;
				}
			}
		}
		fitness *= -0.5;
		
		return fitness;
	}/*  end of calculateFitness method  */
	
	
	
	
	/*  start of calculateAndSetFitness method  */
	/**
	 * @param chromosome the chromosome for which to calculate and set fitness
	 */
	public void calculateAndSetFitness(Chromosome chromosome){
		chromosome.setFitness(this.calculateFitness(chromosome));
	}/*  end of calculateAndSetFitness method  */
	
	
	
	
}/*  end of IFutility class  */
