package modeling;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Random;

import javax.vecmath.Point3d;
import utilities.IFutility;
import utilities.PDButility;


/** ------------------------------------------------------------------------------------------------------------------
 * start of MCMC5C class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class MCMC5C {


	
	
	/*  global variables  */
	private boolean verbose = true;
	
	private String outputFolder = "";
	private String inputPath = "";
	private String outputPath = "";

	private DateFormat dateFormat = new SimpleDateFormat("yyyy/MM/dd\tHH:mm:ss");
	
	private NumberFormat formatter0decimals = new DecimalFormat("0");
	private NumberFormat formatter2decimals = new DecimalFormat("0.00");
	
	private Random random = new Random();
	private PDButility utility_PDB = new PDButility();
	private IFutility utility_IF;
	
	private String filenameIFs = "no_name.txt";
	private String filenameFragments = "no_name.txt";
	private String initialStructureFilename = "no_name.txt";
	
	private double[][] IFvalues;
	private double[][] IFvariances;
	
	private double[][] structure;
	private double[][] bestStructure;
	
	private double probabilityOfBestStructure = -911;
	private double probabilityCurrent = -911;
	private long indexOfBestStructure;
	private String bestStructureFilename = "no_name.txt";
	
	private long runLength = (long)1E7;
	private double numberOfStructuresToOutput = 10;
	private double stepSize = 0.05;
	private double halfStepSize = 0.025;
	private double exponent = 2;
	private int initialStructureCase = 6;
	private double single_point_translation_move_fraction = 0.80;
	private double block_translation_move_fraction = 0.10;
	private double mirror_move_fraction = 0.10;
	
	private double maximum_neighbour_distance = -1;
	
	
	private long[] numberOfChangesAccepted = new long[4];
	private long[] numberOfChangesAttempted = new long[4];
	
	private String[] argsGlobal;
	/*  end of global variables  */
	
	
	
	
	/*  start of constructor  */
	/**
	 * @return void
	 */
	public MCMC5C(){
		this("", "", false);
	}/*  end of constructor  */
	
	
	
	
	/*  start of constructor  */
	/**
	 * @param inputPath the input path to use for the MCMC5C object
	 * @param outputPath the output path prefix to use to generate the output path for the MCMC5C object
	 * @param verbose a boolean indicating whether or not the MCMC5C object should print to the console
	 */
	public MCMC5C(String inputPath, String outputPath, boolean verbose){
		//  set the input path
		this.setInputPath(inputPath);
		
		//  generate and set the output path
		this.setOutputPath(outputPath);
		
		//  set the verbose flag
		this.setVerbose(verbose);
	}/*  end of constructor  */
	
	
	
	
	/*  start of setInputPath method  */
	/**
	 * @param inputPath the input path to set for the current MCMC5C object
	 */
	public void setInputPath(String inputPath){
		if(inputPath == null){
			System.err.println("Error: cannot set the input path to null.\nSystem.exit(0)");
			System.exit(0);
		}
		this.inputPath = inputPath;
	}/*  end of setInputPath method  */
	
	
	
	
	/*  start of setOutputPath method  */
	/**
	 * @param outputPath the output path prefix to use to generate the output path for the current MCMC5C object
	 */
	public void setOutputPath(String outputPath){
		if(outputPath == null){
			System.err.println("Error: cannot set the output path to null.\nSystem.exit(0)");
			System.exit(0);
		}
		
		DateFormat dateFormatOutputFolder = new SimpleDateFormat("yyyy-MM-dd-HH-mm-ss");
		String outputFolderDate = dateFormatOutputFolder.format(new Date());
		
		this.outputFolder = outputPath + outputFolderDate;
		this.outputPath = this.outputFolder + "/";
	}/*  end of setOutputPath method  */
	
	
	
	
	/*  start of getInputPath method  */
	/**
	 * @return the input path for the current MCMC5C object
	 */
	public String getInputPath(){
		return this.inputPath;
	}/*  end of getInputPath method  */
	
	
	
	
	/*  start of getOutputPath method  */
	/**
	 * @return the output path for the current MCMC5C object
	 */
	public String getOutputPath(){
		return this.outputPath;
	}/*  end of getOutputPath method  */
	
	
	
	
	/*  start of setVerbose method  */
	/**
	 * @param verbose a boolean of whether (true) or not (false) the MCMC5C object should print to the console
	 */
	public void setVerbose(boolean verbose){
		this.verbose = verbose;
	}/*  end of setVerbose method  */
	
	
	
	
	/*  start of getNumberOfPoints method  */
	/**
	 * @param fileToCountNumberOfPointsFrom the filename of the fragments file used to determine the number of points
	 * @return the number of points there should be in the structure
	 */
	private int getNumberOfPoints(String fileToCountNumberOfPointsFrom){
		String line = null;
		int numberOfPoints = 0;
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + fileToCountNumberOfPointsFrom));			
			while((line = br.readLine()) != null){
				numberOfPoints++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.out.println("Error: cannot find file with filename \"" + fileToCountNumberOfPointsFrom + "\"\n" +
			"System.exit(0)");
			System.exit(0);
		}
		return numberOfPoints;
	}/*  end of getNumberOfPoints method  */
	
	
	
	
	/*  start of initializeIFutility method  */
	/**
	 * @return void
	 */
	private void initializeIFutility(){
		//  create the IFutility with the IF data for the current run
		utility_IF = new IFutility();
		utility_IF.loadFile(inputPath + filenameIFs);
		utility_IF.setExponent(exponent);		
	}/*  end of initializeIFutility method  */

	
	

	/*  start of getIFsAndVariancesFromFile method  */
	/**
	 * @return void
	 */
	private void getIFsAndVariancesFromFile(){
		
		
		//  make sure that the IFutility has been created
		if(utility_IF == null){
			System.err.println("Error: trying to load the IFs and variances from a null IFutility pointer.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		IFvalues = utility_IF.getIFs();
		IFvariances = utility_IF.getVariances();		
	}/*  end of getIFsAndVariancesFromFile method  */

	
	
	
	/*  start of initializeStructure method  */
	/**
	 * @param bwParameters the buffered writer for the parameters file
	 * @throws IOException throws an IOException if the buffered writer causes an error
	 */
	private void initializeStructure(BufferedWriter bwParameters) throws IOException{
		

		
		//  get the number of points (can pass the fragments file as parameter to count the number of points from this file)
		int numberOfPoints = getNumberOfPoints(filenameFragments);
		
		
		
		switch (initialStructureCase) {
		case 0:			//		INITIALIZE STRUCTURE FROM FILE
			structure = utility_PDB.getStructure(inputPath + initialStructureFilename);
			bwParameters.write("Initializing the structure with a given structure from file \"" + initialStructureFilename + "\"");
			bwParameters.newLine();
			bwParameters.flush();
			break;
		case 1:			//		CUBE WITH SIDE LENGTH [0,0.01]
			structure = new double[numberOfPoints][3];

			Random random = new Random();
			
			//  initialize the structure in a cube with side length [0,0.01]
			for(int i=0; i<structure.length; i++){
				structure[i][0] = random.nextDouble() / 100.0;
				structure[i][1] = random.nextDouble() / 100.0;
				structure[i][2] = random.nextDouble() / 100.0;
			}
			bwParameters.write("Initializing the structure in a cube with side length [0,0.01]");
			bwParameters.newLine();
			bwParameters.flush();
			break;
		case 2:			//		CUBE WITH SIDE LENGTH = 10 * AVERAGE_DISTANCE
			System.err.println("Error: this initial case has been deprecated (2011-09-22).\nSystem.exit(0)");
			System.exit(0);
//			structure = new double[numberOfPoints][3];
//
//			random = new Random();
//			
//			double averageDistanceX10 = 10.0 / (Math.pow(averageIF, exponent));
//			
//			//  initialize the structure in a cube with side length = 10 * average distance of input dataset
//			for(int i=0; i<structure.length; i++){
//				structure[i][0] = random.nextDouble() * averageDistanceX10;
//				structure[i][1] = random.nextDouble() * averageDistanceX10;
//				structure[i][2] = random.nextDouble() * averageDistanceX10;
//				
//			}
//			bwParameters.write("Initializing the structure in a cube with side length = " + averageDistanceX10);
//			bwParameters.newLine();
//			bwParameters.flush();
			break;
		case 3:			//		SPRIAL WITH SMALL RANDOM NOISE
			structure = new double[numberOfPoints][3];

			random = new Random();
			
			//  use counter1 for the x-coordinates 
			int counter1 = 1;
			//  use counter2 for the y-coordinates
			int counter2 = 2;
			//  use zStepSize for the z-coordinates
			double zStepSize = 5.0 / ((double)numberOfPoints);
			
			//  initialize the structure along a spiral with a bit of random noise
			for(int i=0; i<structure.length; i++){
				//  x-coordinate
				if(counter1 == 1 || counter1 == 2){
					structure[i][0] = 10 + random.nextDouble();
					counter1++;
				}
				else{
					structure[i][0] = 11 + random.nextDouble();
					counter1++;
				}
				//  reset the counter1
				if(counter1 == 5){
					counter1 = 1;
				}
				
				//  y-coordinate
				if(counter2 == 1 || counter2 == 2){
					structure[i][1] = 10 + random.nextDouble();
					counter2++;
				}
				else{
					structure[i][1] = 11 + random.nextDouble();
					counter2++;
				}
				//  reset the counter2
				if(counter2 == 5){
					counter2 = 1;
				}
				
				//  z-coordinate
				structure[i][2] = (i * zStepSize) + random.nextDouble();			
			}
			bwParameters.write("Initializing the structure in a spiral (with small random noise)");
			bwParameters.newLine();
			bwParameters.flush();
			break;
		case 4:			//		STRAIGHT LINE WITH SMALL RANDOM NOISE
			structure = new double[numberOfPoints][3];

			random = new Random();
			
			double consecutivePointGapStraightLine = 0.25;
			
			for(int i=0; i<structure.length; i++){
				structure[i][0] = random.nextDouble() * consecutivePointGapStraightLine;
				structure[i][0] = random.nextDouble() * consecutivePointGapStraightLine;
				structure[i][2] = i * consecutivePointGapStraightLine;
			}
			
			bwParameters.write("Initializing the structure in a straight line (with small random noise)");
			bwParameters.newLine();
			bwParameters.flush();
			break;
		case 5:			//		CIRCLE WITH NO NOISE
			structure = new double[numberOfPoints][3];
			
			double consecutivePointGapCircle = 0.1;
			
			double circleConstant = 2 * Math.PI / (double)(structure.length);
			
			for(int i=0; i<structure.length; i++){
				structure[i][0] = Math.cos(circleConstant * i);
				structure[i][1] = Math.sin(circleConstant * i);
				structure[i][2] = i * consecutivePointGapCircle;
			}
			
			bwParameters.write("Initializing the structure in a circle (with no noise)");
			bwParameters.newLine();
			bwParameters.flush();
			break;
		case 6:			//		CIRCLE WITH SMALL NOISE
			structure = new double[numberOfPoints][3];
			
			consecutivePointGapCircle = 0.1;
			
			circleConstant = 2 * Math.PI / (double)(structure.length);
			
			for(int i=0; i<structure.length; i++){
				structure[i][0] = Math.cos(circleConstant * i) + Math.cos(circleConstant*10*i)*0.1;
				structure[i][1] = Math.sin(circleConstant * i) + Math.sin(circleConstant*10*i)*0.1;
				structure[i][2] = i * consecutivePointGapCircle;
			}
			
			bwParameters.write("Initializing the structure in a circle (with small noise)");
			bwParameters.newLine();
			bwParameters.flush();
			break;
		default:		//  Need to specify an initial structure case
			System.out.println("Error: there was no initial structure case provided.\nSystem.exit(0)");
			System.exit(0);
			break;
		}
		

		
		//  calculate the probability of the initial structure
		probabilityCurrent = utility_IF.calculateProbability(structure);
		
		
		
	}/*  end of initializeStructure method  */

	
	
	
	/*  start of outputGenomicPositionMap method  */
	/**
	 * @return void
	 */
	private void outputGenomicPositionMap(){

		
		String outputFilename = outputPath + "genomicPositionMap_" + filenameIFs;

		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filenameFragments));
			BufferedWriter bw = new BufferedWriter(new FileWriter(outputFilename));

			int lineCounter = 0;
			String line;
			while((line = br.readLine()) != null){
				lineCounter++;
				
				String values[] = line.split("\t");
				if(values.length == 1){
					bw.write(lineCounter + "\t" + values[0]);
					bw.newLine();
					bw.flush();
				}
				else{
					System.err.println("Error: there is not one entry in each row in the fragments file\n"
							+ "System.exit(0)");
					System.exit(0);
				}
			}
			bw.flush();
			bw.close();
			
			br.close();			
		}catch (IOException e) {
			System.err.println("Error: cannot find fragment file with filename \"" + filenameFragments + "\"");
			System.err.println("Error: could not output file: " + outputFilename);
			e.printStackTrace();
			System.exit(0);
		}//end of try-catch block
		
		if(verbose){
			System.out.println("\tOutput file: " + outputFilename);
		}
	}/*  end of outputGenomicPositionMap method  */

	
	
	
	/*  start of outputStructure method  */
	/**
	 * @param structure the structure to output to file
	 * @param structureIndex the index of the structure to output to file
	 * @param probability the probability of the structure to output to file
	 * @param filename the filename of the structure to output to file
	 */
	private void outputStructure(double[][] structure, int structureIndex, double probability, String filename){		
		
		
		//  define the structure filename
		String outputFilename;
		
		if(filename.compareTo("default") == 0){			
			outputFilename = utility_PDB.getDefaultFilenameForIndex(structureIndex);			
		}
		else{
			//  use the name passed as a parameter to the method
			outputFilename = filename;
		}
		
		
		//  create the header and write the structure to file
		String header = dateFormat.format(new Date()) + "\tlengthOfRun=\t" + formatter0decimals.format(runLength) + "\tstepSize=\t" + stepSize +
			"\tindex=\t" + structureIndex + "\tln(probability)=\t" + probability;
		
		utility_PDB.writeStructure(outputPath + outputFilename, structure, header);
	}/*  end of outputStructure method  */
	
	
	
	
	/*  start of outputBestStructure method  */
	/**
	 * @retun void
	 */
	private void outputBestStructure(){		
		this.outputStructure(bestStructure, (int)indexOfBestStructure, probabilityOfBestStructure, bestStructureFilename);		
	}/*  end of outputBestStructure method  */
	
	
	
	
	/*  start of calculateProbabilityOfStructure method  */
	/**
	 * @return the probability of the current global structure
	 */
	private double calculateProbabilityOfStructure(){
		return utility_IF.calculateProbability(structure);		
	}/*  end of calculateProbabilityOfStructure method  */
	
	
	
	
	/*  start of acceptanceFunction method  */
	/**
	 * @param probabilityNew the probability of the new structure to test for acceptance or rejection against the current structure
	 * @return true if the new structure is accepted or false if the new structure is rejected
	 */
	private boolean acceptanceFunction(double probabilityNew){
		
		
		
		//  calculate the ratio of the new probability to the current probability
		double lnProbabilityRatio = probabilityNew - probabilityCurrent;
		double probabilityRatio = Math.pow(Math.E, lnProbabilityRatio);
		
		
		
		//  calculate the value of alpha
		double alpha = 1.0;
		if(probabilityRatio < alpha){			//  if the ratio is less than one (1), then the new structure has a lower probability than the current structure
			alpha = probabilityRatio;
		}

		
		
		//  accept the new structure with probability equal to alpha
		if(alpha >= random.nextDouble()){ 
			return true;
		}
		return false;
	}/*  end of acceptanceFunction method  */
	
	
	
	
	/*  start of calculateDistanceBetweenPoints method  */
	/**
	 * @param point1 the first point in the pair of points to calculate the distance between
	 * @param point2 the second point in the pair of points to calculate the distance between
	 * @return the Euclidean distance between the two specified points
	 */
	private double calculateDistanceBetweenPoints(Point3d point1, Point3d point2){
		//  make sure the two pointers are valid
		if((point1 == null) || (point2 == null)){
			System.err.println("Error: both of the pointers need to be non-null to calculate the distance between the points.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		double distance = 0;
		
		double temp = point2.x - point1.x;
		temp *= temp;
		distance += temp;
		
		temp = point2.y - point1.y;
		temp *= temp;
		distance += temp;
		
		temp = point2.z - point1.z;
		temp *= temp;
		distance += temp;
		
		return Math.sqrt(distance);		
	}/*  end of calculateDistanceBetweenPoints method  */
	
	
	
	
	/*  start of calculateDistanceBetweenPoints method  */
	/**
	 * @param point1 the first point in the pair of points to calculate the distance between
	 * @param point2 the second point in the pair of points to calculate the distance between
	 * @return the Euclidean distance between the two specified points
	 */
	private double calculateDistanceBetweenPoints(double[] point1, double[] point2){
		//  make sure the two pointers are valid
		if((point1 == null) || (point2 == null)){
			System.err.println("Error: both of the pointers need to be non-null to calculate the distance between the points.\nSystem.exit(0)");
			System.exit(0);
		}
		
		return calculateDistanceBetweenPoints(new Point3d(point1), new Point3d(point2));
	}/*  end of calculateDistanceBetweenPoints method  */
	
	
	
	
	/*  start of calculateDistanceBetweenPoints method  */
	/**
	 * @param point1 the first point in the pair of points to calculate the distance between
	 * @param point2 the second point in the pair of points to calculate the distance between
	 * @return the Euclidean distance between the two specified points
	 */
	private double calculateDistanceBetweenPoints(double[] point1, Point3d point2){
		//  make sure the two pointers are valid
		if((point1 == null) || (point2 == null)){
			System.err.println("Error: both of the pointers need to be non-null to calculate the distance between the points.\nSystem.exit(0)");
			System.exit(0);
		}
		
		return calculateDistanceBetweenPoints(new Point3d(point1), point2);
	}/*  end of calculateDistanceBetweenPoints method  */
	
	
	

	/*  start of calculateDistanceBetweenPoints method  */
	/**
	 * @param point1 the first point in the pair of points to calculate the distance between
	 * @param point2 the second point in the pair of points to calculate the distance between
	 * @return the Euclidean distance between the two specified points
	 */
	private double calculateDistanceBetweenPoints(Point3d point1, double[] point2){
		//  make sure the two pointers are valid
		if((point1 == null) || (point2 == null)){
			System.err.println("Error: both of the pointers need to be non-null to calculate the distance between the points.\nSystem.exit(0)");
			System.exit(0);
		}
		
		return calculateDistanceBetweenPoints(point1, new Point3d(point2));
	}/*  end of calculateDistanceBetweenPoints method  */
	
	
	
	
	/*  start of calculateDistancePointToPlane method  */
	/**
	 * @param point the point from which to calculate the distance to the plane
	 * @param planeEquation the plance to which to calculate the distance from the point
	 */
	private double calculateDistancePointToPlane(Point3d point, double[] planeEquation){
		if((point == null) || (planeEquation == null)){
			System.err.println("Error: both pointers need to be non-null to calculate the distance from the point to the plane.\nSystem.exit(0)");
			System.exit(0);
		}
		
		return (planeEquation[0] * point.x + planeEquation[1] * point.y + planeEquation[2] * point.z + planeEquation[3]) / this.calculateMagnitudeOfPlaneNormal(planeEquation);
	}/*  end of calculateDistancePointToPlane method  */
	
	
	
	
	/*  start of translateBlock method  */
	/**
	 * @return void
	 */
	private void translateBlock(){
		
		
		//  choose a block of points to move at random
		int blockToMoveStartIndex = random.nextInt(structure.length);
		int blockToMoveEndIndex = random.nextInt(structure.length);
		while(blockToMoveEndIndex == blockToMoveStartIndex){
			blockToMoveEndIndex = random.nextInt(structure.length);
		}
		//  make sure the start index is smaller than the end index
		if(blockToMoveStartIndex > blockToMoveEndIndex){
			int tempIndex = blockToMoveStartIndex;
			blockToMoveStartIndex = blockToMoveEndIndex;
			blockToMoveEndIndex = tempIndex;
		}
		
		
		//  choose a random vector and apply it to the randomly chosen points in the block
		double[] randomVector = new double[3];
		randomVector[0] = random.nextDouble() * stepSize - halfStepSize;
		randomVector[1] = random.nextDouble() * stepSize - halfStepSize;
		randomVector[2] = random.nextDouble() * stepSize - halfStepSize;
		
		
		double[][] candidateStructure = new double[structure.length][3];
		boolean[] pointMoved = new boolean[structure.length];
		for(int i=0; i<structure.length; i++){
			if((i >= blockToMoveStartIndex) && (i <= blockToMoveEndIndex)){
				candidateStructure[i][0] = structure[i][0] + randomVector[0];
				candidateStructure[i][1] = structure[i][1] + randomVector[1];
				candidateStructure[i][2] = structure[i][2] + randomVector[2];
				pointMoved[i] = true;
			}
			else{
				candidateStructure[i][0] = structure[i][0];
				candidateStructure[i][1] = structure[i][1];
				candidateStructure[i][2] = structure[i][2];
			}					
		}
		
		this.testCandidateStructureAndUpdateIfPassAcceptanceTest(candidateStructure, pointMoved, 3);			
	}/*  end of translateBlock method  */
	
	
	
	
	/*  start of translateSinglePoint method  */
	/**
	 * @return void
	 */
	private void translateSinglePoint(){

		
		//  choose a point to move at random
		int pointToMoveIndex = random.nextInt(structure.length);
		Point3d currentPoint = new Point3d(structure[pointToMoveIndex]);
		
		
		//  choose a random vector and apply it to the randomly chosen point
		Point3d newPoint = new Point3d(currentPoint.x, currentPoint.y, currentPoint.z);
		double[] randomVector = new double[3];
		randomVector[0] = random.nextDouble() * stepSize - halfStepSize;
		randomVector[1] = random.nextDouble() * stepSize - halfStepSize;
		randomVector[2] = random.nextDouble() * stepSize - halfStepSize;
		this.applyVectorToPoint(randomVector, newPoint);
		
		
		boolean[] pointMoved = new boolean[structure.length];
		pointMoved[pointToMoveIndex] = true;
		
		
		double[][] candidateStructure = new double[structure.length][3];
		for(int i=0; i<structure.length; i++){
			if(i == pointToMoveIndex){
				candidateStructure[i][0] = newPoint.x;
				candidateStructure[i][1] = newPoint.y;
				candidateStructure[i][2] = newPoint.z;
			}
			else{
				candidateStructure[i][0] = structure[i][0];
				candidateStructure[i][1] = structure[i][1];
				candidateStructure[i][2] = structure[i][2];
			}					
		}
		
		
		this.testCandidateStructureAndUpdateIfPassAcceptanceTest(candidateStructure, pointMoved, 0);
	}/*  end of translateSinglePoint method  */
	
	
	
	
	/*  start of testCandidateStructureAndUpdateIfPassAcceptanceTest method  */
	/**
	 * @param candidateStructure the candidate structure to calculate the probability, test using the acceptance function and if accepted set as the current structure
	 * @param pointMoved an array of boolean indicating which points have been moved in the candidate structure (true=moved; false=not moved from current structure)
	 * @param typeOfMoveIndex the index for the type of move that generated the candidate structure (0=single point translation, 1=segment mirror, 2=end mirror) 
	 */
	private void testCandidateStructureAndUpdateIfPassAcceptanceTest(double[][] candidateStructure, boolean[] pointMoved, int typeOfMoveIndex){

		//  if there is a maximum neighbour distance parameter set, test the candidate structure first
		if(this.maximum_neighbour_distance > 0){			//  a negative maximum distance parameter indicates that the test sould be skipped
			
			if(!testAllNeighboursCloserThanBound(candidateStructure, this.maximum_neighbour_distance)){
				//  if the test fails, then do not accept the candidate structure and return immediately
				return;
			}
		}
		
		
		//  calculate the probability of the candidate structure
		double probabilityCandidate = this.calculateProbabilityOfCandidateStructureOptimized(candidateStructure, pointMoved);
		
//		double probabilityCandidate2 = utility_IF.calculateProbability(candidateStructure);
//		System.out.println("Type of move = " + typeOfMoveIndex);
//		System.out.println("Probability of new candidate structure using the local optimized method =\t" + probabilityCandidate);
//		System.out.println("Probability of new candidate structure using the IF utility =\t\t\t" + probabilityCandidate2 + "\n");
//		System.exit(0);
		
		
		if(this.acceptanceFunction(probabilityCandidate)){
			structure = candidateStructure;
			probabilityCurrent = probabilityCandidate;
			numberOfChangesAccepted[typeOfMoveIndex]++;
		}
	}/*  end of testCandidateStructureAndUpdateIfPassAcceptanceTest method  */
	
	
	
	
	/*  start of testAllNeighboursCloserThanBound method  */
	/**
	 * @param structure the structure to test that all of the neighbour points are at a distance strictly less than the specified bound
	 * @param maximum_distance the maximum distance allowed (exclusive) between any pair of neighbouring points in the structure
	 * @return true if all of the neighbouring points are closer than the specified maximum distance, false otherwise
	 */
	private boolean testAllNeighboursCloserThanBound(double[][] structure, double maximum_distance){
		
		//  go through all of the pairs of points in the structure
		for(int i=0; i<structure.length-1; i++){
			for(int j=i+1; j<structure.length; j++){
				if(this.calculateDistanceBetweenPoints(structure[i], structure[j]) >= maximum_distance){
					return false;
				}
			}
		}
		return true;
	}/*  end of testAllNeighboursCloserThanBound method  */
	
	
	
	
	/*  start of mirrorPointsOverPlane method  */
	/**
	 * @param point the point to mirror over the plane
	 * @param planeEquation the equation of the plane over which to mirror the point
	 */
	private void mirrorPointOverPlane(Point3d point, double[] planeEquation){
		//  make sure the pointers are valid
		if((point == null) || (planeEquation == null)){
			System.err.println("Error: the pointers need to be non-null to mirror the point over the plane.\nSystem.exit(0)");
			System.exit(0);
		}

		//  calculate the base point of the perpendicular (where the perpendicular line meets the plane)
		Point3d basePointOfPerpendicular = this.determineBasePointOfPerpendicular(point, planeEquation);
		
		
//		System.out.println("base point of perpendicular: " + basePointOfPerpendicular);
//		System.out.println("point to mirror before: " + point);
//		System.out.println("distance from point to mirror (before) to plane: " + this.calculateDistanceBetweenPoints(point, basePointOfPerpendicular));
//		System.out.println("distance from point to mirror (before) to plane: " + this.calculateDistancePointToPlane(point, planeEquation));
		
		
		//  calculate the vector to apply to the point to mirror (this is negative twice the vector from the point to mirror to the base of the perpendicular)
		double[] vectorDisplacement = this.determineVectorFromPointAtoPointB(basePointOfPerpendicular, point);
		this.applyScalarToVector(-2.0, vectorDisplacement);
		
		//  mirror the point by applying the displacement vector
		this.applyVectorToPoint(vectorDisplacement, point);
		
		
//		System.out.println("point to mirror after: " + pointToMirror);
//		System.out.println("distance from point to mirror (after) to plane: " + this.calculateDistanceBetweenPoints(point, basePointOfPerpendicular));
//		System.out.println("distance from point to mirror (after) to plane: " + this.calculateDistancePointToPlane(point, planeEquation));
		
		
	}/*  end of mirrorPointsOverPlane method  */
	
	
	
	
	/*  start of testPointsDifferent method  */
	/**
	 * @param point1 the first point to test if different than second point
	 * @param point2 the second point to test if different than first point
	 */
	private boolean testPointsDifferent(double[] point1, double[] point2){
		//  make sure the two pointers are valid
		if((point1 == null) || (point2 == null)){
			System.err.println("Error: the two pointers need to be non-null to test if the points are different.\nSystem.exit(0)");
			System.exit(0);
		}
		return testPointsDifferent(new Point3d(point1), new Point3d(point2));
	}/*  end of testPointsDifferent method  */
	
	
	
	
	/*  start of testPointsDifferent method  */
	/**
	 * @param point1 the first point to test if different than second point
	 * @param point2 the second point to test if different than first point
	 */
	private boolean testPointsDifferent(Point3d point1, Point3d point2){
		//  make sure the two pointers are valid
		if((point1 == null) || (point2 == null)){
			System.err.println("Error: the two pointers need to be non-null to test if the points are different.\nSystem.exit(0)");
			System.exit(0);
		}
		if((point1.x == point2.x) && (point1.y == point2.y) && (point1.z == point2.z)){
			return false;
		}
		return true;
	}/*  end of testPointsDifferent method  */
	
	
	
	
	/*  start of determinePlaneFromThreePoints method  */
	/**
	 * @param pointA the first of the three points from which to determine the plane
	 * @param pointB the second of the three points from which to determine the plane
	 * @param pointC the third of the three points from which to determine the plane
	 * @return an array containing the plane equation as ax + by + cz + d = 0
	 */
	private double[] determinePlaneFromThreePoints(Point3d pointA, Point3d pointB, Point3d pointC){
		//  make sure that the three points are non-null
		if((pointA == null) || (pointB == null) || (pointC == null)){
			System.err.println("Error: the three points need to be non-null to detemrine the equation of the plane.\nSystem.exit(0)");
			System.exit(0);
		}
		
		//  make sure that the three points are unique
		if((!this.testPointsDifferent(pointA, pointB)) || (!this.testPointsDifferent(pointA, pointC)) || (!this.testPointsDifferent(pointB, pointC))){
			System.err.println("Error: the three points must be unique to determine the plane.\nSystem.exit(0)");
			System.err.println("\tPointA: " + pointA);
			System.err.println("\tPointB: " + pointB);
			System.err.println("\tPointC: " + pointC);
			System.exit(0);
		}
		else{
//			System.out.println("Passed test, the three points are unique in determining the plane");
		}
		
		double[] planeEquation = new double[4];

		//  calculate the two vectors AB and AC
		double[] vectorAB = new double[3];
		vectorAB[0] = pointB.x - pointA.x;
		vectorAB[1] = pointB.y - pointA.y;
		vectorAB[2] = pointB.z - pointA.z;
		
		double[] vectorAC = new double[3];
		vectorAC[0] = pointC.x - pointA.x;
		vectorAC[1] = pointC.y - pointA.y;
		vectorAC[2] = pointC.z - pointA.z;
		
		
		//  calculate the cross product of the two vectors AB and AC that is a vector normal to the plane containing the three points A, B, and C
		planeEquation[0] = vectorAB[1] * vectorAC[2] - vectorAC[1] * vectorAB[2];
		planeEquation[1] = -1.0 * (vectorAB[0] * vectorAC[2] - vectorAC[0] * vectorAB[2]);
		planeEquation[2] = vectorAB[0] * vectorAC[1] - vectorAC[0] * vectorAB[1];
		
		
		//  calculate the distance of the plane from the normal ('d' in the equation for a plane given by ax + by + cz + d = 0)
		planeEquation[3] = planeEquation[0] * (-1.0 * pointA.x) + planeEquation[1] * (-1.0 * pointA.y) + planeEquation[2] * (-1.0 * pointA.z); 
		
		
		//  to make sure that the equation of the plane was properly determined test the remaining two points with the resulting equation
		//        Note: casting the double to an int to avoid errors with numerical imprecision causing a zero value to appear slightly off
		double testValueWithPointB = planeEquation[0] * pointB.x + planeEquation[1] * pointB.y + planeEquation[2] * pointB.z + planeEquation[3];
		double testValueWithPointC = planeEquation[0] * pointC.x + planeEquation[1] * pointC.y + planeEquation[2] * pointC.z + planeEquation[3];
		if(((int)testValueWithPointB != 0) || ((int)testValueWithPointC != 0)){
			System.err.println("Error: both " + (int)testValueWithPointB + " and " + (int)testValueWithPointC +" should equal zero (0.0)");
			System.err.println("Error: the equation of the plane was not properly determined.\nSystem.exit(0)");
			System.exit(0);
		}
		
		return planeEquation;		
	}/*  end of determinePlaneFromThreePoints method  */
	
	
	
	
	/*  start of applyVectorToPoint method  */
	/**
	 * @param vector the vector to apply (add) to the point
	 * @param point the point to which the vector is applied (added)
	 */
	private void applyVectorToPoint(double[] vector, Point3d point){
		if((vector == null) || (point == null)){
			System.err.println("Error: both of the pointers must be non-null to apply the vector to the point.\nSystem.exit(0)");
			System.exit(0);
		}
		point.x += vector[0];
		point.y += vector[1];
		point.z += vector[2];
	}/*  end of applyVectorToPoint method  */
	
	
	
	
	/*  start of determineBasePointOfPerpendicular method  */
	/**
	 * @param point the point from which to drop the perpendicular to the plane
	 * @param planeEquation the plane to which to drop the perpendicular from the point
	 */
	private Point3d determineBasePointOfPerpendicular(Point3d point, double[] planeEquation){
		//  make sure that both pointers are non-null
		if((point == null) || (planeEquation == null)){
			System.err.println("Error: both pointers need to be non-null to determine the base point of the perpendicular.\nSystem.exit(0)");
			System.exit(0);
		}
		double magnitudeOfPlaneNormal = this.calculateMagnitudeOfPlaneNormal(planeEquation);
		double magnitudeOfPlaneNormalSquared = magnitudeOfPlaneNormal * magnitudeOfPlaneNormal;
		double scalar = (planeEquation[0] * point.x + planeEquation[1] * point.y + planeEquation[2] * point.z + planeEquation[3]) / magnitudeOfPlaneNormalSquared;
		return new Point3d(point.x - scalar * planeEquation[0], point.y - scalar * planeEquation[1], point.z - scalar * planeEquation[2]);
	}/*  end of determineBasePointOfPerpendicular method  */
	
	
	
	
	/*  start of calculateMagnitudeOfPlaneNormal method  */
	/**
	 * @param planeEquation the equation of the plane for which to calculate the magnitude of the plane normal
	 */
	private double calculateMagnitudeOfPlaneNormal(double[] planeEquation){
		//  make sure that the pointer is non-null
		if(planeEquation == null){
			System.err.println("Error: the pointer must be non-null to calculate the magnitude of the plane normal.\nSystem.exit(0)");
			System.exit(0);
		}
		double magnitude = planeEquation[0] * planeEquation[0] + planeEquation[1] * planeEquation[1] + planeEquation[2] * planeEquation[2];
		return Math.sqrt(magnitude);
	}/*  end of calculateMagnitudeOfPlaneNormal method  */
	
	
	
	
	/*  start of determineVectorFromPointAtoPointB method  */
	/**
	 * @param pointA the origin point from which to start the vector (AB)
	 * @param pointB the destination point where the head of the vector ends (AB)
	 */
	private double[] determineVectorFromPointAtoPointB(Point3d pointA, Point3d pointB){
		//  make sure that both points are valid
		if((pointA == null) || (pointB == null)){
			System.err.println("Error: both points must be non-null to determine the vector between them.\nSystem.exit(0)");
			System.exit(0);
		}
		double[] vector = new double[3];
		vector[0] = pointB.x - pointA.x;
		vector[1] = pointB.y - pointA.y;
		vector[2] = pointB.z - pointA.z;
		return vector;
	}/*  end of determineVectorFromPoint1toPoint2 method  */
	
	
	
	
	/*  start of applyScalarToVector method  */
	/**
	 * @param scalar the scalar value to apply to the vector (multiplication)
	 * @param vector the vector to which the scalar is applied (multiplied)
	 */
	private void applyScalarToVector(double scalar, double[] vector){
		//  make sure the pointer is valid
		if(vector == null){
			System.err.println("Error: the vector pointer must be non-null to apply a scalar to it.\nSystem.exit(0)");
			System.exit(0);
		}
		vector[0] *= scalar;
		vector[1] *= scalar;
		vector[2] *= scalar;
	}/*  end of applyScalarToVector method  */
	
	
	
	
	/*  start of choosePointAtRandom method  */
	/**
	 * @param cubeSideLength the length of the cube from within to choose a point at random (the cube is centered at the origin (0,0,0))
	 */
	private Point3d choosePointAtRandom(int cubeSideLength){
		//  make sure that the cube side length is at least one
		if(cubeSideLength < 1){
			System.err.println("Error: the cube side length must be at least one (1).\nSystem.exit(0)");
			System.exit(0);
		}
		double halfCubeSideLength = cubeSideLength / 2.0;
		return new Point3d(random.nextInt(cubeSideLength)-halfCubeSideLength, random.nextInt(cubeSideLength)-halfCubeSideLength, random.nextInt(cubeSideLength)-halfCubeSideLength);
	}/*  end of choosePointAtRandom method  */
	
	
	
	
	/*  start of mirrorSegment method  */
	/**
	 * @return void
	 */
	private void mirrorSegment(){
		
		
		//  choose two points in the structure at random
		int pointIndexA = random.nextInt(structure.length);
		int pointIndexB = random.nextInt(structure.length);
		//  make sure that there is at least one point between the two end points and the two points are unique
		while((Math.abs(pointIndexA-pointIndexB) < 2) || (!this.testPointsDifferent(structure[pointIndexA], structure[pointIndexB]))){
			pointIndexB = random.nextInt(structure.length);
		}
		//  make pointIndexA the smaller of the two indices
		if(pointIndexA > pointIndexB){
			int tempPointIndex = pointIndexB;
			pointIndexB = pointIndexA;
			pointIndexA = tempPointIndex;
		}
		
		
		
		//  determine the equation of the plane passing through the two points chosen in the structure and a randomly chosen third point
		Point3d pointA = new Point3d(structure[pointIndexA]);
		Point3d pointB = new Point3d(structure[pointIndexB]);
		//  choose a third point at random (needs to be different than the first two points)		
		Point3d randomPointC = this.choosePointAtRandom(10);
		while((!this.testPointsDifferent(pointA, randomPointC)) || (!this.testPointsDifferent(pointB, randomPointC))){
			randomPointC = this.choosePointAtRandom(10);
		}
		double[] planeEquation = this.determinePlaneFromThreePoints(pointA, pointB, randomPointC);
		
		
		//  test case, should give plane defined by 58x + 20y - 26z  + 62 = 0
//		planeEquation = this.determinePlaneFromThreePoints(new Point3d(1,-6,0), new Point3d(-4,2,-5), new Point3d(-2,4,1));
//		System.out.println(planeEquation[0] + "x\t" + planeEquation[1] + "y\t" + planeEquation[2] + "z\t" + planeEquation[3] + "\t=0");
		
		boolean[] pointMoved = new boolean[structure.length];
		
		double[][] candidateStructure = new double[structure.length][3];
		for(int pointIndex=0; pointIndex<structure.length; pointIndex++){
			if((pointIndex >= pointIndexA+1) && (pointIndex < pointIndexB)){
				Point3d mirroredPoint = new Point3d(structure[pointIndex]);
				this.mirrorPointOverPlane(mirroredPoint, planeEquation);
				candidateStructure[pointIndex][0] = mirroredPoint.x;
				candidateStructure[pointIndex][1] = mirroredPoint.y;
				candidateStructure[pointIndex][2] = mirroredPoint.z;
				pointMoved[pointIndex] = true;
			}
			else{
				candidateStructure[pointIndex][0] = structure[pointIndex][0];
				candidateStructure[pointIndex][1] = structure[pointIndex][1];
				candidateStructure[pointIndex][2] = structure[pointIndex][2];
				pointMoved[pointIndex] = false;
			}			
		}

		
		this.testCandidateStructureAndUpdateIfPassAcceptanceTest(candidateStructure, pointMoved, 1);
	}/*  end of mirrorSegment method  */
	
	
	
	
	/*  start of mirrorEnd method  */
	/**
	 * @return void
	 */
	private void mirrorEnd(){

		//  choose one point in the structure at random
		int pointIndexA = random.nextInt(structure.length-2)+1;
		
		
		//  determine the equation of the plane passing through the chosen point and two randomly chosen points
		Point3d pointA = new Point3d(structure[pointIndexA]);
		//  choose a second and third point at random (need to be different than the first point)		
		Point3d randomPointB = this.choosePointAtRandom(10);
		while(!this.testPointsDifferent(pointA, randomPointB)){
			randomPointB = this.choosePointAtRandom(10);
		}
		Point3d randomPointC = this.choosePointAtRandom(10);
		while((!this.testPointsDifferent(pointA, randomPointC)) || (!this.testPointsDifferent(randomPointB, randomPointC))){
			randomPointC = this.choosePointAtRandom(10);
		}
		
		double[] planeEquation = this.determinePlaneFromThreePoints(pointA, randomPointB, randomPointC);
				
		
		boolean[] pointMoved = new boolean[structure.length];
		
		double[][] candidateStructure = new double[structure.length][3];
		for(int pointIndex=0; pointIndex<structure.length; pointIndex++){
			if(pointIndex < pointIndexA){
				Point3d mirroredPoint = new Point3d(structure[pointIndex]);
				this.mirrorPointOverPlane(mirroredPoint, planeEquation);
				candidateStructure[pointIndex][0] = mirroredPoint.x;
				candidateStructure[pointIndex][1] = mirroredPoint.y;
				candidateStructure[pointIndex][2] = mirroredPoint.z;
				pointMoved[pointIndex] = true;
			}
			else{
				candidateStructure[pointIndex][0] = structure[pointIndex][0];
				candidateStructure[pointIndex][1] = structure[pointIndex][1];
				candidateStructure[pointIndex][2] = structure[pointIndex][2];
				pointMoved[pointIndex] = false;
			}			
		}

		
		this.testCandidateStructureAndUpdateIfPassAcceptanceTest(candidateStructure, pointMoved, 2);
	}/*  end of mirrorEnd method  */
	
	
	
	
	/*  start of calculateProbabilityOfCandidateStructureOptimized method  */
	/**
	 * @param candidateStructure the candidate structure for which to calculate the probability
	 * @param pointMoved an array of boolean indicating which points have been moved in the candidate structure (true=moved; false=not moved from current structure)
	 */
	private double calculateProbabilityOfCandidateStructureOptimized(double[][] candidateStructure, boolean[] pointMoved){

		//  calculate the probability of the structure
		double probability = probabilityCurrent;
		
		for(int i=0; i<structure.length; i++){
			if(pointMoved[i]){
				for(int j=0; j<structure.length; j++){
					if(j != i){
						if(IFvalues[i][j] != -1.0){
							Point3d pointICurrent = new Point3d(structure[i]);
							Point3d pointINew = new Point3d(candidateStructure[i]);
							
							Point3d pointJCurrent = new Point3d(structure[j]);
							Point3d pointJNew = new Point3d(structure[j]);
							//  if the point j has not been moved then the new coordinate is the same as the current coordinate
							if(pointMoved[j]){
								pointJNew = new Point3d(candidateStructure[j]);
							}
							
							//  calculate the distance in the current structure between the point to move (current location) and point j
							double distanceCurrent = this.calculateDistanceBetweenPoints(pointICurrent, pointJCurrent);
							//  make sure not to divide by zero
							if(distanceCurrent <= 0){
								System.err.println("Error: the distance between the points was not properly determined\nSystem.exit(0)");
								System.exit(0);
							}
							double IFmodelCurrent = 1.0 / Math.pow(distanceCurrent, exponent);

					
					
							//  calculate the distance in the new structure between the point to move (new location) and point j
							double distanceNew = this.calculateDistanceBetweenPoints(pointINew, pointJNew);
							//  make sure not to divide by zero
							if(distanceNew <= 0){
								System.err.println("Error: the distance between the points was not properly determined\nSystem.exit(0)");
								System.exit(0);
							}
							double IFmodelNew = 1.0 / Math.pow(distanceNew, exponent);
					
					
							double varianceTerm = -1.0 / (2.0 * IFvariances[i][j]);
							double IFvalueExperimental = IFvalues[i][j];
						
							//  first calculate the term corresponding to the old model
							double IFdifferenceOldSquared = IFmodelCurrent - IFvalueExperimental;
							IFdifferenceOldSquared *= IFdifferenceOldSquared;

							//  second calculate the term corresponding to the new model
							double IFdifferenceNewSquared = IFmodelNew - IFvalueExperimental;
							IFdifferenceNewSquared *= IFdifferenceNewSquared;
					
							//  third calculate the difference between the old model and the new model
							double IFnetDifference = IFdifferenceNewSquared - IFdifferenceOldSquared;
					
							//  calculate the probability of the new structure
							probability += varianceTerm * IFnetDifference;
						}
					}
				}
			}
		}
		return probability;
	}/*  end of calculateProbabilityOfCandidateStructureOptimized method  */
	
	
	
	
	/*  start of runMCMC method  */
	/**
	 * @param bwParameters the buffered write used to output the parameters of the current run to file
	 * @throws IOException throws an error if there is a problem with the buffered writer
	 */
	private void runMCMC(BufferedWriter bwParameters) throws IOException{

		
		
		try{
			BufferedWriter bwStats = new BufferedWriter(new FileWriter(outputPath + "stats_" + filenameIFs));

			
			//  write the header line of the stats file
			bwStats.write(outputFolder);
			bwStats.flush();
			bwStats.newLine();
			bwStats.write("command line args=\t");
			bwStats.flush();
			for(String arg: argsGlobal){
				bwStats.write(arg + "\t");
				bwStats.flush();
			}
			bwStats.newLine();
			bwStats.flush();
			bwStats.write("structureIndex\tCurrentProbability\tBestProbability");
			bwStats.newLine();
			bwStats.flush();
			

			
			//  output the genomic position map
			outputGenomicPositionMap();
			
			
			
			//  initialize the IFutility for the current run
			initializeIFutility();
			
			
			
			//  get the IF and variance values from file
			getIFsAndVariancesFromFile();


			
			
			//  calculate and set the half step size
			halfStepSize = stepSize / 2.0;
			
			
			
			//  initialize the structure
			initializeStructure(bwParameters);
			
			
			
			//  output the initial structure
			outputStructure(structure, 0, probabilityCurrent, "structure_0000.pdb");
			bwStats.write(0 + "\t" + probabilityCurrent);
			bwStats.newLine();
			bwStats.flush();

			
			
			//  keep track of the best structure ever seen and set the initial structure as the best structure seen
			bestStructureFilename = "best_structure_" + filenameIFs.substring(0, filenameIFs.lastIndexOf('.')) + ".pdb";
			bestStructure = new double[structure.length][3];
			for(int i=0; i<structure.length; i++){
				bestStructure[i][0] = structure[i][0];
				bestStructure[i][1] = structure[i][1];
				bestStructure[i][2] = structure[i][2];
			}
			probabilityOfBestStructure = probabilityCurrent;
			indexOfBestStructure = 0;
			
			
			
			//  use a separate counter so the structures that are output have consecutive numbers in their filenames
			int structureFilenameCounter = 1;
			
			
			
			//  run the Markov chain Monte Carlo simulation using the Metropolis-Hastings algorithm
			for(long counter=1; counter<=runLength; counter++){

				
				double moveType = random.nextDouble();
				
				if(moveType < single_point_translation_move_fraction){
					this.translateSinglePoint();
					numberOfChangesAttempted[0]++;
				}
				else if(moveType < (single_point_translation_move_fraction + block_translation_move_fraction)){
					this.translateBlock();
					numberOfChangesAttempted[3]++;
				}
				else{
					if(random.nextInt(2) < 1){
						this.mirrorSegment();
						numberOfChangesAttempted[1]++;
					}
					else{
						this.mirrorEnd();
						numberOfChangesAttempted[2]++;
					}
				}
				
				
				
				//  check if the new structure is the best structure ever seen
				if(probabilityCurrent > probabilityOfBestStructure){
					probabilityOfBestStructure = probabilityCurrent;
					indexOfBestStructure = counter;

					//  update the coordinates of the best structure
					for(int i=0; i<structure.length; i++){
						bestStructure[i][0] = structure[i][0];
						bestStructure[i][1] = structure[i][1];
						bestStructure[i][2] = structure[i][2];
					}
				}					
				
				
				
				//  output the structure and statistics to file
				if((counter % (runLength / numberOfStructuresToOutput)) == 0){				

					
					//  update the probability of the current and best structures using the IFutility
					//        (corrects for numerical imprecision caused by using such small probability values)
					probabilityCurrent = utility_IF.calculateProbability(structure);
					probabilityOfBestStructure = utility_IF.calculateProbability(bestStructure);
					
					
					//  output the current structure
					outputStructure(structure, structureFilenameCounter, probabilityCurrent, "default");
					bwStats.write(structureFilenameCounter + "\t" + probabilityCurrent + "\t" + probabilityOfBestStructure);
					bwStats.flush();
					bwStats.newLine();
					structureFilenameCounter++;
				
					
					//  output and overwrite the best structure seen so far
					outputBestStructure();
				}
			}//  end of Markov chain Monte Carlo loop
			
			
			
			//  output the best structure seen during the run
			outputBestStructure();
			bwParameters.write("bestStructureProbability=\t" + probabilityOfBestStructure);
			bwParameters.newLine();
			bwParameters.flush();
			
			
			
			bwStats.close();
			
						
			
		}catch (IOException e) {
			System.out.println("Error: could not output file: " + outputPath + "stats_" + filenameIFs);
			e.printStackTrace();
		}//end of try-catch block
		
		if(verbose){
			System.out.println("\tOutput file: " + outputPath + "stats_" + filenameIFs);
		}
	}/*  end of runMCMC method  */


	
	
	/*  start of run method  */
	/**
	 * @param args the parameters for the MCMC5C run
	 */
	public void run(String[] args) {
		//  make the command line arguments accessible by all methods as a global variable
		argsGlobal = args;
		
		switch (args.length) {
		case 12:
			filenameIFs = args[0];
			filenameFragments = args[1];
			runLength = Long.parseLong(args[2]);
			numberOfStructuresToOutput = Double.parseDouble(args[3]);
			stepSize = Double.parseDouble(args[4]);
			exponent = Double.parseDouble(args[5]);
			initialStructureCase = Integer.parseInt(args[6]);
			initialStructureFilename = args[7];
			single_point_translation_move_fraction = Double.parseDouble(args[8]);
			block_translation_move_fraction = Double.parseDouble(args[9]);
			mirror_move_fraction = Double.parseDouble(args[10]);
			maximum_neighbour_distance = Double.parseDouble(args[11]);
			break;
		default:
			System.err.println("Error: there need to be exactly twelve (12) command line parameters.\n" +
					"\t\tIFs_filename.txt\n" +
					"\t\tfragments_filename.txt\n" +
					"\t\trun_length\n" +
					"\t\tnumber_of_structures_to_output\n" +
					"\t\tstep_size\n" +
					"\t\texponent\n" +
					"\t\tinitial_structure_case\n" +
					"\t\tinitial_structure_filename\n" +
					"\t\tsingle_point_translation_move_fraction\n" +
					"\t\tblock_translation_move_fraction\n" +
					"\t\tmirror_move_fraction\n" +
					"\t\tmaximum_neighbour_distance\n" +
					"System.exit(0)");
			System.exit(0);
			break;
		}

		
		
		if((single_point_translation_move_fraction + block_translation_move_fraction + mirror_move_fraction) != 1){
			System.err.println("Error: the probabilities of each type of move must add up to one.\nSystem.exit(0)");
			System.err.println("\t\tsingle_point_translation_move_fraction = " + (int)(single_point_translation_move_fraction*100) + "%");
			System.err.println("\t\tblock_translation_move_fraction = " + (int)(block_translation_move_fraction*100) + "%");
			System.err.println("\t\tmirror_move_fraction = " + (int)(mirror_move_fraction*100) + "%");
			System.exit(0);
		}
		
		
		
		//  create the output folder (use the experiment name as found in the IFs file)
		String experimentName = filenameIFs.substring(0, filenameIFs.lastIndexOf("."));
		if(experimentName.startsWith("IFs_")){
			experimentName = experimentName.substring(4, experimentName.length());
		}
		this.outputFolder += "_[" + experimentName + "]";
		new File(outputFolder).mkdir();
		
		
		
		//  update the output path
		this.outputPath = this.outputFolder + "/";
		
		
		
		//  calculate how long running the MCMC takes (time point before run)
		long timeBefore = System.currentTimeMillis();


		
		//  output the parameters to file
		try{
			String parameterFilename = "parameters_" + filenameIFs;
			BufferedWriter bwParameters = new BufferedWriter(new FileWriter(outputPath + parameterFilename));
			
			bwParameters.write(outputPath);
			bwParameters.newLine();
			bwParameters.flush();
			bwParameters.write("filenameIFs=\t" + filenameIFs);
			bwParameters.newLine();
			bwParameters.flush();
			bwParameters.write("filenameFragments=\t" + filenameFragments);
			bwParameters.newLine();
			bwParameters.flush();
			bwParameters.write("runLength=\t" + runLength);
			bwParameters.newLine();
			bwParameters.flush();
			bwParameters.write("numberOfStructuresToOutput=\t" + (int)numberOfStructuresToOutput);
			bwParameters.newLine();
			bwParameters.flush();
			bwParameters.write("stepSize=\t" + stepSize);
			bwParameters.newLine();
			bwParameters.flush();
			bwParameters.write("exponent=\t" + exponent);
			bwParameters.newLine();
			bwParameters.flush();
			bwParameters.write("initialStructureCase=\t" + initialStructureCase);
			bwParameters.newLine();
			bwParameters.flush();
			bwParameters.write("initialStructureFilename=\t" + initialStructureFilename);
			bwParameters.newLine();
			bwParameters.flush();
			bwParameters.write("single_point_translation_move_fraction=\t" + single_point_translation_move_fraction);
			bwParameters.newLine();
			bwParameters.flush();
			bwParameters.write("block_translation_move_fraction=\t" + block_translation_move_fraction);
			bwParameters.newLine();
			bwParameters.flush();
			bwParameters.write("mirror_move_fraction=\t" + mirror_move_fraction);
			bwParameters.newLine();
			bwParameters.flush();
			bwParameters.write("maximum neighbour distance=\t" + maximum_neighbour_distance);
			bwParameters.newLine();
			bwParameters.flush();
			bwParameters.write("command line parameters:");
			if(args.length == 0){
				bwParameters.write("\tnone");
			}
			else{
				for(String arg: args){
					bwParameters.write("\t" + arg);
				}
			}
			bwParameters.newLine();
			bwParameters.flush();
			
			
			
			
			//  run the MCMC simulation
			runMCMC(bwParameters);
			
			
			
			
			//  output the time taken by the MCMC run
			long timeTaken = System.currentTimeMillis() - timeBefore;
			timeTaken = timeTaken / (long)(1000.0);			//	convert from milliseconds to seconds		
			if(timeTaken > 3600){		//	(1hr = 3600 sec)
				int numberOfHours = (int)timeTaken / (int)3600;
				long leftOver = timeTaken % 3600;
				int numberOfMinutes = (int)leftOver / (int)60;
				leftOver = leftOver % 60;
				int numberOfSeconds = (int)leftOver;
				bwParameters.write("Amount of time taken = " + numberOfHours + "hr" + numberOfMinutes + "min" + numberOfSeconds + "sec");
				bwParameters.newLine();
				bwParameters.flush();
				if(verbose){
					System.out.println("Amount of time taken = " + numberOfHours + "hr" + numberOfMinutes + "min" + numberOfSeconds + "sec");				
				}
			}
			else if(timeTaken > 60){		//	(1min = 60 sec)
				int numberOfMinutes = (int)timeTaken / (int)60;
				long leftOver = timeTaken % 60;
				int numberOfSeconds = (int)leftOver;
				bwParameters.write("Amount of time taken = " + numberOfMinutes + "min" + numberOfSeconds + "sec");
				bwParameters.newLine();
				bwParameters.flush();
				if(verbose){
					System.out.println("Amount of time taken = " + numberOfMinutes + "min" + numberOfSeconds + "sec");
				}
			}
			else{
				int numberOfSeconds = (int)timeTaken;
				bwParameters.write("Amount of time taken = " + numberOfSeconds + "sec");
				bwParameters.newLine();
				bwParameters.flush();
				if(verbose){
					System.out.println("Amount of time taken = " + numberOfSeconds + "sec");
				}
			}
			
			
			//  output the number of proposed changes accepted during the run for each type of attempted move
			if(numberOfChangesAttempted[0] > 0){
				bwParameters.write("Number of single point translations accepted/attempted = " + numberOfChangesAccepted[0] + "/" + numberOfChangesAttempted[0]);
				bwParameters.write(" [" + formatter2decimals.format(numberOfChangesAccepted[0] * 100.0 / numberOfChangesAttempted[0]) + "%]");
				bwParameters.newLine();
				bwParameters.flush();
			}
			if(numberOfChangesAttempted[1] > 0){
				bwParameters.write("Number of mirror segment moves accepted/attempted = " + numberOfChangesAccepted[1] + "/" + numberOfChangesAttempted[1]);
				bwParameters.write(" [" + formatter2decimals.format(numberOfChangesAccepted[1] * 100.0 / numberOfChangesAttempted[1]) + "%]");
				bwParameters.newLine();
				bwParameters.flush();
			}
			if(numberOfChangesAttempted[2] > 0){
				bwParameters.write("Number of mirror end moves accepted/attempted = " + numberOfChangesAccepted[2] + "/" + numberOfChangesAttempted[2]);
				bwParameters.write(" [" + formatter2decimals.format(numberOfChangesAccepted[2] * 100.0 / numberOfChangesAttempted[2]) + "%]");
				bwParameters.newLine();
				bwParameters.flush();
			}
			if(numberOfChangesAttempted[3] > 0){
				bwParameters.write("Number of block translation moves accepted/attempted = " + numberOfChangesAccepted[3] + "/" + numberOfChangesAttempted[3]);
				bwParameters.write(" [" + formatter2decimals.format(numberOfChangesAccepted[3] * 100.0 / numberOfChangesAttempted[3]) + "%]");
				bwParameters.newLine();
				bwParameters.flush();
			}
			bwParameters.write("Type of move\t\tFraction of attempted moves");
			bwParameters.newLine();
			bwParameters.flush();
			bwParameters.write("Single point\t\t\t" + formatter2decimals.format(numberOfChangesAttempted[0] * 100.0 / runLength) + "%");
			bwParameters.newLine();
			bwParameters.flush();
			bwParameters.write("Mirror segment\t\t\t" + formatter2decimals.format(numberOfChangesAttempted[1] * 100.0 / runLength) + "%");
			bwParameters.newLine();
			bwParameters.flush();
			bwParameters.write("Mirror end\t\t\t" + formatter2decimals.format(numberOfChangesAttempted[2] * 100.0 / runLength) + "%");
			bwParameters.newLine();
			bwParameters.flush();
			bwParameters.write("Block translation\t\t" + formatter2decimals.format(numberOfChangesAttempted[3] * 100.0 / runLength) + "%");
			bwParameters.newLine();
			bwParameters.flush();
			
			bwParameters.close();
			
			if(verbose){
				System.out.println("\tOutput file: " + outputPath + parameterFilename);
			}
		}catch (IOException e) {
			e.printStackTrace();
		}//end of try-catch block
		
	}/*  end of run method  */
	
	
}/*  end of MCMC5C class  */