package utilities;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;


/** ------------------------------------------------------------------------------------------------------------------
 * start of MedianNormalizationToTargetValues class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class MedianNormalizationToTargetValues {

	
	
	
	/*  global parameters  */
	private static boolean verbose = true;
	private static String inputPath = "";
	private static String outputPath = "";
	/*  end of global parameters  */
	
	
	

	/*  start of getFilenames method  */
	/**
	 * @param filename_experiment_list the file containing the list of experiment matched filenames (normalization region and experimental region)
	 * @return a list of paired filenames for matched Gene Desert--Genomic Region files
	 */
	private static ArrayList<String[]> getFilenames(String filename_experiment_list){
		if(filename_experiment_list == null){
			System.err.println("Error: the filename list is null.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<String[]> experimentList = new ArrayList<String[]>();

		String line = null;
		int lineCounter = -1;			//  there is one header line
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename_experiment_list));
			while((line = br.readLine()) != null){
				if(lineCounter >= 0){
					String[] values = line.split("\t");
					if(values.length != 2){
						System.err.println("Error: there are not exactly two (2) entries in each row of the input experiment list file.\nSystem.exit(0)");
						System.exit(0);
					}
					experimentList.add(values);
				}
				else{
					//  skip and the header line
//					System.out.println(line);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		experimentList.trimToSize();
		
		if(verbose){
			System.out.println("Loaded input file: " + inputPath + filename_experiment_list);
		}
		
		return experimentList;
	}/*  end of getFilenames method  */
	
	
	
	
	/*  start of displayFilenames method  */
	/**
	 * @param experimentList a list of matched Gene Desert--Genomic Region filename pairs to display
	 */
	private static void displayFilenames(ArrayList<String[]> experimentList){
		if(experimentList == null){
			System.err.println("Error: the experiment list is null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);	
		}
		
		System.out.println("Gene Desert Region\t\tExperimental Region");
		for(String[] currentPair: experimentList){
			System.out.println(currentPair[0] + "\t\t" + currentPair[1]);
		}
	}/*  end of displayFilenames method  */
	
	
	
	
	/*  start of loadInputData method  */
	/**
	 * @param experimentList a list of matched Gene Desert--Genomic Region filename pairs to load from file
	 * @return a paired list containing the matched lists of Gene Desert experiment data (first) and Genomic Region experiment data (second)
	 */
	private static ArrayList<ArrayList<FiveC>> loadInputData(ArrayList<String[]> experimentList){
		if(experimentList == null){
			System.err.println("Error: the experiment list is null and the data cannot be loaded.\nSystem.exit(0)");
			System.exit(0);	
		}
		
		ArrayList<FiveC> FiveC_GeneDeserts = new ArrayList<FiveC>();
		ArrayList<FiveC> FiveC_ExperimentalRegions = new ArrayList<FiveC>();
		
		for(String[] filenamePair: experimentList){
			FiveC_GeneDeserts.add(new FiveC(filenamePair[0], inputPath, outputPath));
			FiveC_ExperimentalRegions.add(new FiveC(filenamePair[1], inputPath, outputPath));
		}
		
		FiveC_GeneDeserts.trimToSize();
		FiveC_ExperimentalRegions.trimToSize();
		
		ArrayList<ArrayList<FiveC>> returnArray = new ArrayList<ArrayList<FiveC>>();
		returnArray.add(FiveC_GeneDeserts);
		returnArray.add(FiveC_ExperimentalRegions);
		returnArray.trimToSize();
		
		return returnArray;
	}/*  end of loadInputData method  */
	
	
	
	
	/*  start of displayData method  */
	/**
	 * @param dataToDisplay the list of FiveC objects to display
	 * @param title a String to display as the title for the data being displayed
	 */
	private static void displayData(ArrayList<FiveC> dataToDisplay, String title){
		if(dataToDisplay == null){
			System.err.println("The data list is null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		if(title != null){
			System.out.println("\n\n\t" + title + ":");
		}
		for(FiveC currentFiveC: dataToDisplay){
			System.out.println(currentFiveC);
		}
	}/*  end of displayData method  */
	
	
	
	
	/*  start of loadTargetValues method  */
	/**
	 * @param filename_targetValues the name of the file containing the target values
	 * @return an array containing the loaded target values
	 */
	private static double[] loadTargetValues(String filename_targetValues){
		if(filename_targetValues == null){
			System.err.println("Error: the target values filename is null and the values cannot be loaded.\nSystem.exit(0)");
			System.exit(0);
		}
		
		double targetValues[] = new double[8];
		

		String line = null;
		int lineCounter = -1;			//  there is one header line
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename_targetValues));
			while((line = br.readLine()) != null){
				if(lineCounter == 0){
					String[] values = line.split("\t");
					if(values.length != 8){
						System.err.println("Error: there are not exactly eight (8) entries in the target values file.\nSystem.exit(0)");
						System.exit(0);
					}
					for(int i=0; i<8; i++){
						targetValues[i] = Double.parseDouble(values[i]);
					}
				}
				else{
					//  skip and the header line
//					System.out.println(line);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		
		if(verbose){
			System.out.println("Loaded input file: " + inputPath + filename_targetValues);
		}
		
		return targetValues;
	}/*  end of loadTargetValues method  */
	
	
	
	
	/*  start of displayTargetValues method  */
	/**
	 * @param targetValues the target values to display
	 */
	private static void displayTargetValues(double[] targetValues){
		if(targetValues == null){
			System.err.println("Error: the target values are null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		System.out.println();
		System.out.println("##########  Target values  ##########");
		System.out.println("\t15bp\t18bp\t19bp\t20bp\t21bp\t22bp\t23bp\t24bp");
		for(int i=0; i<targetValues.length; i++){
			System.out.print("\t" + targetValues[i]);
		}
		System.out.println();
		System.out.println("#####################################");
	}/*  end of displayTargetValues method  */
	
	
	
	
	/*  start of calculateNormalizationFactors method  */
	/**
	 * @param a list of the FiveC objects containing the Gene Desert data from which to calculate the normalization factors
	 * @param targetValues an array of the target values from which to calculate the normalization factors
	 * @return an ArrayList<double[]> containing the per-probe length normalization factors for each dataset
	 */
	private static ArrayList<double[]> calculateNormalizationFactors(ArrayList<FiveC> FiveC_GeneDeserts, double[] targetValues){
		if(FiveC_GeneDeserts == null){
			System.err.println("The input gene desert data arrays are null and the normalization factors cannot be determined.\nSystem.exit(0)");
			System.exit(0);
		}
//		
		
		ArrayList<double[]> normalizationFactors = new ArrayList<double[]>();
		
		
		//  making all of the Gene Desert averages the same
		for(FiveC currentGeneDesert: FiveC_GeneDeserts){
			
			double[] currentAverages = currentGeneDesert.getAverageValues();
			
			double[] currentNormalizationFactors = new double[8];
			
			for(int i=0; i<targetValues.length; i++){
				
				if(currentAverages[i] == 0){
					currentNormalizationFactors[i] = 1;
				}
				else{
					currentNormalizationFactors[i] = (double)targetValues[i] / (double)currentAverages[i];
				}
			}
			normalizationFactors.add(currentNormalizationFactors);
		}
		normalizationFactors.trimToSize();
		
		
		return normalizationFactors;
	}/*  end of calculateNormalizationFactors method  */
	
	
	
	
	/*  start of applyNormalizationFactors method  */
	/**
	 * @param normalizationFactors a list of double arrays, each containing the per-probe length normalization factors for each dataset
	 * @param FiveC_GeneDeserts a list of FiveC objects for the gene desert regions that will be normalized using the specified normalization factors
	 * @param FiveC_ExperimentalRegions a list of FiveC objects for the genomic regions that will be normalized using the specified normalization factors
	 * @return void
	 */
	private static void applyNormalizationFactors(ArrayList<double[]> normalizationFactors, ArrayList<FiveC> FiveC_GeneDeserts, ArrayList<FiveC> FiveC_ExperimentalRegions){
		if(normalizationFactors == null){
			System.err.println("Error: the normalization factors are null and cannot be applied.\nSystem.exit(0)");
			System.exit(0);
		}
		
		//  make sure that there are a matching number of experiments
		if(FiveC_GeneDeserts.size() != FiveC_ExperimentalRegions.size()){
			System.err.println("Error: there are not a matching number of gene desert and experimental region 5C experiments.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		for(int i=0; i<FiveC_GeneDeserts.size(); i++){
			double[] currentNormalizationArray = normalizationFactors.get(i);
			
			FiveC_GeneDeserts.get(i).normalize(currentNormalizationArray);
			FiveC_ExperimentalRegions.get(i).normalize(currentNormalizationArray);
		}
	}/*  end of applyNormalizationFactors method  */
		
	
	
	
	/*  start of outputDataToFile method  */
	/**
	 * @param experimentList the list of matched Gene Desert--Genomic Region filename pairs
	 * @param FiveC_GeneDeserts the list of FiveC objects for the gene desert regions
	 * @param FiveC_ExperimentalRegions the list of FiveC objects for the genomic regions
	 * @param normalizationFactors the list of double arrays, each containing the per-probe length normalization factors for each dataset
	 * @param targetValues the array of target values (average of medians) that were used as the target for the average gene desert signal in each experiment
	 * @return void
	 */
	private static void outputDataToFile(ArrayList<String[]> experimentList, ArrayList<FiveC> FiveC_GeneDeserts, ArrayList<FiveC> FiveC_ExperimentalRegions, ArrayList<double[]> normalizationFactors, double[] targetValues){
		if(normalizationFactors == null){
			System.err.println("Error: the normalization factors are null and cannot be output to file.\nSystem.exit(0)");
			System.exit(0);
		}
		

		String filename_output = "normalization_factors.txt";
		
		try{
			BufferedWriter bw = new BufferedWriter(new FileWriter(outputPath + filename_output));

			//  write the header lines
			bw.write("Experiment name\t15bp\t18bp\t19bp\t20bp\t21bp\t22bp\t23bp\t24bp");
			bw.newLine();
			bw.flush();
			
			bw.write("Target values");
			for(int i=0; i<targetValues.length; i++){
				bw.write("\t" + targetValues[i]);
			}
			bw.newLine();
			bw.flush();
			
			for(int i=0; i<normalizationFactors.size(); i++){
				bw.write(experimentList.get(i)[0] + "\t");
				
				double[] currentNormalizationFactors = normalizationFactors.get(i);
				
				for(int j=0; j<currentNormalizationFactors.length; j++){
					if(j == (currentNormalizationFactors.length-1)){
						bw.write(String.valueOf(currentNormalizationFactors[j]));
					}
					else{
						bw.write(currentNormalizationFactors[j] + "\t");
					}
				}
				bw.newLine();
				bw.flush();
			}
			bw.flush();
			bw.close();
		}catch (IOException e) {
			e.printStackTrace();
		}//end of try-catch block
		
		if(verbose){
			System.out.println("\nOutput file: " + outputPath + filename_output);
		}
		
		
		for(FiveC currentFiveC: FiveC_GeneDeserts){
			currentFiveC.outputToFile(outputPath + currentFiveC.getFilename());
		}
		for(FiveC currentFiveC: FiveC_ExperimentalRegions){
			currentFiveC.outputToFile(outputPath + currentFiveC.getFilename());
		}
	}/*  end of outputDataToFile method  */
	
	
	
	
	/*  start of main method  */
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		
		
		
		boolean usingCommandLine = false;
		
		String filename_targetValues = "default.txt";
		String filename_experiment_list = "default.txt";
		ArrayList<String[]> experimentList = new ArrayList<String[]>();
		
		
		//  load the parameters from the command line
		if(args.length >= 3){
			filename_targetValues = args[0];
			
			//  there must be a matched number of Gene Desert--Genomic Region files
			if((args.length % 2) != 1){
				System.err.println("Error: there needs to be a matched number of Gene Desert--Genomic Regions filename pairs.\nSystem.exit(0)");
				System.exit(0);
			}
			for(int i=1; i<args.length; i=i+2){
				String[] currentFilenamePair = new String[2];
				currentFilenamePair[0] = args[i];
				currentFilenamePair[1] = args[i+1];
				experimentList.add(currentFilenamePair);
			}
			experimentList.trimToSize();
			
			usingCommandLine = true;
			verbose = false;
		}
		else{
			System.err.println("Error: there need to be at least three (3) command line parameters.\n" +
					"\t\ttargetValues.txt\n" +
					"\t\texperiment_1_gene_desert.txt\n" +
					"\t\texperiment_1_genomic_region.txt\n" +
					"\t\texperiment_2_gene_desert.txt\n" +
					"\t\texperiment_2_genomic_region.txt\n" +
					"\t\texperiment_..._gene_desert.txt\n" +
					"\t\texperiment_..._genomic_region.txt\n" +
					"System.exit(0)");
//			System.exit(0);
		}
		
		
		
		
		/******************************************************
		 * 													  *
		 * 				SET THE PARAMETERS HERE				  *
		 * 													  *
		 ******************************************************/
		
		
		if(!usingCommandLine){
			
			
			inputPath = "MedianNormalizationToTargetValues/Input/";
			outputPath = "MedianNormalizationToTargetValues/Output/";
		
			
			filename_targetValues = "target_values.txt";
			
			filename_experiment_list = "experiment_list.txt";
		
			
		}
		
		
		/******************************************************
		 * 													  *
		 * 				END OF SETTING THE PARAMETERS		  *
		 * 													  * 
		 ******************************************************/
		
		
		
		
		if(!usingCommandLine){
			experimentList = getFilenames(filename_experiment_list);
		}
//		displayFilenames(experimentList);
		
		
		ArrayList<ArrayList<FiveC>> loadedDataLists = loadInputData(experimentList);
		ArrayList<FiveC> FiveC_GeneDeserts = loadedDataLists.get(0);
		ArrayList<FiveC> FiveC_ExperimentalRegions = loadedDataLists.get(1);
//		displayData(FiveC_GeneDeserts, "Gene Deserts");
//		displayData(FiveC_ExperimentalRegions, "Genomic Regions");
		
		
		double[] targetValues = loadTargetValues(filename_targetValues);
		if(verbose){
			displayTargetValues(targetValues);
		}
		
		ArrayList<double[]> normalizationFactors = calculateNormalizationFactors(FiveC_GeneDeserts, targetValues);
		
		applyNormalizationFactors(normalizationFactors, FiveC_GeneDeserts, FiveC_ExperimentalRegions);

		outputDataToFile(experimentList, FiveC_GeneDeserts, FiveC_ExperimentalRegions, normalizationFactors, targetValues);
		
		
		if(verbose){
			System.out.println("Done!");
		}
	}/*  end of main method  */
	

}/*  end of MedianNormalizationToTargetValues class  */
