package utilities;


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


/** ------------------------------------------------------------------------------------------------------------------
 * start of NormalizationGeneDesert class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class NormalizationGeneDesert {

	
	
	
	/*  global parameters  */
	private static boolean verbose = true;
	private static String inputPath = "";
	private static String outputPath = "";
	private static ArrayList<double[]> referenceSignals;
	private static ArrayList<double[]> targetSignalsGeneDesert;
	private static ArrayList<double[]> targetSignalsGenomicRegion;
	private static int count = 0;
	private static double normalizationFactor = 0;
	/*  end of global parameters  */
	
	
	
	
	/*  start of initializeParameters method  */
	/**
	 * @return void
	 */
	private static void initializeParameters(){
		count = 0;
		normalizationFactor = 0;
	}/*  end of initializeParameters method  */
	
	
	
	
	/*  start of loadSignalsFromFile method  */
	/**
	 * @param filename the input filename
	 */
	private static void loadSignalsFromFile(String filename, ArrayList<double[]> signalArrayToFill){		
		String line = null;
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename));			
			while((line = br.readLine()) != null){
				String[] values = line.split("\t");
				if(values.length < 10){
					System.out.println("Error: there is an entry in \"" + inputPath + filename + "\" that does not have ten (10) values" +
							" (i.e. index1, index2, 15bp, 18bp, 19bp, 20bp, 21bp, 22bp, 23bp, 24bp)\nSystem.exit(0)");
					System.exit(0);
				}
				double[] rowEntry = new double[10];
				for(int i=0; i<10; i++){
					rowEntry[i] = Double.parseDouble(values[i]);
				}
				signalArrayToFill.add(rowEntry);
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		signalArrayToFill.trimToSize();
		if(verbose){
			System.out.println("Loaded signals file: " + inputPath + filename);
		}
	}/*  end of loadSignalsFromFile method  */
	
	
	
	
	/*  start of displaySignals method  */
	/**
	 * @param signalsToDisplay the ArrayList of signals to display
	 */
	private static void displaySignals(ArrayList<double[]> signalsToDisplay){
		for(int i=0; i<signalsToDisplay.size(); i++){
			double[] currentRow = signalsToDisplay.get(i);
			for(int j=0; j<10; j++){
				System.out.print(currentRow[j] + "\t");
			}
			System.out.println();
		}
	}/*  end of displaySignals method  */
	
	
	
	
	/*  start of calculateNormalizationFactor method  */
	/**
	 * @return the normalization factor
	 */
	private static double calculateNormalizationFactor(){
		//  make sure that the two lists are the same size
		if(referenceSignals.size() != targetSignalsGeneDesert.size()){
			System.err.println("Error: the reference signals and target signals lists are not the same size.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		for(int i=0; i<referenceSignals.size(); i++){
			double[] referenceCurrentRow = referenceSignals.get(i);
			double[] targetCurrentRow = targetSignalsGeneDesert.get(i);
			
			//  make sure that the fragment indices are the same in both files
			if((referenceCurrentRow[0] != targetCurrentRow[0]) || (referenceCurrentRow[1] != targetCurrentRow[1])){
				System.err.println("Error: the fragment indices do not match in the reference and target files.\nSystem.exit(0)");
				System.exit(0);
			}
			
			for(int j=3; j<10; j++){
				double referenceSignal = referenceCurrentRow[j];
				double targetSignal = targetCurrentRow[j];
				
				if((referenceSignal != 0) && (targetSignal != 0)){
					double ratioReferenceToTarget = referenceSignal / targetSignal;
					normalizationFactor += Math.log10(ratioReferenceToTarget);
//					System.out.println("ratio = " + ratioReferenceToTarget);
					count++;
				}
			}
			 
		}
		
		if(count == 0){
			System.err.println("Error: there are no matching values in the reference and target datasets to use to calculate the normalization factor.\nSystem.exit(0)");
			System.exit(0);
		}
		normalizationFactor /= (double)count;
		normalizationFactor = Math.pow(10, normalizationFactor);
//		System.out.println("Normalization factor = " + normalizationFactor);
//		System.out.println("count = " + count);
		return normalizationFactor;
	}/*  end of calculateNormalizationFactor method  */
	
	
	
	
	/*  start of normalizeTargetSignals method  */
	/**
	 * @param the list of signals to be normalized (every non-zero signal value is multiplied by the normalization factor)
	 */
	private static void normalizeSignals(ArrayList<double[]> signalsToNormalize){
		for(double[] currentRow: signalsToNormalize){
			for(int i=2; i<10; i++){
				if(currentRow[i] != 0){
					currentRow[i] *= normalizationFactor;
				}
			}
		}
	}/*  end of normalizeTargetSignals method  */
	
	
	
	
	/*  start of outputNormalizedSignals method  */
	/**
	 * @param outputFilename the filename for the normalized signal output file
	 * @param signalsToOutput the list of normalized signals to output to file
	 */
	private static void outputNormalizedSignals(String outputFilename, ArrayList<double[]> signalsToOutput){
		try{			
			BufferedWriter bw = new BufferedWriter(new FileWriter(outputPath + outputFilename));

			for(double[] rowEntry: signalsToOutput){
				for(int i=0; i<10; i++){
					if(i < 2){
						bw.write((int)rowEntry[i] + "\t");
					}
					else if(i == 9){
						bw.write(String.valueOf(rowEntry[i]));
					}
					else{
						bw.write(rowEntry[i] + "\t");
					}
				}
				bw.newLine();
				bw.flush();
			}
			bw.close();
		}catch (IOException e) {
			e.printStackTrace();
		}//end of try-catch block
		
		if(verbose){
			System.out.println("\nOutput file: " + outputPath + outputFilename);
		}
	}/*  end of outputNormalizedSignals method  */
	
	
	
	
	/*  start of outputLogFile method  */
	/**
	 * @param outputFilename the filename for the output log file
	 */
	private static void outputLogFile(String outputFilename, String referenceFilename, String targetFilenameGeneDesert, String targetFilenameGenomicRegion){
		try{			
			BufferedWriter bw = new BufferedWriter(new FileWriter(outputPath + outputFilename));

			bw.write("Reference file = " + referenceFilename);
			bw.newLine();
			bw.flush();
			
			bw.write("Target file to be normalized (Gene Desert) = " + targetFilenameGeneDesert);
			bw.newLine();
			bw.flush();
			
			bw.write("Target file to be normalized (Genomic Region) = " + targetFilenameGenomicRegion);
			bw.newLine();
			bw.flush();
			
			bw.write("Normalization factor that the target file signal values need to be multiplied by=\t" + normalizationFactor);
			bw.newLine();
			bw.flush();
			
			bw.write("Number of matched points used to calculate normalization factor = " + count);
			bw.newLine();
			bw.flush();
			
			bw.close();
		}catch (IOException e) {
			e.printStackTrace();
		}//end of try-catch block
		
		if(verbose){
			System.out.println("\nOutput file: " + outputPath + outputFilename);
		}
	}/*  end of outputLogFile method  */
	
	
	
	
	/*  start of main method  */
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		
		
		
		boolean usingCommandLine = false;
		
		String filename_ReferenceGeneDesert = "default.txt";
		ArrayList<String> filenames_DatasetsToNormalizeGeneDesert = new ArrayList<String>();
		ArrayList<String> filenames_DatasetsToNormalizeGenomicRegion = new ArrayList<String>();
		
		
		//  load the parameters from the command line
		if(args.length >= 3){
			//  first parameter is the filename of the reference Gene Desert
			filename_ReferenceGeneDesert = args[0];
			
			//  all subsequent parameters are the filenames of the experiment files
			//  there needs to be a matched number of gene desert and genomic region experiment filenames specified
			if(((args.length-1)%2) != 0){
				System.err.println("Error: there needs to be a matched number of gene desert and genomic region experiment filenames specified.\nSystem.exit(0)");
				System.exit(0);
			}
			for(int i=1; i<args.length; i++){
				//  the even entries are the Gene Desert files
				if((i%2) == 0){
					filenames_DatasetsToNormalizeGeneDesert.add(args[i]);
				}
				//  the odd entries are the Genomic region files
				else{
					filenames_DatasetsToNormalizeGenomicRegion.add(args[i]);
				}
			}
			filenames_DatasetsToNormalizeGeneDesert.trimToSize();
			filenames_DatasetsToNormalizeGenomicRegion.trimToSize();
			
			usingCommandLine = true;
			verbose = false;
		}
		else{
			System.err.println("Error: there need to be at least three (3) command line parameters and a matched number of experiment gene desert and genomic region files.\n" +
					"\t\treference_gene_desert_filename.txt\n" +
					"\t\texperiment_1_gene_desert.txt\n" +
					"\t\texperiment_1_genomic_region.txt\n" +
					"\t\texperiment_2_gene_desert.txt\n" +
					"\t\texperiment_2_genomic_region.txt\n" +
					"\t\texperiment_..._gene_desert.txt\n" +
					"\t\texperiment_..._genomic_region.txt\n" +
					"System.exit(0)");
//			System.exit(0);
		}
		
		
		
		
		/******************************************************
		 * 													  *
		 * 				SET THE PARAMETERS HERE				  *
		 * 													  *
		 ******************************************************/
		
		
		if(!usingCommandLine){
		
			
			inputPath = "NormalizationGeneDesert/Input/";
			outputPath = "NormalizationGeneDesert/Output/";
			
			
//			filename_ReferenceGeneDesert = "cleaned_96Hr_GDGD.txt";
//
//			filenames_DatasetsToNormalizeGeneDesert.add("cleaned_0Hr_GDGD.txt");
//			filenames_DatasetsToNormalizeGenomicRegion.add("cleaned_0Hr_HOXAHOXA.txt");
//
//			filenames_DatasetsToNormalizeGeneDesert.add("cleaned_0Hr_2_GDGD.txt");
//			filenames_DatasetsToNormalizeGenomicRegion.add("cleaned_0Hr_2_HOXAHOXA.txt");
//
//			filenames_DatasetsToNormalizeGeneDesert.add("cleaned_4Hr_GDGD.txt");
//			filenames_DatasetsToNormalizeGenomicRegion.add("cleaned_4Hr_HOXAHOXA.txt");
//
//			filenames_DatasetsToNormalizeGeneDesert.add("cleaned_4Hr_2_GDGD.txt");
//			filenames_DatasetsToNormalizeGenomicRegion.add("cleaned_4Hr_2_HOXAHOXA.txt");
//
//			filenames_DatasetsToNormalizeGeneDesert.add("cleaned_12Hr_GDGD.txt");
//			filenames_DatasetsToNormalizeGenomicRegion.add("cleaned_12Hr_HOXAHOXA.txt");
//
//			filenames_DatasetsToNormalizeGeneDesert.add("cleaned_48Hr_GDGD.txt");
//			filenames_DatasetsToNormalizeGenomicRegion.add("cleaned_48Hr_HOXAHOXA.txt");
//		
//			filenames_DatasetsToNormalizeGeneDesert.add("DMSO_GD.txt");
//			filenames_DatasetsToNormalizeGenomicRegion.add("DMSO_HoxA.txt");
//			
//			filenames_DatasetsToNormalizeGeneDesert.add("PMA_GD.txt");
//			filenames_DatasetsToNormalizeGenomicRegion.add("PMA_HoxA.txt");
			
			
			//  Hox in motion  --  replicates
			
			filename_ReferenceGeneDesert = "cleaned_96Hr_GDGD.txt";
			
			filenames_DatasetsToNormalizeGeneDesert.add("0h_biological2_GDGD.txt");
			filenames_DatasetsToNormalizeGenomicRegion.add("0h_biological2_HOXAHOXA.txt");
			filenames_DatasetsToNormalizeGeneDesert.add("0h_technical_GDGD.txt");
			filenames_DatasetsToNormalizeGenomicRegion.add("0h_technical_HOXAHOXA.txt");
			filenames_DatasetsToNormalizeGeneDesert.add("4h_biological2_GDGD.txt");
			filenames_DatasetsToNormalizeGenomicRegion.add("4h_biological2_HOXAHOXA.txt");
			filenames_DatasetsToNormalizeGeneDesert.add("4h_technical_3_GDGD.txt");
			filenames_DatasetsToNormalizeGenomicRegion.add("4h_technical_3_HOXAHOXA.txt");
			filenames_DatasetsToNormalizeGeneDesert.add("12h_biological2_GDGD.txt");
			filenames_DatasetsToNormalizeGenomicRegion.add("12h_biological2_HOXAHOXA.txt");
			filenames_DatasetsToNormalizeGeneDesert.add("12h_technical_GDGD.txt");
			filenames_DatasetsToNormalizeGenomicRegion.add("12h_technical_HOXAHOXA.txt");
			filenames_DatasetsToNormalizeGeneDesert.add("48h_biological2_GDGD.txt");
			filenames_DatasetsToNormalizeGenomicRegion.add("48h_biological2_HOXAHOXA.txt");
			filenames_DatasetsToNormalizeGeneDesert.add("48h_technical_GDGD.txt");
			filenames_DatasetsToNormalizeGenomicRegion.add("48h_technical_HOXAHOXA.txt");
			filenames_DatasetsToNormalizeGeneDesert.add("96h_biological2_GDGD.txt");
			filenames_DatasetsToNormalizeGenomicRegion.add("96h_biological2_HOXAHOXA.txt");
			filenames_DatasetsToNormalizeGeneDesert.add("96h_technical_GDGD.txt");
			filenames_DatasetsToNormalizeGenomicRegion.add("96h_technical_HOXAHOXA.txt");
			
		}

		
		/******************************************************
		 * 													  *
		 * 				END OF SETTING THE PARAMETERS		  *
		 * 													  * 
		 ******************************************************/
		
		
		
		
		filenames_DatasetsToNormalizeGeneDesert.trimToSize();
		filenames_DatasetsToNormalizeGenomicRegion.trimToSize();
		
		if(filenames_DatasetsToNormalizeGeneDesert.size() != filenames_DatasetsToNormalizeGenomicRegion.size()){
			System.err.println("Error: the filename lists for the Gene Deserts and the Genomic Regions must have the same number of matched entries.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		for(int i=0; i<filenames_DatasetsToNormalizeGeneDesert.size(); i++){
			
			String filename_DatasetToNormalizeGeneDesert = filenames_DatasetsToNormalizeGeneDesert.get(i);
			String filename_DatasetToNormalizeGenomicRegion = filenames_DatasetsToNormalizeGenomicRegion.get(i);
			
		
			//  display the parameters
			if(verbose){
				System.out.println("Reference file: " + filename_ReferenceGeneDesert);
				System.out.println("Dataset to normalize (Gene Desert file): " + filename_DatasetToNormalizeGeneDesert);
				System.out.println("Dataset to normalize (Genomic Region file): " + filename_DatasetToNormalizeGenomicRegion);
			}
		
		
			//  initialize the parameters
			initializeParameters();
		
		
			//  load the reference signals from file
			referenceSignals = new ArrayList<double[]>();
			loadSignalsFromFile(filename_ReferenceGeneDesert, referenceSignals);
		
		
			//  display the reference signals
//			displaySignals(referenceSignals);
		
		
			//  load the Gene Desert signals used to calculate the normalization factor (and to be normalized) from file
			targetSignalsGeneDesert = new ArrayList<double[]>();
			loadSignalsFromFile(filename_DatasetToNormalizeGeneDesert, targetSignalsGeneDesert);
		

			//  display the target signals
//			displaySignals(targetSignalsGeneDesert);
		
		
			//  calculate the normalization factor
			calculateNormalizationFactor();
		
		
			//  normalize the target signals (Gene Desert)
			normalizeSignals(targetSignalsGeneDesert);

		
			//  load the Genomic Region signals to be normalized from file
			targetSignalsGenomicRegion = new ArrayList<double[]>();
			loadSignalsFromFile(filename_DatasetToNormalizeGenomicRegion, targetSignalsGenomicRegion);
		
		
			//  normalize the target signals (Genomic Region)
			normalizeSignals(targetSignalsGenomicRegion);

		
			//  output the normalized target signals for the Gene Desert
			outputNormalizedSignals(filename_DatasetToNormalizeGeneDesert, targetSignalsGeneDesert);
		
		
			//  output the normalized target signals for the Gene Desert
			outputNormalizedSignals(filename_DatasetToNormalizeGenomicRegion, targetSignalsGenomicRegion);
		
		
			//  output log file
			outputLogFile("logFile_" + filename_DatasetToNormalizeGeneDesert + "_normalizedTo_" + filename_ReferenceGeneDesert, filename_ReferenceGeneDesert, filename_DatasetToNormalizeGeneDesert, filename_DatasetToNormalizeGenomicRegion);
		}

		
		if(verbose){
			System.out.println("Done!");
		}
	}/*  end of main method  */

}/*  end of NormalizationGeneDesert class  */
