package utilities;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;


/** ------------------------------------------------------------------------------------------------------------------
 * start of NormalizeBACs class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class NormalizeBACs {

	

	
	/*  global parameters  */
	private static boolean verbose = true;
	private static String inputPath = "";
	private static String outputPath = "";
	/*  end of global parameters  */
	

	
	
	/*  start of getFilenames method  */
	/**
	 * @param filename_file_list the file containing the list of files
	 * @return a list containing the filenames in the file specified as input parameter
	 */
	private static ArrayList<String> getFilenames(String filename_file_list){
		if(filename_file_list == null){
			System.err.println("Error: the filename list is null.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<String> file_list = new ArrayList<String>();

		String line = null;
		int lineCounter = -1;			//  there is one header line
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename_file_list));
			while((line = br.readLine()) != null){
				if(lineCounter >= 0){
					file_list.add(line);
				}
				else{
					//  skip and the header line
//					System.out.println(line);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		file_list.trimToSize();
		
		if(verbose){
			System.out.println("Loaded input file: " + inputPath + filename_file_list);
		}
		
		return file_list;
	}/*  end of getFilenames method  */
	
	

	
	/*  start of displayFilenames method  */
	/**
	 * @param file_list the list of BAC filenames to display
	 * @return void
	 */
	private static void displayFilenames(ArrayList<String> file_list){
		if(file_list == null){
			System.err.println("Error: the file list is null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);	
		}
		
		System.out.println("Filename");
		for(String currentFilename: file_list){
			System.out.println(currentFilename);
		}
	}/*  end of displayFilenames method  */
	
	

	
	/*  start of loadInputData method  */
	/**
	 * @param file_list the list of BAC filenames
	 * @return a list of FiveC objects containing the BAC input data
	 */
	private static ArrayList<FiveC> loadInputData(ArrayList<String> file_list){
		if(file_list == null){
			System.err.println("Error: the file list is null and the data cannot be loaded.\nSystem.exit(0)");
			System.exit(0);	
		}
		
		ArrayList<FiveC> FiveC_Datasets = new ArrayList<FiveC>();
		
		for(String currentFilename: file_list){
			FiveC_Datasets.add(new FiveC(currentFilename, inputPath, outputPath));
		}
		
		FiveC_Datasets.trimToSize();
		
		return FiveC_Datasets;
	}/*  end of loadInputData method  */
	
	
	
	
	/*  start of displayData method  */
	/**
	 * @param FiveC_Datasets the list of FiveC objects to display
	 * @return void
	 */
	private static void displayData(ArrayList<FiveC> FiveC_Datasets){
		if(FiveC_Datasets == null){
			System.err.println("The input data are null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		for(FiveC currentFiveC: FiveC_Datasets){
			System.out.println(currentFiveC);
		}
	}/*  end of displayData method  */
	

	
	
	/*  start of calculateNormalizationFactors method  */
	/**
	 * @param FiveC_Datasets the list of FiveC objects containing the BAC data from which to calculate the normalization factors
	 * @return a list containing the normalization factors (ArrayList<double[]>, first entry) and the target values (double[], second entry)
	 */
	private static ArrayList<Object> calculateNormalizationFactors(ArrayList<FiveC> FiveC_Datasets){
		if(FiveC_Datasets == null){
			System.err.println("The input data arrays are null and the normalization factors cannot be determined.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<double[]> medianArrays = new ArrayList<double[]>();
		
		for(FiveC currentFiveC: FiveC_Datasets){
			double[] medianValues = currentFiveC.getMedianValues();
			
//			for(int i=0; i<medianValues.length; i++){
//				System.out.print(medianValues[i] + "\t");
//			}
//			System.out.println();
			
			medianArrays.add(medianValues);
		}
		medianArrays.trimToSize();
		
		
		double[] averageMedians = new double[8];
		for(double[] currentMedians: medianArrays){
			for(int i=0; i<currentMedians.length; i++){
				averageMedians[i] += currentMedians[i];
			}
		}
		if(medianArrays.size() > 0){
			for(int i=0; i<averageMedians.length; i++){
				averageMedians[i] /= medianArrays.size();
			}
		}
//		for(int i=0; i<averageMedians.length; i++){
//			System.out.print(averageMedians[i] + "\t");
//		}
//		System.out.println();
		
		double[] targetValues = averageMedians;
		
		
		ArrayList<double[]> normalizationFactors = new ArrayList<double[]>();
		
		
		//  making all of the averages the same
		for(FiveC currentFiveC: FiveC_Datasets){
			
			double[] currentAverages = currentFiveC.getAverageValues();
			
			double[] currentNormalizationFactors = new double[8];
			
			for(int i=0; i<targetValues.length; i++){
				
				if(currentAverages[i] == 0){
					currentNormalizationFactors[i] = 1;
				}
				else{
					currentNormalizationFactors[i] = (double)targetValues[i] / (double)currentAverages[i];
				}
			}
			normalizationFactors.add(currentNormalizationFactors);
		}
		normalizationFactors.trimToSize();
		
		
		
		
		//  making all of the medians the same
//		for(double[] currentMedians: medianArrays){
//			
//			double[] currentNormalizationFactors = new double[8];
//			
//			for(int i=0; i<targetValues; i++){
//				if(currentMedians[i] == 0){
//					currentNormalizationFactors[i] = 1;
//				}
//				else{
//					currentNormalizationFactors[i] = (double)targetValues[i] / (double)currentMedians[i];
//				}
//			}
//			normalizationFactors.add(currentNormalizationFactors);
//		}
//		normalizationFactors.trimToSize();
		
		ArrayList<Object> returnList = new ArrayList<Object>();
		returnList.add(normalizationFactors);
		returnList.add(targetValues);
		returnList.trimToSize();
		
		return returnList;
	}/*  end of calculateNormalizationFactors method  */
	
	
	

	/*  start of displayNormalizationFactors method  */
	/**
	 * @param normalizationFactors the list of per-probe length normalization factors for each BAC experiment
	 * @param file_list the list of input BAC filenames
	 * @return void
	 */
	private static void displayNormalizationFactors(ArrayList<double[]> normalizationFactors, ArrayList<String> file_list){
		if(normalizationFactors == null){
			System.err.println("Error: the normalization factors are null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		for(int i=0; i<normalizationFactors.size(); i++){
			
			System.out.println("\n" + file_list.get(i));
			
			double[] currentNormalizationFactors = normalizationFactors.get(i);
			for(int j=0; j<currentNormalizationFactors.length; j++){
				System.out.print(currentNormalizationFactors[j] + "\t");
			}
			System.out.println();
		}
	}/*  end of displayNormalizationFactors method  */
	
	
	

	/*  start of applyNormalizationFactors method  */
	/**
	 * @param FiveC_Datasets the list of FiveC objects for the BAC input data that need to be normalized
	 * @param normalizationFactors the list of per-probe length normalization factors for each BAC experiment
	 * @return void
	 */
	private static void applyNormalizationFactors(ArrayList<FiveC> FiveC_Datasets, ArrayList<double[]> normalizationFactors){
		if(normalizationFactors == null){
			System.err.println("Error: the normalization factors are null and cannot be applied.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		for(int i=0; i<FiveC_Datasets.size(); i++){
			double[] currentNormalizationArray = normalizationFactors.get(i);
			
			FiveC_Datasets.get(i).normalize(currentNormalizationArray);
		}
	}/*  end of applyNormalizationFactors method  */
	
	
	
	
	/*  start of outputDataToFile method  */
	/**
	 * @param FiveC_Datasets the list of FiveC objects for the normalized BAC data to be output to file
	 * @param normalizationFactors the list of per-probe length normalization factors for each back to be output to file
	 * @param targetValues the array of target values (average per-probe length median) that were used to calculate the normalization factors to be output to file
	 * @param file_list the list of BAC filenames
	 * @return void
	 */
	private static void outputDataToFile(ArrayList<FiveC> FiveC_Datasets, ArrayList<double[]> normalizationFactors, double[] targetValues, ArrayList<String> file_list){
		if(normalizationFactors == null){
			System.err.println("Error: the normalization factors are null and cannot be output to file.\nSystem.exit(0)");
			System.exit(0);
		}
		

		String filename_output = "normalization_factors.txt";
		
		try{
			BufferedWriter bw = new BufferedWriter(new FileWriter(outputPath + filename_output));

			//  write the header lines
			bw.write("Experiment name\t15bp\t18bp\t19bp\t20bp\t21bp\t22bp\t23bp\t24bp");
			bw.newLine();
			bw.flush();
			
			bw.write("Target medians");
			for(int i=0; i<targetValues.length; i++){
				bw.write("\t" + targetValues[i]);
			}
			bw.newLine();
			bw.flush();
			
			for(int i=0; i<normalizationFactors.size(); i++){
				bw.write(file_list.get(i) + "\t");
				
				double[] currentNormalizationFactors = normalizationFactors.get(i);
				
				for(int j=0; j<currentNormalizationFactors.length; j++){
					if(j == (currentNormalizationFactors.length-1)){
						bw.write(String.valueOf(currentNormalizationFactors[j]));
					}
					else{
						bw.write(currentNormalizationFactors[j] + "\t");
					}
				}
				bw.newLine();
				bw.flush();
			}
			bw.flush();
			bw.close();
		}catch (IOException e) {
			e.printStackTrace();
		}//end of try-catch block
		
		if(verbose){
			System.out.println("\nOutput file: " + outputPath + filename_output);
		}
		
		
		for(FiveC currentFiveC: FiveC_Datasets){
			currentFiveC.outputToFile(outputPath + currentFiveC.getFilename());
		}
	}/*  end of outputDataToFile method  */
		
	
	
	
	/*  start of main method  */
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		
		
		
		boolean usingCommandLine = false;
		
		String filename_file_list = "default.txt";
		ArrayList<String> file_list = new ArrayList<String>();
		
		
		//  load the parameters from the command line
		if(args.length >= 2){
			//  all the parameters are the filenames of the BAC files
			for(int i=0; i<args.length; i++){
				file_list.add(args[i]);
			}
			file_list.trimToSize();
			
			usingCommandLine = true;
			verbose = false;
		}
		else{
			System.err.println("Error: there need to be at least two (2) command line parameters.\n" +
					"\t\tBAC_1_filename.txt\n" +
					"\t\tBAC_2_filename.txt\n" +
					"\t\tBAC_..._filename.txt\n" +
					"System.exit(0)");
//			System.exit(0);
		}
		
		
		
		
		/******************************************************
		 * 													  *
		 * 				SET THE PARAMETERS HERE				  *
		 * 													  *
		 ******************************************************/
		
		
		if(!usingCommandLine){
			
			
			inputPath = "NormalizeBACs/Input/";
			outputPath = "NormalizeBACs/Output/";
			
			
			filename_file_list = "file_list.txt";
			
			
//			verbose = false;
			
			
		}
		
		
		/******************************************************
		 * 													  *
		 * 				END OF SETTING THE PARAMETERS		  *
		 * 													  * 
		 ******************************************************/
		
		
		

		if(!usingCommandLine){
			file_list = getFilenames(filename_file_list);
		}
//		displayFilenames(file_list);
		
		

		ArrayList<FiveC> FiveC_Datasets = loadInputData(file_list);
//		displayData(FiveC_Datasets);
		
		
		ArrayList<Object> normalizationFactorsAndTargetValuesList = calculateNormalizationFactors(FiveC_Datasets);
		ArrayList<double[]> normalizationFactors = (ArrayList<double[]>)normalizationFactorsAndTargetValuesList.get(0);
		double[] targetValues = (double[])normalizationFactorsAndTargetValuesList.get(1);
//		displayNormalizationFactors(normalizationFactors, file_list);
		
		
		applyNormalizationFactors(FiveC_Datasets, normalizationFactors);
//		displayData();
		
		
		outputDataToFile(FiveC_Datasets, normalizationFactors, targetValues, file_list);
		
		
		if(verbose){
			System.out.println("Done!");
		}
	}/*  end of main method  */

}/*  end of NormalizeBACs class  */
