package utilities;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


/** ------------------------------------------------------------------------------------------------------------------
 * start of PDB class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class PDB {

	

	
	/*  global parameters  */
	private String path_filename;
	private ArrayList<String> header;
	private double[][] coordinates;
	/*  end of global parameters  */
	
	
	
	
	/*  start of constructor method  */
	/**
	 * @param path_filename the filename (including the prefix path) of the input PDB file to load 
	 */
	public PDB(String path_filename){
		if(path_filename == null){
			System.err.println("Error: the specified filename is null and the input PDB object cannot be created.\nSystem.exit(0)");
			System.exit(0);
		}
		
		this.loadFromFile(path_filename);
	}/*  end of constructor method  */
	


	
	/*  start of constructor method  */
	/**
	 * @param path_filename the filename (including the prefix path) to set for the new PDB object (nothing is loaded from this file)
	 * @param header the ArrayList of String objects that are deep copied to use for the new PDB object
	 * @param coordinates the array of coordinates to deep copy to use for the new PDB object
	 */
	public PDB(String path_filename, ArrayList<String> header, double[][] coordinates){
		if((path_filename == null) || (header == null) || (coordinates == null)){
			System.err.println("Error: the filename, header and coordinates must all be non-null to create a new PDB object.\nSystem.exit(0)");
			System.exit(0);
		}
		
		this.setFilename(path_filename);
		this.setHeader(header);
		this.setCoordinates(coordinates);
	}/*  end of constructor method  */
	
	
	
	
	/*  start of setFilename method  */
	/**
	 * @param path_filename the filename (including the prefix path) to deep-copy and set for the current PDB object (nothing is loaded from this file)
	 */
	public void setFilename(String path_filename){
		if(path_filename == null){
			System.err.println("Error: the filename parameter is null and cannot be set.\nSystem.exit(0)");
			System.exit(0);
		}
		
		this.path_filename = new String(path_filename);
	}/*  end of setFilename method  */
	

	
	
	/*  start of setHeader method  */
	/**
	 * @param header the list of header Strings to deep-copy and set for the current PDB object
	 */
	public void setHeader(ArrayList<String> header){
		if(header == null){
			System.err.println("Error: the header parameter is null and cannot be set.\nSystem.exit(0)");
			System.exit(0);
		}
		
		this.header = new ArrayList<String>();
		for(String currentHeaderString: header){
			this.header.add(new String(currentHeaderString));
		}
		this.header.trimToSize();
	}/*  end of setHeader method  */
	
	
	
	
	/*  start of setCoordinates method  */
	/**
	 * @param coordinates the array of coordinates to deep-copy and set for the current PDB object
	 */
	public void setCoordinates(double[][] coordinates){
		if(coordinates == null){
			System.err.println("Error: the coordinates parameter is null and cannot be set.\nSystem.exit(0)");
			System.exit(0);
		}
		
		this.coordinates = new double[coordinates.length][3];
		for(int i=0; i<coordinates.length; i++){
			this.coordinates[i][0] = coordinates[i][0];
			this.coordinates[i][1] = coordinates[i][1];
			this.coordinates[i][2] = coordinates[i][2];
		}
	}/*  end of setCoordinates method  */
	
	
	
	
	/*  start of loadFromFile method  */
	/**
	 * @param path_filename the filename (including the prefix path) of the input PDB file to load
	 */
	private void loadFromFile(String path_filename){
		if(path_filename == null){
			System.err.println("Error: the specified filename is null and the input PDB file cannot be loaded.\nSystem.exit(0)");
			System.exit(0);
		}
		
		this.path_filename = path_filename;
		
		
		PDButility utility_PDB = new PDButility();
		
		
		this.header = utility_PDB.getHeaderLines(path_filename);
		
		this.coordinates = utility_PDB.getStructure(path_filename);
		
		System.out.println("Loaded input file: " + path_filename);
	}/*  end of loadFromFile method  */
	
	
	
	
	/*  start of getFilename method  */
	/**
	 * @return the String containing the filename (excluding the path prefix if there is one) for the current PDB object
	 */
	public String getFilename(){
		if(this.path_filename == null){
			System.err.println("Error: the filename is null and cannot be returned.\nSystem.exit(0)");
			System.exit(0);
		}
		if(this.path_filename.contains("/")){
			return this.path_filename.substring(this.path_filename.lastIndexOf("/")+1);
		}
		return this.path_filename;
	}/*  end of getFilename method  */
	
	
	

	/*  start of getFullHeader method  */
	/**
	 * @return the list of String objects specifying the header for the current PDB object
	 */
	public ArrayList<String> getFullHeader(){
		if(this.header == null){
			System.err.println("Error: the header is null and cannot be returned.\nSystem.exit(0)");
			System.exit(0);
		}
		return this.header;
	}/*  end of getFullHeader method  */
	

	
	
	/*  start of getHeaderValueOnly method  */
	/**
	 * @return a single String object specifying value of header (stripped of AUTHOR line) for the current PDB object
	 */
	public String getHeaderValueOnly(){
		if(this.header == null){
			System.err.println("Error: the header is null and cannot be returned.\nSystem.exit(0)");
			System.exit(0);
		}
		String headerFirstLine = this.header.get(0);
		String[] values = headerFirstLine.split("    ");		//  remove the leading COMPND
		if(values.length != 2){
			System.err.println("Error splitting the header line:\"" + headerFirstLine + "\"\nSystem.exit(0)");
			System.exit(0);
		}
		return values[1];
	}/*  end of getHeaderValueOnly method  */
	
	
	
	
	/*  start of getCoordinates method  */
	/**
	 * @return the array containing the coordinates of the points for the current PDB object
	 */
	public double[][] getCoordinates(){
		if(this.coordinates == null){
			System.err.println("Error: the coordinates are null and cannot be returned.\nSystem.exit(0)");
			System.exit(0);
		}
		return this.coordinates;
	}/*  end of getCoordinates method  */
	
	
	
	
	/*  start of toString method  */
	/**
	 * @return a String representation of the PDB object
	 */
	public String toString(){
		if((this.header == null) || (this.coordinates == null)){
			System.err.println("Error: both the header and coordinates must be non-null to return a String representation of the current PDB object.\nSystem.exit(0)");
			System.exit(0);
		}
		
		String returnString = "";
		
		//  first add the header lines
		for(String currentHeaderLine: this.header){
			returnString += currentHeaderLine + "\n";
		}
		
		//  then add the coordinates
		for(int i=0; i<this.coordinates.length; i++){
			double[] currentCoordinates = this.coordinates[i];
			returnString += currentCoordinates[0] + "\t" + currentCoordinates[1] + "\t" + currentCoordinates[2] + "\n";
		}
		
		return returnString;
	}/*  end of toString method  */
	
	
	
	
}/*  end of PDB class  */
