package utilities;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 **/


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;

import javax.vecmath.Point3d;

import modeling.Chromosome;


/** ------------------------------------------------------------------------------------------------------------------
 * start of PDButility class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class PDButility {

	
	
	
	/*  global variables  */
	private static boolean verbose = true;
	private final static String AUTHOR = "ROUSSEAU";
	private final NumberFormat formatterPDB0decimals = new DecimalFormat("0");
	private final NumberFormat formatterPDB1decimals = new DecimalFormat("0.0");
	private final NumberFormat formatterPDB2decimals = new DecimalFormat("0.00");
	private final NumberFormat formatterPDB3decimals = new DecimalFormat("0.000");
	/*  end of global variables  */
	
	
	
	
	/*  start of constructor  */
	/**
	 * @return void
	 */
	public PDButility(){
		this(true);
	}/*  end of constructor  */
	
	
	
	
	/*  start of constructor  */
	/**
	 * @param verbose the verbose boolean flag to specify if output should be sent to the console (true) or not (false)
	 * @return void
	 */
	public PDButility(boolean verbose){
		
		//  set the verbose flag
		this.setVerbose(verbose);
		
	}/*  end of constructor  */
	
	
	
	
	/*  start of setVerbose method  */
	/**
	 * @param verbose the verbose boolean flag to specify if output should be sent to the console (true) or not (false)
	 * @return void
	 */
	public void setVerbose(boolean verbose){
		this.verbose = verbose;
	}/*  end of setVerbose method  */
	
	
	
	
	/*  start of getVerbose method  */
	/**
	 * @return the verbose boolean flag that specifies if output is sent to the console (true) or not (false)
	 */
	public boolean getVerbose(){
		return this.verbose;
	}/*  end of getVerbose method  */
	
	
	
	
	/*  start of getAuthor method  */
	/**
	 * @return the author used in the output PDB files
	 */
	public String getAuthor(){
		return this.AUTHOR;
	}/*  end of getAuthor method  */
	
	
	
	
	/*  start of getNumberOfPoints method  */
	public int getNumberOfPoints(String pathFilename){
		String line = null;
		int numberOfPoints = 0;
		try{
			BufferedReader br = new BufferedReader(new FileReader(pathFilename));
			int headerLinesLeft = 2;		//  there are two (2) header lines that need to be skipped
			boolean doneHeader = false;
			boolean doneStructure = false;
			while(((line = br.readLine()) != null) && (!doneStructure)){
				if(doneHeader){
					String[] values = line.split(" ");
					if(values[0].compareTo("ATOM") == 0){
						numberOfPoints++;
					}
					else{
						doneStructure = true;
					}
				}
				else{
//					System.out.println(line);	//  display the header lines
					headerLinesLeft--;
					if(headerLinesLeft == 0){
						doneHeader = true;
					}
				}
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error: cannot find file with filename \"" + pathFilename + "\"\n" +
			"System.exit(0)");
			System.exit(0);
		}
		return numberOfPoints;
	}/*  end of getNumberOfPoints method  */
	
	
	
	
	/*  start of translateTo000 method  */
	/**
	 * @param coordinates an array containing the coordinates to make positive
	 * @return an array containing a copy of the coordinates translated to be positive
	 */
	public double[][] translateTo000(double[][] coordinates){
		if(coordinates == null){
			System.err.println("Error: the coordinates are null and cannot be translated.\nSystem.exit(0)");
			System.exit(0);
		}
		
		//  make the minimum x-, y-, and z-coordinates of the coordinates equal to zero (0)
		double translationX = Double.MAX_VALUE;
		double translationY = Double.MAX_VALUE;
		double translationZ = Double.MAX_VALUE;
	
		//  find the minimum x,y,z coordinate values
		for(int i=0; i<coordinates.length; i++){
			if(coordinates[i][0] < translationX){
				translationX = coordinates[i][0];
			}
			if(coordinates[i][1] < translationY){
				translationY = coordinates[i][1];
			}
			if(coordinates[i][2] < translationZ){
				translationZ = coordinates[i][2];
			}
		}
		
		double[][] translated_coordinates = new double[coordinates.length][3];
	
		//  translate all of the coordinates
		for(int i=0; i<coordinates.length; i++){
			translated_coordinates[i][0] = coordinates[i][0] - translationX;
			translated_coordinates[i][1] = coordinates[i][1] - translationY;
			translated_coordinates[i][2] = coordinates[i][2] - translationZ;
		}
		
		return translated_coordinates;
	}/*  end of translateTo000 method  */
	
	
	
	
	/*  start of writeStructureNoTranslate method  */
	/**
	 * @param pathFilename the filename (including the output path prefix) to use to write the structure to file with
	 * @param structure the structure to write to file
	 * @param header the header line to use for the output file
	 */
	public void writeStructureNoTranslate(String pathFilename, Chromosome structure, String header){
		if(structure == null){
			System.err.println("Error: the structure pointer is null and the structure cannot be written to file.\nSystem.exit(0)");
			System.exit(0);
		}
		writeStructure(pathFilename, structure.getAsDoubleArray(), header, false);
	}/*  end of writeStructureNoTranslate method  */
	
	
	
	
	/*  start of writeStructure method  */
	/**
	 * @param pathFilename the filename (including the output path prefix) to use to write the structure to file with
	 * @param structure the structure to write to file
	 * @param header the header line to use for the output file
	 */
	public void writeStructure(String pathFilename, Chromosome structure, String header){
		if(structure == null){
			System.err.println("Error: the structure pointer is null and the structure cannot be written to file.\nSystem.exit(0)");
			System.exit(0);
		}
		writeStructure(pathFilename, structure.getAsDoubleArray(), header, true);
	}/*  end of writeStructure method  */
	
	
	
	
	/*  start of writeStructure method  */
	/**
	 * @param pathFilename the filename (including the output path prefix) to use to write the structure to file with
	 * @param structure the structure to write to file
	 * @param header the header line to use for the output file
	 */
	public void writeStructure(String pathFilename, double[][] structure, String header){
		if(structure == null){
			System.err.println("Error: the structure pointer is null and the structure cannot be written to file.\nSystem.exit(0)");
			System.exit(0);
		}
		writeStructure(pathFilename, structure, header, true);
	}/*  end of writeStructure method  */
	
	
	
	
	/*  start of writePDB method  */
	/**
	 * @param PDBtoOutput the PDB object to output to file
	 * @param filename the filename to use to output the current PDB object (if it is null, the filename stored in the PDB object will be used instead)
	 * @param outputPath the outputPath to use as a prefix for the output filename (ignored if null)
	 */
	public void writePDB(PDB PDBtoOutput, String filename, String outputPath){
		if(PDBtoOutput == null){
			System.err.println("Error: the PDB object is null and cannot be output to file.\nSystem.exit(0)");
			System.exit(0);
		}
		
		//  if the filename is null, use the filename stored in the PDB object
		if(filename == null){
			filename = PDBtoOutput.getFilename();
		}
		
		//  if the output path is not null, then if the filename does not already start with the output path, add it as the prefix to the filename
		if(outputPath != null){
			if(!filename.startsWith(outputPath)){
				filename = outputPath + filename;
			}
		}
		
		
		this.writeStructure(filename, PDBtoOutput.getCoordinates(), PDBtoOutput.getHeaderValueOnly());
	}/*  end of writePDB method  */
	
	
	
	
	/*  start of writeHeader method  */
	/**
	 * @param pathFilename the filename (including the output path prefix) to use to write the structure to file with
	 * @param header the header line to use for the output file
	 */
	private void writeHeader(String pathFilename, String header){
		
		try{						
			BufferedWriter bw = new BufferedWriter(new FileWriter(pathFilename, false));
			
			//  write the header block
			bw.write("COMPND    " + header.toUpperCase() + "\nAUTHOR    " + AUTHOR + "\n");
			bw.flush();
			

			//  close the writer			
			bw.close();
		}catch (IOException e) {
			System.err.println("Error: could not output file: " + pathFilename + "\nSystem.exit(0)");
			e.printStackTrace();
			System.exit(0);
		}//end of try-catch block
	}/*  end of writeHeader method  */
	
	
	
	
	/*  start of writeStructure method  */
	/**
	 * @param pathFilename the filename (including the output path prefix) to use to write the structure to file with
	 * @param structure the structure to write to file
	 * @param header the header line to use for the output file
	 * @param translateTo000 a flag to indicate whether or not to translate the structure so the minimum x,y,z coordinates are equal to zero
	 */
	public void writeStructure(String pathFilename, double[][] structure, String header, boolean translateTo000){

		if(translateTo000){
			structure = this.translateTo000(structure);
		}
		
		
		//  write the structure to file
		
		//  first the header
		this.writeHeader(pathFilename, header);
		
		//  then the rest of the file
		try{						
			BufferedWriter bw = new BufferedWriter(new FileWriter(pathFilename, true));				//  append to the existing file that already contains the header
			
			
			//  write the structure
			for(int i=0; i<structure.length; i++){
				
				//  write the type of particle
				String lineToWrite = "ATOM  ";
				
				
				//  write the index of the atom
				int atomIndex = i+1;
				if(atomIndex < 10){
					lineToWrite += "    " + atomIndex;
				}
				else if(atomIndex < 100){
					lineToWrite += "   " + atomIndex;
				}
				else if(atomIndex < 1000){
					lineToWrite += "  " + atomIndex;
				}
				else if(atomIndex < 10000){
					lineToWrite += " " + atomIndex;
				}
				else{
					System.err.println("Error: the index value (" + atomIndex + ") is too large to convert to PDB format.\nSystem.exit(0)");
					System.exit(0);
				}
				
				
				//  write the atom used to represent each point
				lineToWrite += " C";
				
				
				//  write the type of atom present
				lineToWrite += "    LIG";
				
				
				//  write the chain identifier
				lineToWrite += " A";
				
				
				//  write the white-space before the coordinates
				//  if all of the coordinates are positive values then this should be nine (9) spaces
				lineToWrite += "         ";
				
								
				//  write the x-coordinate
				//  add the requisite amount of white-space and adjust the number of digits after the decimal (if required)
				if(structure[i][0] < 10){
					lineToWrite += "  ";
					lineToWrite += formatterPDB3decimals.format(structure[i][0]) + " ";
				}
				else if(structure[i][0] < 100){
					lineToWrite += " ";
					lineToWrite += formatterPDB3decimals.format(structure[i][0]) + " ";
				}
				else if(structure[i][0] < 1000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB3decimals.format(structure[i][0]) + " ";
				}
				else if(structure[i][0] < 10000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB2decimals.format(structure[i][0]) + " ";
				}
				else if(structure[i][0] < 100000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB1decimals.format(structure[i][0]) + " ";
				}
				else if(structure[i][0] < 1000000){
					lineToWrite += " ";				//  need to add an extra space to make up for the decimal that is removed
					lineToWrite += formatterPDB0decimals.format(structure[i][0]) + " ";
				}
				else if(structure[i][0] < 10000000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB0decimals.format(structure[i][0]) + " ";
				}
				else{
					System.err.println("Error: the x-coordinate value is greater than 10000000 and cannot be converted to PDB.\nSystem.exit(0)");
					System.exit(0);
				}
				
				
				//  write the y-coordinate
				//  add the requisite amount of white-space and adjust the number of digits after the decimal (if required)
				if(structure[i][1] < 10){
					lineToWrite += "  ";
					lineToWrite += formatterPDB3decimals.format(structure[i][1]) + " ";
				}
				else if(structure[i][1] < 100){
					lineToWrite += " ";
					lineToWrite += formatterPDB3decimals.format(structure[i][1]) + " ";
				}
				else if(structure[i][1] < 1000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB3decimals.format(structure[i][1]) + " ";
				}
				else if(structure[i][1] < 10000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB2decimals.format(structure[i][1]) + " ";
				}
				else if(structure[i][1] < 100000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB1decimals.format(structure[i][1]) + " ";
				}
				else if(structure[i][1] < 1000000){
					lineToWrite += " ";				//  need to add an extra space to make up for the decimal that is removed
					lineToWrite += formatterPDB0decimals.format(structure[i][1]) + " ";
				}
				else if(structure[i][1] < 10000000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB0decimals.format(structure[i][1]) + " ";
				}
				else{
					System.err.println("Error: the y-coordinate value is greater than 10000000 and cannot be converted to PDB.\nSystem.exit(0)");
					System.exit(0);
				}
				
				
				//  write the z-coordinate
				//  add the requisite amount of white-space and adjust the number of digits after the decimal (if required)
				if(structure[i][2] < 10){
					lineToWrite += "  ";
					lineToWrite += formatterPDB3decimals.format(structure[i][2]) + " ";
				}
				else if(structure[i][2] < 100){
					lineToWrite += " ";
					lineToWrite += formatterPDB3decimals.format(structure[i][2]) + " ";
				}
				else if(structure[i][2] < 1000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB3decimals.format(structure[i][2]) + " ";
				}
				else if(structure[i][2] < 10000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB2decimals.format(structure[i][2]) + " ";
				}
				else if(structure[i][2] < 100000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB1decimals.format(structure[i][2]) + " ";
				}
				else if(structure[i][2] < 1000000){
					lineToWrite += " ";				//  need to add an extra space to make up for the decimal that is removed
					lineToWrite += formatterPDB0decimals.format(structure[i][2]) + " ";
				}
				else if(structure[i][2] < 10000000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB0decimals.format(structure[i][2]) + " ";
				}
				else{
					System.err.println("Error: the z-coordinate value is greater than 10000000 and cannot be converted to PDB.\nSystem.exit(0)");
					System.exit(0);
				}
				
				
				//  write the number for the atom
				lineToWrite += " 1.00";
				
				
				//  write the radius for the atom
				lineToWrite += " 75.00";
				
				
				//  write the trailing white-space (if all other formatting conditions have been met then this should be four (4) spaces) 
				lineToWrite += "    ";
				
				
				//  output the line to write to file
				bw.write(lineToWrite);
				bw.newLine();
				bw.flush();
			}
			
			
			//  connect all of the successive points
			for(int i=1; i<structure.length; i++){
				bw.write("CONECT ");
				//  write the first index
				if(i < 10){
					bw.write("   ");
				}
				else if(i < 100){
					bw.write("  ");
				}
				else if(i < 1000){
					bw.write(" ");
				}
				else if(i < 10000){
					//  don't need to add any white-space
				}
				else{
					System.err.println("Error: the first index is greater than 10000, cannot convert to PDB.\nSystem.exit(0)");
					System.exit(0);
				}
				bw.write(String.valueOf(i));
				
				//  write the second index
				int secondIndex = i+1;
				if(secondIndex < 10){
					bw.write("   ");
				}
				else if(secondIndex < 100){
					bw.write("  ");
				}
				else if(secondIndex < 1000){
					bw.write(" ");
				}
				else if(secondIndex < 10000){
					//  don't need to add any white-space
				}
				else{
					System.err.println("Error: the second index is greater than 10000, cannot convert to PDB.\nSystem.exit(0)");
					System.exit(0);
				}
				bw.write(String.valueOf(secondIndex));
				
				bw.newLine();
				bw.flush();
			}
			
			
			//  write the END of the PDB file
			bw.write("END");
			bw.newLine();
			bw.flush();
			
			//  close the writer			
			bw.close();
		}catch (IOException e) {
			System.err.println("Error: could not output file: " + pathFilename + "\nSystem.exit(0)");
			e.printStackTrace();
			System.exit(0);
		}//end of try-catch block
	}/*  end of writeStructure method  */
	
	
	
	
	/*  start of writeEmptyPDBfile method  */
	/**
	 * @param pathFilename the filename to use for the output file including the output path
	 */
	public void writeEmptyPDBfile(String pathFilename){
		if(pathFilename == null){
			System.err.println("Error: the filename is null and the output file cannot be created.\nSystem.exit(0)");
			System.exit(0);
		}
		
		this.writeStructure(pathFilename, new double[0][0], "BLANK");
	}/*  end of writeEmptyPDBfile method  */

	
	
	
	/*  start of getStructure method  */
	public double[][] getStructure(String pathFilename){
		//  declare and initialize the structure (get the number of points from the structure file)
		double[][] structure = new double[getNumberOfPoints(pathFilename)][3];
		
		String line = null;
		int lineCounter = 0;
		try{
			BufferedReader br = new BufferedReader(new FileReader(pathFilename));
			int headerLinesLeft = 2;		//  there are two (2) header lines that need to be skipped
			boolean doneHeader = false;
			boolean doneStructure = false;
			while(((line = br.readLine()) != null) && (!doneStructure)){
				if(doneHeader){
					String[] values = line.split(" ");
					if(values[0].compareTo("ATOM") == 0){
						structure[lineCounter][0] = Double.parseDouble(line.substring(31, 38));
						structure[lineCounter][1] = Double.parseDouble(line.substring(39, 46));
						structure[lineCounter][2] = Double.parseDouble(line.substring(47, 54));
						lineCounter++;
					}
					else{
						doneStructure = true;
					}
				}
				else{
//					System.out.println(line);	//  display the header lines
					headerLinesLeft--;
					if(headerLinesLeft == 0){
						doneHeader = true;
					}
				}
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error: cannot find structure file with filename \"" + pathFilename + "\"\n" +
			"System.exit(0)");
			System.exit(0);
		}
		return structure;
	}/*  end of getStructure method  */
	
	
	
	
	/*  start of getStructureProbabilityFromMCMC5C method  */
	/**
	 * @param pathFilename the filename of the pdb structure to load, including the input path
	 * @return the probability of the structure as saved in the pdb file
	 */
	public double getStructureProbabilityFromMCMC5C(String pathFilename){
		return this.getStructureProbabilityFromHeader(pathFilename, 9);			//  the output from MCMC5C has the probability at index nine (9)
	}/*  end of getStructureProbabilityFromMCMC5C method  */
	
	
	
	
	/*  start of getStructureProbabilityFromGenetic method  */
	/**
	 * @param pathFilename the filename of the pdb structure to load, including the input path
	 * @return the probability of the structure as saved in the pdb file
	 */
	public double getStructureProbabilityFromGenetic(String pathFilename){
		return this.getStructureProbabilityFromHeader(pathFilename, 3);			//  the output from Genetic has the probability at index three (3)
	}/*  end of getStructureProbabilityFromGenetic method  */
	
	
	
	
	
	/*  start of getStructureProbabilityFromHeader method  */
	/**
	 * @param pathFilename the filename of the pdb structure to load, including the input path
	 * @param probabilityIndex the index of the location in the header that has the structure probability
	 * @return the probability of the structure as saved in the pdb file
	 */
	public double getStructureProbabilityFromHeader(String pathFilename, int probabilityIndex){
		//  the probability is stored in the header of the pdb file
		String[] values = this.getHeader(pathFilename).split("\t");
//		for(int i=0; i<values.length; i++){
//			System.out.println(i + ": " + values[i]);
//		}
		return(Double.parseDouble(values[probabilityIndex]));
	}/*  end of getStructureProbabilityFromHeader method  */
	
	
	
	
	/*  start of getStructureFromMCMC5CasChromosome method  */	
	/**
	 * @param pathFilename the filename of the pdb structure to load (as output by MCMC5C), including the input path
	 * @return the structure stored in the pdb file as a Chromosome object, with the fitness set according to the probability in the pdb file
	 */
	public Chromosome getStructureFromMCMC5CasChromosome(String pathFilename){
		Chromosome newChromosome = new Chromosome(this.getStructure(pathFilename));
		newChromosome.setFitness(this.getStructureProbabilityFromMCMC5C(pathFilename));
		return newChromosome;
	}/*  end of getStructureFromMCMC5CasChromosome method  */
	
	
	
	
	/*  start of getStructureFromGeneticAsChromosome method  */	
	/**
	 * @param pathFilename the filename of the pdb structure to load (as output by Genetic), including the input path
	 * @return the structure stored in the pdb file as a Chromosome object, with the fitness set according to the probability in the pdb file
	 */
	public Chromosome getStructureFromGeneticAsChromosome(String pathFilename){
		Chromosome newChromosome = new Chromosome(this.getStructure(pathFilename));
		newChromosome.setFitness(this.getStructureProbabilityFromGenetic(pathFilename));
		return newChromosome;
	}/*  end of getStructureFromGeneticAsChromosome method  */
	
	
	
	
	/*  start of getStructureFrom5C3DasChromosome method  */	
	/**
	 * @param pathFilename the filename of the pdb structure to load (as output by 5C3D), including the input path
	 * @return the structure stored in the pdb file as a Chromosome object, with the fitness set according to the probability in the pdb file
	 */
	public Chromosome getStructureFrom5C3DasChromosome(String pathFilename){
		Chromosome newChromosome = new Chromosome(this.getStructure(pathFilename));
		newChromosome.setFitness(0);
		return newChromosome;
	}/*  end of getStructureFrom5C3DasChromosome method  */
	
	
	
	
	/*  start of getHeader method  */
	public String getHeader(String pathFilename){
		String header = null;
		try{
			BufferedReader br = new BufferedReader(new FileReader(pathFilename));
			header = br.readLine();		//  the header is the first line of the pdb file
			String[] values = header.split("    ");		//  remove the leading COMPND
			if(values.length != 2){
				System.err.println("Error reading the header line of " + pathFilename);
				System.err.println("System.exit(0)");
				System.exit(0);
			}
			header = values[1];
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error: cannot find structure file with filename \"" + pathFilename + "\"\n" +
			"System.exit(0)");
			System.exit(0);
		}
		return header;
	}/*  end of getHeader method  */
	
	
	
	
	/*  start of getHeaderLines method  */
	/**
	 * @param pathFilename the filename (including the prefix path) of the PDB file from which to load the header lines
	 * @return a list containing the header lines stored as String objects
	 */
	public ArrayList<String> getHeaderLines(String pathFilename){
		if(pathFilename == null){
			System.err.println("Error: the filename is null and the header cannot be loaded.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<String> headerLines = new ArrayList<String>();
		String line = null;
		try{
			BufferedReader br = new BufferedReader(new FileReader(pathFilename));
			while((line = br.readLine()) != null){
				if(line.startsWith("ATOM") || line.startsWith("CONECT") || (line.startsWith("END"))){
					break;
				}
				else{
					headerLines.add(line);
				}
			}			
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error: cannot find PDB file with filename \"" + pathFilename + "\"\n" +
			"System.exit(0)");
			System.exit(0);
		}
		
		headerLines.trimToSize();
		
		if(headerLines.size() != 2){
			System.err.println("Error: the header of a PDB file must be exactly two (2) lines (found " + headerLines.size() + " lines in \"" + pathFilename +"\").\nSystem.exit(0)");
			System.exit(0);
		}
		
		return headerLines;
	}/*  end of getHeaderLines method  */
	
	
	
	
	/*  start of getDefaultFilenameForIndex method  */
	/**
	 * @param index the index for which to determine the default structure filename
	 * @return the full default structure filename for the specified index
	 */
	public String getDefaultFilenameForIndex(int index){
		String filename = "";
		if(index<10){
			filename = "structure_000" + index + ".pdb";
		}
		else if(index<100){
			filename = "structure_00" + index + ".pdb";
		}
		else if(index<1000){
			filename = "structure_0" + index + ".pdb";
		}
		else{
			filename = "structure_" + index + ".pdb";
		}
		return filename;
	}/*  end of getDefaultFilenameForIndex method  */
	
	
	
	
	/*  start of zoomStructure  method  */
	public double[][] zoomStructure(double[][] structure, double zoomFactor){
		if(zoomFactor == 0){
			System.err.println("Warning: trying to zoom by a factor of zero (0). No zoom operation performed.");
			return structure;
		}
		
		double[][] zoomedStructure = new double[structure.length][structure[0].length];
		
		for(int i=0; i<structure.length; i++){
			zoomedStructure[i][0] = structure[i][0] * zoomFactor;
			zoomedStructure[i][1] = structure[i][1] * zoomFactor;
			zoomedStructure[i][2] = structure[i][2] * zoomFactor;
		}
		
		return zoomedStructure;
	}/*  end of zoomStructure method  */
	
	
	
	
	/*  start of mirrorStructure method  */
	public double[][] mirrorStructure(double[][] structure){
		double[][] mirrorStructure = new double[structure.length][structure[0].length];
		
		for(int i=0; i<structure.length; i++){
			mirrorStructure[i][0] = -1.0 * structure[i][0];
			mirrorStructure[i][1] = structure[i][1];
			mirrorStructure[i][2] = structure[i][2];
		}
		
		return mirrorStructure;
	}/*  end of mirrorStructure method  */
	
	
	
	
	/*  start of writeAnnotation method  */
	/**
	 * @param pathFilename the filename to output the annotation
	 * @param annotations the list of coordinates of the annotations
	 * @param labels the list of labels of the annotations (in the same order as the annotations list)
	 * @param header the header line for the output file
	 * @param useLabels a boolean indicating whether or not to use the labels in the output PDB file
	 */
	public void writeAnnotation(String pathFilename, ArrayList<Point3d> annotations, ArrayList<String> labels, String header, boolean useLabels){
		//  make sure that all of the pointers are valid
		if((pathFilename == null) || (annotations == null) || (labels == null) || (header == null)){
			System.err.println("Error: null pointer when trying to write the annotations.\nSystem.exit(0)");
			System.exit(0);
		}
		
		annotations.trimToSize();
		double[][] annotationArray = new double[annotations.size()][3];
		for(int i=0; i<annotations.size(); i++){
			Point3d currentPoint = annotations.get(i);
			annotationArray[i][0] = currentPoint.x;
			annotationArray[i][1] = currentPoint.y;
			annotationArray[i][2] = currentPoint.z;
		}
		writeAnnotation(pathFilename, annotationArray, labels, header, useLabels);
	}/*  end of writeAnnotation method  */
	
	
	
	
	/*  start of writeAnnotation method  */
	/**
	 * @param pathFilename the filename to output the annotation
	 * @param annotations the list of coordinates of the annotations
	 * @param labels the list of labels of the annotations (in the same order as the annotations list)
	 * @param header the header line for the output file
	 * @param useLbels a boolean indicating whether or not to use the labels in the output PDB file
	 */
	public void writeAnnotation(String pathFilename, double[][] annotations, ArrayList<String> labels, String header, boolean useLabels){
		//  make sure that all of the pointers are valid
		if((pathFilename == null) || (annotations == null) || (labels == null) || (header == null)){
			System.err.println("Error: null pointer when trying to write the annotations.\nSystem.exit(0)");
			System.exit(0);
		}
		
		//  make sure that there are the same number of annotations and labels
		if(annotations.length != labels.size()){
			System.err.println("Error: there are not the same number of annotations and labels.\nSystem.exit(0)");
			System.exit(0);
		}
		
		//  make the annotations and labels each two atoms
//		double[][] doubleAnnotations = new double[annotations.length*2][3];
//		ArrayList<String> doubleLabels = new ArrayList<String>(labels.size());
//		for(int i=0; i<annotations.length; i++){
//			doubleAnnotations[i*2][0] = annotations[i][0];
//			doubleAnnotations[i*2][1] = annotations[i][1];
//			doubleAnnotations[i*2][2] = annotations[i][2];
//			doubleAnnotations[i*2 + 1][0] = annotations[i][0] + 0.001;
//			doubleAnnotations[i*2 + 1][1] = annotations[i][1];
//			doubleAnnotations[i*2 + 1][2] = annotations[i][2];
//			
//			doubleLabels.add(labels.get(i));
//			doubleLabels.add(labels.get(i));
//		}
//		doubleLabels.trimToSize();
		
		
		//  write the structure to file
		try{						
			BufferedWriter bw = new BufferedWriter(new FileWriter(pathFilename));
			
			
			//  write the header block
			bw.write("COMPND    " + header.toUpperCase() + "\nAUTHOR    " + AUTHOR + "\n");
			bw.flush();
			

			//  write the annotations
			for(int i=0; i<annotations.length; i++){
				
				//  write the type of particle
				String lineToWrite = "ATOM  ";
				
				
				//  write the index of the atom
				int atomIndex = i+1;
				if(atomIndex < 10){
					lineToWrite += "    " + atomIndex;
				}
				else if(atomIndex < 100){
					lineToWrite += "   " + atomIndex;
				}
				else if(atomIndex < 1000){
					lineToWrite += "  " + atomIndex;
				}
				else if(atomIndex < 10000){
					lineToWrite += " " + atomIndex;
				}
				else{
					System.err.println("Error: the index value (" + atomIndex + ") is too large to convert to PDB format.\nSystem.exit(0)");
					System.exit(0);
				}
				
				
				//  write the atom used to represent each point
				lineToWrite += " H";
				
				
				if(useLabels){
					//  write the annotation label (it can be at most three characters in length)
					switch(labels.get(i).length()){
						case 1:
							lineToWrite += "      " + labels.get(i);
							break;
						case 2:
							lineToWrite += "     " + labels.get(i);
							break;
						case 3:
							lineToWrite += "    " + labels.get(i);
							break;
						case 4:
							lineToWrite += "   " + labels.get(i);
							break;
						case 5:
							lineToWrite += "  " + labels.get(i);
							break;
						case 6:
							lineToWrite += " " + labels.get(i);
							break;
						default:
							System.err.println("Error: the label can not be longer than three (3) characters.\nSystem.exit(0)");
							System.exit(0);
							break;
					}
				}
				else{
					lineToWrite += "       ";
				}
				
				//  write the chain identifier
				lineToWrite += " A";
				
				
				//  write the white-space before the coordinates
				//  if all of the coordinates are positive values then this should be nine (9) spaces
				lineToWrite += "         ";
				
								
				//  write the x-coordinate
				//  add the requisite amount of white-space and adjust the number of digits after the decimal (if required)
				if(annotations[i][0] < 10){
					lineToWrite += "  ";
					lineToWrite += formatterPDB3decimals.format(annotations[i][0]) + " ";
				}
				else if(annotations[i][0] < 100){
					lineToWrite += " ";
					lineToWrite += formatterPDB3decimals.format(annotations[i][0]) + " ";
				}
				else if(annotations[i][0] < 1000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB3decimals.format(annotations[i][0]) + " ";
				}
				else if(annotations[i][0] < 10000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB2decimals.format(annotations[i][0]) + " ";
				}
				else if(annotations[i][0] < 100000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB1decimals.format(annotations[i][0]) + " ";
				}
				else if(annotations[i][0] < 1000000){
					lineToWrite += " ";				//  need to add an extra space to make up for the decimal that is removed
					lineToWrite += formatterPDB0decimals.format(annotations[i][0]) + " ";
				}
				else if(annotations[i][0] < 10000000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB0decimals.format(annotations[i][0]) + " ";
				}
				else{
					System.err.println("Error: the x-coordinate value is greater than 10000000 and cannot be converted to PDB.\nSystem.exit(0)");
					System.exit(0);
				}
				
				
				//  write the y-coordinate
				//  add the requisite amount of white-space and adjust the number of digits after the decimal (if required)
				if(annotations[i][1] < 10){
					lineToWrite += "  ";
					lineToWrite += formatterPDB3decimals.format(annotations[i][1]) + " ";
				}
				else if(annotations[i][1] < 100){
					lineToWrite += " ";
					lineToWrite += formatterPDB3decimals.format(annotations[i][1]) + " ";
				}
				else if(annotations[i][1] < 1000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB3decimals.format(annotations[i][1]) + " ";
				}
				else if(annotations[i][1] < 10000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB2decimals.format(annotations[i][1]) + " ";
				}
				else if(annotations[i][1] < 100000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB1decimals.format(annotations[i][1]) + " ";
				}
				else if(annotations[i][1] < 1000000){
					lineToWrite += " ";				//  need to add an extra space to make up for the decimal that is removed
					lineToWrite += formatterPDB0decimals.format(annotations[i][1]) + " ";
				}
				else if(annotations[i][1] < 10000000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB0decimals.format(annotations[i][1]) + " ";
				}
				else{
					System.err.println("Error: the y-coordinate value is greater than 10000000 and cannot be converted to PDB.\nSystem.exit(0)");
					System.exit(0);
				}
				
				
				//  write the z-coordinate
				//  add the requisite amount of white-space and adjust the number of digits after the decimal (if required)
				if(annotations[i][2] < 10){
					lineToWrite += "  ";
					lineToWrite += formatterPDB3decimals.format(annotations[i][2]) + " ";
				}
				else if(annotations[i][2] < 100){
					lineToWrite += " ";
					lineToWrite += formatterPDB3decimals.format(annotations[i][2]) + " ";
				}
				else if(annotations[i][2] < 1000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB3decimals.format(annotations[i][2]) + " ";
				}
				else if(annotations[i][2] < 10000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB2decimals.format(annotations[i][2]) + " ";
				}
				else if(annotations[i][2] < 100000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB1decimals.format(annotations[i][2]) + " ";
				}
				else if(annotations[i][2] < 1000000){
					lineToWrite += " ";				//  need to add an extra space to make up for the decimal that is removed
					lineToWrite += formatterPDB0decimals.format(annotations[i][2]) + " ";
				}
				else if(annotations[i][2] < 10000000){
					//  don't need to add any more white-space
					lineToWrite += formatterPDB0decimals.format(annotations[i][2]) + " ";
				}
				else{
					System.err.println("Error: the z-coordinate value is greater than 10000000 and cannot be converted to PDB.\nSystem.exit(0)");
					System.exit(0);
				}
				
				
				//  write the number for the atom
				lineToWrite += " 1.00";
				
				
				//  write the radius for the atom
				lineToWrite += " 75.00";
				
				
				//  write the trailing white-space (if all other formatting conditions have been met then this should be four (4) spaces) 
				lineToWrite += "    ";
				
				
				//  output the line to write to file
				bw.write(lineToWrite);
				bw.newLine();
				bw.flush();
			}
			
			
			//  connect all of the successive points
//			for(int i=1; i<annotations.length; i+=2){
//				bw.write("CONECT ");
//				//  write the first index
//				if(i < 10){
//					bw.write("   ");
//				}
//				else if(i < 100){
//					bw.write("  ");
//				}
//				else if(i < 1000){
//					bw.write(" ");
//				}
//				else if(i < 10000){
//					//  don't need to add any white-space
//				}
//				else{
//					System.err.println("Error: the first index is greater than 10000, cannot convert to PDB.\nSystem.exit(0)");
//					System.exit(0);
//				}
//				bw.write(String.valueOf(i));
//				
//				//  write the second index
//				int secondIndex = i+1;
//				if(secondIndex < 10){
//					bw.write("   ");
//				}
//				else if(secondIndex < 100){
//					bw.write("  ");
//				}
//				else if(secondIndex < 1000){
//					bw.write(" ");
//				}
//				else if(secondIndex < 10000){
//					//  don't need to add any white-space
//				}
//				else{
//					System.err.println("Error: the second index is greater than 10000, cannot convert to PDB.\nSystem.exit(0)");
//					System.exit(0);
//				}
//				bw.write(String.valueOf(secondIndex));
//				
//				bw.newLine();
//				bw.flush();
//			}
			
			
			//  write the END of the PDB file
			bw.write("END");
			bw.newLine();
			bw.flush();
			
			//  close the writer			
			bw.close();
		}catch (IOException e) {
			System.err.println("Error: could not output file: " + pathFilename);
			e.printStackTrace();
		}//end of try-catch block
		
		if(verbose){
			System.out.println("Output file: " + pathFilename);
		}
	}/*  end of writeAnnotation method  */
	
	
	/*  start of toString method  */
	public String toString(double[][] structure){
		String structureString = "index\t  x\t  y\t  z\n";
		
		for(int i=0; i<structure.length; i++){
			if(i == structure.length-1){
				structureString += "  " + (i+1) + "\t" + formatterPDB3decimals.format(structure[i][0]) + 
				"\t" + formatterPDB3decimals.format(structure[i][1]) +
				"\t" + formatterPDB3decimals.format(structure[i][2]);
			}
			else{
				structureString += "  " + (i+1) + "\t" + formatterPDB3decimals.format(structure[i][0]) + 
					"\t" + formatterPDB3decimals.format(structure[i][1]) +
					"\t" + formatterPDB3decimals.format(structure[i][2]) + "\n";
			}
		}
		
		return structureString;
	}/*  end of toString method  */
}/*  end of PDButility class  */
