package utilities;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;


/** ------------------------------------------------------------------------------------------------------------------
 * start of RawMicroarraySignalCleaner class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class RawMicroarraySignalCleaner {

	

	
	/*  global parameters  */
	private static boolean verbose = true;
	private static String inputPath = "";
	private static String outputPath = "";
	
	private static ArrayList<double[]> signals;
	private static int countValuesTotalInput = 0;
	private static int countValuesRemovedForSaturation = 0;
	private static int countValuesRemovedForBackground = 0;
	private static int countValuesRemovedForTrend = 0;
	private static boolean saturationCleanDone = false;
	private static boolean backgroundCleanDone = false;
	/*  end of global parameters  */
	
	
	
	
	/*  start of resetCountersAndFlags method  */
	/**
	 * @return void
	 */
	private static void resetCountersAndFlags(){
		countValuesTotalInput = 0;
		countValuesRemovedForSaturation = 0;
		countValuesRemovedForBackground = 0;
		countValuesRemovedForTrend = 0;
		saturationCleanDone = false;
		backgroundCleanDone = false;
	}/*  end of resetCountersAndFlags method  */
	
	
	
	
	/*  start of loadRawSignalsFromFile method  */
	/**
	 * @param filename the filename containing the raw microarray signals to be cleaned
	 */
	private static void loadRawSignalsFromFile(String filename){
		
		//  reset all of the values before cleaning the current input file
		resetCountersAndFlags();
		
		signals = new ArrayList<double[]>();
		String line = null;
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename));			
			while((line = br.readLine()) != null){
				String[] values = line.split("\t");
				if(values.length < 10){
					System.err.println("Error: there is an entry in \"" + inputPath + filename + "\" that does not have ten (10) values" +
							" (i.e. index1, index2, 15bp, 18bp, 19bp, 20bp, 21bp, 22bp, 23bp, 24bp)\nSystem.exit(0)");
					System.exit(0);
				}
				double[] rowEntry = new double[10];
				for(int i=0; i<10; i++){
					rowEntry[i] = Double.parseDouble(values[i]);
				}
				signals.add(rowEntry);
				countValuesTotalInput += 7;			//  do not count the 15bp fragment which will be removed as background
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.exit(0);
		}
		signals.trimToSize();
		if(verbose){
			System.out.println("Loaded raw microarray signal file: " + inputPath + filename);
		}
	}/*  end of loadRawSignalsFromFile method  */
	
	
	
	
	/*  start of displayRawData method  */
	/**
	 * void
	 */
	private static void displayRawData(){
		for(int i=0; i<signals.size(); i++){
			double[] currentRow = signals.get(i);
			for(int j=0; j<10; j++){
				System.out.print(currentRow[j] + "\t");
			}
			System.out.println();
		}
	}/*  end of displayRawData method  */
	
	
	
	
	/*  start of cleanRawDataForBackground method  */
	/**
	 * @param background_cutoff the background cutoff percentage (the values must be at least this percentage of the shortest feature length (15bp) value)
	 */
	private static void cleanRawDataForBackground(int background_cutoff){
		if(backgroundCleanDone){
			System.err.println("Error: the input data has already been cleaned for background.\nSystem.exit(0)");
			System.exit(0);
		}
		//  make sure that the signals have been cleaned for saturation before performing the background signal clean
		if(!saturationCleanDone){
			System.err.println("Error: the input data must first be cleaned for satured signals before being cleaned for background.\nSystem.exit(0)");
			System.exit(0);
		}
		
		if(background_cutoff < 0){
			if(verbose){
				System.out.println("\nSkipping the background cleaning step...\n");
			}
		}
		else{
			for(int i=0; i<signals.size(); i++){
				double[] currentRow = signals.get(i);

				//  remove all of the values that are less than the background cutoff
				for(int j=9; j>2; j--){
					if(currentRow[j] != 0){			//  if the value was removed due to failing the saturation test then it will be equal to zero and should be skipped here
						if(currentRow[j] < (background_cutoff * currentRow[2] / 100.0)){
							while(j > 2){
								currentRow[j] = 0;
								j--;
								countValuesRemovedForBackground++;
							}
							break;
						}
					}
				}
				//  subtract the background value from all of the values that passed the background cutoff
				for(int j=9; j>1; j--){
					if(currentRow[j] != 0){
						currentRow[j] -= currentRow[2];
					}
				}
			}
		}
		backgroundCleanDone = true;
	}/*  end of cleanRawDataForBackground method  */
	
	
	

	/*  start of cleanRawDataForSaturation method  */
	/**
	 * @param saturation_cutoff the saturation cutoff value (specify a negative value to skip the saturation cleaning step)
	 */
	private static void cleanRawDataForSaturation(double saturation_cutoff){
		if(saturationCleanDone){
			System.err.println("Error: the input data has already been cleaned for saturation.\nSystem.exit(0)");
			System.exit(0);
		}
		
		if(saturation_cutoff < 0){
			if(verbose){
				System.out.println("\nSkipping the saturation cleaning step...\n");
			}
		}
		else{
			for(int i=0; i<signals.size(); i++){
				double[] currentRow = signals.get(i);

				//  remove all of the values that are greater than the saturation cutoff
				for(int j=3; j<10; j++){
					if(currentRow[j] > saturation_cutoff){
						while(j < 10){
							currentRow[j] = 0;
							j++;
							countValuesRemovedForSaturation++;
						}
						break;
					}
				}
			}
		}
		saturationCleanDone = true;
	}/*  end of cleanRawDataForSaturation method  */
	
	
	
	
	/*  start of cleanRawDataForTrend method  */
	/**
	 * @param trend_cutoff the trend cutoff value
	 */
	private static void cleanRawDataForTrend(int trend_cutoff){
		//  make sure that the signals have been cleaned for both saturation and background before performing the trend signal clean
		if(!saturationCleanDone || !backgroundCleanDone){
			System.err.println("Error: the input data must first be cleaned for both satured signals and background before being cleaned for trend.\nSystem.exit(0)");
			System.exit(0);
		}
		
		if(trend_cutoff < 0){
			if(verbose){
				System.out.println("\nSkipping the trend cleaning step...\n");
			}
		}
		else{
			for(int i=0; i<signals.size(); i++){
				double[] currentRow = signals.get(i);


				//  find the start and end indices of the non-zero entries
				int startIndex = -1;
				int endIndex = -2;
				boolean foundStartIndex = false;
				boolean foundEndIndex = false;
				for(int j=3; j<10; j++){
					if((!foundEndIndex) && (foundStartIndex) && (currentRow[j] == 0)){
						endIndex = j-1;
						foundEndIndex = true;
					}
					if((!foundStartIndex) && (currentRow[j] > 0)){
						startIndex = j;
						foundStartIndex = true;
					}
					if((foundStartIndex) && (!foundEndIndex) && (j == 9)){
						endIndex = j;
						foundEndIndex = true;
					}
				}
				int numberOfEntries = endIndex - startIndex + 1;

				if(numberOfEntries > 0){
					//  find the longest stretch of values that respect the trend cutoff

					int maximumLengthOfStretch = 1;
					int stretchStartIndex = startIndex;
					int stretchEndIndex = startIndex;

					for(int j=startIndex; j<=endIndex; j++){

						for(int k=j; k<endIndex; k++){
							if( currentRow[k+1] < (currentRow[k]/100.0*trend_cutoff)  ){
								int currentLengthOfStretch = k - j + 1;
								if(currentLengthOfStretch > maximumLengthOfStretch){
									maximumLengthOfStretch = currentLengthOfStretch;
									stretchStartIndex = j;
									stretchEndIndex = k;
								}
								break;
							}
							else if(k+1 == endIndex){
								int currentLengthOfStretch = k+1 - j + 1;
								if(currentLengthOfStretch > maximumLengthOfStretch){
									maximumLengthOfStretch = currentLengthOfStretch;
									stretchStartIndex = j;
									stretchEndIndex = k+1;
								}
								break;
							}
						}				
					}


					//  only keep the values contained in the longest stretch
					for(int j=startIndex; j<=endIndex; j++){

						if((j>=stretchStartIndex) && (j<=stretchEndIndex)){
							//  keep the value
							//  do nothing
						}
						else{
							//  remove the value
							currentRow[j] = 0;
							countValuesRemovedForTrend++;
						}

					}


				}
			}
		}
	}/*  end of cleanRawDataForTrend method  */
	
	
	
	
	/*  start of outputCleanedData method  */
	/**
	 * @param inputFilename the filename of the input raw data, will be renamed for the output file
	 */
	private static void outputCleanedData(String inputFilename){
		String outputFilename = outputPath + inputFilename;
		try{			
			BufferedWriter bw = new BufferedWriter(new FileWriter(outputFilename));

			for(double[] rowEntry: signals){
				for(int i=0; i<10; i++){
					if(i < 2){
						bw.write((int)rowEntry[i] + "\t");
					}
					else if(i == 9){
						bw.write(String.valueOf(rowEntry[i]));
					}
					else{
						bw.write(rowEntry[i] + "\t");
					}
				}
				bw.newLine();
				bw.flush();
			}
			bw.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.exit(0);
		}//end of try-catch block
		
		if(verbose){
			System.out.println("\nOutput file: " + outputFilename);
		}
	}/*  end of outputCleanedData method  */
	
	
	
	
	/*  start of outputStats method  */
	/**
	 * @param outputFilename the filename to use for the stats output file
	 * @param backgroundCutoff the background cutoff used to clean the raw data
	 * @param saturationCutoff the saturation cutoff used to clean the raw data
	 * @param trendCutoff the trend cutoff used to clean the raw data
	 */
	private static void outputStats(String outputFilename, int backgroundCutoff, double saturationCutoff, int trendCutoff){
		//  make sure that the output filename is valid
		if(outputFilename == null){
			System.err.println("Error: the pointer to the output filename is null.\nSystem.exit(0)");
			System.exit(0);
		}
		
		outputFilename = outputPath + outputFilename;
		
		try{			
			BufferedWriter bw = new BufferedWriter(new FileWriter(outputFilename));

			bw.write("saturation cutoff =\t" + saturationCutoff);
			bw.newLine();
			bw.flush();
			bw.write("background cutoff =\t" + backgroundCutoff + "%");
			bw.newLine();
			bw.flush();
			bw.write("trend cutoff =\t" + trendCutoff + "%");
			bw.newLine();
			bw.newLine();
			bw.flush();
			bw.write("initial total number of values in input file =\t" + countValuesTotalInput);			
			bw.newLine();
			bw.flush();
			int countValuesTotalInputAfter = countValuesTotalInput - countValuesRemovedForSaturation - countValuesRemovedForBackground - countValuesRemovedForTrend;
			bw.write("final total number of values in input file =\t" + countValuesTotalInputAfter + "[" + (int)(100*countValuesTotalInputAfter/countValuesTotalInput) + "%]");
			bw.newLine();
			bw.newLine();
			bw.flush();
			bw.write("number of values removed for failing saturation test =\t" + countValuesRemovedForSaturation + "[" + (int)(100*countValuesRemovedForSaturation/countValuesTotalInput) + "%]");
			bw.newLine();
			bw.flush();
			bw.write("number of values removed for failing background test =\t" + countValuesRemovedForBackground + "[" + (int)(100*countValuesRemovedForBackground/countValuesTotalInput) + "%]");
			bw.newLine();
			bw.flush();
			bw.write("number of values removed for failing trend test =\t" + countValuesRemovedForTrend + "[" + (int)(100*countValuesRemovedForTrend/countValuesTotalInput) + "%]");
			bw.newLine();
			bw.flush();
			
			bw.newLine();
			bw.flush();
			
			int[] countersEntriesPerRow = new int[8];
			for(int i=0; i<signals.size(); i++){
				double[] currentRow = signals.get(i);
				int currentRowNumberOfEntries = 0;
				for(int j=3; j<10; j++){
					if(currentRow[j] > 0){
						currentRowNumberOfEntries++;
					}
				}
				countersEntriesPerRow[currentRowNumberOfEntries]++;
			}
			for(int i=0; i<8; i++){
				if(i == 1){
					bw.write("number of rows with " + i + " entry =\t" + countersEntriesPerRow[i]);
				}
				else{
					bw.write("number of rows with " + i + " entries =\t" + countersEntriesPerRow[i]);
				}
				bw.newLine();
				bw.flush();
			}
			
			//  make sure that the stats for the frequency of rows with each number of output values were properly calculated
			int totalNumberOfRows = 0;
			for(int i=0; i<8; i++){
				totalNumberOfRows += countersEntriesPerRow[i];
			}
			if(totalNumberOfRows != signals.size()){
				System.err.println("Error: the number of rows in the output data does not match the number of rows in the input signal file.\nSystem.exit(0)");
				System.exit(0);
			}
			
			
			bw.flush();
			bw.close();
		}catch (IOException e) {
			e.printStackTrace();
			System.exit(0);
		}//end of try-catch block
	
		if(verbose){
			System.out.println("\nOutput file: " + outputFilename);
		}
	}/*  end of outputStats method  */
	
		
	

	/*  start of getFilenames method  */
	/**
	 * @param filename_experiment_list the file containing the list of experiment matched filenames (normalization region and experimental region)
	 * @return an ArrayList containing the filenames specified in the file passed as a parameter
	 */
	private static ArrayList<String> getFilenames(String filename_experiment_list){
		if(filename_experiment_list == null){
			System.err.println("Error: the filename list is null.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<String> experimentList = new ArrayList<String>();

		String line = null;
		int lineCounter = -1;			//  there is one header line
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename_experiment_list));
			while((line = br.readLine()) != null){
				if(lineCounter >= 0){
					experimentList.add(line);
				}
				else{
					//  skip and the header line
//					System.out.println(line);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		experimentList.trimToSize();
		
		if(verbose){
			System.out.println("Loaded input file: " + inputPath + filename_experiment_list);
		}
		
		return experimentList;
	}/*  end of getFilenames method  */
	
	
	
	
	/*  start of displayFilenames method  */
	/**
	 * @param filenames_RawMicroarraySignals an ArrayList of Strings specifying the input filenames
	 */
	private static void displayFilenames(ArrayList<String> filenames_RawMicroarraySignals){
		if(filenames_RawMicroarraySignals == null){
			System.err.println("Error: the filename list is null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);	
		}
		
		System.out.println("Filename");
		for(String currentFilename: filenames_RawMicroarraySignals){
			System.out.println(currentFilename);
		}
	}/*  end of displayFilenames method  */
	
	
	
	
	/*  start of main method  */
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		
		
		
		boolean usingCommandLine = false;
		
		double saturationCutoff = -1;
		int backgroundCutoff = 150;
		int trendCutoff = 100;
		String filename_experiment_list = "experiment_list.txt";
		ArrayList<String> filenames_RawMicroarraySignals = new ArrayList<String>(); 
		
		
		//  load the parameters from the command line
		if(args.length >= 4){
			saturationCutoff = Double.parseDouble(args[0]);
			
			backgroundCutoff = Integer.parseInt(args[1]);
			
			trendCutoff = Integer.parseInt(args[2]);
			
			for(int i=3; i<args.length; i++){
				filenames_RawMicroarraySignals.add(args[i]);
			}
			filenames_RawMicroarraySignals.trimToSize();
			
			
			usingCommandLine = true;
			verbose = false;
		}
		else{
			System.err.println("Error: there need to be at least four (4) command line parameters.\n" +
					"\t\tsaturation_cutoff\n" +
					"\t\tbackground_cutoff\n" +
					"\t\ttrend_cutoff\n" +
					"\t\texperiment_1_filename.txt\n" +
					"\t\texperiment_2_filename.txt\n" +
					"\t\texperiment_..._filename.txt\n" +
					"System.exit(0)");
//			System.exit(0);
		}
		
		
		
		
		/******************************************************
		 * 													  *
		 * 				SET THE PARAMETERS HERE				  *
		 * 													  *
		 ******************************************************/
		
		
		if(!usingCommandLine){
			
			
			inputPath = "RawMicroarraySignalCleaner/Input/";
			outputPath = "RawMicroarraySignalCleaner/Output/";
		
		
			saturationCutoff = -1;
//			saturationCutoff = 55000;
			backgroundCutoff = 150;
			trendCutoff = 100;
		
		
			filename_experiment_list = "experiment_list.txt";
			
			
		}
		
		
		/******************************************************
		 * 													  *
		 * 				END OF SETTING THE PARAMETERS		  *
		 * 													  * 
		 ******************************************************/
		
		
		
		
		if(!usingCommandLine){
			filenames_RawMicroarraySignals = getFilenames(filename_experiment_list);
		}
//		displayFilenames(filenames_RawMicroarraySignals);
		
		
		//  display the parameters
		if(verbose){
			System.out.println("saturation cutoff = " + saturationCutoff);
			System.out.println("background cutoff = " + backgroundCutoff + "%");
			System.out.println("trend cutoff = " + trendCutoff + "%");
		}
		
		
		for(String filename_CurrentRawMicroarraySignals: filenames_RawMicroarraySignals){
			//  load the raw microarray signals from file
			loadRawSignalsFromFile(filename_CurrentRawMicroarraySignals);
		
		
			//  display the raw microarray signals
//			displayRawData();
		
		
			//  clean the raw microarray signals using the specified saturation cutoff value
			cleanRawDataForSaturation(saturationCutoff);
		
		
			//  clean the raw microarray signals using the specified background cutoff value
			cleanRawDataForBackground(backgroundCutoff);
		
		
			//  test the trend
			cleanRawDataForTrend(trendCutoff);
		
		
			//  display the cleaned microarray data
//			displayRawData();
		
		
			//  output the cleaned signals to file
			outputCleanedData(filename_CurrentRawMicroarraySignals);
		
		
			//  output the stats of the cleaning run to file
			if(verbose){
				outputStats("stats_" + filename_CurrentRawMicroarraySignals, backgroundCutoff, saturationCutoff, trendCutoff);
			}
		}
		
		
		if(verbose){
			System.out.println("Done!");
		}
	}/*  end of main method  */

}/*  end of RawMicroarraySignalCleaner class  */
