package utilities;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;


/** ------------------------------------------------------------------------------------------------------------------
 * start of SaturationFilter class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class SaturationFilter {

		


	/*  global parameters  */
	private static boolean verbose = true;
	private static String inputPath = "";
	private static String outputPath = "";
	/*  end of global parameters  */
	
	

	
	/*  start of getFilenames method  */
	/**
	 * @param filename_experiment_list the file containing the list of experiments
	 * @return a list of String objects representing the loaded filenames
	 */
	private static ArrayList<String> getFilenames(String filename_experiment_list){
		if(filename_experiment_list == null){
			System.err.println("Error: the filename list is null.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<String> experimentList = new ArrayList<String>();

		String line = null;
		int lineCounter = -1;			//  there is one header line
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename_experiment_list));
			while((line = br.readLine()) != null){
				if(lineCounter >= 0){
					experimentList.add(line);
				}
				else{
					//  skip and the header line
//					System.out.println(line);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		experimentList.trimToSize();
		
		if(verbose){
			System.out.println("Loaded input file: " + inputPath + filename_experiment_list);
		}
		
		return experimentList;
	}/*  end of getFilenames method  */
	
	

	
	/*  start of displayFilenames method  */
	/**
	 * @param experimentList the list of filenames to display
	 */
	private static void displayFilenames(ArrayList<String> experimentList){
		if(experimentList == null){
			System.err.println("Error: the experiment list is null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);	
		}
		
		System.out.println("Filename");
		for(String currentFilename: experimentList){
			System.out.println(currentFilename);
		}
	}/*  end of displayFilenames method  */
	
	
	
	
	/*  start of loadInputData method  */
	/**
	 * @param experimentList the list of filenames to load
	 * @return a list containing the loaded data loaded as FiveC objects
	 */
	private static ArrayList<FiveC> loadInputData(ArrayList<String> experimentList){
		if(experimentList == null){
			System.err.println("Error: the experiment list is null and the data cannot be loaded.\nSystem.exit(0)");
			System.exit(0);	
		}
		
		ArrayList<FiveC> FiveC_inputData = new ArrayList<FiveC>();
		
		
		for(String filename: experimentList){
			FiveC_inputData.add(new FiveC(filename, inputPath, outputPath));
		}
		
		FiveC_inputData.trimToSize();
		
		return FiveC_inputData;
	}/*  end of loadInputData method  */
	
	

	
	/*  start of displayData method  */
	/**
	 * @param FiveC_inputData the list of FiveC objects to display
	 */
	private static void displayData(ArrayList<FiveC> FiveC_inputData){
		if(FiveC_inputData == null){
			System.err.println("The input data arrays are null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		for(FiveC currentFiveC: FiveC_inputData){
			System.out.println("\n\n" + currentFiveC);
		}
	}/*  end of displayData method  */
	
	
	
	
	/*  start of applySaturationFilter method  */
	/**
	 * @param FiveC_inputData the list of FiveC objects to clean
	 * @param saturationLimit the maximum value (inclusive) for the signal intensities, anything larger is removed from the dataset
	 * @return the list of FiveC objects cleaned for saturation
	 */
	private static ArrayList<FiveC> applySaturationFilter(ArrayList<FiveC> FiveC_inputData, double saturationLimit){
		for(FiveC currentFiveC: FiveC_inputData){
			currentFiveC.applySaturationFilter(saturationLimit);
		}
		return FiveC_inputData;
	}/*  end of applySaturationFilter method  */
	
	
	
	
	/*  start of outputDataToFile method  */
	/**
	 * @param FiveCdata the list of FiveC objects to output to file
	 */
	private static void outputDataToFile(ArrayList<FiveC> FiveCdata){
		if(FiveCdata == null){
			System.err.println("Error: the FiveC data is null and cannot be output to file.\nSystem.exit(0)");
			System.exit(0);
		}
		
		for(FiveC currentFiveC: FiveCdata){
			currentFiveC.outputToFile();
		}
	}/*  end of outputDataToFile method  */
	
	
	
	
	/*  start of main method  */
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		

		
		
		boolean usingCommandLine = false;
		
		ArrayList<String> experimentList = new ArrayList<String>();
		double saturationLimit = 40000;
		String filename_experiment_list = "default.txt";
		
		
		//  load the parameters from the command line
		if(args.length >= 2){
			saturationLimit = Double.parseDouble(args[0]);
			
			for(int i=1; i<args.length; i++){
				experimentList.add(args[i]);
			}
			experimentList.trimToSize();
			
			
			usingCommandLine = true;
			verbose = false;
		}
		else{
			System.err.println("Error: there need to be at least two (2) command line parameters.\n" +
					"\t\tsaturation_limit\n" +
					"\t\texperiment_1_filename.txt\n" +
					"\t\texperiment_2_filename.txt\n" +
					"\t\texperiment_..._filename.txt\n" +
					"System.exit(0)");
//			System.exit(0);
		}
		
		
		
		/******************************************************
		 * 													  *
		 * 				SET THE PARAMETERS HERE				  *
		 * 													  *
		 ******************************************************/
		
		
		if(!usingCommandLine){
		
			
			inputPath = "SaturationFilter/Input/";
			outputPath = "SaturationFilter/Output/";

		
			saturationLimit = 40000;
		
		
			filename_experiment_list = "experiment_list.txt";
		
			
		}
		
		
		/******************************************************
		 * 													  *
		 * 				END OF SETTING THE PARAMETERS		  *
		 * 													  * 
		 ******************************************************/
		
		
		
		
		if(!usingCommandLine){
			experimentList = getFilenames(filename_experiment_list);
		}
//		displayFilenames(experimentList);


		ArrayList<FiveC> FiveC_inputData = loadInputData(experimentList);
//		displayData(FiveC_inputData);
		
		
		ArrayList<FiveC> FiveC_cleaned = applySaturationFilter(FiveC_inputData, saturationLimit);
//		displayData(FiveC_cleaned);
		

		outputDataToFile(FiveC_cleaned);
		
		
		if(verbose){
			System.out.println("Done!");
		}
	}/*  end of main method  */

}/*  end of SaturationFilter class  */
