package utilities;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;


/** ------------------------------------------------------------------------------------------------------------------
 * start of SplitIntoTestTrainingSets class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class SplitIntoTestTrainingSets {

	
	
	
	/*  global parameters  */
	private static final String inputPath = "SplitIntoTestTrainingSets/Input/";
	private static final String outputPath = "SplitIntoTestTrainingSets/Output/";
	/*  end of global parameters  */
	
	
	
	
	/*  start of loadTestSetSpecifications method  */
	/**
	 * @param filename_test_set_indices the filename for the file containing the description of each of the test sets
	 */
	private static ArrayList<ArrayList<String>> loadTestSetSpecifications(String filename_test_set_indices){
		if(filename_test_set_indices == null){
			System.err.println("Error: the filename for the test set indices is null and cannot be loaded.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<ArrayList<String>> test_set_indices = new ArrayList<ArrayList<String>>();
		

		String line = null;
		int lineCounter = -1;			//  there is one header line
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename_test_set_indices));
			while((line = br.readLine()) != null){
				if(lineCounter >= 0){
					
					String[] values = line.split("\t");
					if(values.length != 3){
						System.err.println("Error: there are not exactly three entries in each row of the test set indices file.\nSystem.exit(0)");
						System.exit(0);
					}
					
					ArrayList<String> currentRow = new ArrayList<String>();
					
					currentRow.add(values[0]);
					currentRow.add(values[1]);
					currentRow.add(values[2]);
					
					test_set_indices.add(currentRow);
				}
				else{
					//  skip and print the header line
//					System.out.println(line);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		test_set_indices.trimToSize();
		System.out.println("Loaded input file: " + inputPath + filename_test_set_indices);
		
		
		return test_set_indices;
	}/*  end of loadTestSetSpecifications method  */
	
	
	
	
	/*  start of displayTestSetSpecifications method  */
	/**
	 * @param testSetSpecifications the list containing the specifications for each test set
	 */
	private static void displayTestSetSpecifications(ArrayList<ArrayList<String>> testSetSpecifications){
		if(testSetSpecifications == null){
			System.err.println("Error: the list of test set specifications is null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		System.out.println("===  Test Set Specifications  ===");
		for(ArrayList<String> currentTestSetSpecifications: testSetSpecifications){
			String currentLabel = currentTestSetSpecifications.get(0);
			int currentStartIndex = Integer.parseInt(currentTestSetSpecifications.get(1));
			int currentEndIndex = Integer.parseInt(currentTestSetSpecifications.get(2));
			System.out.println(currentLabel + ":\t" + currentStartIndex + "-" + currentEndIndex);
		}
	}/*  end of displayTestSetSpecifications method  */
	
	
	
	
	/*  start of outputTrainingTestSets method  */
	/**
	 * @param fullInputARFF the ARFFfile object containing all of the input data
	 * @param testSetSpecificationsList a list containing the index range and label for each test set to be split and output separately
	 */
	private static void outputTrainingTestSets(ARFFfile fullInputARFF, ArrayList<ArrayList<String>> testSetSpecificationsList){
		if((fullInputARFF == null) || (testSetSpecificationsList == null)){
			System.err.println("Error: both pointers need to be non-null to output the split training-test sets to file.\nSytem.exit(0)");
			System.exit(0);
		}
		
		for(int i=1; i<=testSetSpecificationsList.size(); i++){
			ArrayList<String> currentTestSetSpecifications = testSetSpecificationsList.get(i-1);			//  need to correct the index to start counting at zero
			
			String currentLabel = currentTestSetSpecifications.get(0);
			int currentTestSetStartIndex = Integer.parseInt(currentTestSetSpecifications.get(1));
			int currentTestSetEndIndex = Integer.parseInt(currentTestSetSpecifications.get(2));
			
			String current_filename_training_set = i + "_training_" + "[removed_" + currentTestSetStartIndex + "-" + currentTestSetEndIndex + "]_[" + currentLabel  + "]_"+ fullInputARFF.getFilename();
			String current_filename_test_set = i + "_test_" + "[" + currentTestSetStartIndex + "-" + currentTestSetEndIndex + "]_[" + currentLabel + "]_" + fullInputARFF.getFilename();
			
			//  output the training set by skipping the vectors in the test set range
			fullInputARFF.outputToFileSkippingSpecifiedVectors(outputPath + current_filename_training_set, currentTestSetStartIndex, currentTestSetEndIndex);
			
			//  output the test set by specifying only the vectors in the test set range
			fullInputARFF.outputToFileSpecifiedVectors(outputPath + current_filename_test_set, currentTestSetStartIndex, currentTestSetEndIndex);
		}
	}/*  end of outputTrainingTestSets method  */
	
	
	
	
	/*  start of main method  */
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		
		
		
		/******************************************************
		 * 													  *
		 * 				SET THE PARAMETERS HERE				  *
		 * 													  *
		 ******************************************************/
		
		
//		String filename_ARFF = "MLL_internal_distances_[fusion_type].arff";
//		String filename_test_set_indices = "test_set_indices_[fusion_type].txt";
		
		
		String filename_ARFF = "MLL_internal_distances_[disease_state].arff";
		String filename_test_set_indices = "test_set_indices_[disease_state].txt";
		
		
		/******************************************************
		 * 													  *
		 * 				END OF SETTING THE PARAMETERS		  *
		 * 													  * 
		 ******************************************************/
		
		
		
		
		ARFFfile fullInputARFF = new ARFFfile(inputPath + filename_ARFF);
		
		
		ArrayList<ArrayList<String>> testSetSpecificationsList = loadTestSetSpecifications(filename_test_set_indices);
//		displayTestSetSpecifications(testSetSpecificationsList);
		
		
		outputTrainingTestSets(fullInputARFF, testSetSpecificationsList);
		
		
		System.out.println("Done");
	}/*  end of main method  */

}/*  end of SplitIntoTestTrainingSets class  */
