package utilities;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Set;
import javax.vecmath.Point3d;

import org.openscience.cdk.Atom;

import modeling.Chromosome;



/** ------------------------------------------------------------------------------------------------------------------
 * start of StructureUtility class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class StructureUtility {
	
	
	
	
	/*  global parameters  */
	private boolean verbose = true;
	private String inputPath = "";
	private String outputPath = "";
	private final int SPHERE_GAP_SIZE = 100;								//  put a sphere every 'x' bases to cover a region
	private PDButility utility_PDB;
	/*  end of global parameters  */
	
	
	
	
	/*  start of constructor  */
	/**
	 * @return void
	 */
	public StructureUtility(){
		this("", "", true);
	}/*  end of constructor  */
	
	
	
	
	/*  start of constructor  */
	/**
	 * @param inputPath the String to use as the input path for the current StructureUtility object
	 * @param outputPath the String to use as the output path for the current StructureUtility object
	 * @param verbose the verbose boolean flag to specify if output should be sent to the console (true) or not (false)
	 * @return void
	 */
	public StructureUtility(String inputPath, String outputPath, boolean verbose){
		utility_PDB = new PDButility(verbose);
		
		//  set the inputPath
		this.setInputPath(inputPath);
		
		//  set the outputPath
		this.setOutputPath(outputPath);
		
		//  set the verbose flag
		this.setVerbose(verbose);
		
	}/*  end of constructor  */
	
	

	
	/*  start of setInputPath method  */
	/**
	 * @param inputPath the String to use as the input path for the current StructureUtility object
	 * @return void
	 */
	public void setInputPath(String inputPath){
		if(inputPath == null){
			System.err.println("Error: cannot set the input path to null.\nSystem.exit(0)");
			System.exit(0);
		}
		this.inputPath = inputPath;
	}/*  end of setInputPath method  */
	
	
	
	
	/*  start of getInputPath method  */
	/**
	 * @return the String used as the input path for the current StructureUtility object
	 */
	public String getInputPath(){
		return this.inputPath;
	}/*  end of getInputPath method  */
	
	
	
	
	/*  start of setOutputPath method  */
	/**
	 * @param outputPath the String to use as the output path for the current StructureUtility object
	 * @return void
	 */
	public void setOutputPath(String outputPath){
		if(outputPath == null){
			System.err.println("Error: cannot set the output path to null.\nSystem.exit(0)");
			System.exit(0);
		}
		this.outputPath = outputPath;
	}/*  end of setOutputPath method  */
	
	
	
	
	/*  start of getOutputPath method  */
	/**
	 * @return the String used as the output path for the current StructureUtility object
	 */
	public String getOutputPath(){
		return this.outputPath;
	}/*  end of getOutputPath method  */
	
	
	
	
	/*  start of setVerbose method  */
	/**
	 * @param verbose the verbose boolean flag to specify if output should be sent to the console (true) or not (false)
	 * @return void
	 */
	public void setVerbose(boolean verbose){
		this.verbose = verbose;
	}/*  end of setVerbose method  */
	
	
	
	
	/*  start of getVerbose method  */
	/**
	 * @return the verbose boolean flag that specifies if output is sent to the console (true) or not (false)
	 */
	public boolean getVerbose(){
		return this.verbose;
	}/*  end of getVerbose method  */
	
	
	
	
	/*  start of annotateStructure method  */
	/**
	 * @param filename_Structure the filename for the structure to annotate
	 * @param filename_GenomicPositionMap the filename of the mapping between the structure indices and their corresponding genomic indices
	 * @param filename_BED_features the filename of the features for which to create an annotation file
	 */
	public void annotateStructure(String filename_Structure, String filename_GenomicPositionMap, String filename_BED_features){
		annotateStructure(filename_Structure, filename_GenomicPositionMap, filename_BED_features, null);
	}/*  end of annotateStructure method  */
	

	
	
	/*  start of annotateStructure method  */
	/**
	 * @param filename_Structure the filename for the structure to annotate
	 * @param filename_GenomicPositionMap the filename of the mapping between the structure indices and their corresponding genomic indices
	 * @param filename_BED_features the filename of the features for which to create an annotation file
	 * @param filename_output the filename to use for the output file
	 */
	public void annotateStructure(String filename_Structure, String filename_GenomicPositionMap, String filename_BED_features, String filename_output){
		//  load the PDB structure from file
		double[][] structure = utility_PDB.getStructure(inputPath + filename_Structure);
//		System.out.println(utility_PDB.toString(structure));
		
		annotateStructure(structure, filename_GenomicPositionMap, filename_BED_features, filename_output);
	}/*  end of annotateStructure method  */
	
	
	
	
	/*  start of annotateStructure method  */
	/**
	 * @param structure the structure to annotate
	 * @param filename_GenomicPositionMap the filename of the mapping between the structure indices and their corresponding genomic indices
	 * @param filename_BED_features the filename of the features for which to create an annotation file
	 * @param filename_output the filename to use for the output file
	 */
	public void annotateStructure(Chromosome structure, String filename_GenomicPositionMap, String filename_BED_features, String filename_output){
		if((structure == null) || (filename_GenomicPositionMap == null) || (filename_BED_features == null)){
			System.err.println("Error: all of the pointers must be non-null to annotate the structure.\nSystem.exit(0)");
			System.exit(0);
		}
		annotateStructure(structure.getAsDoubleArray(), filename_GenomicPositionMap, filename_BED_features, filename_output);
	}/*  end of annotateStructure method  */
	
	
	
	
	/*  start of annotateStructure method  */
	/**
	 * @param structure the structure to annotate
	 * @param filename_GenomicPositionMap the filename of the mapping between the structure indices and their corresponding genomic indices
	 * @param filename_features the filename of the features for which to create an annotation file
	 * @param filename_output the filename to use for the output file
	 */
	public void annotateStructure(double[][] structure, String filename_GenomicPositionMap, String filename_features, String filename_output){
		//  load the genomic positions map from file
		HashMap<String, int[]> genomicMapping = loadPositionsFromBED(filename_GenomicPositionMap);
		
		if(genomicMapping.size() != structure.length){
			System.err.println("Error: there are not the same number of entries in the structure (" + structure.length + ") and genomic positions map files (" + genomicMapping.size() + ").\nSystem.exit(0)");
			System.exit(0);
		}
		
//		displayHashMap(genomicMapping);
		//  augment the structure to include the genomic mapping of each point
		double[][] structureWithGenomicPositions = new double[structure.length][4];
		for(int i=0; i<structure.length; i++){
			structureWithGenomicPositions[i][0] = structure[i][0];
			structureWithGenomicPositions[i][1] = structure[i][1];
			structureWithGenomicPositions[i][2] = structure[i][2];
			String currentKey = String.valueOf(i+1);
			if(!genomicMapping.containsKey(currentKey)){
				System.err.println("Error: position " + currentKey + " does not have a mapped entry in the genomic positions file.\nSystem.exit(0)");
				System.exit(0);
			}
			structureWithGenomicPositions[i][3] = genomicMapping.get(String.valueOf((i+1)))[0];			
		}
//		display2Darray(structureWithGenomicPositions);
		
		
		//  load the features from file
		HashMap<String, int[]> features = new HashMap<String, int[]>();
		
		if(filename_features.endsWith(".bed")){
			features = loadPositionsFromBED(filename_features);
		}
		else if(filename_features.endsWith(".wig")){
			features = loadPositionsFromWIG(filename_features);
		}
		else{
			System.err.println("Error: the features filename does not have a recognized extension (\".bed\" or \".wig\"\nSystem.exit(0)");
			System.exit(0);
		}
		
		if(features.size() < 1){
			System.err.println("Error: there are no features specified to be annotated.\nSystem.exit(0)");
			System.exit(0);
		}
//		displayHashMap(features);
		
		
		ArrayList<String> featureLabels = new ArrayList<String>();
		ArrayList<Point3d> featureCoordinates = new ArrayList<Point3d>();
		
		Set<String> allKeysSetFeatures = features.keySet();
		ArrayList<String> allKeysFeatures = new ArrayList<String>(allKeysSetFeatures);
		Collections.sort(allKeysFeatures);
		for(String currentKey: allKeysFeatures){
			int[] currentEntry = features.get(currentKey);
			int index_start = currentEntry[0];
			int index_end = currentEntry[1];
			
			
			//  if there is only one base, then only add one sphere for the location
			int regionSize = index_end - index_start + 1;
			for(int i=0; i<regionSize; i=i+SPHERE_GAP_SIZE){
				if(i == 0){
					featureLabels.add(currentKey);
				}
				else{
					featureLabels.add(" ");
				}
				featureCoordinates.add(new Point3d(getCoordinatesOfGenomicIndex(index_start+i, structureWithGenomicPositions)));
			}
		}
		
		String featuresPrefix = (String)filename_features.subSequence(0, filename_features.lastIndexOf("."));
		
		if(filename_output == null){
			filename_output = featuresPrefix + ".pdb";
		}
		else{
			//  make sure that the output file ending is '.pdb'
			String outputPrefix = (String)filename_output.subSequence(0, filename_output.lastIndexOf("."));
			filename_output = outputPrefix + ".pdb";
		}
		
		utility_PDB.writeAnnotation(outputPath + filename_output, featureCoordinates, featureLabels, featuresPrefix, true);
	}/*  end of annotateStructure method  */
	
	
	
	
	/*  start of getCoordinatesOfGenomicIndex method  */
	/**
	 * @param genomicPosition the genomic position for which to find the geometric coordinates in the specified structure
	 */
	private double[] getCoordinatesOfGenomicIndex(int genomicPosition, double[][] structureWithGenomicPositions){
		//  NOTE: this method assumes that the genomic mappings are in monotonically increasing order
		
		//  make sure that the structure pointer is valid
		if(structureWithGenomicPositions == null){
			System.err.println("Error: the structure pointer is null.\nSystem.exit(0)");
			System.exit(0);
		}
		
		
		double[] coordinates = new double[3];
		
		//  make sure that the genomic position is within the range of the structure
		if((genomicPosition < structureWithGenomicPositions[0][3]) || (genomicPosition > structureWithGenomicPositions[structureWithGenomicPositions.length-1][3])){
			System.err.println("Error: the genomic position " + genomicPosition + " is not contained within the structure.");
			System.err.println("System.exit(0)");
			System.exit(0);
		}
		else{
//			System.out.println("The genomic position " + genomicPosition + " is contained within the structure.");
		}
		
		//  find the endpoints of the fragment that contains the genomic position
		if(structureWithGenomicPositions.length < 2){
			System.err.println("Error: there is less than one fragment in the structure.\nSystem.exit(0)");
			System.exit(0);
		}
		
		int fragmentStartIndex = 0;
		for(int i=0; i<structureWithGenomicPositions.length-1; i++){
			if((genomicPosition >= (int)structureWithGenomicPositions[i][3]) && (genomicPosition <= (int)structureWithGenomicPositions[i+1][3])){
				fragmentStartIndex = i;
				break;
			}			
		}
		//  at this point the genomic position MUST be contained within the region defined by [startPosition,endPosition]
		int fragmentEndIndex = fragmentStartIndex + 1;
		int sizeOfFragment = (int)structureWithGenomicPositions[fragmentEndIndex][3] - (int)structureWithGenomicPositions[fragmentStartIndex][3];
//		System.out.println("size of fragment = " + sizeOfFragment);
		int distanceIntoFragment = genomicPosition - (int)structureWithGenomicPositions[fragmentStartIndex][3];
//		System.out.println("distance into fragment = " + distanceIntoFragment);
		double fractionIntoFragment = distanceIntoFragment / (double)sizeOfFragment;
//		System.out.println("fraction into fragment = " + fractionIntoFragment);
		
		
		double xVector = structureWithGenomicPositions[fragmentEndIndex][0] - structureWithGenomicPositions[fragmentStartIndex][0];
		double yVector = structureWithGenomicPositions[fragmentEndIndex][1] - structureWithGenomicPositions[fragmentStartIndex][1];
		double zVector = structureWithGenomicPositions[fragmentEndIndex][2] - structureWithGenomicPositions[fragmentStartIndex][2];
		
		coordinates[0] = structureWithGenomicPositions[fragmentStartIndex][0] + fractionIntoFragment * xVector;
		coordinates[1] = structureWithGenomicPositions[fragmentStartIndex][1] + fractionIntoFragment * yVector;
		coordinates[2] = structureWithGenomicPositions[fragmentStartIndex][2] + fractionIntoFragment * zVector;
		
		return coordinates;
	}/*  end of getCoordinatesOfGenomicIndex method  */
	
	
	
	
	/*  start of loadPositionsFromBED method  */
	/**
	 * @param filename_BEDfile the bed file containing the positions to be loaded
	 * @return a HashMap containing the positions specified in the bed file
	 */
	private HashMap<String, int[]> loadPositionsFromBED(String filename_BEDfile){
		HashMap<String, int[]> positions = new HashMap<String, int[]>();
		String line = null;
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename_BEDfile));
			int lineCounter = 0;
			while((line = br.readLine()) != null){
				//  there are two (2) header lines to skip
				if(lineCounter < 2){
					//  do nothing
//					System.out.println(line);					
				}
				else{
					String[] values = line.split("\t");
					//  there should be at least three entries, the first is the chromosome, the second is the genomic index start, third is the stop, fourth is the label
					if(values.length < 3){
						System.err.println("Error: not at least three (3) entries per row in genomic mappings file. Not a valid bed file.\nSystem.exit(0)");
						System.exit(0);
					}
					String label = String.valueOf(lineCounter-1);
					//  use the label specified in the file if it exists
					if(values.length > 3){
						label = values[3];
					}
//					System.out.println(line);
					int[] index_start_stop = new int[2];
					index_start_stop[0] = Integer.parseInt(values[1]);
					index_start_stop[1] = Integer.parseInt(values[2]);
					positions.put(label, index_start_stop);
				}
				lineCounter++;
			}
			br.close();			
		}catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error: cannot find bed file with filename \"" + filename_BEDfile + "\"\n" +
			"System.exit(0)");
			System.exit(0);
		}
		
		if(verbose){
			System.out.println("Opened and read file: " + filename_BEDfile);
		}
		
		return positions;
	}/*  end of loadPositionsFromBED method  */
	
	
	
	
	/*  start of loadPositionsFromWIG method  */
	/**
	 * @param filename_WIGfile the Wiggle file containing the positions to be loaded
	 * @return a HashMap containing the positions specified in the bed file
	 */
	private HashMap<String, int[]> loadPositionsFromWIG(String filename_WIGfile){
		HashMap<String, int[]> positions = new HashMap<String, int[]>();
		String line = null;
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename_WIGfile));
			int lineCounter = 0;
			int span = 0;
			while((line = br.readLine()) != null){
				//  there are three (3) header lines to skip
				if(lineCounter < 3){
					//  do nothing
//					System.out.println(line);
					if(lineCounter == 2){
						String[] values = line.split(" ");
						String[] spanArray = values[2].split("=");
						span = Integer.parseInt(spanArray[1]);
						if(verbose){
							System.out.println("Span = " + span);
						}
					}
				}
				else{
					String[] values = line.split("\t");
					//  there should be two entries, the first is the genomic index start and the second is the signal level
					if(values.length != 2){
						System.err.println("Error: there are not exactly two (2) entries per row in the Wiggle file. Not a valid Wiggle file.\nSystem.exit(0)");
						System.exit(0);
					}
					String label = String.valueOf(lineCounter-1);
					//  use the label specified in the file if it exists
					if(values.length > 3){
						label = values[3];
					}
//					System.out.println(line);
					int[] index_start_stop = new int[2];
					index_start_stop[0] = Integer.parseInt(values[0]);
					index_start_stop[1] = index_start_stop[0] + span - 1;
					positions.put(label, index_start_stop);
				}
				lineCounter++;
			}
			br.close();			
		}catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error: cannot find Wiggle file with filename \"" + filename_WIGfile + "\"\n" +
			"System.exit(0)");
			System.exit(0);
		}
		
		if(verbose){
			System.out.println("Opened and read file: " + filename_WIGfile);
		}
		
		return positions;
	}/*  end of loadPositionsFromWIG method  */
	
	
	
	
	/*  start of zoomStructure  method  */
	/**
	 * @param structure the structure to zoom
	 * @param zoomFactor the factor by which to zoom the structure
	 */
	public double[][] zoomStructure(double[][] structure, double zoomFactor){
		if(zoomFactor == 0){
			if(verbose){
				System.err.println("Warning: trying to zoom by a factor of zero (0). No zoom operation performed.");
			}
			return structure;
		}
		
		double[][] zoomedStructure = new double[structure.length][structure[0].length];
		
		for(int i=0; i<structure.length; i++){
			zoomedStructure[i][0] = structure[i][0] * zoomFactor;
			zoomedStructure[i][1] = structure[i][1] * zoomFactor;
			zoomedStructure[i][2] = structure[i][2] * zoomFactor;
		}
		
		return zoomedStructure;
	}/*  end of zoomStructure method  */
	
	
	
	
	/*  start of mirrorStructure method  */
	/**
	 * @param structure the structure to mirror
	 * @return a new double array containing the mirrored structure
	 */
	public double[][] mirrorStructure(double[][] structure){
		double[][] mirrorStructure = new double[structure.length][structure[0].length];
		
		for(int i=0; i<structure.length; i++){
			mirrorStructure[i][0] = -1.0 * structure[i][0];
			mirrorStructure[i][1] = structure[i][1];
			mirrorStructure[i][2] = structure[i][2];
		}
		
		return mirrorStructure;
	}/*  end of mirrorStructure method  */
	
	
	
	
	/*  start of mirrorStructure method  */
	/**
	 * @param structure the structure to mirror
	 * @return a new Chromosome object containing a deep copy mirror of the original structure
	 */
	public Chromosome mirrorStructure(Chromosome structure){
		//  create a new (deep-copy) of the chromosome to mirror
		Chromosome mirroredStructure = new Chromosome(structure);
		
		//  mirror the chromosome by changing the sign of all the x-coordinates
		for(int i=0; i<mirroredStructure.getSize(); i++){
			Atom currentAtom = mirroredStructure.getAtom(i);
			currentAtom.getPoint3d().x *= -1;
		}
		return mirroredStructure;
	}/*  end of mirrorStructure method  */
	
	
	
	
	/*  start of calculatePairwiseDistance method  */
	/**
	 * @param chromosomeOne the first structure in the pair to calculate the pairwise distance
	 * @param chromosomeTwo the second structure in the pair to calculate the pairwise distance
	 * @return the pairwise distance between the two specified structures
	 */
	public double calculatePairwiseDistance(Chromosome chromosomeOne, Chromosome chromosomeTwo){
		//  make sure that the two structures are non null
		if((chromosomeOne == null) || (chromosomeTwo == null)){
			System.err.println("Error: both structures need to have non-null pointers to calculate the pairwise distance.\nSystem.exit(0)");
			System.exit(0);
		}
		//  make sure that the two structures have the same number of points
		if(chromosomeOne.getSize() != chromosomeTwo.getSize()){
			System.err.println("Error: the two structures do not have the same size, cannot calculate the pairwise distance.\nSystem.exit(0)");
			System.exit(0);
		}
		double pairwiseDistance = 0;
		
		
		for(int i=0; i<chromosomeOne.getSize()-1; i++){
			
			Point3d chromosomeOne_point1 = chromosomeOne.getAtomCoordinates(i);
			Point3d chromosomeTwo_point1 = chromosomeTwo.getAtomCoordinates(i);
			
			for(int j=i+1; j<chromosomeOne.getSize(); j++){
			
				Point3d chromosomeOne_point2 = chromosomeOne.getAtomCoordinates(j);
				Point3d chromosomeTwo_point2 = chromosomeTwo.getAtomCoordinates(j);
				
				double chromosomeOneIntraDistance = Math.sqrt(Math.pow((chromosomeOne_point2.x - chromosomeOne_point1.x), 2) + Math.pow((chromosomeOne_point2.y - chromosomeOne_point1.y), 2) + Math.pow((chromosomeOne_point2.z - chromosomeOne_point1.z), 2));
				double chromosomeTwoIntraDistance = Math.sqrt(Math.pow((chromosomeTwo_point2.x - chromosomeTwo_point1.x), 2) + Math.pow((chromosomeTwo_point2.y - chromosomeTwo_point1.y), 2) + Math.pow((chromosomeTwo_point2.z - chromosomeTwo_point1.z), 2));
				
				pairwiseDistance += Math.pow(chromosomeOneIntraDistance - chromosomeTwoIntraDistance, 2);				
			}
		}
		return Math.sqrt(pairwiseDistance / chromosomeOne.getSize());
	}/*  end of calculatePairwiseDistance method  */
	
	
	
	
	/*  start of display1DintArray method  */
	/**
	 * @param arrayToDisplay the array to display
	 */
	private void display1DintArray(int[] arrayToDisplay){
		if(arrayToDisplay == null){
			System.err.println("Error: trying to display an array with a null pointer.\nSystem.exit(0)");
			System.exit(0);
		}
		for(int i=0; i<arrayToDisplay.length; i++){
			System.out.println((i+1) + "\t" + arrayToDisplay[i]);
		}
	}/*  end of display1DintArray method  */
	
	
	
	
	/*  start of display2Darray method  */
	/**
	 * @param arrayToDisplay the array to display
	 */
	private void display2Darray(double[][] arrayToDisplay){
		if(arrayToDisplay == null){
			System.err.println("Error: trying to display an array with a null pointer.\nSystem.exit(0)");
			System.exit(0);
		}
		for(int i=0; i<arrayToDisplay.length; i++){
			System.out.print((i+1) + "\t");
			for(int j=0; j<arrayToDisplay[i].length; j++){
				System.out.print("\t" + arrayToDisplay[i][j]);
			}
			System.out.println();
		}
	}/*  end of display2Darray method  */
	
	
	
	
	/*  start of display2Darray method  */
	/**
	 * @param arrayToDisplay the array to display
	 */
	private void display3Darray(double[][][] arrayToDisplay){
		if(arrayToDisplay == null){
			System.err.println("Error: trying to display an array with a null pointer.\nSystem.exit(0)");
			System.exit(0);
		}
		for(int i=0; i<arrayToDisplay.length; i++){
			System.out.println("start:\t" + arrayToDisplay[i][0][3] + "\t" + arrayToDisplay[i][0][0] + "\t" + arrayToDisplay[i][0][1] + "\t" + arrayToDisplay[i][0][2]);
			System.out.println("end:\t" + arrayToDisplay[i][1][3] + "\t" + arrayToDisplay[i][1][0] + "\t" + arrayToDisplay[i][1][1] + "\t" + arrayToDisplay[i][1][2]);
			System.out.println();
		}
	}/*  end of display2Darray method  */
	
	
	
	
	/*  start of displayHashMap method  */
	/**
	 * @param arrayToDisplay the array to display
	 */
	private void displayHashMap(HashMap<String, int[]> HashMapToDisplay){
		if(HashMapToDisplay == null){
			System.err.println("Error: trying to display a HashMap with a null pointer.\nSystem.exit(0)");
			System.exit(0);
		}
		Set<String> allKeysSet = HashMapToDisplay.keySet();
		ArrayList<String> allKeys = new ArrayList<String>(allKeysSet);
		Collections.sort(allKeys);
		for(String currentKey: allKeys){
			int[] currentEntry = HashMapToDisplay.get(currentKey);
			System.out.println(currentKey + "\t" + currentEntry[0] + "\t" + currentEntry[1]);
		}
	}/*  end of displayHashMap method  */
	
	
	
	
	/*  start of displayArrayList method  */
	/**
	 * @param displayArrayList the array to display
	 */
	private void displayArrayList(ArrayList<String> arrayListToDisplay){
		if(arrayListToDisplay == null){
			System.err.println("Error: trying to display an ArrayList with a null pointer.\nSystem.exit(0)");
			System.exit(0);
		}
		for(String currentEntry: arrayListToDisplay){
			System.out.println(currentEntry);
		}
	}/*  end of displayArrayList method  */
	
	
	
	
	/*  start of calculateIntraStructureDistanceMatrix method  */
	/**
	 * @param structure the structure for which to calculate the intra structure distance matrix
	 */
	public double[][] calculateIntraStructureDistanceMatrix(double[][] structure){
		if(structure == null){
			System.err.println("Error: the structure is null, cannot calculate the intra-structure distances.\nSystem.exit(0)");
			System.exit(0);
		}
		
		double[][] distanceMatrix = new double[structure.length][structure.length];
		
		
		for(int i=0; i<structure.length-1; i++){
			for(int j=i+1; j<structure.length; j++){
				double x1 = structure[i][0];
				double y1 = structure[i][1];
				double z1 = structure[i][2];
				
				double x2 = structure[j][0];
				double y2 = structure[j][1];
				double z2 = structure[j][2];
				
				double pairwiseDistance = ((x2-x1)*(x2-x1))  +  ((y2-y1)*(y2-y1))  +  ((z2-z1)*(z2-z1));
				pairwiseDistance = Math.sqrt(pairwiseDistance);
				
				if(pairwiseDistance <= 0){
					System.err.println("Error: the pairwise distance must be greater than zero, something must have gone wrong in the calculations.\nSystem.exit(0)");
					System.exit(0);
				}
				
				distanceMatrix[i][j] = pairwiseDistance;
				distanceMatrix[j][i] = distanceMatrix[i][j];			//  keep the matrix symmetric
			}
		}
		return distanceMatrix;
	}/*  end of calculateIntraStructureDistanceMatrix method  */
	
	
	
	
	/*  start of calculateIntraStructureDistanceMatrix method  */
	/**
	 * @param structure the structure for which to calculate the intra structure distance matrix
	 */
	public double[][] calculateIntraStructureDistanceMatrix(Chromosome structure){
		if(structure == null){
			System.err.println("Error: the structure is null, cannot calculate the intra-structure distances.\nSystem.exit(0)");
			System.exit(0);
		}
		return calculateIntraStructureDistanceMatrix(structure.getAsDoubleArray());
	}/*  end of calculateIntraStructureDistanceMatrix method  */
	
	
	
	
	/*  start of calculateIntraStructureDistanceMatrixAsVector method  */
	/**
	 * @param structure the structure for which to calculate the intra structure distance matrix
	 */
	public double[] calculateIntraStructureDistanceMatrixAsVector(double[][] structure){
		if(structure == null){
			System.err.println("Error: the structure is null, cannot calculate the intra-structure distances.\nSystem.exit(0)");
			System.exit(0);
		}
		double[][] matrix = calculateIntraStructureDistanceMatrix(structure);
		
		int numberOfEntries = matrix.length * (matrix.length - 1) / 2;			//  this is (n choose 2)
		
		double[] vector = new double[numberOfEntries];
		
		int vectorIndex = 0;
		for(int i=0; i<matrix.length-1; i++){
			for(int j=i+1; j<matrix.length; j++){
				vector[vectorIndex] = matrix[i][j];
				vectorIndex++;
			}
		}
		
		return vector;
	}/*  end of calculateIntraStructureDistanceMatrixAsVector method  */
	


	
	/*  start of calculateIntraStructureDistanceMatrixAsVector method  */
	/**
	 * @param structure the structure for which to calculate the intra structure distance matrix
	 */
	public double[] calculateIntraStructureDistanceMatrixAsVector(Chromosome structure){
		if(structure == null){
			System.err.println("Error: the structure is null, cannot calculate the intra-structure distances.\nSystem.exit(0)");
			System.exit(0);
		}
		return calculateIntraStructureDistanceMatrixAsVector(structure.getAsDoubleArray());
	}/*  end of calculateIntraStructureDistanceMatrixAsVector method  */
	
	
	
	
}/*  end of StructureUtility class  */
