package analysis;


/**
 * @author Mathieu Rousseau
 * @email mathieu.rousseau3@mail.mcgill.ca
 */


import java.awt.BorderLayout;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Random;

import javax.swing.JFrame;

import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;
import weka.classifiers.functions.SMO;
import weka.classifiers.trees.J48;
import weka.core.converters.ConverterUtils.DataSink;
import weka.core.converters.ConverterUtils.DataSource;
import weka.core.Instances;
import weka.gui.treevisualizer.PlaceNode2;
import weka.gui.treevisualizer.TreeVisualizer;


/** ------------------------------------------------------------------------------------------------------------------
 * start of WEKALOOCV class
 * ------------------------------------------------------------------------------------------------------------------  **/
public class WEKALOOCV {

	
	
	
	/*  global parameters  */
	private static final String inputPath = "WEKALOOCV/Input/";
	private static final String outputPath = "WEKALOOCV/Output/";
	/*  end of global parameters  */
	
	
	
	
	/*  start of loadFilenames method  */
	/**
	 * @param filename_file_list the filename for the file containing the list of paired training-test filenames
	 */
	private static ArrayList<String[]> loadFilenames(String filename_file_list){
		if(filename_file_list == null){
			System.err.println("Error: the filename of the file list is null and the filenames cannot be loaded.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<String[]> filenamePairs = new ArrayList<String[]>();		

		String line = null;
		int lineCounter = -1;			//  there is one header line
		try{
			BufferedReader br = new BufferedReader(new FileReader(inputPath + filename_file_list));
			while((line = br.readLine()) != null){
				if(lineCounter >= 0){
					String[] values = line.split("\t");
					if(values.length != 2){
						System.err.println("Error: there are not exactly two (2) entries in each row of the file list.\nSystem.exit(0)");
						System.exit(0);
					}
					String[] currentPair = new String[2];
					currentPair[0] = values[0];
					currentPair[1] = values[1];
					filenamePairs.add(currentPair);
				}
				else{
					//  skip and print the header line
//					System.out.println(line);
				}
				lineCounter++;
			}
			br.close();
		}catch (IOException e) {
			e.printStackTrace();
		}
		filenamePairs.trimToSize();
		
		System.out.println("Loaded input file: " + inputPath + filename_file_list);
		
		return filenamePairs;
	}/*  end of loadFilenames method  */
	
	
	
	
	/*  start of displayFilenamePairs method  */
	/**
	 * @param filename_pairs the list of filenames for the training-test pairs
	 */
	private static void displayFilenamePairs(ArrayList<String[]> filename_pairs){
		if(filename_pairs == null){
			System.err.println("Error: the filename pairs list is null and cannot be dispalyed.\nSystem.exit(0)");
			System.exit(0);
		}
		System.out.println("=====\t\tTraining\t\t===\t\tTest\t\t=====");
		for(String[] currentPair: filename_pairs){
			System.out.println("\"" + currentPair[0] + "\"\t\"" + currentPair[1] + "\"");
		}
	}/*  end of displayFilenamePairs method  */
	
	
	
	
	/*  start of loadData method  */
	/**
	 * @param filename_pairs the list of filenames for the training-test pairs
	 */
	private static ArrayList<Instances[]> loadData(ArrayList<String[]> filename_pairs) throws Exception{
		if(filename_pairs == null){
			System.err.println("Error: the filename pairs list is null and cannot be dispalyed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		ArrayList<Instances[]> dataPairs = new ArrayList<Instances[]>();
		
		for(String[] currentFilenamePair: filename_pairs){
			//  load the training data
			Instances current_training_data = DataSource.read(inputPath + currentFilenamePair[0]);
			if(current_training_data.classIndex() == -1){
				current_training_data.setClassIndex(current_training_data.numAttributes()-1);
			}
			

			//  load the test data
			Instances current_test_data = DataSource.read(inputPath + currentFilenamePair[1]);
			if(current_test_data.classIndex() == -1){
				current_test_data.setClassIndex(current_test_data.numAttributes()-1);
			}
			
			Instances[] currentDataPair = new Instances[2];
			currentDataPair[0] = current_training_data;
			currentDataPair[1] = current_test_data;
			
			dataPairs.add(currentDataPair);
		}
		
		dataPairs.trimToSize();
		
		return dataPairs;
	}/*  end of loadData method  */
	
	
	
	
	/*  start of displayDataPairs method  */
	/**
	 * @param dataPairs the list of paired data Instances objects containing the training-test pairs
	 * @param filenamePairs the list of paired String objects containing the filenames of the training-test pairs
	 */
	private static void displayDataPairs(ArrayList<Instances[]> dataPairs, ArrayList<String[]> filenamePairs){
		if(dataPairs == null){
			System.err.println("Error: the data pointer is null and the data pairs cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		for(int i=0; i<dataPairs.size(); i++){
			
			Instances[] currentDataPair = dataPairs.get(i);
			String[] currentFilenamePair = filenamePairs.get(i);
			
			System.out.println("\n\n===  Current Pair  ===");
			System.out.println("===  Training  (" + currentFilenamePair[0] + ")  ===");
			System.out.println(currentDataPair[0].toSummaryString());
			System.out.println("\n===  Test  (" + currentFilenamePair[1] + ")  ===");
			System.out.println(currentDataPair[1].toSummaryString());
		}
	}/*  end of displayDataPairs method  */
	
	
	
	
	/*  start of decisionTreeLOOCV method  */
	/**
	 * @param trainingTestPairs the list of paired Instances objects containing the training-test data pairs
	 * @param filenamePairs the list of paired String objects containing the filenames of the training-test pairs
	 * @return the confusion matrix summed over all of the leave-one-out cross-validation evaluations
	 */
	private static double[][] decisionTreeLOOCV(ArrayList<Instances[]> trainingTestPairs, ArrayList<String[]> filenamePairs) throws Exception{
		if(trainingTestPairs == null){
			System.err.println("Error: the data pairs are null and the LOOCV cannot be performed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		int numberOfClasses = trainingTestPairs.get(0)[0].numClasses();
		
		double[][] confusionMatrix = new double[numberOfClasses][numberOfClasses];
		
		for(int i=0; i<trainingTestPairs.size(); i++){
			
			Instances[] currentPair = trainingTestPairs.get(i);
			
			Instances trainingSet = currentPair[0];
			Instances testSet = currentPair[1];
			

			J48 decisionTreeClassifier = new J48();
			decisionTreeClassifier.buildClassifier(trainingSet);
			
			Evaluation evaluationOfClassifier = new Evaluation(trainingSet);
			
			evaluationOfClassifier.evaluateModel(decisionTreeClassifier, testSet);
			
			System.out.println(evaluationOfClassifier.toSummaryString("\nResults\n\n", false));
			System.out.println(evaluationOfClassifier.toMatrixString("\nResults\n\n"));
			displayDecisionTree(decisionTreeClassifier, filenamePairs.get(i)[0]);
			
			double[][] currentConfusionMatrix = evaluationOfClassifier.confusionMatrix();
			
			for(int m=0; m<confusionMatrix.length; m++){
				for(int n=0; n<confusionMatrix.length; n++){
					confusionMatrix[m][n] += currentConfusionMatrix[m][n];
				}
			}
		}
		return confusionMatrix;
	}/*  end of decisionTreeLOOCV method  */

	
	
	
	/*  start of SVMLOOCV method  */
	/**
	 * @param trainingTestPairs the list of paired Instances objects containing the training-test data pairs
	 * @param filenamePairs the list of paired String objects containing the filenames of the training-test pairs
	 * @return the confusion matrix summed over all of the leave-one-out cross-validation evaluations
	 */
	private static double[][] SVMLOOCV(ArrayList<Instances[]> trainingTestPairs, ArrayList<String[]> filenamePairs) throws Exception{
		if(trainingTestPairs == null){
			System.err.println("Error: the data pairs are null and the LOOCV cannot be performed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		int numberOfClasses = trainingTestPairs.get(0)[0].numClasses();
		
		double[][] confusionMatrix = new double[numberOfClasses][numberOfClasses];
		
		for(int i=0; i<trainingTestPairs.size(); i++){
			
			Instances[] currentPair = trainingTestPairs.get(i);
			
			Instances trainingSet = currentPair[0];
			Instances testSet = currentPair[1];
			
			
			SMO SVMclassifier = new SMO();
			SVMclassifier.buildClassifier(trainingSet);
			

//			J48 decisionTreeClassifier = new J48();
//			decisionTreeClassifier.buildClassifier(trainingSet);
			
			Evaluation evaluationOfClassifier = new Evaluation(trainingSet);
			
			evaluationOfClassifier.evaluateModel(SVMclassifier, testSet);
			
			System.out.println(evaluationOfClassifier.toSummaryString("\nResults\n\n", false));
			System.out.println(evaluationOfClassifier.toMatrixString("\nResults\n\n"));
			
			double[][] currentConfusionMatrix = evaluationOfClassifier.confusionMatrix();
			
			for(int m=0; m<confusionMatrix.length; m++){
				for(int n=0; n<confusionMatrix.length; n++){
					confusionMatrix[m][n] += currentConfusionMatrix[m][n];
				}
			}
		}
		return confusionMatrix;
	}/*  end of SVMLOOCV method  */
	
	
	
	
	/*  start of displayConfusionMatrix method  */
	/**
	 * @param confusionMatrix the confusion matrix to display
	 */
	private static void displayConfusionMatrix(double[][] confusionMatrix){
		displayConfusionMatrix(confusionMatrix, null);
	}/*  end of displayConfusionMatrix method  */
	
	
	
	
	/*  start of displayConfusionMatrix method  */
	/**
	 * @param confusionMatrix the confusion matrix to display
	 * @param title a String representation of the title to use for the confusion matrix
	 */
	private static void displayConfusionMatrix(double[][] confusionMatrix, String title){
		if(confusionMatrix == null){
			System.err.println("Error: the confusion matrix is null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		if(title == null){
			title = "";
		}
		
		System.out.println("\n===  Confusion Matrix (" + title + ")  ===");
		for(int i=0; i<confusionMatrix.length; i++){
			for(int j=0; j<confusionMatrix.length; j++){
				System.out.print((int)confusionMatrix[i][j] + "\t");
			}
			System.out.println();
		}
	}/*  end of displayConfusionMatrix method  */
	
	
	
	
	/*  start of displayDecisionTree method  */
	/**
	 * @param decisionTreeClassifier the J48 decision tree classifier to display to screen
	 */
	private static void displayDecisionTree(J48 decisionTreeClassifier) throws Exception{
		displayDecisionTree(decisionTreeClassifier, null);
	}/*  end of displayDecisionTree method  */
	
	
	
	
	/*  start of displayDecisionTree method  */
	/**
	 * @param decisionTreeClassifier the J48 decision tree classifier to display to screen
	 * @param title a String containing the title to be used for the display window
	 */
	private static void displayDecisionTree(J48 decisionTreeClassifier, String title) throws Exception{
		if(decisionTreeClassifier == null){
			System.err.println("Error: the decision tree classifier is null and cannot be displayed.\nSystem.exit(0)");
			System.exit(0);
		}
		
		if(title == null){
			title = "Weka Classifier Tree Visualizer: J48";
		}
		
		TreeVisualizer tv = new TreeVisualizer(null, decisionTreeClassifier.graph(), new PlaceNode2());
		JFrame jf = new JFrame(title);
		jf.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		jf.setSize(800, 600);
		jf.getContentPane().setLayout(new BorderLayout());
		jf.getContentPane().add(tv, BorderLayout.CENTER);
		jf.setVisible(true);
		// adjust tree
		tv.fitToScreen();
	}/*  end of displayDecisionTree method  */
	
	
	
	
	/*  start of writeInstancesToFile method  */
	/**
	 * @param dataInstances the Instances object containing the data to be output to file
	 * @param filename the name of the file to use for the output data
	 */
	private static void writeDataInstanceToFile(Instances dataInstances, String filename) throws Exception{
		if((dataInstances == null) || (filename == null)){
			System.err.println("Error: the Instances and filename must be non-null to output to file.\nSystem.exit(0)");
		}
		DataSink.write(outputPath + filename, dataInstances);
	}/*  end of writeDataInstanceToFile method  */
	
	
	
	
	/*  start of main method  */
	/**
	 * @param args
	 * @throws Exception 
	 */
	public static void main(String[] args) throws Exception {
		
		
		boolean runDecisionTree = false;
		boolean runSVM = false;
		
		
		
		
		/******************************************************
		 * 													  *
		 * 				SET THE PARAMETERS HERE				  *
		 * 													  *
		 ******************************************************/
		
		
//		String filename_file_list = "file_list_[test].txt";
//		String filename_file_list = "file_list_[test3].txt";
//		String filename_file_list = "file_list_[test4].txt";
//		String filename_file_list = "file_list_[fusion_type].txt";
//		String filename_file_list = "file_list_[disease_state].txt";
//		String filename_file_list = "file_list_[fusion_type_no_NB-4].txt";
		String filename_file_list = "file_list_[disease_state_no_NB-4].txt";
		
		
		runDecisionTree = false;
		runSVM = true;
		
		
		/******************************************************
		 * 													  *
		 * 				END OF SETTING THE PARAMETERS		  *
		 * 													  * 
		 ******************************************************/
		
		
		
		
		ArrayList<String[]> training_test_pair_filenames = loadFilenames(filename_file_list);
//		displayFilenamePairs(training_test_pair_filenames);
		
		
		ArrayList<Instances[]> training_test_pair_data = loadData(training_test_pair_filenames);
//		displayDataPairs(training_test_pair_data, training_test_pair_filenames);
		
		
		if(runDecisionTree){
			double[][] decisionTreeConfusionMatrix = decisionTreeLOOCV(training_test_pair_data, training_test_pair_filenames);
			displayConfusionMatrix(decisionTreeConfusionMatrix, "Decision Tree");
		}
		
		
		if(runSVM){
			double[][] SVMconfusionMatrix = SVMLOOCV(training_test_pair_data, training_test_pair_filenames);
			displayConfusionMatrix(SVMconfusionMatrix, "SVM");
		}
		
		
		System.out.println("Done!");
	}/*  end of main method  */

}/*  end of WEKALOOCV class  */
